/*
** fpga.c - Infrastructure used by the Win32 sample applications for parsing
**          command-line arguments and generating filenames for .BIT files.
**
** (C) Copyright 2011 Alpha Data
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _vstprintf_s(buf, max, format, ...) _vsntprintf(buf, max, format, __VA_ARGS__)
# define memcpy_s(pDst, max, pSrc, n) memcpy(pDst, pSrc, n)
#   define _tcscat_s(pDstStr, max, pStr) strcat(pDstStr, pStr)
#   define _tcscpy_s(pDstStr, max, pSrcStr) strcpy(pDstStr, pSrcStr)
#   define _tcslwr_s(pStr, max) strlwr_s(pStr, max)
# endif

#else

/* Linux or VxWorks */
# define _T(c) c
# if (defined(__VXWORKS__) || defined(__vxworks)) && (_WRS_VXWORKS_MAJOR <= 5)
  /* VxWorks 5.x doesn't have snprintf() or vsnprintf() */
#   define _vstprintf_s(buf, max, format, ...) vsprintf(buf, format, __VA_ARGS__)
# else
#   define _vstprintf_s(buf, max, format, ...) vsnprintf(buf, max, format, __VA_ARGS__)
# endif
# define memcpy_s(pDst, max, pSrc, n) memcpy(pDst, pSrc, n)
# define _tcscat_s(pDstStr, max, pStr) strcat(pDstStr, pStr)
# define _tcscpy_s(pDstStr, max, pSrcStr) strcpy(pDstStr, pSrcStr)
# define _tcslen(pStr) strlen(pStr)
# define _tcslwr_s(pStr, max) strlwr_s(pStr, max)
# define _tprintf printf

#endif

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <ctype.h>

#include "sdk_common.h"

#define NUL _T('\0')

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/*
** The environment variable that the sample applications
** attempt to pick up and use to find the base of the ADM-XRC
** SDK and hence the FPGA bitstream files.
*/
#define ENVIRONMENT_VARIABLE (_T("ADMXRC3_SDK"))

#if _WIN32
/* Windows */
static const TCHAR nativeSlash = _T('\\');
static const TCHAR nativeSlashStr[] = _T("\\");
static const TCHAR nonNativeSlash = _T('/');
static const TCHAR nonNativeSlashStr[] = _T("/");
#else
/* Linux or VxWorks */
static const TCHAR nativeSlash = _T('/');
static const TCHAR nativeSlashStr[] = _T("/");
static const TCHAR nonNativeSlash = _T('/');
static const TCHAR nonNativeSlashStr[] = _T("/");
#endif

#define SET_STRING(STRING, VALUE) { \
  if ((STRING) != NULL) { \
		free(STRING); \
    (STRING) = NULL; \
  } \
  if ((VALUE) != NULL) { \
    size_t n = _tcslen(VALUE) + 1; \
    (STRING) = (TCHAR*)malloc(sizeof(TCHAR) * n); \
    memcpy_s((STRING), sizeof(TCHAR) * n, VALUE, sizeof(TCHAR) * n); \
  } \
}

static void
fatal(
  TCHAR* pFmt, ...)
{
  TCHAR buffer[1024];
  va_list ap;
  /* size_t n; */

  va_start(ap, pFmt);
  /* n = */ _vstprintf_s(buffer, ARRAY_LENGTH(buffer), pFmt, ap);
  va_end(ap);

  _tprintf(_T("%s"), buffer);
}

#if !defined(_MSC_VER)
static void
strlwr_s(
  TCHAR* pStr,
  size_t max)
{
  TCHAR c;

  while (max) {
    c = *pStr;
    if (c == NUL) {
      break;
    }
    *pStr++ = tolower(c);
    max--;
  }
}
#endif

const TCHAR*
sampleGetFpgaFilename(
  ADMXRC3_HANDLE hCard,
  unsigned int targetIndex,
  const TCHAR* pDir,
  const TCHAR* pDesign,
  Arguments* pArgs,
  bool_t* pbShouldConfigure)
{
  const TCHAR ext[] = _T(".bit");
  TCHAR* pModelName;
  TCHAR* pBitFilename;
  ADMXRC3_CARD_INFO cardInfo;
  ADMXRC3_FPGA_INFO fpgaInfo;
  ADMXRC3_STATUS status;
  bool_t bShouldConfigure = TRUE;

  if (pDesign == NULL) {
    fatal(_T("get_fpga_filename_dir(): argument 'pDesign' is NULL (must not be NULL)\n"));
    return NULL;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (ADMXRC3_SUCCESS != status) {
    fatal(_T("get_fpga_filename_dir(): cannot get card information\n"));
    return NULL;
  }

  status = ADMXRC3_GetFpgaInfo(hCard, targetIndex, &fpgaInfo);
  if (ADMXRC3_SUCCESS != status) {
    fatal(_T("get_fpga_filename_dir(): cannot get information for target FPGA %lu(0x%lX)\n"), (unsigned long)targetIndex, (unsigned long)targetIndex);
    return NULL;
  }
  _tcslwr_s(fpgaInfo.Device, ARRAY_LENGTH(fpgaInfo.Device));

  switch(cardInfo.Model)
  {
  case ADMXRC3_MODEL_GENERIC:
    fatal(_T("Cannot get .BIT filename for generic model\n"));
    return NULL;

  case ADMXRC3_MODEL_ADMXRC2:
    pModelName = _T("admxrc2");
    break;

  case ADMXRC3_MODEL_ADPEXRC5T:
    pModelName = _T("adpexrc5t");
    break;

  case ADMXRC3_MODEL_ADMXRC6TL:
    pModelName = _T("admxrc6tl");
    break;

  case ADMXRC3_MODEL_ADMXRC6T1:
    pModelName = _T("admxrc6t1");
    break;

  case ADMXRC3_MODEL_ADMXRC6TGE:
    pModelName = _T("admxrc6tge");
    break;

  case ADMXRC3_MODEL_ADMXRC6TADV8:
    /* Target FPGA is not configurable via ADMXRC3 API */
    bShouldConfigure = FALSE;
    pModelName = _T("admxrc6tadv8");
    break;

  case ADMXRC3_MODEL_ADPEXRC6T:
    pModelName = _T("adpexrc6t");
    break;

  case ADMXRC3_MODEL_ADPEXRC6TL:
    pModelName = _T("adpexrc6tl");
    break;

  default:
    fatal(_T("Unknown model: %d\n"), cardInfo.Model);
    return NULL;
  }

  if (pArgs->bitPath != NULL) {
    size_t n;

    n = _tcslen(pArgs->bitPath) + 1 +
      _tcslen(pDir) + 1 +
      _tcslen(pDesign) + 1 +
      _tcslen(pModelName) + 1 +
      _tcslen(fpgaInfo.Device) +
      _tcslen(ext) + 1;
    pBitFilename = malloc(n * sizeof(TCHAR));
    if (NULL == pBitFilename) {
      fatal(_T("*** get_fpga_filename(): failed to allocate 'bitfile'\n"));
      return NULL;
    }
    _tcscpy_s(pBitFilename, n, pArgs->bitPath);
    _tcscat_s(pBitFilename, n, nativeSlashStr);
    _tcscat_s(pBitFilename, n, pDir);
    _tcscat_s(pBitFilename, n, nativeSlashStr);
    _tcscat_s(pBitFilename, n, pDesign);
    _tcscat_s(pBitFilename, n, _T("-"));
    _tcscat_s(pBitFilename, n, pModelName);
    _tcscat_s(pBitFilename, n, _T("-"));
    _tcscat_s(pBitFilename, n, fpgaInfo.Device);
    _tcscat_s(pBitFilename, n, ext);
    SET_STRING(pArgs->bitFile, pBitFilename);
    free(pBitFilename);
  } else {
    fatal(
      _T("The path to the FPGA bitstream directory cannot be determined via the ")
      _T("%s environment variable and has not been specified via the ")
      _T("via the /sdk or /bitdir switches. Please correct this and ")
      _T("try again.\n"),
      ENVIRONMENT_VARIABLE);
    return NULL;
  }

  if (NULL != pbShouldConfigure) {
    *pbShouldConfigure = bShouldConfigure;
  }
  return pArgs->bitFile;
}
