#include <math.h>
#include <iostream>
#include <cairomm/context.h>

#include "scopewnd.h"

#if !defined(CLEAR_TRACES_ON_PERIOD_CHANGE)
# define CLEAR_TRACES_ON_PERIOD_CHANGE (0)
#endif

static const TCHAR* g_prefixStr[] = {
  // From atto- (10^-18) up to exa- (10^15)
  _T("a"), _T("f"), _T("p"), _T("n"), _T("u"), _T("m"), _T(""), _T("k"), _T("M"), _T("G"), _T("T"), _T("P"), _T("E")
};
static const float g_factor[] = {
  1.0e18f, 1.0e15f, 1.0e12f, 1.0e9f, 1.0e6f, 1.0e3f, 1.0f, 1.0e-3f, 1.0e-6f, 1.0e-9f, 1.0e-12f, 1.0e-15f, 1.0e-18f
};

// -------------------------------------------------------------------------
// Start of IScopeWnd interface
// -------------------------------------------------------------------------

void
CScopeWnd::addTrace(
  unsigned int sensorIndex,
  int at) {
  unsigned int i, pos, n = m_numTrace + 1;
  CTraceBuffer** ppNew;
  CTraceBuffer* pNew;

  pos = (at == -1) ? m_numTrace : (unsigned int)at;
  ppNew = new CTraceBuffer*[n];
  ASSERT(NULL != ppNew);
  for (i = 0; i < pos; i++) {
    ppNew[i] = m_ppTraceBuffers[i];
  }
  pNew = new CTraceBuffer(sensorIndex);
  ppNew[pos] = pNew;
  pNew->setSize(m_layout.m_width);
  for (i = pos + 1; i < n; i++) {
    ppNew[i] = m_ppTraceBuffers[i - 1];
  }
  delete[] m_ppTraceBuffers;
  m_ppTraceBuffers = ppNew;
  setNumTrace(n);
}

void
CScopeWnd::adjustScale(
  float value)
{
  bool_t bChanged = FALSE;

  if (value < m_minimum) {
    m_minimum = value;
    bChanged = TRUE;
  }
  if (value > m_maximum) {
    m_maximum = value;
    bChanged = TRUE;
  }
  if (bChanged) {
    calcYScaling(m_scaling, m_ppTraceBuffers, m_layout, m_numTrace);
    needRedraw();
  }
}

void
CScopeWnd::removeAllTraces() {
  if (m_numTrace) {
    for (unsigned int i = 0; i < m_numTrace; i++) {
      delete m_ppTraceBuffers[i];
      m_ppTraceBuffers[i] = NULL;
    }
    delete[] m_ppTraceBuffers;
    m_ppTraceBuffers = NULL;
  }
  setNumTrace(0);
  m_minimum = 1.0e20f;
  m_maximum = -1.0e20f;
}

void
CScopeWnd::deviceChanged() {
  removeAllTraces();
  needRedraw();
}

void
CScopeWnd::idle() {
}

void
CScopeWnd::setNumTrace(
  unsigned int numTrace) {
  m_numTrace = numTrace;
  if (m_numTrace) {
    calcYScaling(m_scaling, m_ppTraceBuffers, m_layout, m_numTrace);
  }
  m_pParent->getToolbar()->numTraceChanged();
  needRedraw();
}

void
CScopeWnd::update(timestamp_t timestamp) {
  DeviceInfo devInfo = m_pParent->getParent()->getDeviceInfo();
  SensorInfo* pSensorInfo;
  CTraceBuffer* pTraceBuffer;
  unsigned int sensorIndex;
  CTraceBuffer::Point p;
  float point = 0.0f; // Prevent compiler whine

  if (m_numTrace) {
    p.m_timestamp = timestamp;
    p.m_bDiscontinuity = false;
    for (unsigned int i = 0; i < m_numTrace; i++) {
      pTraceBuffer = m_ppTraceBuffers[i];
      sensorIndex = pTraceBuffer->getSensorIndex();
      pSensorInfo = &devInfo.m_pSensors[sensorIndex];
      ASSERT(pSensorInfo->m_bKnownDataType);
      switch (pSensorInfo->m_info.DataType) {
      case ADMXRC3_DATA_BOOL:
        point = pSensorInfo->m_current.Bool ? 1.0f : 0.0f;
        break;

      case ADMXRC3_DATA_DOUBLE:
        point = (float)pSensorInfo->m_current.Double * pSensorInfo->m_exponent;
        break;

      case ADMXRC3_DATA_INT32:
        point = (float)pSensorInfo->m_current.Int32 * pSensorInfo->m_exponent;
        break;

      case ADMXRC3_DATA_UINT32:
        point = (float)pSensorInfo->m_current.UInt32 * pSensorInfo->m_exponent;
        break;

      default:
        ASSERT(FALSE);
        break;
      }
      p.m_value = point;
      pTraceBuffer->addPoint(p);
    }
    calcYScaling(m_scaling, m_ppTraceBuffers, m_layout, m_numTrace);
    needRedraw();
  }
}

void
CScopeWnd::updatePeriodChanged() {
  if (m_numTrace) {
#if CLEAR_TRACES_ON_PERIOD_CHANGE
    for (unsigned int i = 0; i < m_numTrace; i++) {
      CTraceBuffer* pTraceBuffer = m_ppTraceBuffers[i];

      pTraceBuffer->clear();
    }
    calcYScaling(m_scaling, m_ppTraceBuffers, m_layout, m_numTrace);
#endif
    needRedraw();
  }
}

// -------------------------------------------------------------------------
// End of IScopeWnd interface
// -------------------------------------------------------------------------

CScopeWnd::CScopeWnd(
  IGraphWnd* pParent) : m_cursor(Gdk::CROSSHAIR) {
  m_pParent = pParent;
  m_ppTraceBuffers = NULL;
  m_numTrace = 0;
  m_bMouseOver = FALSE;
  m_minimum = 1.0e20f;
  m_maximum = -1.0e20f;

  signal_expose_event().connect(sigc::mem_fun(*this, &CScopeWnd::onExpose), false);
  signal_enter_notify_event().connect(sigc::mem_fun(*this, &CScopeWnd::onMouseEnter), false);
  signal_leave_notify_event().connect(sigc::mem_fun(*this, &CScopeWnd::onMouseLeave), false);
  signal_motion_notify_event().connect(sigc::mem_fun(*this, &CScopeWnd::onMouseMove), false);
  signal_realize().connect(sigc::mem_fun(*this, &CScopeWnd::onRealize), false);
  signal_size_allocate().connect(sigc::mem_fun(*this, &CScopeWnd::onSizeAllocate), false);
  add_events(Gdk::POINTER_MOTION_MASK | Gdk::ENTER_NOTIFY_MASK | Gdk::LEAVE_NOTIFY_MASK);
}

CScopeWnd::~CScopeWnd() {
  if (NULL != m_ppTraceBuffers) {
    for (unsigned int i = 0; i < m_numTrace; i++) {
      delete m_ppTraceBuffers[i];
      m_ppTraceBuffers[i] = NULL;
    }
    delete[] m_ppTraceBuffers;
    m_ppTraceBuffers = NULL;
  }
}

void
CScopeWnd::calcLayout(
  Layout& layout) {
  Gtk::Allocation a = get_allocation();
  int w = a.get_width(), h = a.get_height();

  layout.m_width = w;
  layout.m_height = h;
  layout.m_axisYTop = 15;
  layout.m_axisXYBottom = h - 25;
  layout.m_axisXRight = w - 20;
  layout.m_axisXYLeft = 110;
  layout.m_axisYHeight = m_layout.m_axisXYBottom - m_layout.m_axisYTop;
  layout.m_axisXWidth = m_layout.m_axisXRight - m_layout.m_axisXYLeft;
}

void
CScopeWnd::calcYScaling(
  Scaling& /* scaling */,
  const CTraceBufferConstPtr* m_pTraceData,
  const Layout& /* layout */,
  unsigned int numTrace) {
  float min = 1.0e20f, max = -1.0e20f;
  float delta, tmp;
  const CTraceBuffer* pTraceBuffer;

  if (m_minimum < min) {
    min = m_minimum;
  }
  if (m_maximum > max) {
    max = m_maximum;
  }

  assert(numTrace > 0);
  m_scaling.m_bYValid = FALSE;
  for (unsigned int i = 0; i < numTrace; i++) {
    pTraceBuffer = m_pTraceData[i];
    if (pTraceBuffer->getLength()) {
      tmp = pTraceBuffer->getMin();
      if (tmp < min) {
        min = tmp;
      }
      tmp = pTraceBuffer->getMax();
      if (tmp > max) {
        max = tmp;
      }
      tmp = pTraceBuffer->getMean();
      if (tmp < min) {
        min = tmp;
      }
      if (tmp > max) {
        max = tmp;
      }
    }
  }

  delta = max - min;
  if (delta < 1.0e-9f) {
    return;
  }

  int maxYDivisions = (m_layout.m_axisYHeight + 1) / 20;
  float fl = log10f(delta / maxYDivisions);
  int exp = (int)floorf(fl);
  float tmp1 = fl - (float)exp;
  float tmp2 = powf(10, tmp1);
  float division;
  if (tmp2 < 2.001f) {
    division = 2.0f;
  } else if (tmp2 < 5.001f) {
    division = 5.0f;
  } else {
    division = 10.0f;
  }
  division = division * powf(10, (float)exp);
  int qMin = (int)floorf(min / division);
  int qMax = (int)ceilf(max / division - 0.01f);
  m_scaling.m_numYDivision = qMax - qMin;
  m_scaling.m_yDivision = division;
  m_scaling.m_yMin = (float)qMin * division;
  m_scaling.m_yMax = (float)qMax * division;
  m_scaling.m_yScaling = 1.0f / (m_scaling.m_yMax - m_scaling.m_yMin);
  m_scaling.m_bYValid = TRUE;
}

bool
CScopeWnd::xToTimestamp(
  int mouseX,
  timestamp_t t0,
  unsigned int updatePeriod,
  timestamp_t& timestamp)
{
  if (mouseX > m_layout.m_axisXYLeft && mouseX <= m_layout.m_axisXRight) {
    timestamp = t0 - (m_layout.m_axisXRight - mouseX) * (updatePeriod >> 1);
    return true;
  } else {
    return false;
  }
}

bool
CScopeWnd::timestampToX(
  timestamp_t timestamp,
  timestamp_t t0,
  unsigned int updatePeriod,
  int& x)
{
  float fx = (float)(t0 - timestamp) / (updatePeriod >> 1) + 0.5f;
  x = m_layout.m_axisXRight - (int)floorf(fx);
  return (x > m_layout.m_axisXYLeft && x <= m_layout.m_axisXRight);
}

bool
CScopeWnd::valueToY(
  float value,
  int& y)
{
  float fy = (value - m_scaling.m_yMin) * m_scaling.m_yScaling * (float)m_layout.m_axisYHeight;
  y = m_layout.m_axisXYBottom - (int)floorf(fy + 0.5f);
  return true;
}

void
CScopeWnd::drawAxes(
  Cairo::RefPtr<Cairo::Context> cr) {
  cr->set_source_rgb(1.0, 1.0, 1.0);
  cr->set_antialias(Cairo::ANTIALIAS_NONE);
  cr->set_line_width(1.0);
  cr->move_to((double)m_layout.m_axisXYLeft, (double)m_layout.m_axisYTop);
  cr->line_to((double)m_layout.m_axisXYLeft, (double)m_layout.m_axisXYBottom);
  cr->line_to((double)m_layout.m_axisXRight, (double)m_layout.m_axisXYBottom);
  cr->stroke();
}

void
CScopeWnd::drawGuides(
  Cairo::RefPtr<Cairo::Context> cr,
  const DeviceInfo& devInfo) {
  SensorInfo* pSensorInfo;
  CTraceBuffer* pTraceBuffer;
  unsigned int sensorIndex;
  float point;
  int y;
  timestamp_t timestamp;

  cr->set_source_rgb(0.25, 0.25, 0.25);
  cr->set_antialias(Cairo::ANTIALIAS_NONE);
  cr->set_line_width(1.0);
  cr->move_to((double)m_mousePos.m_x, 0.0);
  cr->line_to((double)m_mousePos.m_x, (double)m_layout.m_height);
  cr->move_to(0.0, (double)m_mousePos.m_y);
  cr->line_to((double)m_layout.m_width, (double)m_mousePos.m_y);
  cr->stroke();

  if (m_scaling.m_bYValid && xToTimestamp(m_mousePos.m_x, m_pParent->getCurrentTimestamp(), m_pParent->getParent()->getUpdatePeriod(), timestamp)) {
    for (unsigned int i = 0; i < m_numTrace; i++) {
      pTraceBuffer = m_ppTraceBuffers[i];
      sensorIndex = pTraceBuffer->getSensorIndex();
      pSensorInfo = &devInfo.m_pSensors[sensorIndex];
      if (pTraceBuffer->getValueAt(timestamp, point)) {
        valueToY(point, y);
        cr->set_source_rgb((double)RED(pSensorInfo->m_traceColor) / 255.0, (double)GREEN(pSensorInfo->m_traceColor) / 255.0, (double)BLUE(pSensorInfo->m_traceColor) / 255.0);
        cr->set_line_width(0.0);
        cr->move_to((double)m_layout.m_axisXYLeft - 3.0, (double)y);
        cr->line_to((double)m_layout.m_axisXYLeft - 10.0, (double)y - 3.0);
        cr->line_to((double)m_layout.m_axisXYLeft - 10.0, (double)y + 3.0);
        cr->fill();
      }
    }
  }
}

void
CScopeWnd::drawMessage(
  Cairo::RefPtr<Cairo::Context> cr,
  const DeviceInfo& devInfo)
{
  static const TCHAR* pMessageNoDevice[] = {
    "No device selected", "", "Please select a device.", NULL
  };
  static const TCHAR* pMessageNoSensor[] = {
    "No sensors selected", "", "Use the toolbar buttons to", "add sensors to the 'scope.", NULL
  };
  static const TCHAR* pMessageNoSensorAvail[] = {
    "No sensors available", "", "Please select a device", "with some sensors.", NULL
  };
  static const TCHAR** pMessage;

  Glib::RefPtr<Gdk::Window> window = get_window();
  if (window) {
    if (devInfo.m_hDevice == ADMXRC3_HANDLE_INVALID_VALUE) {
      pMessage = pMessageNoDevice;
    } else {
      if (devInfo.m_numAvailableSensor == 0) {
        pMessage = pMessageNoSensorAvail;
      } else {
        pMessage = pMessageNoSensor;
      }
    }

    cr->set_font_size(16.0);
    cr->set_source_rgb(1.0, 1.0, 1.0);
    unsigned int i, n = 0;
    double tot = 0.0, y;
    Cairo::TextExtents ex;
    while (pMessage[n] != NULL) {
      if (n == 0 || pMessage[n][0] != '\0') {
        cr->get_text_extents(pMessage[n], ex);
      }
      tot += ex.height + 6.0;
      n++;
    }
    tot -= 6.0;
    y = ((double)m_layout.m_height - tot) * 0.5 - ex.y_bearing;
    for (i = 0; i < n; i++) {
      if (i == 0 || pMessage[i][0] != '\0') {
        cr->get_text_extents(pMessage[i], ex);
        cr->move_to(((double)m_layout.m_width - ex.width) * 0.5, (double)y);
        cr->show_text(pMessage[i]);
      }
      y += ex.height + 6.0;
    }
  }
}

void
CScopeWnd::drawMinimalTicks(
  Cairo::RefPtr<Cairo::Context> cr) {
  cr->set_source_rgb(1.0, 1.0, 1.0);
  cr->set_antialias(Cairo::ANTIALIAS_NONE);
  cr->set_line_width(1.0);
  cr->move_to((double)m_layout.m_axisXYLeft, (double)m_layout.m_axisYTop);
  cr->line_to((double)m_layout.m_axisXYLeft - 5.0, (double)m_layout.m_axisYTop);
  cr->move_to((double)m_layout.m_axisXYLeft, (double)m_layout.m_axisXYBottom);
  cr->line_to((double)m_layout.m_axisXYLeft - 5.0, (double)m_layout.m_axisXYBottom);
  cr->stroke();
}

void
CScopeWnd::drawTicksAndLabels(
  Cairo::RefPtr<Cairo::Context> cr,
  const SensorInfo* pSensorInfo)
{
  static const TCHAR* pOrderStr;
  static const TCHAR* pUnitStr;
  std::ostringstream ss;
  Cairo::TextExtents ex;
  float fy, fdy, val, logVal, factor;
  int y;

  cr->set_source_rgb(1.0, 1.0, 1.0);
  cr->set_antialias(Cairo::ANTIALIAS_NONE);
  cr->set_line_width(1.0);
  cr->set_font_size(14.0);

  // Figure out unit mulitplier string ("k", "n", "G" etc.)
  val = m_scaling.m_yMax;
  logVal = log10f(fabs(m_scaling.m_yMax));
  int index = ((int)floorf(logVal + 0.5f) - 2) / 3 + 6;
  if (index < 0) {
    factor = 1.0e18f;
    index = 0;
  } else if (index > 12) {
    factor = 1.0e-18f;
    index = 12;
  } else {
    factor = g_factor[index];
  }
  pOrderStr = g_prefixStr[index];

  // Get unit as a string
  pUnitStr = CAdmxrc3Help::getUnitString(pSensorInfo->m_info.Unit);
  if (NULL == pUnitStr) {
    pUnitStr = "?";
  }

  // Draw unit with multiplier near top of Y-axis
  ss << pOrderStr << pUnitStr;
  cr->get_text_extents(ss.str(), ex);
  cr->move_to((double)m_layout.m_axisXYLeft - ex.width - 50.0, (double)m_layout.m_axisYTop + ex.height);
  cr->show_text(ss.str());

  // Draw Y-axis ticks & labels
  ss.precision(1);
  ss << fixed;
  fy = (float)m_layout.m_axisYTop;
  fdy = (float)m_layout.m_axisYHeight / (float)m_scaling.m_numYDivision;
  for (int i = 0; i <= m_scaling.m_numYDivision; i++) {
    y = (int)floorf(fy + 0.5f);
    cr->move_to((double)m_layout.m_axisXYLeft, (double)y);
    cr->line_to((double)m_layout.m_axisXYLeft - 5.0, (double)y);
    ss.str("");
    ss << (val * factor);
    cr->get_text_extents(ss.str(), ex);
    cr->move_to((double)m_layout.m_axisXYLeft - 10.0 - ex.width, (double)y + ex.height * 0.5);
    cr->show_text(ss.str());
    fy += fdy;
    val -= m_scaling.m_yDivision;
  }
  cr->stroke();

  // Draw time unit right of X-axis
  cr->get_text_extents("s", ex);
  cr->move_to((double)m_layout.m_axisXRight + 5.0 - ex.width * 0.5, (double)m_layout.m_axisXYBottom + ex.height * 0.5);
  cr->show_text("s");

  // Draw X-axis labels and ticks
  timestamp_t t, t0 = m_pParent->getCurrentTimestamp();
  if (t0 != TIMESTAMP_INVALID) {
    const unsigned int minorPerMajor = 5;
    unsigned int updatePeriod = m_pParent->getParent()->getUpdatePeriod();
    unsigned int tickPeriodMajor = updatePeriod * 50;
    unsigned int tickPeriodMinor = tickPeriodMajor / minorPerMajor;
    unsigned int n;
    int x;
    ASSERT(t0 >= 0);
    t = t0 - (t0 % tickPeriodMinor);
    n = (int)(t % tickPeriodMajor) / tickPeriodMinor;
    timestampToX(t, t0, updatePeriod, x);
    while (x > m_layout.m_axisXYLeft - 4 && t >= 0) {
      if (0 == n) {
        ss.str("");
        ss << t / 1000;
        cr->move_to((double)x, (double)m_layout.m_axisXYBottom);
        cr->line_to((double)x, (double)m_layout.m_axisXYBottom + 5.0);
        cr->get_text_extents(ss.str(), ex);
        cr->move_to((double)x - ex.width * 0.5 - ex.x_bearing, (double)m_layout.m_axisXYBottom + 6.0f + ex.height);
        cr->show_text(ss.str());
        n = minorPerMajor;
      } else {
        cr->move_to((double)x, (double)m_layout.m_axisXYBottom);
        cr->line_to((double)x, (double)m_layout.m_axisXYBottom + 3.0);
      }
      n--;
      t -= tickPeriodMinor;
      timestampToX(t, t0, updatePeriod, x);
    }
    cr->stroke();
  }
}

void
CScopeWnd::drawTrace(
  Cairo::RefPtr<Cairo::Context> cr,
  SensorInfo* pSensorInfo,
  CTraceBuffer* pTraceBuffer,
  timestamp_t t0)
{
  std::vector<double> dashes = std::vector<double>(2, 2.0);
  unsigned int length;
  unsigned int updatePeriod = m_pParent->getParent()->getUpdatePeriod();
  int xLow, xHigh;
  timestamp_t timestampLow = TIMESTAMP_INVALID, timestampHigh = TIMESTAMP_INVALID; /* init. to prevent compiler whine */
  unsigned int indexLow = 0, indexHigh = 0; /* init. to prevent compiler whine */
  timestamp_t minTimestamp = TIMESTAMP_INVALID, maxTimestamp = TIMESTAMP_INVALID; /* init. to prevent compiler whine */
  float valueLow = 0.0f, valueHigh = 0.0f; /* init. to prevent compiler whine */
  unsigned int index;
  int x1, x2;
  int y1, y2;

  /* Skip this trace if it doesn't yet contain enough data points */
  length = pTraceBuffer->getLength();
  if (length < 2) {
    return;
  }

  cr->set_source_rgb((double)RED(pSensorInfo->m_traceColor) / 255.0, (double)GREEN(pSensorInfo->m_traceColor) / 255.0, (double)BLUE(pSensorInfo->m_traceColor) / 255.0);
  cr->set_antialias(Cairo::ANTIALIAS_NONE);
  cr->set_line_width(1.0);

  xLow = m_layout.m_axisXYLeft + 1;
  xHigh = m_layout.m_axisXRight;
  xToTimestamp(xLow, t0, updatePeriod, timestampLow);
  xToTimestamp(xHigh, t0, updatePeriod, timestampHigh);
  ASSERT(xLow < xHigh);
  minTimestamp = pTraceBuffer->getMinTime();
  maxTimestamp = pTraceBuffer->getMaxTime();
  ASSERT(minTimestamp < maxTimestamp);
  if (minTimestamp <= timestampHigh && maxTimestamp >= timestampLow) {
    /* Dataset is partially or fully inside visible X-axis interval */
    if (minTimestamp < timestampLow) {
      pTraceBuffer->findTimeLE(timestampLow, indexLow);
      pTraceBuffer->getValueAt(timestampLow, valueLow);
    } else {
      indexLow = 0;
      timestampToX(minTimestamp, t0, updatePeriod, xLow);
      valueLow = pTraceBuffer->getValue(indexLow);
    }
    if (maxTimestamp > timestampHigh) {
      pTraceBuffer->findTimeGE(timestampLow, indexHigh);
      pTraceBuffer->getValueAt(timestampLow, valueHigh);
    } else {
      indexHigh = length - 1;
      timestampToX(maxTimestamp, t0, updatePeriod, xHigh);
      valueHigh = pTraceBuffer->getValue(indexHigh);
    }
    x1 = xLow;
    index = indexLow;
    valueToY(valueLow, y1);
    cr->move_to((double)x1, (double)y1);
    while (index < indexHigh) {
      index++;
      const CTraceBuffer::Point& p = pTraceBuffer->getPoint(index);
      if (index == indexHigh) {
        x2 = xHigh;
        valueToY(valueHigh, y2);
      } else {
        timestampToX(p.m_timestamp, t0, updatePeriod, x2);
        valueToY(p.m_value, y2);
      }
      cr->line_to((double)x2, (double)y2);
      x1 = x2;
    }
    cr->stroke();
  } else {
    /* Dataset is wholly outside visible x-axis interval */
  }

  // Draw min and max
  cr->set_dash(dashes, 0.0f);
  valueToY(pTraceBuffer->getMin(), y1);
  cr->move_to((double)m_layout.m_axisXYLeft + 1, (double)y1);
  cr->line_to((double)m_layout.m_axisXRight, (double)y1);
  valueToY(pTraceBuffer->getMax(), y2);
  cr->move_to((double)m_layout.m_axisXYLeft + 1, (double)y2);
  cr->line_to((double)m_layout.m_axisXRight, (double)y2);
  cr->stroke();
  cr->unset_dash();
}

void
CScopeWnd::needRedraw() {
  Glib::RefPtr<Gdk::Window> window = get_window();
  if (window) {
    Gdk::Rectangle r(0, 0, m_layout.m_width, m_layout.m_height);
    window->invalidate_rect(r, false);
  }
}

bool
CScopeWnd::onExpose(GdkEventExpose* pEvent) {
  ISysMonWnd* pMainWnd = m_pParent->getParent();
  const DeviceInfo& devInfo = pMainWnd->getDeviceInfo();
  Glib::RefPtr<Gdk::Window> window = get_window();
  timestamp_t t0, minShow = TIMESTAMP_INVALID; /* init. to prevent compiler whine */

  if (window) {
    Cairo::RefPtr<Cairo::Context> cr = window->create_cairo_context();

    if (pEvent) {
      cr->rectangle(pEvent->area.x, pEvent->area.y, pEvent->area.width, pEvent->area.height);
      cr->clip();
    }

    cr->set_source_rgb(0.0, 0.0, 0.0);
    cr->paint();

    if (m_numTrace) {
      if (m_bMouseOver) {
        drawGuides(cr, devInfo);
      }
      drawAxes(cr);
      if (!m_scaling.m_bYValid) {
        // Don't have any valid traces to draw, so just draw ticks at top and bottom of Y-axis
        drawMinimalTicks(cr);
      } else {
        SensorInfo* pSensorInfo;
        CTraceBuffer* pTraceBuffer;
        unsigned int sensorIndex;

        // Draw the Y-axis ticks and labels
        pTraceBuffer = m_ppTraceBuffers[0];
        sensorIndex = pTraceBuffer->getSensorIndex();
        pSensorInfo = &devInfo.m_pSensors[sensorIndex];
        drawTicksAndLabels(cr, pSensorInfo);

        // Draw the traces
        t0 = m_pParent->getCurrentTimestamp();
        xToTimestamp(m_layout.m_axisXYLeft + 1, t0, m_pParent->getParent()->getUpdatePeriod(), minShow);
        for (unsigned int i = 0; i < m_numTrace; i++) {
          pTraceBuffer = m_ppTraceBuffers[i];
          sensorIndex = pTraceBuffer->getSensorIndex();
          pSensorInfo = &devInfo.m_pSensors[sensorIndex];
          drawTrace(cr, pSensorInfo, pTraceBuffer, t0);
          if (i == 0 && pTraceBuffer->getLength() > 0) {
            timestamp_t min = pTraceBuffer->getMinTime();
            timestamp_t max = pTraceBuffer->getMaxTime();
            m_pParent->viewChanged(max - min, minShow - min, t0 - min);
          }
        }
      }
    } else {
      drawMessage(cr, devInfo);
    }
  }

  return false;
}

bool
CScopeWnd::onMouseEnter(
  GdkEventCrossing* pEvent)
{
  m_bMouseOver = TRUE;
  if (m_numTrace) {
    needRedraw();
  }

  return true;
}

bool
CScopeWnd::onMouseLeave(
  GdkEventCrossing* pEvent)
{
  m_bMouseOver = FALSE;
  if (m_numTrace) {
    needRedraw();
  }

  return true;
}

bool
CScopeWnd::onMouseMove(
  GdkEventMotion* pEvent)
{
  m_mousePos.m_x = pEvent->x;
  m_mousePos.m_y = pEvent->y;
  if (m_numTrace) {
    needRedraw();
  }

  return true;
}

void
CScopeWnd::onRealize()
{
  // Change cursor to a crosshair
  Glib::RefPtr <Gdk::Window> ref_window = get_window();
  ref_window->set_cursor(m_cursor);
}

void
CScopeWnd::onSizeAllocate(
  Gtk::Allocation& allocation)
{
  // Determine new layout of axes etc.
  calcLayout(m_layout);
  if (m_numTrace) {
    // Resize trace buffers for new window size
    for (unsigned int i = 0; i < m_numTrace; i++) {
      m_ppTraceBuffers[i]->setSize(m_layout.m_width);
    }
    // Determine Y-axis scale, tick spacing etc.
    calcYScaling(m_scaling, m_ppTraceBuffers, m_layout, m_numTrace);
  }
}
