/*
** vpd.c - utility to display / modify Vital Product Data for an ADM-XRC
**         series card.
**
** (C) Copyright 2009-2011 Alpha Data
**
** NOTE: for fill and write commands to succeed, VPD write-protection must be
**       disabled as detailed in the ADB3 Driver for VxWorks release notes.
*/

#if defined(_WIN32)

/* Windows */
# include <windows.h>
# include <tchar.h>

# if !defined(_MSC_VER)
/* Non-MSVC - won't have secure strings library */
#   define _tcsnset_s(buffer, max, c, n) _tcsnset(buffer, c, n)
#   define _stprintf_s _sntprintf
#   define _stscanf_s _stscanf
# endif

#else

/* Linux or VxWorks */
# define _T(x) x
# define _gettc(x) getc(x)
# define _istprint(x) isprint(x)
# define _istspace(x) isspace(x)
# define _stprintf_s snprintf
# define _stscanf_s sscanf
# define _tcscpy_s(x, y, z) strcpy(x, z)
# if defined(__VXWORKS__) || defined(__vxworks)
#   define _tcsicmp(pStr1, pStr2) ourStrCaseCmp(pStr1, pStr2)
# else
#   define _tcsicmp(pStr1, pStr2) strcasecmp(pStr1, pStr2)
# endif
# define _tcslen(x) strlen(x)
# define _tcsnset_s(buffer, max, c, n) strset_s(buffer, max, c, n)
# define _tmain main
# define _tprintf printf
# define _tscanf(buffer, ...) scanf(buffer, __VA_ARGS__)
# define FALSE (0)
# define TRUE (1)

#endif

#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>

#include <admxrc3.h>

#include <sdk_common.h>
#include <platform_endian.h>

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define MIN(a, b) ((a) < (b) ? (a) : (b))

typedef enum _CommandType {
  CommandFill,
  CommandRead,
  CommandWrite
} CommandType;

typedef enum _TokenStateType {
  TokenStateInToken,
  TokenStateInSpace
} TokenStateType;

typedef enum _EscapeStateType {
  EscapeStateNormal,
  EscapeStateInEscape
} EscapeStateType;

typedef struct _State {
  TokenStateType state;
  EscapeStateType escapeState;
  unsigned int tokenPos;
  TCHAR tokenBuf[128];
} State;

const State g_initialState = {
  TokenStateInSpace,
  EscapeStateNormal,
  0
};

#if !defined(_WIN32)
static void
strset_s(
  TCHAR* pBuffer,
  size_t max,
  TCHAR c,
  size_t n)
{
  if (!max) {
    return;
  }
  max--;
  while (max) {
    *pBuffer++ = c;
    max--;
  }
  *pBuffer++ = _T('\0');
}
#endif

#if defined(__VXWORKS__) || defined(__vxworks)
static int
ourStrCaseCmp(
  const char* pS1,
  const char* pS2)
{
  char c1, c2;

  while (1) {
    c1 = *pS1;
    c2 = *pS2;
    if (c1 == '\0' || tolower(c1) != tolower(c2)) {
      break;
    }
    pS1++;
    pS2++;
  }
  return tolower(c1) - tolower(c2);
}
#endif

static char
toAscii(
  TCHAR tc)
{
  if (tc >= _T('0') && tc <= _T('9')) {
    return tc - _T('0') + '0';
  } else if (tc >= _T('A') && tc <= _T('Z')) {
    return tc - _T('A') + 'A';
  } else if (tc >= _T('a') && tc <= _T('z')) {
    return tc - _T('a') + 'a';
  } else {
    switch (tc) {
      case _T(' '):
        return ' ';
      case _T('!'):
        return '!';
      case _T('"'):
        return '"';
      case _T('#'):
        return '#';
      case _T('$'):
        return '$';
      case _T('%'):
        return '%';
      case _T('&'):
        return '&';
      case _T('\''):
        return '\'';
      case _T('('):
        return '(';
      case _T(')'):
        return ')';
      case _T('*'):
        return '*';
      case _T('+'):
        return '+';
      case _T(','):
        return ',';
      case _T('-'):
        return '-';
      case _T('.'):
        return '.';
      case _T('/'):
        return '/';
      case _T(':'):
        return ':';
      case _T(';'):
        return ';';
      case _T('<'):
        return '<';
      case _T('='):
        return '=';
      case _T('>'):
        return '>';
      case _T('?'):
        return '?';
      case _T('@'):
        return '@';
      case _T('['):
        return '[';
      case _T('\\'):
        return '\\';
      case _T(']'):
        return ']';
      case _T('^'):
        return '^';
      case _T('_'):
        return '_';
      case _T('`'):
        return '`';
      case _T('{'):
        return '{';
      case _T('|'):
        return '|';
      case _T('}'):
        return '}';
      case _T('~'):
        return '~';
      default:
        return '?';
    }
  }
}

static ADMXRC3_STATUS
dumpVpd(
  ADMXRC3_HANDLE hCard,
  unsigned int indent,
  unsigned int nColumn,
  unsigned int radix,
  bool_t bAlign,
  uint32_t offset,
  size_t n)
{
  ADMXRC3_STATUS status;
  uint32_t address, limit, addressCol, m;
  unsigned int addressLo, columnMask;
  uint32_t tmpOffs;
  uint8_t* p8;
  TCHAR line[257];
  TCHAR* pPos;
  uint8_t columnBuf[256];
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;
  unsigned int i, j, k, chunk;
  unsigned int rowBytes;
  int nwritten;
  size_t remaining;
  
  if (n > (size_t)0xffffffffU) {
    _tprintf(_T("*** dumpVpd: invalid value '0x%lx' for n\n"), (unsigned long)n);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  if (indent > 128) {
    _tprintf(_T("*** dumpVpd: invalid value '%lu' for indent\n"), (unsigned long)indent);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  switch (nColumn) {
  case 1:
  case 2:
  case 4:
  case 8:
  case 16:
  case 32:
    break;
  default:
    _tprintf(_T("*** dumpVpd: invalid value '%lu' for nColumn\n"), (unsigned long)nColumn);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  if (radix != 1 && radix != 2 && radix != 4 && radix != 8) {
    _tprintf(_T("*** dumpVpd: invalid value '%lu' for radix\n"), (unsigned long)radix);
    return ADMXRC3_UNEXPECTED_ERROR;
  }
  
  rowBytes = nColumn * radix;

  /* Make sure everything is aligned */
  columnMask = rowBytes - 1U;
  address = offset;
  limit = address + (uint32_t)n;
  if (bAlign) {
    addressCol = address & ~(uint32_t)columnMask;
    addressLo = ((unsigned int)address) & columnMask;
  } else {
    addressCol = address;
    addressLo = 0;
  }
  
  m = limit - address;
  tmpOffs = offset;
  
  pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent);
  pPos += MIN(indent, remaining);
  remaining -= MIN(indent, remaining);
  nwritten =  _stprintf_s(pPos, remaining, _T("Dump of VPD at 0x%lx + %lu(0x%lx) bytes:\n"),
    (unsigned long)offset, (unsigned long)n, (unsigned long)n);
  if (nwritten > 0) {
    pPos += nwritten;
    remaining -= nwritten;
  }
  line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);
  
  pPos = line;
  line[0] = _T('\0');
  remaining = ARRAY_LENGTH(line);
  _tcsnset_s(pPos, remaining, _T(' '), indent + 12);
  pPos += MIN(indent + 12, remaining);
  remaining -= MIN(indent + 12, remaining);
  for (i = 0; i < rowBytes; i += radix) {
    switch (radix) {
    case 0x1U:
      nwritten = _stprintf_s(pPos, remaining, _T("%02lx "), (unsigned long)i);
      break;
      
    case 0x2U:
      nwritten = _stprintf_s(pPos, remaining, _T("  %02lx "), (unsigned long)i);
      break;
      
    case 0x4U:
      nwritten = _stprintf_s(pPos, remaining, _T("      %02lx "), (unsigned long)i);
      break;
      
    case 0x8U:
      nwritten = _stprintf_s(pPos, remaining, _T("               %02lx "), (unsigned long)i);
      break;
    }
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
  }
   if (remaining) {
	  *pPos++ = _T('\n');
    remaining--;
  }
  if (remaining) {
	  *pPos++ = _T('\0');
    remaining--;
  }
 line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
  _tprintf(_T("%s"), line);
  
  while (m) {
    chunk = rowBytes - addressLo;
    if (chunk > m) {
      chunk = (unsigned int)m;
    }
    
    memset(columnBuf, 0, rowBytes);
    status = ADMXRC3_ReadVPD(hCard, 0, tmpOffs, chunk, columnBuf + addressLo);
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Failed to read VPD chunk at offset 0x%lx, size 0x%lx: %s\n"),
        (unsigned long)tmpOffs, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }

    p8 = (uint8_t*)columnBuf;
    
    pPos = line;
    line[0] = _T('\0');
    remaining = ARRAY_LENGTH(line);
    _tcsnset_s(pPos, remaining, _T(' '), indent);
    pPos += MIN(indent, remaining);
    remaining -= MIN(indent, remaining);
    nwritten = _stprintf_s(pPos, remaining, _T("0x%08lx: "), (unsigned long)(uint32_t)addressCol);
    if (nwritten > 0) {
      pPos += nwritten;
      remaining -= nwritten;
    }
    
    j = addressLo + chunk;
    for (i = 0; i < rowBytes; i += radix) {
      switch (radix) {
      case 0x1U:
        val8 = *p8++;
        if (i >= addressLo && i < j) {
          nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)val8);
          if (nwritten > 0) {
            pPos += nwritten;
            remaining -= nwritten;
          }
        } else {
          if (remaining) {
            *pPos++ = _T('.');
            remaining--;
          }
          if (remaining) {
            *pPos++ = _T('.');
            remaining--;
          }
        }
        break;
        
      case 0x2U:
        val16 = (((uint16_t)p8[0]) << 0U) | (((uint16_t)p8[1]) << 8U);
        p8 += 2;
        for (k = 2; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
            nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)(uint8_t)(val16 >> (k << 3)));
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
          }
          if (k == 3) {
            if (remaining) {
              *pPos++ = _T('_');
              remaining--;
            }
          }
        }
        break;
        
      case 0x4U:
        val32 = (((uint32_t)p8[0]) <<  0U) | (((uint32_t)p8[1]) <<  8U) | (((uint32_t)p8[2]) << 16U) | (((uint32_t)p8[3]) << 24U);
        p8 += 4;
        for (k = 4; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
            nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)(uint8_t)(val32 >> (k << 3)));
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
          }
        }
        break;
        
      case 0x8U:
        val64  = (((uint64_t)p8[0]) <<  0U) | (((uint64_t)p8[1]) <<  8U) | (((uint64_t)p8[2]) << 16U) | (((uint64_t)p8[3]) << 24U);
        val64 |= (((uint64_t)p8[4]) << 32U) | (((uint64_t)p8[5]) << 40U) | (((uint64_t)p8[6]) << 48U) | (((uint64_t)p8[7]) << 56U);
        p8 += 8;
        for (k = 8; k != 0;) {
          k--;
          if (i + k >= addressLo && i + k < j) {
            nwritten = _stprintf_s(pPos, remaining, _T("%02lx"), (unsigned long)(uint8_t)(val64 >> (k << 3)));
            if (nwritten > 0) {
              pPos += nwritten;
              remaining -= nwritten;
            }
          } else {
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
            if (remaining) {
              *pPos++ = _T('.');
              remaining--;
            }
          }
          if (k == 4) {
            if (remaining) {
              *pPos++ = _T('_');
              remaining--;
            }
          }
        }
        break;
      }
      if (remaining) {
        *pPos++ = _T(' ');
        remaining--;
      }
    }
    
    p8 = (uint8_t*)columnBuf;
    for (i = 0; i < rowBytes; i++) {
      char c = (char)*p8++;
      
      if (remaining) {
        if (_istprint(c)) {
          *pPos++ = c;
        } else {
          *pPos++ = _T('.');
        }
        remaining--;
      }
    }
    if (remaining) {
      *pPos++ = _T('\n');
      remaining--;
    }
    if (remaining) {
      *pPos++ = _T('\0');
      remaining--;
    }
    line[ARRAY_LENGTH(line) - 1] = _T('\0'); /* Ensure line is nul-terminated */
    _tprintf(_T("%s"), line);
    addressLo = 0;
    addressCol += rowBytes;
    m -= chunk;
    tmpOffs += chunk;
  }
  
  return ADMXRC3_SUCCESS;
}

static ADMXRC3_STATUS
fillVpd(
  ADMXRC3_HANDLE hCard,
  uint32_t offset,
  unsigned int count,
  unsigned int width,
  bool_t bInterpretString,
  bool_t bInterpretHex,
  const TCHAR* pString)
{
  ADMXRC3_STATUS status;
  uint8_t buf[65536];
  const unsigned int bufSize = ARRAY_LENGTH(buf);
  unsigned int bufPos, strPos;
  size_t length;
  unsigned int chunk;
  unsigned long long tmp;
  uint8_t* p8;
  uint16_t* p16;
  uint32_t* p32;
  uint64_t* p64;
  uint8_t val8;
  uint16_t val16;
  uint32_t val32;
  uint64_t val64;

  length = _tcslen(pString);
  if (bInterpretString) {
    strPos = 0;
    while (count) {
      chunk = (count > bufSize) ? bufSize : count;
      bufPos = 0;
      while (bufPos < bufSize && bufPos < chunk) {
        buf[bufPos] = toAscii(pString[strPos]);
        bufPos++;
        strPos++;
        if (strPos >= length) {
          strPos = 0;
        }
      }
      status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
      if (ADMXRC3_SUCCESS != status) {
        _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
          (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
        return status;
      }
      offset += chunk;
      count -= chunk;
    }
  } else {
    if (length >= 2 && pString[0] == _T('0') && (pString[1] == _T('x') || pString[1] == _T('X'))) {
      _stscanf_s(pString + 2, _T("%llx"), &tmp);
    } else {
      if (bInterpretHex) {
        _stscanf_s(pString, _T("%llx"), &tmp);
      } else {
        _stscanf_s(pString, _T("%llu"), &tmp);
      }
    }
    val64 = (uint64_t)tmp;
    switch (width) {
      case 1U:
        val8 = (uint8_t)val64;
        p8 = (uint8_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p8++ = val8;
            bufPos++;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      case 2U:
        val16 = (uint16_t)val64;
        p16 = (uint16_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p16++ = cpu_to_le16(val16);
            bufPos += 2;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      case 4U:
        val32 = (uint32_t)val64;
        p32 = (uint32_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p32++ = cpu_to_le32(val32);
            bufPos += 4;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      case 8U:
        p64 = (uint64_t*)buf;
        while (count) {
          chunk = (count > bufSize) ? bufSize : count;
          bufPos = 0;
          while (bufPos < chunk) {
            *p64++ = cpu_to_le64(val64);
            bufPos += 8;
          }
          status = ADMXRC3_WriteVPD(hCard, 0, offset, chunk, buf);
          if (ADMXRC3_SUCCESS != status) {
            _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
              (unsigned long)offset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
            return status;
          }
          offset += chunk;
          count -= chunk;
        }
        break;

      default:
        assert(FALSE);
        break;
    }
  }

  return status;
}

static ADMXRC3_STATUS
writeVpdToken(
  ADMXRC3_HANDLE hCard,
  uint32_t* pOffset,
  unsigned int* pCount,
  unsigned int width,
  bool_t bInterpretString,
  bool_t bInterpretHex,
  const TCHAR* pString,
  size_t length)
{
  ADMXRC3_STATUS status;
  uint8_t buf[65536];
  const unsigned int bufSize = ARRAY_LENGTH(buf);
  uint64_t val64;
  uint32_t val32;
  uint16_t val16;
  uint8_t val8;
  unsigned long long tmp;
  unsigned int i, chunk;

  if (bInterpretString) {
    while (*pCount) {
      if (length > *pCount) {
        chunk = *pCount;
      } else {
        chunk = (unsigned int)length; /* Cast is safe due to above check */
      }
      if (chunk > bufSize) {
        chunk = bufSize;
      }
      for (i = 0; i < chunk; i++) {
        buf[i] = toAscii(*pString++);
      }
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, chunk, buf);
      if (ADMXRC3_SUCCESS != status) {
        _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
          (unsigned long)*pOffset, (unsigned long)chunk, ADMXRC3_GetStatusString(status, TRUE));
        return status;
      }
      *pOffset += chunk;
      *pCount -= chunk;
    }
  } else {
    if (_tcslen(pString) >= 2 && pString[0] == _T('0') && (pString[1] == _T('x') || pString[1] == _T('X'))) {
      _stscanf_s(pString + 2, _T("%llx"), &tmp);
    } else {
      if (bInterpretHex) {
        _stscanf_s(pString, _T("%llx"), &tmp);
      } else {
        _stscanf_s(pString, _T("%llu"), &tmp);
      }
    }
    val64 = (uint64_t)tmp;
    switch (width) {
    case 1U:
      val8 = (uint8_t)val64;
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val8);
      break;

    case 2U:
      val16 = cpu_to_le16((uint16_t)val64);
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val16);
      break;

    case 4U:
      val32 = cpu_to_le32((uint32_t)val64);
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val32);
      break;

    case 8U:
      val64 = cpu_to_le64(val64);
      status = ADMXRC3_WriteVPD(hCard, 0, *pOffset, width, &val64);
      break;

    default:
      assert(FALSE);
      break;
    }
    if (ADMXRC3_SUCCESS != status) {
      _tprintf(_T("*** Failed to write VPD at offset 0x%lx, size 0x%lx: %s\n"),
        (unsigned long)*pOffset, (unsigned long)width, ADMXRC3_GetStatusString(status, TRUE));
      return status;
    }
    *pOffset += width;
    *pCount -= width;
  }

  return status;
}

static ADMXRC3_STATUS
writeVpd(
  ADMXRC3_HANDLE hCard,
  uint32_t *pOffset,
  unsigned int* pCount,
  State* pState,
  unsigned int width,
  bool_t bInterpretString,
  bool_t bInterpretHex,
  TCHAR* pString)
{
  ADMXRC3_STATUS status = ADMXRC3_SUCCESS;
  TCHAR c;

  while (*pCount) {
    c = *pString++;
    if (c == _T('\0')) {
      break;
    }
    if (bInterpretString) {
      switch (pState->state) {
      case TokenStateInToken:
        if (c == _T('\n')) {
          status = writeVpdToken(hCard, pOffset, pCount, width, bInterpretString, bInterpretHex, pState->tokenBuf, pState->tokenPos);
          if (ADMXRC3_SUCCESS != status) {
            return status;
          }
          pState->state = TokenStateInSpace;
          pState->tokenPos = 0;
        } else {
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      case TokenStateInSpace:
        if (c != _T('\n')) {
          pState->state = TokenStateInToken;
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      default:
        assert(FALSE);
        break;
      }
    } else {
      switch (pState->state) {
      case TokenStateInToken:
        if (_istspace(c)) {
          status = writeVpdToken(hCard, pOffset, pCount, width, bInterpretString, bInterpretHex, pState->tokenBuf, pState->tokenPos);
          if (ADMXRC3_SUCCESS != status) {
            return status;
          }
          pState->state = TokenStateInSpace;
          pState->tokenPos = 0;
        } else {
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      case TokenStateInSpace:
        if (!_istspace(c)) {
          pState->state = TokenStateInToken;
          pState->tokenBuf[pState->tokenPos] = c;
          pState->tokenPos++;
        }
        break;

      default:
        assert(FALSE);
        break;
      }
    }
  }

  return status;
}

int
_tmain(
  int argc,
  TCHAR* argv[])
{
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_STATUS status;
  Option options[3];
  Arguments arguments = g_initArguments;
  OptionType paramType[4];
  TCHAR* paramInfo[4];
  TCHAR* paramMeaning[4];
  CommandType command;
  bool_t bInterpretString, bInterpretHex;
  unsigned int n, width, index, serial;
  uint32_t offset;
  TCHAR* pCommand;
  int i, ret = 0;
#if defined(_UNICODE)
  wint_t c;
#else
  int c;
#endif
  TCHAR str[128];
  State state = g_initialState;

  options[0].key = _T("index");
  options[0].type = OptionUInt;
  options[0].def.uintVal = 0;
  options[0].meaning = _T("index of card to use");
  options[1].key = _T("sn");
  options[1].type = OptionUInt;
  options[1].def.uintVal = 0;
  options[1].meaning = _T("serial number of card to use");
	options[2].key = _T("hex");
	options[2].type = OptionBoolean;
  options[2].def.booleanVal = FALSE;
	options[2].meaning = _T("TRUE => interpret values as being hexadecimal");
  arguments.nOption = 3;
  arguments.option = options;
  paramType[0] = OptionString;
  paramInfo[0] = _T("command");
  paramMeaning[0] = _T("fb, fw, rd, fq, fs, rb, rw, rd, rq, wb, ww, wd, wq or ws");
  paramType[1] = OptionUInt;
  paramInfo[1] = _T("offset");
  paramMeaning[1] = _T("Offset in VPD region");
  paramType[2] = OptionUInt;
  paramInfo[2] = _T("length");
  paramMeaning[2] = _T("Number of bytes to read or write");
  paramType[3] = OptionString;
  paramInfo[3] = _T("data ...");
  paramMeaning[3] = _T("Data items (f / w commands)");
  arguments.nParamType = 4;
  arguments.minNParam = 3;
  arguments.paramType = paramType;
  arguments.paramInfo = paramInfo;
  arguments.paramMeaning = paramMeaning;
  
  if (sampleParseCommandLine(_T("vpd"), argc, argv, &arguments) < 0) {
    ret = -1;
    goto done;
  }
  index = options[0].value.uintVal;
  serial = options[1].value.uintVal;
  bInterpretHex = options[2].value.booleanVal;
  pCommand = arguments.param[0].pStringVal;
  offset = arguments.param[1].uintVal;

  bInterpretString = FALSE;
  if (_tcsicmp(pCommand, _T("rb")) == 0) {
    width = 1;
    command = CommandRead;
  } else if (_tcsicmp(pCommand, _T("rw")) == 0) {
    width = 2;
    command = CommandRead;
  } else if (_tcsicmp(pCommand, _T("rd")) == 0) {
    width = 4;
    command = CommandRead;
  } else if (_tcsicmp(pCommand, _T("rq")) == 0) {
    width = 8;
    command = CommandRead;
  } else if (_tcsicmp(pCommand, _T("wb")) == 0) {
    width = 1;
    command = CommandWrite;
  } else if (_tcsicmp(pCommand, _T("ww")) == 0) {
    width = 2;
    command = CommandWrite;
  } else if (_tcsicmp(pCommand, _T("wd")) == 0) {
    width = 4;
    command = CommandWrite;
  } else if (_tcsicmp(pCommand, _T("wq")) == 0) {
    width = 8;
    command = CommandWrite;
  } else if (_tcsicmp(pCommand, _T("ws")) == 0) {
    width = 1;
    bInterpretString = TRUE;
    command = CommandWrite;
  } else if (_tcsicmp(pCommand, _T("fb")) == 0) {
    width = 1;
    command = CommandFill;
  } else if (_tcsicmp(pCommand, _T("fw")) == 0) {
    width = 2;
    command = CommandFill;
  } else if (_tcsicmp(pCommand, _T("fd")) == 0) {
    width = 4;
    command = CommandFill;
  } else if (_tcsicmp(pCommand, _T("fq")) == 0) {
    width = 8;
    command = CommandFill;
  } else if (_tcsicmp(pCommand, _T("fs")) == 0) {
    width = 1;
    bInterpretString = TRUE;
    command = CommandFill;
  } else {
    _tprintf(_T("*** Invalid command: %s\n"), pCommand);
    ret = -1;
    goto done;
  }

  if (arguments.nParam < 3) {
    n = width;
  } else {
    n = arguments.param[2].uintVal;
  }
  if (n % width) {
    _tprintf(_T("+++ Warning: byte count %u(0x%x) is not an integer multiple of width (%u)\n"), n, n, width);
    n &= ~(width - 1U);
    _tprintf(_T("+++          byte count adjusted to %u(0x%x)\n"), n, n);
  }

  status = sampleOpenCard(index, options[0].specified, serial, options[1].specified, FALSE, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    _tprintf(_T("*** Failed to open card with index %lu: %s\n"),
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = -1;
    goto done;
  }

  switch (command) {
  case CommandFill:
    if (arguments.nParam < 4) {
      _tprintf(_T("*** Too few arguments for fill command.\n"));
      ret = -1;
      goto done;
    }
    status = fillVpd(hCard, offset, n, width, bInterpretString, bInterpretHex, arguments.param[3].pStringVal);
    if (ADMXRC3_SUCCESS != status) {
      ret = -1;
      goto done;
    }
    break;

  case CommandRead:
    status = dumpVpd(hCard, 0, 16 / width, width, TRUE, offset, n);
    if (ADMXRC3_SUCCESS != status) {
      ret = -1;
      goto done;
    }
    break;

  case CommandWrite:
    for (i = 3; i < arguments.nParam && 0 != n; i++) {
      _tprintf(_T("0x%08lx: %s\n"), (unsigned long)offset, arguments.param[i].pStringVal);
      status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, arguments.param[i].pStringVal);
      if (status != ADMXRC3_SUCCESS) {
        ret = -1;
        goto done;
      }
      /* Complete current token (if necessary) */ 
      status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, _T("\n"));
      if (status != ADMXRC3_SUCCESS) {
        ret = -1;
        goto done;
      }
    }
    if (0 != n) {
      _tprintf(_T("0x%08lx: "), (unsigned long)offset);
      fflush(stdout);
    }
    while (n) {
      c = _gettc(stdin);
#if defined(_UNICODE)
      if (c == WEOF) {
#else
      if (c == EOF) {
#endif
        /* Complete last token (if necessary) */
        status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, _T("\n"));
        if (status != ADMXRC3_SUCCESS) {
          ret = -1;
          goto done;
        }
        break;
      }
      str[0] = c;
      str[1] = _T('\0');
      status = writeVpd(hCard, &offset, &n, &state, width, bInterpretString, bInterpretHex, str);
      if (status != ADMXRC3_SUCCESS) {
        ret = -1;
        goto done;
      }
      if (c == _T('\n') && 0 != n) {
        _tprintf(_T("0x%08lx: "), (unsigned long)offset);
        fflush(stdout);
      }
    }
    break;

  default:
    assert(FALSE);
    break;
  }

done:
  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  sampleCleanupCommandLine(&arguments);
  return ret;
}
