/*
** itest.c - example demonstrating interrupt wait functionality in
**           the ADM-XRC API via VxWorks semaphores.
**
** (C) Copyright 2010 Alpha Data
*/

#include <vxWorks.h>
#include <taskLib.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <semaphore.h>

#include <admxrc3.h>

#include "sdk_common.h"
#include "platform_endian.h"

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

/*
** FPGA register addresses on local bus. These are also the offsets from the
** base of space 0, since space 0 begins at absolute local bus address 0x0.
*/
#define REGADDR_ISET  (0xC0U)
#define REGADDR_ISTAT (0xC4U)
#define REGADDR_IMASK (0xC8U)
#define REGADDR_IARM  (0xCCU)

typedef struct _ThreadContext {
  ADMXRC3_HANDLE hCard;
  SEM_ID semId;
  BOOL bTerminate;
  unsigned long signalled;
} ThreadContext;

static int
interruptThread(
  void* pArgument)
{
  ThreadContext* pContext = (ThreadContext*)pArgument;
  ADMXRC3_STATUS status;
  STATUS result;
  uint32_t istat;
  uint32_t dummy = 0; /* value doesn't matter, but avoid compiler whine */

  printf("Interrupt thread started\n");
	
  taskSafe();

  while (!pContext->bTerminate) {
    result = semTake(pContext->semId, WAIT_FOREVER);
    if (result != OK) {
      printf("*** Interrupt thread failed to wait for semaphore (errno %u(0x%x))\n", errno, errno);
      break;
    }

    if (!pContext->bTerminate) {
      pContext->signalled++;

      /* Get the pending FPGA interrupts */
      status = ADMXRC3_Read(pContext->hCard,	0, 0, REGADDR_ISTAT, sizeof(istat), &istat);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to read ISTAT register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      /* Clear the pending FPGA interrupts */
      status = ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_ISTAT, sizeof(istat), &istat);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to write ISTAT register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      /* Rearm the 'FINTI_L' line driven by the FPGA by forcing it high for 1 cycle */
      status = ADMXRC3_Write(pContext->hCard, 0, 0, REGADDR_IARM, sizeof(dummy), &dummy);
      if (ADMXRC3_SUCCESS != status) {
        printf("*** Interrupt thread failed to write IARM register: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
    }
  }

  taskUnsafe();

  return 0;
}

STATUS
admxrc3ITest(
  unsigned int index)
{
  const double clockFreq = 33.0e6;
  STATUS ret = OK;
  ADMXRC3_STATUS status;
  ADMXRC3_CARD_INFO cardInfo;
  unsigned long numInterruptGenerated;
  char filename[256];
  uint32_t val32;
  ThreadContext threadContext = {
    ADMXRC3_HANDLE_INVALID_VALUE, /* hCard */
    NULL, /* semId */
    FALSE, /* bTerminate */
    0 /* signalled */
  };
  bool_t bSemRegistered = FALSE;
  int interruptThreadId = ERROR;
  bool_t bInterruptThreadRunning = FALSE;
  bool_t bShouldConfigure;
  int ourPriority, ch;

  sampleTaskNice();

  status = ADMXRC3_Open(index, &threadContext.hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(threadContext.hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }
	
  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(threadContext.hCard, ADMXRC3_CLOCK_LCLK, 0, clockFreq, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1fMHz: %s\n", clockFreq / 1.0e6f, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  }

  status = sampleGetFpgaFilename(threadContext.hCard, 0, "uber", "uber", ARRAY_LENGTH(filename), filename, &bShouldConfigure);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to determine bitstream filename: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (bShouldConfigure) {
    status = ADMXRC3_ConfigureFromFile(threadContext.hCard, 0, 0, filename);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to load the bitstream '%s': %s\n", filename, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  } else{
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }
  
  /* Create a binary semaphore */
  threadContext.semId = semBCreate(0, 0);
  if (threadContext.semId == NULL) {
    printf("*** Failed to create a VxWorks binary semaphore\n");
    ret = ERROR;
    goto done;
  }

  /*
  ** Register the semaphore with the card, so that when an FPGA interrupt occurs,
  ** the semaphore will be given.
  */
  status = ADMXRC3_RegisterVxwSem(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.semId);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to register Win32 event for FPGA interrupt: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }
  bSemRegistered = TRUE;

  /* Start the thread that waits for interrupts. */
  threadContext.signalled = 0;
  threadContext.bTerminate = FALSE;
  taskPriorityGet(taskIdSelf(), &ourPriority);
  interruptThreadId = taskSpawn("admxrc3InterruptThread", ourPriority - 1, 0, 0x1000, interruptThread, (int)&threadContext, 0, 0, 0, 0, 0, 0, 0, 0, 0);
  if (interruptThreadId == ERROR) {
    printf("*** Failed to start interrupt thread\n");
    ret = ERROR;
    goto done;
  }
  bInterruptThreadRunning = TRUE;

  /* Enable interrupts in the FPGA */
  val32 = cpu_to_le32(0x0U);
  status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_IMASK, sizeof(val32), &val32);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Write to FPGA_IMASK register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  /*
  ** Allow the user to generate interrupts by pressing the enter key.
  ** Ctrl-Z or entering 'q' / 'Q' exits this loop.
  */
  numInterruptGenerated = 0;
  do {
    printf("Enter 'q' to quit, or anything else to generate an interrupt:\n");
    ch = getchar();
    if (EOF == ch || 'q' == tolower(ch)) {
      ch = EOF;
    } else {
      val32 = cpu_to_le32(0x1U);
      status = ADMXRC3_Write(threadContext.hCard, 0, 0, REGADDR_ISET, sizeof(val32), &val32);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Write to ISET register failed: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }

      numInterruptGenerated++;
    }
  } while (EOF != ch);

  printf("Generated %lu interrupts\n", numInterruptGenerated);
  printf("Interrupt thread saw %lu interrupt(s)\n", threadContext.signalled);

done:
  if (bInterruptThreadRunning) {
    /* Force the interrupt thread to terminate */
    threadContext.bTerminate = TRUE;
    semGive(threadContext.semId);
    /* Wait for the interrupt thread to terminate */
    taskDelete(interruptThreadId);
    interruptThreadId = ERROR;
  }

  if (threadContext.semId != NULL) {
    if (bSemRegistered) {
      /*
      ** Unregister the semaphore, so that it will no longer be signalled
      ** on an FPGA interrupt.
      */
      status = ADMXRC3_UnregisterVxwSem(threadContext.hCard, ADMXRC3_EVENT_FPGAINTERRUPT(0), threadContext.semId);
      if (status != ADMXRC3_SUCCESS) {
        printf("*** Failed to unregister interrupt semaphore: %s\n", ADMXRC3_GetStatusString(status, TRUE));
      }
    }
    semDelete(threadContext.semId);
    threadContext.semId = NULL;
  }

  if (threadContext.hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(threadContext.hCard);
    threadContext.hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
