/*
** simple.c - Trivial example showing how to read and write registers
**            in the FPGA.
**
** (C) Copyright 2010 Alpha Data
*/

#include <vxWorks.h>
#include <stdio.h>
#include <string.h>

#include <admxrc3.h>
#include "sdk_common.h"
#include "platform_endian.h"

#ifndef FALSE
# define FALSE (0)
#endif
#ifndef TRUE
# define TRUE (1)
#endif

#define ARRAY_LENGTH(x) (sizeof(x) / sizeof((x)[0]))

#define FLAG_USEUBER (0x10U) /* Use UBER bitstream instead of SIMPLE */

STATUS
admxrc3Simple(
  unsigned int index, /* Device index */
  unsigned int flags) /* See above flags */
{
  STATUS ret = OK;
  int err;
  ADMXRC3_STATUS status;
  ADMXRC3_HANDLE hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  ADMXRC3_CARD_INFO cardInfo;
  volatile uint32_t* pFpgaSpace = NULL;
  unsigned long dataOut, dataIn;
  char filename[256];
  bool_t bConfigured = FALSE, bUseUber, bShouldConfigure;

  sampleTaskNice();

  bUseUber = (flags & FLAG_USEUBER) ? TRUE : FALSE;

  status = ADMXRC3_Open(index, &hCard);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to open card with index %lu: %s\n",
      (unsigned long)index, ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  status = ADMXRC3_GetCardInfo(hCard, &cardInfo);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to get card info: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  /* Get the address of FPGA space */
  status = ADMXRC3_MapWindow(hCard, 0, 0x0, 0x10, (void*)&pFpgaSpace);
  if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to map FPGA space: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }

  if (cardInfo.Model < 0x100 && cardInfo.NumClockGen > 0) {
    /* Set local bus clock to nominal frequency */
    status = ADMXRC3_SetClockFrequency(hCard, ADMXRC3_CLOCK_LCLK, 0, 33.0e6, NULL);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to set LCLK to %.1f MHz: %s\n", 33.0, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
  }

  if (bUseUber) {
    status = sampleGetFpgaFilename(hCard, 0, "uber", "uber", ARRAY_LENGTH(filename), filename, &bShouldConfigure);
  } else {
    status = sampleGetFpgaFilename(hCard, 0, "simple", "simple", ARRAY_LENGTH(filename), filename, &bShouldConfigure);
  }
	if (status != ADMXRC3_SUCCESS) {
    printf("*** Failed to determine bitstream filename: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    ret = ERROR;
    goto done;
  }
  
  if (bShouldConfigure) {
    printf("Configuring target FPGA %u with '%s'...\n", 0, filename);
    status = ADMXRC3_ConfigureFromFile(hCard, 0, 0, filename);
    if (status != ADMXRC3_SUCCESS) {
      printf("*** Failed to configure target FPGA with '%s': %s\n", filename, ADMXRC3_GetStatusString(status, TRUE));
      ret = ERROR;
      goto done;
    }
    bConfigured = TRUE;
  } else {
    /* Running on a model such as the ADM-XRC-6T-ADV8, which doesn't have an FPGA that is configurable via the ADMXRC3 API */
  }
  
  printf("=========================\n");
  printf("Enter values for I/O\n");
  printf("(CTRL-D / CTRL-Z to exit)\n");
  printf("=========================\n");

  do {
    int c;

    err = scanf("%lX", &dataOut);
    if (err >= 1) {
      pFpgaSpace[0] = cpu_to_le32((uint32_t)dataOut);
      dataIn = (unsigned long)pFpgaSpace[0];
      dataIn = le32_to_cpu(dataIn);
      printf("OUT = 0x%08lx, IN = 0x%08lx\n", (unsigned long)dataOut, (unsigned long)dataIn);
    } else {
      if (0 == err) {
        /* Skip to next line */
        do {
          c = getchar();
          if (EOF == c) {
            goto done;
          }
        } while (c != '\n');
      }
    }
  } while (EOF != err);

done:
  if (NULL != pFpgaSpace) {
    status = ADMXRC3_UnmapWindow(hCard, (void*)pFpgaSpace);
    if (status != ADMXRC3_SUCCESS) {
      printf("+++ Failed to unmap FPGA space: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    }
    pFpgaSpace = NULL;
  }

  if (bConfigured) {
    status = ADMXRC3_Unconfigure(hCard, 0, 0);
    if (status != ADMXRC3_SUCCESS) {
      printf("+++ Failed to unconfigure target FPGA: %s\n", ADMXRC3_GetStatusString(status, TRUE));
    }
    bConfigured = FALSE;
  }

  if (hCard != ADMXRC3_HANDLE_INVALID_VALUE) {
    ADMXRC3_Close(hCard);
    hCard = ADMXRC3_HANDLE_INVALID_VALUE;
  }

  return ret;
}
