--==============================================================================
--
--     File Name: adb3_ocp_comp_pkg.vhd
--       Version: 0.3
--          Date: 25 Jul 2011
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package adb3_ocp_comp_pkg
--
--   Description: Package of OCP components used in target FPGAs using ADB3.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, TW,  25 Jan 2011: Initial version
--                     - Created by extracting certain declarations from
--                       adb3_target_pkg.
--                0.2, KDC, 04 Jul 2011:
--                     - Added adb3_ocp_simple_bus_if_nb component
--                0.3, KDC, 25 Jul 2011:
--                     - Moved and renamed functions from adb3_target_pkg:
--                       adb3_ocp_base, adb3_ocp_mask, adb3_ocp_off, and
--                       adb3_ocp_mask_width
--
--==============================================================================

--
-- Exported functions:
--
--   adb3_ocp_base
--   adb3_ocp_mask
--   adb3_ocp_off
--   adb3_ocp_mask_width
--
--
-- Exported components:
--
--   adb3_ocp_split_b
--   adb3_ocp_split_nb
--   adb3_ocp_simple_bus_if
--   adb3_ocp_simple_bus_if_nb
--   adb3_ocp_mux_b
--   adb3_ocp_mux_nb
--   adb3_ocp_cross_ck_dom
--   adb3_ocp_retime_nb
--   adb3_ocp_ocp2ddr3_nb
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

package adb3_ocp_comp is

  -- Define adb3_ocp_split_(n)b types
  type adb3_ocp_addr_range_t is record
    addr_base : adb3_ocp_addr_s;
    addr_mask : adb3_ocp_addr_s;
  end record;

  type adb3_ocp_addr_range_table_t is array (natural range <>) of adb3_ocp_addr_range_t;

  function adb3_ocp_base(
    constant x : in std_logic_vector;
    constant w : in natural range 1 to ADB3_OCP_ADDR_WIDTH)
  return adb3_ocp_addr_s;

  function adb3_ocp_mask(
    constant x : in std_logic_vector;
    constant w : in natural range 1 to ADB3_OCP_ADDR_WIDTH)
  return adb3_ocp_addr_s;

  function adb3_ocp_off(
    constant x : in natural)
  return adb3_ocp_addr_s;

  function adb3_ocp_mask_width(
    constant x : in adb3_ocp_addr_s)
  return natural;


  component adb3_ocp_simple_bus_if
  generic(
    addr_width   : natural range 1 to ADB3_OCP_ADDR_WIDTH;
    data_width   : natural range 8 to ADB3_OCP_DATA_WIDTH;
    read_latency : natural := 1);
  port(
    -- OCP interface
    ocp_rst : in    std_logic;
    ocp_clk : in    std_logic;
    ocp_m2s : in    adb3_ocp_m2sT;
    ocp_s2m : out   adb3_ocp_s2mT;
    -- Simple bus interface
    d       : out   std_logic_vector(data_width-1 downto 0);
    q       : in    std_logic_vector(data_width-1 downto 0);
    a       : out   std_logic_vector(addr_width-1 downto 0);
    w       : out   std_logic;
    r       : out   std_logic;
    we      : out   std_logic_vector((data_width/8)-1 downto 0));
  end component;

  component adb3_ocp_simple_bus_if_nb
  generic(
    addr_width   : natural range 1 to ADB3_OCP_ADDR_WIDTH;
    data_width   : natural range 8 to ADB3_OCP_DATA_WIDTH;
    read_latency : natural := 1);
  port(
    -- OCP interface
    ocp_rst    : in    std_logic;
    ocp_clk    : in    std_logic;
    ocp_m2s    : in    adb3_ocp_m2sT;
    ocp_s2m    : out   adb3_ocp_s2mT;
    -- Bus interface
    d          : out   std_logic_vector(data_width-1 downto 0);
    q          : in    std_logic_vector(data_width-1 downto 0);
    a          : out   std_logic_vector(addr_width-1 downto 0);
    w          : out   std_logic;
    r          : out   std_logic;
    we         : out   std_logic_vector((data_width/8)-1 downto 0);
    -- Debug
    debug_info : out   std_logic_vector(31 downto 0));
  end component;

  component adb3_ocp_split_b
  generic(
    addr_range_table : adb3_ocp_addr_range_table_t;
    retime_slave     : boolean := false;
    retime_masters   : boolean := false);
  port(
    ocp_rst     : in    std_logic;
    ocp_clk     : in    std_logic;
    -- OCP primary port
    slave_m2s   : in    adb3_ocp_m2sT;
    slave_s2m   : out   adb3_ocp_s2mT;
    -- OCP secondary ports
    masters_m2s : out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1);
    masters_s2m : in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1);
    -- Debug
    debug_info  : out   std_logic_vector(31 downto 0));
  end component;

  component adb3_ocp_split_nb
  generic(
    addr_range_table : adb3_ocp_addr_range_table_t;
    error_data       : adb3_ocp_data_s := X"DEADC0DEDEADC0DEDEADC0DEDEADC0DE");
  port(
    ocp_rst     : in    std_logic;
    ocp_clk     : in    std_logic;
    -- OCP primary port
    slave_m2s   : in    adb3_ocp_m2sT;
    slave_s2m   : out   adb3_ocp_s2mT;
    -- OCP secondary ports
    masters_m2s : out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1);
    masters_s2m : in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1);
    -- Debug
    debug_info  : out   std_logic_vector(31 downto 0));
  end component;

  component adb3_ocp_mux_b
  generic(
    mux_inputs : natural := 2);
  port(
    ocp_rst    : in    std_logic;
    ocp_clk    : in    std_logic;
    -- OCP primary ports
    slaves_m2s : in    adb3_ocp_m2sT_array(0 to mux_inputs-1);
    slaves_s2m : out   adb3_ocp_s2mT_array(0 to mux_inputs-1);
    -- OCP secondary port
    master_m2s : out   adb3_ocp_m2sT;
    master_s2m : in    adb3_ocp_s2mT;
    -- Debug
    debug_info : out   std_logic_vector(31 downto 0));
  end component;

  component adb3_ocp_mux_nb
  generic(
    mux_inputs    : natural := 2);
  port(
    ocp_rst    : in    std_logic;
    ocp_clk    : in    std_logic;
    -- OCP primary ports
    slaves_m2s : in    adb3_ocp_m2sT_array(0 to mux_inputs-1);
    slaves_s2m : out   adb3_ocp_s2mT_array(0 to mux_inputs-1);
    -- OCP secondary port
    master_m2s : out   adb3_ocp_m2sT;
    master_s2m : in    adb3_ocp_s2mT;
    -- Debug
    debug_info : out   std_logic_vector(31 downto 0));
  end component;

  component adb3_ocp_cross_clk_dom
  port(
    -- OCP primary port
    slave_rst  : in    std_logic;
    slave_clk  : in    std_logic;
    slave_m2s  : in    adb3_ocp_m2sT;
    slave_s2m  : out   adb3_ocp_s2mT;
    -- OCP secondary port
    master_rst : in    std_logic;
    master_clk : in    std_logic;
    master_m2s : out   adb3_ocp_m2sT;
    master_s2m : in    adb3_ocp_s2mT);
  end component;

  component adb3_ocp_retime_nb
  port(
    ocp_rst    : in    std_logic;
    ocp_clk    : in    std_logic;
    -- OCP primary port
    slave_m2s  : in    adb3_ocp_m2sT;
    slave_s2m  : out   adb3_ocp_s2mT;
    -- OCP secondary port
    master_m2s : out   adb3_ocp_m2sT;
    master_s2m : in    adb3_ocp_s2mT);
  end component;

  component adb3_ocp_ocp2ddr3_nb
  generic(
    app_row_width  : natural;
    app_col_width  : natural;
    app_bank_width : natural;
    app_addr_width : natural);
  port(
    -- OCP interface
    ocp_rst           : in    std_logic;
    ocp_clk           : in    std_logic;
    ocp_m2s           : in    adb3_ocp_m2sT;
    ocp_s2m           : out   adb3_ocp_s2mT;
    -- MIG user interface
    mig_rst           : in    std_logic;
    mig_clk           : in    std_logic;
    phy_init_done     : in    std_logic;
    app_rdy           : in    std_logic;
    app_wdf_rdy       : in    std_logic;
    app_rd_data       : in    adb3_ocp_data_s;
    app_rd_data_valid : in    std_logic;
    app_en            : out   std_logic;
    app_cmd           : out   std_logic_vector(2 downto 0);
    app_addr          : out   std_logic_vector(app_addr_width-1 downto 0);
    app_sz            : out   std_logic;
    app_wdf_wren      : out   std_logic;
    app_wdf_data      : out   adb3_ocp_data_s;
    app_wdf_mask      : out   adb3_ocp_be_s;
    app_wdf_end       : out   std_logic;
    -- Debug
    debug_info        : out   std_logic_vector(31 downto 0));
  end component;

end;

package body adb3_ocp_comp is

  -- Convert base address x of width w to adb3_ocp_addr_s
  function adb3_ocp_base(
    constant x : in std_logic_vector;
    constant w : in natural range 1 to ADB3_OCP_ADDR_WIDTH)
  return adb3_ocp_addr_s is
    variable n : natural := x'length;                  -- Size of x in bits (min size of 1)
    variable m : std_logic_vector(n-1 downto 0) := x;  -- New vector of size n bits
  begin
    if (n > w) then
      assert not((OR_reduce(m(n-1 downto w))) = '1')
        report "(base_conv): Address larger than allowed width = " & integer'image(w)
        severity failure;
    end if;

    return std_logic_vector(RESIZE(unsigned(x),ADB3_OCP_ADDR_WIDTH));
  end;

  -- Convert mask address x of width w to adb3_ocp_addr_s
  function adb3_ocp_mask(
    constant x : in std_logic_vector;
    constant w : in natural range 1 to ADB3_OCP_ADDR_WIDTH)
  return adb3_ocp_addr_s is
    variable n : natural := x'length;                  -- Size of x in bits (min size of 1)
    variable m : std_logic_vector(n-1 downto 0) := x;  -- New vector of size n bits
    variable p : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto w) := (others => '1');
    variable q : adb3_ocp_addr_s;                      -- Define direction of vector
  begin
    if (n > w) then
      assert not((OR_reduce(m(n-1 downto w))) = '1')
        report "(mask_conv): Address larger than allowed width = " & integer'image(w)
        severity failure;
    end if;

    q := p & std_logic_vector(RESIZE(unsigned(x),w));
    return q;
  end;

  -- Convert address offset x to adb3_ocp_addr_s
  function adb3_ocp_off(
    constant x : in natural)
  return adb3_ocp_addr_s is
  begin
    return std_logic_vector(to_unsigned(x,ADB3_OCP_ADDR_WIDTH));
  end;

  -- Convert address mask width to natural
  function adb3_ocp_mask_width(
    constant x : in adb3_ocp_addr_s)
  return natural is
    variable n : natural;
  begin
    n := 0;
    for j in ADB3_OCP_ADDR_WIDTH-1 downto 0 loop
      if (x(j) = '0') then
        n := j;
      end if;
    end loop;

    return n;
  end;

end;