--==============================================================================
--
--     File Name: adb3_ocp_mux_nb.vhd
--       Version: 0.2
--          Date: 09 Feb 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP mux (non-blocking)
--
--   Description: Component giving multiple OCP slave ports control over a
--                single OCP master port
--
--   Limitations:
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 05 Nov 2010: Initial design
--                0.2, TW,  09 Feb 2011:
--                  - References adb3_ocp_comp package instead of
--                    adb3_target_pkg, and removed an unused signal.
--                  - Rewritten from scratch, removing retime blocks and
--                    including FIFOs to fully buffer command and data paths.
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

--
-- A synchronous FIFO used to buffer commands, command attributes, data, byte
-- enables etc. in 'adb3_ocp_mux_nb'.
--
-- 'ORDER' is log2 of the FIFO depth, e.g. ORDER = 4 => depth is 16 locations.
--
-- 'WIDTH' is the width of FIFO locations, in bits.
--
-- 'EMPTY_VAL' is the value driven on 'q(EMPTY_LO:EMPTY_HI)' when 'valid' is
-- not asserted; map both 'EMPTY_LO' and 'EMPTY_HI' to 0 if this feature is
-- not required.
--
entity adb3_ocp_mux_nb_fifo is
generic(
  ORDER       : in    natural range 2 to 6;
  WIDTH       : in    natural;
  EMPTY_VAL   : in    std_logic := '0';
  EMPTY_LO    : in    natural := 0;
  EMPTY_HI    : in    natural := 0);
port(
  rst         : in    std_logic;                             -- Asynchronous reset
  clk         : in    std_logic;                             -- Clock
  wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
  d           : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
  adv         : in    std_logic;                             -- '1' => present next item on 'q' to next item
  full        : out   std_logic;                             -- '1' => full
  not_full    : out   std_logic;                             -- '1' => not full
  valid       : out   std_logic;                             -- '1' => not empty
  q           : out   std_logic_vector(WIDTH - 1 downto 0)); -- Current item, qualified by 'valid'
end;

architecture syn of adb3_ocp_mux_nb_fifo is

  constant DEPTH : natural := 2**ORDER;
  constant MAX_LEVEL : natural := DEPTH - 1;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal wptr, rptr : std_logic_vector(ORDER - 1 downto 0);
  signal rn : std_logic_vector(ORDER - 1 downto 0);
  signal not_empty : std_logic;
  signal i_full, i_not_full : std_logic;

  signal qvalid : std_logic;
  signal qce : std_logic;
  signal rd : std_logic;

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  valid <= qvalid;
  full <= i_full;
  not_full <= i_not_full;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_not_full <= '1';
      i_full <= '0';
    elsif rising_edge(clk) then
      if wr = '1' then
        wptr <= wptr + 1;
      end if;

      if wr = '1' then
        assert CONV_INTEGER(wptr - rptr) < MAX_LEVEL severity failure;
        tmp := wptr - rptr + 1;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      else
        tmp := wptr - rptr;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr)) <= d;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= adv OR NOT qvalid;
  rd <= (adv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rptr <= rptr + 1;
      end if;
    end if;

    if rising_edge(clk) then
      if rd = '1' then
        assert CONV_INTEGER(wptr - rptr) > 0 severity failure;
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
    elsif rising_edge(clk) then
      if qce = '1' then
        qvalid <= not_empty;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : for i in 0 to WIDTH - 1 generate
    with_reset : if EMPTY_LO <= i AND i < EMPTY_HI generate
      U0 : process(rst, clk)
      begin
        if rst = '1' then
          q(i) <= EMPTY_VAL;
        elsif rising_edge(clk) then
          if qce = '1' then
            if not_empty = '1' then
              q(i) <= sel_ram(i);
            else
              q(i) <= EMPTY_VAL;
            end if;
          end if;
        end if;
      end process;
    end generate;

    without_reset : if NOT(EMPTY_LO <= i AND i < EMPTY_HI) generate
      U0 : process(clk)
      begin
        if rising_edge(clk) then
          if qce = '1' then
            q(i) <= sel_ram(i);
          end if;
        end if;
      end process;
    end generate;
  end generate;

end;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;

entity adb3_ocp_mux_nb is
generic(
  mux_inputs    : natural := 2);
port(
  ocp_rst    : in    std_logic;
  ocp_clk    : in    std_logic;
  -- OCP primary ports
  slaves_m2s : in    adb3_ocp_m2sT_array(0 to mux_inputs - 1);
  slaves_s2m : out   adb3_ocp_s2mT_array(0 to mux_inputs - 1);
  -- OCP secondary port
  master_m2s : out   adb3_ocp_m2sT;
  master_s2m : in    adb3_ocp_s2mT;
  -- Debug
  debug_info : out   std_logic_vector(31 downto 0));
end;

architecture syn of adb3_ocp_mux_nb is

  function priority(
    constant v : in std_logic_vector)
  return std_logic_vector is
    variable tmp : std_logic_vector(v'range);
  begin
    tmp := (others => '0');
    for i in v'range loop
      if i > 0  then
        if v'ascending then
          tmp(i) := v(i) AND NOT OR_reduce(v(0 to i - 1));
        else
          tmp(i) := v(i) AND NOT OR_reduce(v(i - 1 downto 0));
        end if;
      else
        tmp(i) := v(i);
      end if;
    end loop;
    return tmp;
  end;

  component adb3_ocp_mux_nb_fifo
  generic(
    ORDER       : in    natural range 1 to 6;
    WIDTH       : in    natural;
    EMPTY_VAL   : in    std_logic := '0';
    EMPTY_LO    : in    natural := 0;
    EMPTY_HI    : in    natural := 0);
  port(
    rst         : in    std_logic;
    clk         : in    std_logic;
    wr          : in    std_logic;
    d           : in    std_logic_vector(WIDTH - 1 downto 0);
    adv         : in    std_logic;
    full        : out   std_logic;
    not_full    : out   std_logic;
    valid       : out   std_logic;
    q           : out   std_logic_vector(WIDTH - 1 downto 0));
  end component;

  subtype port_1h_t is std_logic_vector(mux_inputs - 1 downto 0);

  -- One hot vector indicating which slave port's command was last accepted
  signal cmd_port_1h : port_1h_t;
  -- One hot vector indicating which slave port's command is accepted in the current cycle
  signal n_cmd_port_1h : port_1h_t;

  -- Slave command FIFOs
  constant SLV_CMD_ATTR_WIDTH : natural := ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;
  constant SLV_CMD_WIDTH : natural := ADB3_OCP_CMD_WIDTH + SLV_CMD_ATTR_WIDTH;
  type slv_cmd_array_t is array(natural range <>) of std_logic_vector(SLV_CMD_WIDTH - 1 downto 0);
  signal slv_cmd_d, slv_cmd_q : slv_cmd_array_t(0 to mux_inputs - 1);
  signal slv_cmd_wr : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_cmd_adv : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_cmd_not_full : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_cmd_valid : std_logic_vector(mux_inputs - 1 downto 0);
  type slv_cmd_attr_t is record
    Cmd : ocp_CmdT;
    Addr : adb3_ocp_addr_s;
    BurstLength : adb3_ocp_burst_s;
    Tag : adb3_ocp_tag_s;
  end record;
  type slv_cmd_attr_array_t is array(natural range <>) of slv_cmd_attr_t;
  signal slv_cmd_vec : slv_cmd_attr_array_t(0 to mux_inputs - 1);
  signal slv_cmd_sel : slv_cmd_attr_t;

  -- Master command FIFO
  constant MST_CMD_ATTR_WIDTH : natural := ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;
  constant MST_CMD_WIDTH : natural := ADB3_OCP_CMD_WIDTH + MST_CMD_ATTR_WIDTH;
  signal mst_cmd_d, mst_cmd_q : std_logic_vector(MST_CMD_WIDTH - 1 downto 0);
  signal mst_cmd_wr : std_logic;
  signal mst_cmd_adv : std_logic;
  signal mst_cmd_not_full : std_logic;
  signal mst_cmd_valid : std_logic;

  -- Slave write burst tracking FIFO
  constant WBT_WIDTH : natural := ADB3_OCP_BURST_WIDTH + mux_inputs;
  type wbt_attr_t is record
    blen : adb3_ocp_burst_s;
    port_1h : port_1h_t;
  end record;
  signal wbt_d, wbt_q : std_logic_vector(WBT_WIDTH - 1 downto 0);
  signal wbt_wr : std_logic;
  signal wbt_adv : std_logic;
  signal wbt_not_full : std_logic;
  signal wbt_valid : std_logic;
  signal wbt_q_rec : wbt_attr_t;

  -- Slave read burst tracking FIFO
  constant RBT_WIDTH : natural := ADB3_OCP_BURST_WIDTH + mux_inputs;
  type rbt_attr_t is record
    blen : adb3_ocp_burst_s;
    port_1h : port_1h_t;
  end record;
  signal rbt_d, rbt_q : std_logic_vector(RBT_WIDTH - 1 downto 0);
  signal rbt_wr : std_logic;
  signal rbt_adv : std_logic;
  signal rbt_not_full : std_logic;
  signal rbt_valid : std_logic;
  signal rbt_q_rec : rbt_attr_t;

  -- Write data transfer control
  type wdx_state_t is (wdx_s_idle, wdx_s_burst);
  signal wdx_state, n_wdx_state : wdx_state_t;
  signal wdx_blen, n_wdx_blen : adb3_ocp_burst_s;
  signal wdx_blast, n_wdx_blast : std_logic;
  signal wdx_port_1h, n_wdx_port_1h : port_1h_t;
  signal n_wdx_bstart : std_logic;

  -- Read data transfer control
  type rdx_state_t is (rdx_s_idle, rdx_s_burst);
  signal rdx_state, n_rdx_state : rdx_state_t;
  signal rdx_blen, n_rdx_blen : adb3_ocp_burst_s;
  signal rdx_blast, n_rdx_blast : std_logic;
  signal rdx_port_1h, n_rdx_port_1h : port_1h_t;
  signal n_rdx_bstart : std_logic;

  -- Slave write data FIFOs
  constant SLV_WDATA_WIDTH : natural := ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;
  type slv_wdata_d_array_t is array(natural range <>) of std_logic_vector(SLV_WDATA_WIDTH - 1 downto 0);
  signal slv_wdata_d, slv_wdata_q : slv_wdata_d_array_t(0 to mux_inputs - 1);
  signal slv_wdata_wr : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_wdata_adv : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_wdata_not_full : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_wdata_valid : std_logic_vector(mux_inputs - 1 downto 0);
  type slv_wdata_t is record
    Data : adb3_ocp_data_s;
    DataByteEn : adb3_ocp_be_s;
  end record;
  type slv_wdata_array_t is array(natural range <>) of slv_wdata_t;
  signal slv_wdata_vec : slv_wdata_array_t(0 to mux_inputs - 1);

  -- Master write data FIFO
  constant MST_WDATA_WIDTH : natural := ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;
  signal mst_wdata_d, mst_wdata_q : std_logic_vector(MST_WDATA_WIDTH - 1 downto 0);
  signal mst_wdata_wr : std_logic;
  signal mst_wdata_adv : std_logic;
  signal mst_wdata_not_full : std_logic;
  signal mst_wdata_valid : std_logic;

  -- Slave read data FIFOs
  constant SLV_RDATA_WIDTH : natural := ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH;
  type slv_rdata_d_array_t is array(natural range <>) of std_logic_vector(SLV_RDATA_WIDTH - 1 downto 0);
  signal slv_rdata_d, slv_rdata_q : slv_rdata_d_array_t(0 to mux_inputs - 1);
  signal slv_rdata_wr : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_rdata_adv : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_rdata_not_full : std_logic_vector(mux_inputs - 1 downto 0);
  signal slv_rdata_valid : std_logic_vector(mux_inputs - 1 downto 0);

  -- Master read data FIFO
  constant MST_RDATA_WIDTH : natural := ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH;
  signal mst_rdata_d, mst_rdata_q : std_logic_vector(MST_RDATA_WIDTH - 1 downto 0);
  signal mst_rdata_wr : std_logic;
  signal mst_rdata_adv : std_logic;
  signal mst_rdata_not_full : std_logic;
  signal mst_rdata_valid : std_logic;

begin

  debug_info <= (others => '0');

  -- ---------------------------------------------------------------
  -- Command path
  -- ---------------------------------------------------------------

  gen_cmd_port_1h : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      cmd_port_1h(0) <= '1';
      cmd_port_1h(mux_inputs - 1 downto 1) <= (others => '0');
    elsif rising_edge(ocp_clk) then
      if mst_cmd_wr = '1' then
        cmd_port_1h <= n_cmd_port_1h;
      end if;
    end if;
  end process;

  gen_n_cmd_port_1h : process(
    cmd_port_1h,
    slv_cmd_valid)
    variable requesting, winner, winner_rot, tmp : std_logic_vector(mux_inputs - 1 downto 0);
  begin
    tmp := (others => '0');
    for i in 0 to mux_inputs - 1 loop
      if cmd_port_1h(i) = '1' then
        for j in 0 to mux_inputs - 1 loop
          requesting(j) := slv_cmd_valid((i + 1 + j) mod mux_inputs);
        end loop;
        winner := priority(requesting);
        for j in 0 to mux_inputs - 1 loop
          winner_rot((i + 1 + j) mod mux_inputs) := winner(j);
        end loop;
        tmp := tmp OR winner_rot;
      end if;
    end loop;
    n_cmd_port_1h <= tmp;
  end process;

  slv_cmd_fifos : for i in 0 to mux_inputs - 1 generate
    slv_cmd_wr(i) <= '1' when slv_cmd_not_full(i) = '1' AND slaves_m2s(i).Cmd /= OCP_CMD_IDLE else '0';
    slv_cmd_d(i) <= slaves_m2s(i).Cmd & slaves_m2s(i).Tag & slaves_m2s(i).BurstLength & slaves_m2s(i).Addr;
    slv_cmd_adv(i) <= n_cmd_port_1h(i) AND mst_cmd_not_full AND rbt_not_full AND wbt_not_full;

    U0 : adb3_ocp_mux_nb_fifo
    generic map(
      ORDER => 3,
      WIDTH => SLV_CMD_WIDTH)
    port map(
      rst => ocp_rst,
      clk => ocp_clk,
      wr => slv_cmd_wr(i),
      d => slv_cmd_d(i),
      adv => slv_cmd_adv(i),
      full => open,
      not_full => slv_cmd_not_full(i),
      valid => slv_cmd_valid(i),
      q => slv_cmd_q(i));

    slv_cmd_vec(i) <= (Cmd => slv_cmd_q(i)(SLV_CMD_WIDTH - 1 downto SLV_CMD_ATTR_WIDTH),
                       Tag => slv_cmd_q(i)(SLV_CMD_ATTR_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH),
                       BurstLength => slv_cmd_q(i)(ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH),
                       Addr => slv_cmd_q(i)(ADB3_OCP_ADDR_WIDTH - 1 downto 0));

    slaves_s2m(i).CmdAccept <= slv_cmd_not_full(i);
  end generate;

  mst_cmd_wr <= OR_reduce(n_cmd_port_1h) AND mst_cmd_not_full AND rbt_not_full AND wbt_not_full;

  select_slv_cmd : process(n_cmd_port_1h, slv_cmd_vec)
    variable tmp : slv_cmd_attr_t;
  begin
    tmp := (Cmd => (others => '0'), Tag => (others => '0'), BurstLength => (others => '0'), Addr => (others => '0'));
    for i in 0 to mux_inputs - 1 loop
      if n_cmd_port_1h(i) = '1' then
        tmp.Cmd := tmp.Cmd OR slv_cmd_vec(i).Cmd;
        tmp.Tag := tmp.Tag OR slv_cmd_vec(i).Tag;
        tmp.BurstLength := tmp.BurstLength OR slv_cmd_vec(i).BurstLength;
        tmp.Addr := tmp.Addr OR slv_cmd_vec(i).Addr;
      end if;
    end loop;
    slv_cmd_sel <= tmp;
  end process;

  mst_cmd_d <= slv_cmd_sel.Cmd & slv_cmd_sel.Tag & slv_cmd_sel.BurstLength & slv_cmd_sel.Addr;
  mst_cmd_adv <= mst_cmd_valid AND master_s2m.CmdAccept;

  mst_cmd_fifo : adb3_ocp_mux_nb_fifo
  generic map(
    ORDER => 3,
    WIDTH => MST_CMD_WIDTH,
    EMPTY_VAL => '0',
    EMPTY_LO => MST_CMD_ATTR_WIDTH,
    EMPTY_HI => MST_CMD_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => mst_cmd_wr,
    d => mst_cmd_d,
    adv => mst_cmd_adv,
    full => open,
    not_full => mst_cmd_not_full,
    valid => mst_cmd_valid,
    q => mst_cmd_q);

  master_m2s.Cmd <= mst_cmd_q(MST_CMD_WIDTH - 1 downto MST_CMD_ATTR_WIDTH);
  master_m2s.Tag <= mst_cmd_q(MST_CMD_ATTR_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH);
  master_m2s.BurstLength <= mst_cmd_q(ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH);
  master_m2s.Addr <= mst_cmd_q(ADB3_OCP_ADDR_WIDTH - 1 downto 0);

  wbt_d <= n_cmd_port_1h & slv_cmd_sel.BurstLength;
  wbt_wr <= '1' when mst_cmd_wr = '1' AND (slv_cmd_sel.Cmd = OCP_CMD_WRITE or slv_cmd_sel.Cmd = OCP_CMD_WRITENONPOST) else '0';
  wbt_adv <= n_wdx_bstart;

  wbt_fifo : adb3_ocp_mux_nb_fifo
  generic map(
    ORDER => 3,
    WIDTH => WBT_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => wbt_wr,
    d => wbt_d,
    adv => wbt_adv,
    full => open,
    not_full => wbt_not_full,
    valid => wbt_valid,
    q => wbt_q);

  wbt_q_rec <= (port_1h => wbt_q(mux_inputs + ADB3_OCP_BURST_WIDTH - 1 downto ADB3_OCP_BURST_WIDTH),
                blen => wbt_q(ADB3_OCP_BURST_WIDTH - 1 downto 0));

  rbt_d <= n_cmd_port_1h & slv_cmd_sel.BurstLength;
  rbt_wr <= '1' when mst_cmd_wr = '1' AND slv_cmd_sel.Cmd = OCP_CMD_READ else '0';
  rbt_adv <= n_rdx_bstart;

  rbt_fifo : adb3_ocp_mux_nb_fifo
  generic map(
    ORDER => 3,
    WIDTH => RBT_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => rbt_wr,
    d => rbt_d,
    adv => rbt_adv,
    full => open,
    not_full => rbt_not_full,
    valid => rbt_valid,
    q => rbt_q);

  rbt_q_rec <= (port_1h => rbt_q(mux_inputs + ADB3_OCP_BURST_WIDTH - 1 downto ADB3_OCP_BURST_WIDTH),
                blen => rbt_q(ADB3_OCP_BURST_WIDTH - 1 downto 0));

  -- ---------------------------------------------------------------
  -- Write data path
  -- ---------------------------------------------------------------

  gen_wdx_state : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      wdx_state <= wdx_s_idle;
      wdx_blen <= (others => '-');
      wdx_blast <= '-';
      wdx_port_1h <= (others => '0');
    elsif rising_edge(ocp_clk) then
      wdx_state <= n_wdx_state;
      wdx_blen <= n_wdx_blen;
      wdx_blast <= n_wdx_blast;
      wdx_port_1h <= n_wdx_port_1h;
    end if;
  end process;

  gen_n_wdx_state : process(
    wdx_state,
    wdx_blen,
    wdx_blast,
    wdx_port_1h,
    mst_wdata_wr,
    wbt_valid,
    wbt_q_rec)
  begin
    case wdx_state is
      when wdx_s_idle =>
        if wbt_valid = '1' then
          n_wdx_state <= wdx_s_burst;
          n_wdx_blen <= wbt_q_rec.blen;
          n_wdx_blast <= NOT OR_reduce(wbt_q_rec.blen(wbt_q_rec.blen'high downto 1));
          n_wdx_port_1h <= wbt_q_rec.port_1h;
          n_wdx_bstart <= '1';
        else
          n_wdx_state <= wdx_s_idle;
          n_wdx_blen <= (others => '-');
          n_wdx_blast <= '-';
          n_wdx_port_1h <= (others => '0');
          n_wdx_bstart <= '0';
        end if;

      when wdx_s_burst =>
        if mst_wdata_wr = '1' then
          if wdx_blast = '1' then
            if wbt_valid = '1' then
              n_wdx_state <= wdx_s_burst;
              n_wdx_blen <= wbt_q_rec.blen;
              n_wdx_blast <= NOT OR_reduce(wbt_q_rec.blen(wbt_q_rec.blen'high downto 1));
              n_wdx_port_1h <= wbt_q_rec.port_1h;
              n_wdx_bstart <= '1';
            else
              n_wdx_state <= wdx_s_idle;
              n_wdx_blen <= (others => '-');
              n_wdx_blast <= '-';
              n_wdx_port_1h <= (others => '0');
              n_wdx_bstart <= '0';
            end if;
          else
            n_wdx_state <= wdx_s_burst;
            n_wdx_blen <= wdx_blen - 1;
            n_wdx_blast <= NOT OR_reduce(wdx_blen(wdx_blen'high downto 2) & wdx_blen(0));
            n_wdx_port_1h <= wdx_port_1h;
            n_wdx_bstart <= '0';
          end if;
        else
          n_wdx_state <= wdx_s_burst;
          n_wdx_blen <= wdx_blen;
          n_wdx_blast <= wdx_blast;
          n_wdx_port_1h <= wdx_port_1h;
          n_wdx_bstart <= '0';
        end if;
    end case;
  end process;

  slv_wdata_fifos : for i in 0 to mux_inputs - 1 generate
    slv_wdata_wr(i) <= '1' when slv_wdata_not_full(i) = '1' AND slaves_m2s(i).DataValid = '1' else '0';
    slv_wdata_d(i) <= slaves_m2s(i).DataByteEn & slaves_m2s(i).Data;
    slv_wdata_adv(i) <= wdx_port_1h(i) AND slv_wdata_valid(i) AND mst_wdata_not_full;

    U0 : adb3_ocp_mux_nb_fifo
    generic map(
      ORDER => 4,
      WIDTH => SLV_WDATA_WIDTH)
    port map(
      rst => ocp_rst,
      clk => ocp_clk,
      wr => slv_wdata_wr(i),
      d => slv_wdata_d(i),
      adv => slv_wdata_adv(i),
      full => open,
      not_full => slv_wdata_not_full(i),
      valid => slv_wdata_valid(i),
      q => slv_wdata_q(i));

    slv_wdata_vec(i) <= (DataByteEn => slv_wdata_q(i)(SLV_WDATA_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH),
                         Data => slv_wdata_q(i)(ADB3_OCP_DATA_WIDTH - 1 downto 0));

    slaves_s2m(i).DataAccept <= slv_wdata_not_full(i);
  end generate;

  mst_wdata_wr <= OR_reduce(wdx_port_1h AND slv_wdata_valid) AND mst_wdata_not_full;

  select_slv_wdata : process(wdx_port_1h, slv_wdata_vec)
    variable tmp : slv_wdata_t;
  begin
    tmp := (Data => (others => '0'), DataByteEn => (others => '0'));
    for i in 0 to mux_inputs - 1 loop
      if wdx_port_1h(i) = '1' then
        tmp.Data := tmp.Data OR slv_wdata_vec(i).Data;
        tmp.DataByteEn := tmp.DataByteEn OR slv_wdata_vec(i).DataByteEn;
      end if;
    end loop;
    mst_wdata_d <= tmp.DataByteEn & tmp.Data;
  end process;

  mst_wdata_adv <= mst_wdata_valid AND master_s2m.DataAccept;

  mst_wdata_fifo : adb3_ocp_mux_nb_fifo
  generic map(
    ORDER => 4,
    WIDTH => MST_WDATA_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => mst_wdata_wr,
    d => mst_wdata_d,
    adv => mst_wdata_adv,
    full => open,
    not_full => mst_wdata_not_full,
    valid => mst_wdata_valid,
    q => mst_wdata_q);

  master_m2s.Data <= mst_wdata_q(ADB3_OCP_DATA_WIDTH - 1 downto 0);
  master_m2s.DataByteEn <= mst_wdata_q(ADB3_OCP_DATA_WIDTH + ADB3_OCP_BE_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH);
  master_m2s.DataValid <= mst_wdata_valid;

  -- ---------------------------------------------------------------
  -- Read data path
  -- ---------------------------------------------------------------

  gen_rdx_state : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      rdx_state <= rdx_s_idle;
      rdx_blen <= (others => '-');
      rdx_blast <= '-';
      rdx_port_1h <= (others => '0');
    elsif rising_edge(ocp_clk) then
      rdx_state <= n_rdx_state;
      rdx_blen <= n_rdx_blen;
      rdx_blast <= n_rdx_blast;
      rdx_port_1h <= n_rdx_port_1h;
    end if;
  end process;

  gen_n_rdx_state : process(
    rdx_state,
    rdx_blen,
    rdx_blast,
    rdx_port_1h,
    mst_rdata_adv,
    rbt_valid,
    rbt_q_rec)
  begin
    case rdx_state is
      when rdx_s_idle =>
        if rbt_valid = '1' then
          n_rdx_state <= rdx_s_burst;
          n_rdx_blen <= rbt_q_rec.blen;
          n_rdx_blast <= NOT OR_reduce(rbt_q_rec.blen(rbt_q_rec.blen'high downto 1));
          n_rdx_port_1h <= rbt_q_rec.port_1h;
          n_rdx_bstart <= '1';
        else
          n_rdx_state <= rdx_s_idle;
          n_rdx_blen <= (others => '-');
          n_rdx_blast <= '-';
          n_rdx_port_1h <= (others => '0');
          n_rdx_bstart <= '0';
        end if;

      when rdx_s_burst =>
        if mst_rdata_adv = '1' then
          if rdx_blast = '1' then
            if rbt_valid = '1' then
              n_rdx_state <= rdx_s_burst;
              n_rdx_blen <= rbt_q_rec.blen;
              n_rdx_blast <= NOT OR_reduce(rbt_q_rec.blen(rbt_q_rec.blen'high downto 1));
              n_rdx_port_1h <= rbt_q_rec.port_1h;
              n_rdx_bstart <= '1';
            else
              n_rdx_state <= rdx_s_idle;
              n_rdx_blen <= (others => '-');
              n_rdx_blast <= '-';
              n_rdx_port_1h <= (others => '0');
              n_rdx_bstart <= '0';
            end if;
          else
            n_rdx_state <= rdx_s_burst;
            n_rdx_blen <= rdx_blen - 1;
            n_rdx_blast <= NOT OR_reduce(rdx_blen(rdx_blen'high downto 2) & rdx_blen(0));
            n_rdx_port_1h <= rdx_port_1h;
            n_rdx_bstart <= '0';
          end if;
        else
          n_rdx_state <= rdx_s_burst;
          n_rdx_blen <= rdx_blen;
          n_rdx_blast <= rdx_blast;
          n_rdx_port_1h <= rdx_port_1h;
          n_rdx_bstart <= '0';
        end if;
    end case;
  end process;

  mst_rdata_wr <= '1' when mst_rdata_not_full = '1' AND master_s2m.Resp = OCP_RESP_VALID else '0';
  mst_rdata_d <= master_s2m.Tag & master_s2m.Data;
  mst_rdata_adv <= OR_reduce(rdx_port_1h AND slv_rdata_not_full) AND mst_rdata_valid;

  mst_rdata_fifo : adb3_ocp_mux_nb_fifo
  generic map(
    ORDER => 4,
    WIDTH => MST_RDATA_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => mst_rdata_wr,
    d => mst_rdata_d,
    adv => mst_rdata_adv,
    full => open,
    not_full => mst_rdata_not_full,
    valid => mst_rdata_valid,
    q => mst_rdata_q);

  master_m2s.RespAccept <= mst_rdata_not_full;

  slv_rdata_fifos : for i in 0 to mux_inputs - 1 generate
    slv_rdata_wr(i) <= rdx_port_1h(i) AND slv_rdata_not_full(i) AND mst_rdata_valid;
    slv_rdata_d(i) <= mst_rdata_q;
    slv_rdata_adv(i) <= slv_rdata_valid(i) AND slaves_m2s(i).RespAccept;

    U0 : adb3_ocp_mux_nb_fifo
    generic map(
      ORDER => 4,
      WIDTH => SLV_RDATA_WIDTH)
    port map(
      rst => ocp_rst,
      clk => ocp_clk,
      wr => slv_rdata_wr(i),
      d => slv_rdata_d(i),
      adv => slv_rdata_adv(i),
      full => open,
      not_full => slv_rdata_not_full(i),
      valid => slv_rdata_valid(i),
      q => slv_rdata_q(i));

    slaves_s2m(i).Data <= slv_rdata_q(i)(ADB3_OCP_DATA_WIDTH - 1 downto 0);
    slaves_s2m(i).Tag <= slv_rdata_q(i)(ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH);
    slaves_s2m(i).Resp <= "0" & slv_rdata_valid(i);
  end generate;

end;