--==============================================================================
--
--     File Name: adb3_ocp_retime_nb.vhd
--       Version: 0.4
--          Date: 13 Jul 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP re-time (non-blocking)
--
--   Description: Component for re-time of OCP port using SRL16s
--
--   Limitations:
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 04 Nov 2010: Initial design
--                0.2, KDC  09 Feb 2011: Replaced AFIFO design with SRL16 design
--                0.3, TW   09 Feb 2011:
--                      - Replaced FDCE primitives with inferred flip-flops and
--                        applied SHREG_EXTRACT = "no" attributes, to eliminate
--                        unused output warnings when outputs of FDCEs are
--                        optimized away.
--                0.4, KDC  13 Jul 2011:
--                      - Re-structured design to use common SRL16 block
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library unisim;
use unisim.vcomponents.all;

entity adb3_ocp_srl16_ret is
generic(
  WIDTH : in    natural);
port(
  clk           : in    std_logic;                           -- Clock
  -- Slave side
  slave_d       : in    std_logic_vector(WIDTH-1 downto 0);  -- SRL16 d
  slave_valid   : in    std_logic;                           -- Group valid SRL16 output
  slave_accept  : out   std_logic;                           -- OCP slave accept
  slave_q       : out   std_logic_vector(WIDTH-1 downto 0);  -- SRL16 q
  -- Master side
  master_valid  : in    std_logic;                           -- Group valid FDCE output
  master_accept : in    std_logic;                           -- OCP master accept
  master_q      : out   std_logic_vector(WIDTH-1 downto 0)); -- FDCE q
end;

architecture syn of adb3_ocp_srl16_ret is

  signal srl_sel : std_logic := '0'; -- Select SRL16E(0) on reset
  signal srl_ce  : std_logic := '1'; -- Enable SRL16E on reset
  signal fdce_ce : std_logic := '1'; -- Enable FDCE on reset

  signal srl_q  : std_logic_vector(WIDTH-1 downto 0) := (others => '0');
  signal fdce_q : std_logic_vector(WIDTH-1 downto 0) := (others => '0');

begin

  srl16e_ctrl_p : process(clk)
  begin
    if rising_edge(clk) then
      if (srl_ce = '1') then
        -- Select SRL(0) as input to FDCE
        -- Shift SRL16E, load data in SRL(0)
        -- Continue to shift data until FDCE holds and SRL group valid is true
        if (fdce_ce = '0')and(slave_valid = '1') then
          -- Select SRL(1) as input to FDCE
          -- Hold SRL16E with data in SRL(0) and SRL(1)
          srl_sel <= '1';
          srl_ce  <= '0';
        end if;
      else
        -- Select SRL(1) as input to FDCE
        -- Hold SRL16E with data in SRL(0) and SRL(1)
        -- Continue to hold data until FDCE shifts
        if (fdce_ce = '1') then
          -- Select SRL(0) as input to FDCE
          -- Shift SRL16E, load data in SRL(0)
          srl_sel <= '0';
          srl_ce  <= '1';
        end if;
      end if;
    end if;
  end process;

  -- Generate SRL16Es and FDCEs for signal group
  srl16e_group : for i in 0 to WIDTH-1 generate
    srl16e_i : SRL16E
      generic map(
        INIT => X"0000")
      port map(
        CLK => clk,
        D   => slave_d(i),
        A0  => srl_sel,
        A1  => '0',
        A2  => '0',
        A3  => '0',
        CE  => srl_ce,
        Q   => srl_q(i));

    -- Shift FDCE if master accepts valid data
    -- Shift FDCE if master data is not valid
    -- Hold FDCE if master does not accept valid data
    fdce_ce <= master_accept or not(master_valid);

    fdce_p : process(clk)
    begin
      if rising_edge(clk) then
        if (fdce_ce = '1') then
          fdce_q(i) <= srl_q(i);
        end if;
      end if;
    end process;
  end generate;

  slave_accept <= srl_ce;
  slave_q      <= srl_q;
  master_q     <= fdce_q;
end;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_retime_nb is
port(
  ocp_rst    : in  std_logic;
  ocp_clk    : in  std_logic;
  -- OCP primary port
  slave_m2s  : in  adb3_ocp_m2sT;
  slave_s2m  : out adb3_ocp_s2mT;
  -- OCP secondary port
  master_m2s : out adb3_ocp_m2sT;
  master_s2m : in  adb3_ocp_s2mT);
end;

architecture syn of adb3_ocp_retime_nb is

  signal slave_m2s_int  : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal master_m2s_int : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;

  signal slave_s2m_int  : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;
  signal master_s2m_int : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  -- Command group signals
  constant SLV_CMD_WIDTH : natural := ADB3_OCP_CMD_WIDTH + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;

  signal slv_cmd_d      : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);
  signal slv_cmd_valid  : std_logic;
  signal slv_cmd_accept : std_logic;
  signal slv_cmd_q      : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);

  signal mst_cmd_valid  : std_logic;
  signal mst_cmd_accept : std_logic;
  signal mst_cmd_q      : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);

  -- Write data group signals
  constant SLV_WDATA_WIDTH : natural := 1 + ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;

  signal slv_wdata_d      : std_logic_vector(SLV_WDATA_WIDTH-1 downto 0);
  signal slv_wdata_valid  : std_logic;
  signal slv_wdata_accept : std_logic;
  signal slv_wdata_q      : std_logic_vector(SLV_WDATA_WIDTH-1 downto 0);

  signal mst_wdata_valid  : std_logic;
  signal mst_wdata_accept : std_logic;
  signal mst_wdata_q      : std_logic_vector(SLV_WDATA_WIDTH-1 downto 0);

  -- Read resp group signals
  constant SLV_RDATA_WIDTH : natural := ADB3_OCP_RESP_WIDTH + ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH;

  signal slv_rdata_d      : std_logic_vector(SLV_RDATA_WIDTH-1 downto 0);
  signal slv_rdata_valid  : std_logic;
  signal slv_rdata_accept : std_logic;
  signal slv_rdata_q      : std_logic_vector(SLV_RDATA_WIDTH-1 downto 0);

  signal mst_rdata_valid  : std_logic;
  signal mst_rdata_accept : std_logic;
  signal mst_rdata_q      : std_logic_vector(SLV_RDATA_WIDTH-1 downto 0);

  attribute SHREG_EXTRACT : string;
  attribute SHREG_EXTRACT of master_m2s_int : signal is "no";
  attribute SHREG_EXTRACT of slave_s2m_int  : signal is "no";

  component adb3_ocp_srl16_ret
  generic(
    WIDTH : in    natural);
  port(
    clk           : in    std_logic;
    -- Slave side
    slave_d       : in    std_logic_vector(WIDTH-1 downto 0);
    slave_valid   : in    std_logic;
    slave_accept  : out   std_logic;
    slave_q       : out   std_logic_vector(WIDTH-1 downto 0);
    -- Master side
    master_valid  : in    std_logic;
    master_accept : in    std_logic;
    master_q      : out   std_logic_vector(WIDTH-1 downto 0));
  end component;

begin

  -- Connect internal OCP output signals
  master_m2s <= master_m2s_int;
  slave_s2m  <= slave_s2m_int;


  -- Command group signals
  slv_cmd_d <= slave_m2s.Cmd & slave_m2s.Tag & slave_m2s.BurstLength & slave_m2s.Addr;

  slave_m2s_int.Cmd         <= slv_cmd_q(SLV_CMD_WIDTH-1 downto ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH);
  slave_m2s_int.Tag         <= slv_cmd_q(ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH);
  slave_m2s_int.BurstLength <= slv_cmd_q(ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_ADDR_WIDTH);
  slave_m2s_int.Addr        <= slv_cmd_q(ADB3_OCP_ADDR_WIDTH-1 downto 0);

  slv_cmd_valid <= '1' when (OR_reduce(slave_m2s_int.Cmd) = '1') else '0';
  slave_s2m_int.CmdAccept <= slv_cmd_accept;

  slv_cmd_i : adb3_ocp_srl16_ret
  generic map(
    WIDTH => SLV_CMD_WIDTH)
  port map(
    clk           => ocp_clk,         -- in    std_logic
    -- Slave side
    slave_d       => slv_cmd_d,       -- in    std_logic_vector(WIDTH-1 downto 0)
    slave_valid   => slv_cmd_valid,   -- in    std_logic
    slave_accept  => slv_cmd_accept,  -- out   std_logic
    slave_q       => slv_cmd_q,       -- out   std_logic_vector(WIDTH-1 downto 0)
    -- Master side
    master_valid  => mst_cmd_valid,   -- in    std_logic
    master_accept => mst_cmd_accept,  -- in    std_logic
    master_q      => mst_cmd_q);      -- out   std_logic_vector(WIDTH-1 downto 0)

  mst_cmd_valid  <= '1' when (OR_reduce(master_m2s_int.Cmd) = '1') else '0';
  mst_cmd_accept <= master_s2m.CmdAccept;

  master_m2s_int.Cmd         <= mst_cmd_q(SLV_CMD_WIDTH-1 downto ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH);
  master_m2s_int.Tag         <= mst_cmd_q(ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH);
  master_m2s_int.BurstLength <= mst_cmd_q(ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_ADDR_WIDTH);
  master_m2s_int.Addr        <= mst_cmd_q(ADB3_OCP_ADDR_WIDTH-1 downto 0);


  -- Write data group signals
  slv_wdata_d <= slave_m2s.DataValid & slave_m2s.DataByteEn & slave_m2s.Data;

  slave_m2s_int.DataValid  <= slv_wdata_q(SLV_WDATA_WIDTH-1);
  slave_m2s_int.DataByteEn <= slv_wdata_q(ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  slave_m2s_int.Data       <= slv_wdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0);

  slv_wdata_valid <= '1' when (slave_m2s_int.DataValid = '1') else '0';
  slave_s2m_int.DataAccept <= slv_wdata_accept;

  slv_wdata_i : adb3_ocp_srl16_ret
  generic map(
    WIDTH => SLV_WDATA_WIDTH)
  port map(
    clk           => ocp_clk,           -- in    std_logic
    -- Slave side
    slave_d       => slv_wdata_d,       -- in    std_logic_vector(WIDTH-1 downto 0)
    slave_valid   => slv_wdata_valid,   -- in    std_logic
    slave_accept  => slv_wdata_accept,  -- out   std_logic
    slave_q       => slv_wdata_q,       -- out   std_logic_vector(WIDTH-1 downto 0)
    -- Master side
    master_valid  => mst_wdata_valid,   -- in    std_logic
    master_accept => mst_wdata_accept,  -- in    std_logic
    master_q      => mst_wdata_q);      -- out   std_logic_vector(WIDTH-1 downto 0)

  mst_wdata_valid  <= '1' when (master_m2s_int.DataValid = '1') else '0';
  mst_wdata_accept <= master_s2m.DataAccept;

  master_m2s_int.DataValid  <= mst_wdata_q(SLV_WDATA_WIDTH-1);
  master_m2s_int.DataByteEn <= mst_wdata_q(ADB3_OCP_BE_WIDTH+ADB3_OCP_DATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  master_m2s_int.Data       <= mst_wdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0);


  -- Read resp group signals
  slv_rdata_d <= master_s2m.Resp & master_s2m.Tag & master_s2m.Data;

  master_s2m_int.Resp <= slv_rdata_q(SLV_RDATA_WIDTH-1 downto ADB3_OCP_TAG_WIDTH+ADB3_OCP_DATA_WIDTH);
  master_s2m_int.Tag  <= slv_rdata_q(ADB3_OCP_TAG_WIDTH+ADB3_OCP_DATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  master_s2m_int.Data <= slv_rdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0);

  slv_rdata_valid <= '1' when (OR_reduce(master_s2m_int.Resp) = '1') else '0';
  master_m2s_int.RespAccept <= slv_rdata_accept;

  slv_rdata_i : adb3_ocp_srl16_ret
  generic map(
    WIDTH => SLV_RDATA_WIDTH)
  port map(
    clk           => ocp_clk,           -- in    std_logic
    -- Slave side
    slave_d       => slv_rdata_d,       -- in    std_logic_vector(WIDTH-1 downto 0)
    slave_valid   => slv_rdata_valid,   -- in    std_logic
    slave_accept  => slv_rdata_accept,  -- out   std_logic
    slave_q       => slv_rdata_q,       -- out   std_logic_vector(WIDTH-1 downto 0)
    -- Master side
    master_valid  => mst_rdata_valid,   -- in    std_logic
    master_accept => mst_rdata_accept,  -- in    std_logic
    master_q      => mst_rdata_q);      -- out   std_logic_vector(WIDTH-1 downto 0)

  mst_rdata_valid  <= '1' when (OR_reduce(slave_s2m_int.Resp) = '1') else '0';
  mst_rdata_accept <= slave_m2s.RespAccept;

  slave_s2m_int.Resp <= mst_rdata_q(SLV_RDATA_WIDTH-1 downto ADB3_OCP_TAG_WIDTH+ADB3_OCP_DATA_WIDTH);
  slave_s2m_int.Tag  <= mst_rdata_q(ADB3_OCP_TAG_WIDTH+ADB3_OCP_DATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH);
  slave_s2m_int.Data <= mst_rdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0);

end;