--==============================================================================
--
--     File Name: adb3_ocp_simple_bus_if.vhd
--       Version: 0.1
--          Date: 05 May 2010
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 OCP simple bus interface
--
--   Description: ADB3 OCP simple bus interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 05 May 10: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;
use ieee.math_real.all;       -- Used for CEIL and LOG2 functions only

library work;
use work.adb3_ocp.all;

entity adb3_ocp_simple_bus_if is
generic(
  addr_width   : natural range 1 to ADB3_OCP_ADDR_WIDTH;
  data_width   : natural range 8 to ADB3_OCP_DATA_WIDTH;
  read_latency : natural := 1);
port(
  -- OCP interface
  ocp_rst : in    std_logic;
  ocp_clk : in    std_logic;
  ocp_m2s : in    adb3_ocp_m2sT;
  ocp_s2m : out   adb3_ocp_s2mT;
  -- Simple bus interface
  d       : out   std_logic_vector(data_width-1 downto 0);
  q       : in    std_logic_vector(data_width-1 downto 0);
  a       : out   std_logic_vector(addr_width-1 downto 0);  -- Byte address
  w       : out   std_logic;
  r       : out   std_logic;
  we      : out   std_logic_vector((data_width/8)-1 downto 0));
end;

architecture syn of adb3_ocp_simple_bus_if is

  constant WORD_WIDTH     : natural := ADB3_OCP_DATA_WIDTH/data_width;
  -- We must ensure that when WORD_WIDTH = 1, WORD_CNT_WIDTH = 1
  -- round(A/B) == trunc((A+B/2)/B)
  -- WORD_CNT_WIDTH = 1 + log2(round(WORD_WIDTH/2)) = 1 + log2((WORD_WIDTH+1)/2)
  constant WORD_CNT_WIDTH : natural := 1 + NATURAL(CEIL(LOG2(REAL((WORD_WIDTH+1)/2))));

  subtype word_cnt_s   is std_logic_vector(WORD_CNT_WIDTH-1 downto 0);
  subtype local_addr_s is std_logic_vector(addr_width-1 downto 0);
  subtype local_data_s is std_logic_vector(data_width-1 downto 0);
  subtype local_be_s   is std_logic_vector((data_width/8)-1 downto 0);

  subtype burst_cnt_s  is std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
  subtype tag_s        is std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
  subtype data_s       is std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
  subtype be_s         is std_logic_vector((ADB3_OCP_DATA_WIDTH/8)-1 downto 0);

  constant WORD_CNT_MAX    : word_cnt_s   := std_logic_vector(to_unsigned(WORD_WIDTH-1,WORD_CNT_WIDTH));
  constant LOCAL_DATA_ZERO : local_data_s := (others => '0');
  constant LOCAL_BE_ZERO   : local_be_s   := (others => '0');

  signal SCmdAccept  : std_logic := '0';
  signal SDataAccept : std_logic := '0';

  -- Write command
  signal write_burst_cnt   : burst_cnt_s;
  signal write_word_cnt    : word_cnt_s;
  signal writing           : std_logic := '0';
  signal write_addr        : local_addr_s;

  -- Read command
  signal read_burst_cnt    : burst_cnt_s;
  signal read_word_cnt     : word_cnt_s;
  signal reading           : std_logic := '0';
  signal read_tag          : tag_s;

  -- Combined read/write address
  signal rw_addr           : local_addr_s := (others => '0');

  -- Write data
  signal write_data_sr     : data_s := (others => '0');
  signal write_data_v      : std_logic := '0';
  signal write_we_sr       : be_s := (others => '0');

  -- Read data
  signal read_data_sr      : data_s;
  signal read_data_v       : ocp_RespT := OCP_RESP_NONE;
  signal read_latency_1hot : std_logic_vector(read_latency downto 0) := (others => '0');

begin

  assert not(read_latency < 1)
    report "(adb3_ocp_simple_bus_if): Mimium read_latency is 1 clock cycle"
    severity error;

  assert not((data_width /= 128)and(data_width /= 64)and(data_width /= 32)and(data_width /= 16)and(data_width /= 8))
    report "(adb3_ocp_simple_bus_if): Generic data_width must be 8/16/32/64/128 bits"
    severity error;

  process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      SCmdAccept   <= '0';
      SDataAccept  <= '0';
      write_data_v <= '0';
      write_we_sr  <= (others => '0');

      if (writing = '0')and(reading = '0') then
        if (ocp_m2s.Cmd = OCP_CMD_WRITE) then
          SCmdAccept      <= '1';
          write_burst_cnt <= ocp_m2s.BurstLength;
          write_addr      <= ocp_m2s.Addr(addr_width-1 downto 0);
          write_word_cnt  <= (others => '0');
          writing         <= '1';
        end if;

        if (ocp_m2s.Cmd = OCP_CMD_READ) then
          SCmdAccept     <= '1';
          read_burst_cnt <= ocp_m2s.BurstLength;
          rw_addr        <= ocp_m2s.Addr(addr_width-1 downto 0);
          read_word_cnt  <= (others => '0');
          reading        <= '1';
          read_data_v    <= OCP_RESP_NONE;
          read_tag       <= ocp_m2s.Tag;
          read_latency_1hot(0) <= '1';
          read_latency_1hot(read_latency downto 1) <= (others => '0');
        end if;

      elsif (writing = '1')and(WORD_WIDTH > 1) then
        rw_addr <= write_addr;
        if (OR_reduce(write_word_cnt) = '0') then
          if (ocp_m2s.DataValid = '1')and(SDataAccept = '0') then
            write_data_sr   <= ocp_m2s.Data;
            write_we_sr     <= ocp_m2s.DataByteEn;
            SDataAccept     <= '1';
            write_data_v    <= '1';
            write_burst_cnt <= write_burst_cnt - 1;
            write_addr      <= write_addr + data_width/8;
            write_word_cnt  <= write_word_cnt + 1;
          end if;
        else
          write_data_sr  <= LOCAL_DATA_ZERO & write_data_sr(ADB3_OCP_DATA_WIDTH-1 downto data_width);
          write_we_sr    <= LOCAL_BE_ZERO & write_we_sr((ADB3_OCP_DATA_WIDTH/8)-1 downto (data_width/8));
          write_word_cnt <= write_word_cnt + 1;
          write_addr     <= write_addr + data_width/8;
          write_data_v   <= '1';
          if (write_word_cnt = WORD_CNT_MAX)and(OR_reduce(write_burst_cnt) = '0') then
            writing <= '0';
          end if;
        end if;

      elsif (writing = '1')and(WORD_WIDTH = 1) then
        rw_addr <= write_addr;
        if (ocp_m2s.DataValid = '1')and(SDataAccept = '0') then
          write_data_sr   <= ocp_m2s.Data;
          write_we_sr     <= ocp_m2s.DataByteEn;
          SDataAccept     <= '1';
          write_data_v    <= '1';
          write_burst_cnt <= write_burst_cnt - 1;
          write_addr      <= write_addr + data_width/8;
        else
          if (OR_reduce(write_burst_cnt) = '0')then
            writing <= '0';
          end if;
        end if;

      elsif (reading = '1') then
        if (read_data_v = OCP_RESP_NONE) then
          read_latency_1hot <= read_latency_1hot(read_latency-1 downto 0) & '0';

          if (read_latency_1hot(read_latency) = '1') then
            rw_addr <= rw_addr + data_width/8;
            if (WORD_WIDTH > 1) then
              read_data_sr  <= q & read_data_sr(ADB3_OCP_DATA_WIDTH-1 downto data_width);
              read_word_cnt <= read_word_cnt + 1;
            else
              read_data_sr  <= q;
            end if;

            if (read_word_cnt = WORD_CNT_MAX) then
              read_data_v <= OCP_RESP_VALID;
            else
              read_latency_1hot(read_latency downto 1) <= (others => '0');
              read_latency_1hot(0) <= '1';
            end if;
          end if;
        else
          if (ocp_m2s.RespAccept = '1') then
            read_data_v    <= OCP_RESP_NONE;
            read_burst_cnt <= read_burst_cnt - 1;
            if (OR_reduce(read_burst_cnt(ADB3_OCP_BURST_WIDTH-1 downto 1)) = '0') then
              reading <= '0';
            else
              read_latency_1hot(read_latency downto 1) <= (others => '0');
              read_latency_1hot(0) <= '1';
            end if;
          end if;
        end if;
      end if;
    end if;
  end process;

  a  <= rw_addr;
  d  <= write_data_sr(data_width-1 downto 0);
  we <= write_we_sr((data_width/8)-1 downto 0);
  w  <= write_data_v;
  r  <= read_latency_1hot(0);

  ocp_s2m.CmdAccept  <= SCmdAccept;
  ocp_s2m.DataAccept <= SDataAccept;
  ocp_s2m.Tag        <= read_tag;
  ocp_s2m.Data       <= read_data_sr;
  ocp_s2m.Resp       <= read_data_v;
end;