--==============================================================================
--
--     File Name: adb3_ocp_simple_bus_if_nb.vhd
--       Version: 0.1
--          Date: 24 Jun 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP to bus interface (non-blocking)
--
--   Description: Component giving OCP port access to parallel bus interface
--
--   Limitations: Up to 8 read commands in flight
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 24 Jun 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

--
-- A synchronous FIFO used to buffer commands, command attributes, data, byte
-- enables etc. in 'adb3_ocp_split_nb'.
--
-- 'ORDER' is log2 of the FIFO depth, e.g. ORDER = 4 => depth is 16 locations.
--
-- 'WIDTH' is the width of FIFO locations, in bits.
--
-- 'EMPTY_VAL' is the value driven on 'q(EMPTY_LO:EMPTY_HI)' when 'valid' is
-- not asserted; map both 'EMPTY_LO' and 'EMPTY_HI' to 0 if this feature is
-- not required.
--
entity adb3_ocp_bus_if_fifo is
generic(
  ORDER       : in    natural range 2 to 6;
  WIDTH       : in    natural;
  EMPTY_VAL   : in    std_logic := '0';
  EMPTY_LO    : in    natural := 0;
  EMPTY_HI    : in    natural := 0);
port(
  rst         : in    std_logic;                             -- Asynchronous reset
  clk         : in    std_logic;                             -- Clock
  wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
  d           : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
  adv         : in    std_logic;                             -- '1' => present next item on 'q' to next item
  full        : out   std_logic;                             -- '1' => full
  not_full    : out   std_logic;                             -- '1' => not full
  valid       : out   std_logic;                             -- '1' => not empty
  q           : out   std_logic_vector(WIDTH - 1 downto 0)); -- Current item, qualified by 'valid'
end;

architecture syn of adb3_ocp_bus_if_fifo is

  constant DEPTH : natural := 2**ORDER;
  constant MAX_LEVEL : natural := DEPTH - 1;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal wptr, rptr : std_logic_vector(ORDER - 1 downto 0);
  signal rn : std_logic_vector(ORDER - 1 downto 0);
  signal not_empty : std_logic;
  signal i_full, i_not_full : std_logic;

  signal qvalid : std_logic;
  signal qce : std_logic;
  signal rd : std_logic;

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  valid <= qvalid;
  full <= i_full;
  not_full <= i_not_full;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_not_full <= '1';
      i_full <= '0';
    elsif rising_edge(clk) then
      if wr = '1' then
        wptr <= wptr + 1;
      end if;

      if wr = '1' then
        assert CONV_INTEGER(wptr - rptr) < MAX_LEVEL severity failure;
        tmp := wptr - rptr + 1;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      else
        tmp := wptr - rptr;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr)) <= d;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= adv OR NOT qvalid;
  rd <= (adv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rptr <= rptr + 1;
      end if;
    end if;

    if rising_edge(clk) then
      if rd = '1' then
        assert CONV_INTEGER(wptr - rptr) > 0 severity failure;
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
    elsif rising_edge(clk) then
      if qce = '1' then
        qvalid <= not_empty;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : for i in 0 to WIDTH - 1 generate
    with_reset : if EMPTY_LO <= i AND i < EMPTY_HI generate
      U0 : process(rst, clk)
      begin
        if rst = '1' then
          q(i) <= EMPTY_VAL;
        elsif rising_edge(clk) then
          if qce = '1' then
            if not_empty = '1' then
              q(i) <= sel_ram(i);
            else
              q(i) <= EMPTY_VAL;
            end if;
          end if;
        end if;
      end process;
    end generate;

    without_reset : if NOT(EMPTY_LO <= i AND i < EMPTY_HI) generate
      U0 : process(clk)
      begin
        if rising_edge(clk) then
          if qce = '1' then
            q(i) <= sel_ram(i);
          end if;
        end if;
      end process;
    end generate;
  end generate;

end;


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;

entity adb3_ocp_simple_bus_if_nb is
generic(
  addr_width   : natural range 1 to ADB3_OCP_ADDR_WIDTH;
  data_width   : natural range 8 to ADB3_OCP_DATA_WIDTH;
  read_latency : natural := 1);
port(
  -- OCP interface
  ocp_rst    : in    std_logic;
  ocp_clk    : in    std_logic;
  ocp_m2s    : in    adb3_ocp_m2sT;
  ocp_s2m    : out   adb3_ocp_s2mT;
  -- Bus interface
  d          : out   std_logic_vector(data_width-1 downto 0);
  q          : in    std_logic_vector(data_width-1 downto 0);
  a          : out   std_logic_vector(addr_width-1 downto 0);  -- Byte address
  w          : out   std_logic;
  r          : out   std_logic;
  we         : out   std_logic_vector((data_width/8)-1 downto 0);
  -- Debug
  debug_info : out   std_logic_vector(31 downto 0));
end;

architecture syn of adb3_ocp_simple_bus_if_nb is

  -- Connect debug output
  constant DEBUG_ON : boolean := false;

  -- Calculate number of bus commands for each OCP command
  constant MAX_CMDS : natural := ADB3_OCP_DATA_WIDTH/8;
  constant BUS_CMDS : natural range 1 to MAX_CMDS := ADB3_OCP_DATA_WIDTH/data_width;

  -- Slave command FIFO signals
  constant SLV_CMD_DEPTH : natural := 3;
  constant SLV_CMD_WIDTH : natural := ADB3_OCP_CMD_WIDTH + ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;

  signal slv_cmd_wr       : std_logic;
  signal slv_cmd_adv      : std_logic;
  signal slv_cmd_not_full : std_logic;
  signal slv_cmd_valid    : std_logic;
  signal slv_cmd_d        : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);
  signal slv_cmd_q        : std_logic_vector(SLV_CMD_WIDTH-1 downto 0);

  type slv_cmd_attr_t is record
    Cmd         : ocp_CmdT;
    Addr        : adb3_ocp_addr_s;
    BurstLength : adb3_ocp_burst_s;
    Tag         : adb3_ocp_tag_s;
  end record;

  signal slv_cmd : slv_cmd_attr_t;
  signal slv_cmd_q_wr : std_logic;
  signal slv_cmd_q_rd : std_logic;


  -- Slave write data FIFO signals
  constant SLV_WDATA_DEPTH : natural := SLV_CMD_DEPTH + 1;
  constant SLV_WDATA_WIDTH : natural := ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;

  signal slv_wdata_wr       : std_logic;
  signal slv_wdata_adv      : std_logic;
  signal slv_wdata_not_full : std_logic;
  signal slv_wdata_valid    : std_logic;
  signal slv_wdata_d        : std_logic_vector(SLV_WDATA_WIDTH - 1 downto 0);
  signal slv_wdata_q        : std_logic_vector(SLV_WDATA_WIDTH - 1 downto 0);

  type slv_wdata_attr_t is record
    DataByteEn : adb3_ocp_be_s;
    Data       : adb3_ocp_data_s;
  end record;

  signal slv_wdata : slv_wdata_attr_t;


  -- Bus command/write data
  signal mst_a  : std_logic_vector(addr_width-1 downto 0);
  signal mst_w  : std_logic;
  signal mst_r  : std_logic;
  signal mst_d  : adb3_ocp_data_s;
  signal mst_we : adb3_ocp_be_s;
  signal slv_q  : adb3_ocp_data_s;

  signal mst_cmd_del   : std_logic;
  signal mst_cmd_sr    : std_logic_vector(MAX_CMDS-1 downto 0);
  signal mst_cmd_busy  : std_logic;
  signal mst_cmd_conv  : std_logic;
  signal mst_cmd_index : natural;


  -- Command translation FSM
  type mcx_state_t is (mcx_s_idle, mcx_s_burst);

  signal mcx_state   : mcx_state_t;
  signal mcx_cmd_wr  : std_logic;
  signal mcx_cmd_rd  : std_logic;
  signal mcx_mst_wr  : std_logic;
  signal mcx_blen    : adb3_ocp_burst_s;
  signal mcx_tag     : adb3_ocp_tag_s;
  signal mcx_blast   : std_logic;
  signal mcx_addr    : adb3_ocp_addr_s;
  signal mcx_data    : adb3_ocp_data_s;
  signal mcx_we      : adb3_ocp_be_s;

  signal n_mcx_state   : mcx_state_t;
  signal n_mcx_cmd_wr  : std_logic;
  signal n_mcx_cmd_rd  : std_logic;
  signal n_mcx_mst_wr  : std_logic;
  signal n_mcx_blen    : adb3_ocp_burst_s;
  signal n_mcx_tag     : adb3_ocp_tag_s;
  signal n_mcx_blast   : std_logic;
  signal n_mcx_addr    : adb3_ocp_addr_s;
  signal n_mcx_data    : adb3_ocp_data_s;
  signal n_mcx_we      : adb3_ocp_be_s;

  signal n_mcx_bstart  : std_logic;


  -- Read data
  signal slv_resp_sr    : std_logic_vector(MAX_CMDS-1 downto 0);
  signal slv_resp_busy  : std_logic;
  signal slv_resp_conv  : std_logic;
  signal slv_resp_val   : std_logic;
  signal slv_resp_index : natural;


  -- Slave read resp FIFO signals
  -- Depth of 5 => 32 entries
  -- 32 is required so they don't fill up due to max latency = 32
  constant SLV_RDATA_DEPTH : natural := 5;
  constant SLV_RDATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH;

  signal slv_rdata_wr       : std_logic;
  signal slv_rdata_adv      : std_logic;
  signal slv_rdata_not_full : std_logic;
  signal slv_rdata_valid    : std_logic;
  signal slv_rdata_d        : std_logic_vector(SLV_RDATA_WIDTH - 1 downto 0);
  signal slv_rdata_q        : std_logic_vector(SLV_RDATA_WIDTH - 1 downto 0);

  -- Read burst tracking FIFO signals
  -- FIFO should be same size as the read resp data FIFO
  constant RBT_DEPTH : natural := SLV_RDATA_DEPTH;
  constant RBT_WIDTH : natural := ADB3_OCP_TAG_WIDTH;

  signal rbt_wr       : std_logic;
  signal rbt_adv      : std_logic;
  signal rbt_not_full : std_logic;
  signal rbt_valid    : std_logic;
  signal rbt_d        : std_logic_vector(RBT_WIDTH-1 downto 0);
  signal rbt_q        : std_logic_vector(RBT_WIDTH-1 downto 0);
  signal rbt_q_tag    : adb3_ocp_tag_s;

  -- Delayed resp_valid
  signal resp_valid_vec : std_logic_vector(read_latency-1 downto 0);
  signal resp_valid_del : std_logic;


  -- Debug
  signal fifo_err : std_logic_vector(31 downto 0) := (others => '0');

  component adb3_ocp_bus_if_fifo
  generic(
    ORDER     : in    natural range 1 to 6;
    WIDTH     : in    natural;
    EMPTY_VAL : in    std_logic := '0';
    EMPTY_LO  : in    natural := 0;
    EMPTY_HI  : in    natural := 0);
  port(
    rst       : in    std_logic;
    clk       : in    std_logic;
    wr        : in    std_logic;
    d         : in    std_logic_vector(WIDTH - 1 downto 0);
    adv       : in    std_logic;
    full      : out   std_logic;
    not_full  : out   std_logic;
    valid     : out   std_logic;
    q         : out   std_logic_vector(WIDTH - 1 downto 0));
  end component;

begin

  ---------------------------------------------------------------------------
  -- Check generics
  ---------------------------------------------------------------------------

  -- Check that read_latency is > 0
  assert not(read_latency < 1)
    report "(adb3_ocp_simple_bus_if_nb): Mimium read_latency is 1 clock cycle"
    severity failure;

  -- Check that data width is power of 2
  assert not((data_width /= 128)and(data_width /= 64)and(data_width /= 32)and(data_width /= 16)and(data_width /= 8))
    report "(adb3_ocp_simple_bus_if_nb): Generic data_width must be 8/16/32/64/128 bits"
    severity failure;


  ---------------------------------------------------------------------------
  -- Command path
  ---------------------------------------------------------------------------

  -- Slave command FIFO
  -- Advance read pointer if:
  --   slave command FIFO is not empty
  --   command translation FSM indicates start of OCP command processing
  slv_cmd_wr  <= '1' when (slv_cmd_not_full = '1')and(ocp_m2s.Cmd /= OCP_CMD_IDLE) else '0';
  slv_cmd_d   <= ocp_m2s.Cmd & ocp_m2s.Tag & ocp_m2s.BurstLength & ocp_m2s.Addr;
  slv_cmd_adv <= slv_cmd_valid and n_mcx_bstart;

  slv_cmd_fifo_i : adb3_ocp_bus_if_fifo
  generic map(
    ORDER     => SLV_CMD_DEPTH,
    WIDTH     => SLV_CMD_WIDTH)
  port map(
    rst      => ocp_rst,
    clk      => ocp_clk,
    wr       => slv_cmd_wr,
    d        => slv_cmd_d,
    adv      => slv_cmd_adv,
    full     => open,
    not_full => slv_cmd_not_full,
    valid    => slv_cmd_valid,
    q        => slv_cmd_q);

  ocp_s2m.CmdAccept <= slv_cmd_not_full;

  slv_cmd <= (Cmd         => slv_cmd_q(SLV_CMD_WIDTH-1 downto ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH),
              Tag         => slv_cmd_q(ADB3_OCP_TAG_WIDTH+ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH),
              BurstLength => slv_cmd_q(ADB3_OCP_BURST_WIDTH+ADB3_OCP_ADDR_WIDTH-1 downto ADB3_OCP_ADDR_WIDTH),
              Addr        => slv_cmd_q(ADB3_OCP_ADDR_WIDTH-1 downto 0));

  slv_cmd_q_wr <= slv_cmd.Cmd(0);
  slv_cmd_q_rd <= slv_cmd.Cmd(1);

  -- Error checking
  slv_cmd_err_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (slv_cmd_valid = '0')and(n_mcx_bstart = '1') then
        fifo_err(0) <= '1';
        assert false
          report "(adb3_ocp_simple_bus_if_nb): Attempt to read from slv_cmd_fifo when empty"
          severity error;
      end if;
    end if;
  end process;


  -- Command/write data width conversion
  -- There are 2 cases: BUS_CMDS = 1 (full width), and BUS_CMDS > 1
  conv_width_cmd_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      mst_cmd_sr  <= (others => '0');
      mst_cmd_del <= '0';
      mst_a  <= (others => '-');
      mst_w  <= '0';
      mst_r  <= '0';
      mst_d  <= (others => '-');
      mst_we <= (others => '0');
    elsif rising_edge(ocp_clk) then
      -- Hold mcx_mst_wr until we are not busy
      if (mcx_mst_wr = '1')and(mst_cmd_busy = '1') then
        mst_cmd_del <= '1';
      elsif (mst_cmd_busy = '0') then
        mst_cmd_del <= '0';
      end if;

      -- Use direct or delayed command when not busy to initiate conversion
      if ((mcx_mst_wr = '1')or(mst_cmd_del = '1'))and(mst_cmd_busy = '0') then -- Convert 1st command
        if (BUS_CMDS = 1) then
          mst_cmd_sr <= (others => '0'); -- Full width, always zero
        else
          mst_cmd_sr <= std_logic_vector(to_unsigned(1,16)); -- Conv width
        end if;
        mst_a  <= mcx_addr(addr_width-1 downto 0);
        mst_w  <= mcx_cmd_wr;
        mst_r  <= mcx_cmd_rd;
        mst_d  <= mcx_data;
        mst_we <= mcx_we;
      elsif ((mst_w = '1')or(mst_r = '1'))and(mst_cmd_busy = '1') then -- Convert remaining commands
        mst_cmd_sr <= mst_cmd_sr(14 downto 0) & not(mst_cmd_conv);
        mst_a  <= mst_a + data_width/8;
        mst_w  <= mst_w;
        mst_r  <= mst_r;
        mst_d  <= std_logic_vector(to_unsigned(0,data_width)) & mst_d(ADB3_OCP_DATA_WIDTH-1 downto data_width);
        mst_we <= std_logic_vector(to_unsigned(0,data_width/8)) & mst_we((ADB3_OCP_DATA_WIDTH/8)-1 downto data_width/8);
      else -- Idle
        mst_cmd_sr <= (others => '0');
        mst_a  <= (others => '-');
        mst_w  <= '0';
        mst_r  <= '0';
        mst_d  <= (others => '-');
        mst_we <= (others => '0');
      end if;
    end if;
  end process;

  mst_cmd_index <= 0 when (BUS_CMDS = 1) else BUS_CMDS-2;
  mst_cmd_conv  <= mst_cmd_sr(mst_cmd_index);
  mst_cmd_busy  <= mst_cmd_sr(0);  -- Full width, always zero

  a  <= mst_a;
  w  <= mst_w;
  r  <= mst_r;
  d  <= mst_d(data_width-1 downto 0);
  we <= mst_we((data_width/8)-1 downto 0);

  -- Delay response valid by 'read_latency' cycles
  del_resp_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      resp_valid_vec <= (others => '0');
    elsif rising_edge(ocp_clk) then
      resp_valid_vec <= resp_valid_vec(read_latency-2 downto 0) & mst_r;
    end if;
  end process;

  resp_valid_del <= resp_valid_vec(read_latency-1);


  -- Read burst tracking FIFO
  -- Stores FSM output mcx_tag on mcx_cmd_rd during read command
  -- Advance read when we have read resp data from slv_rdata FIFO
  -- rbt_not_full used to enable command translation FSM generation of read commands
  -- rbt FIFO will become full if the number of read commands in flight is the same as its size
  -- rbt FIFO is the same size as the slave rdata FIFO
  rbt_wr  <= n_mcx_mst_wr and n_mcx_cmd_rd;
  rbt_d   <= n_mcx_tag;
  rbt_adv <= rbt_valid and slv_rdata_adv;

  rbt_fifo : adb3_ocp_bus_if_fifo
  generic map(
    ORDER => RBT_DEPTH,
    WIDTH => RBT_WIDTH)
  port map(
    rst      => ocp_rst,
    clk      => ocp_clk,
    wr       => rbt_wr,
    d        => rbt_d,
    adv      => rbt_adv,
    full     => open,
    not_full => rbt_not_full,
    valid    => rbt_valid,
    q        => rbt_q);

  rbt_q_tag <= rbt_q(ADB3_OCP_TAG_WIDTH-1 downto 0);

  -- Error checking
  rbt_err_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (rbt_wr = '1')and(rbt_not_full = '0') then
        fifo_err(4) <= '1';
        assert false
          report "(adb3_ocp_simple_bus_if_nb): Attempt to write to rbt_fifo when full"
          severity error;
      end if;

      if (rbt_valid = '0')and(slv_rdata_adv = '1') then
        fifo_err(5) <= '1';
        assert false
          report "(adb3_ocp_simple_bus_if_nb): Attempt to read from rbt_fifo when empty"
          severity error;
      end if;
    end if;
  end process;


  ---------------------------------------------------------------------------
  -- Write data path
  ---------------------------------------------------------------------------

  -- Slave write data FIFO
  -- Advance read pointer if:
  --   slave write data FIFO is not empty
  --   command translation FSM indicates we are writing master write data
  slv_wdata_wr  <= slv_wdata_not_full and ocp_m2s.DataValid;
  slv_wdata_d   <= ocp_m2s.DataByteEn & ocp_m2s.Data;
  slv_wdata_adv <= slv_wdata_valid and n_mcx_mst_wr and n_mcx_cmd_wr;

  slv_wdata_fifo_i : adb3_ocp_bus_if_fifo
  generic map(
    ORDER     => SLV_WDATA_DEPTH,
    WIDTH     => SLV_WDATA_WIDTH)
  port map(
    rst      => ocp_rst,
    clk      => ocp_clk,
    wr       => slv_wdata_wr,
    d        => slv_wdata_d,
    adv      => slv_wdata_adv,
    full     => open,
    not_full => slv_wdata_not_full,
    valid    => slv_wdata_valid,
    q        => slv_wdata_q);

  ocp_s2m.DataAccept <= slv_wdata_not_full;

  slv_wdata <= (DataByteEn => slv_wdata_q(SLV_WDATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH),
                Data       => slv_wdata_q(ADB3_OCP_DATA_WIDTH-1 downto 0));

  -- Error checking
  slv_wdata_err_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (slv_wdata_valid = '0')and(n_mcx_mst_wr = '1')and(n_mcx_cmd_wr = '1') then
        fifo_err(9) <= '1';
        assert false
          report "(adb3_ocp_simple_bus_if_nb): Attempt to read from slv_wdata_fifo when empty"
          severity error;
      end if;
    end if;
  end process;


  ---------------------------------------------------------------------------
  -- Command Translation FSM
  ---------------------------------------------------------------------------

  gen_mcx_state : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      mcx_state   <= mcx_s_idle;
      mcx_cmd_wr  <= '0';
      mcx_cmd_rd  <= '0';
      mcx_blen    <= (others => '-');
      mcx_tag     <= (others => '-');
      mcx_blast   <= '-';
      mcx_mst_wr  <= '0';
      mcx_addr    <= (others => '-');
      mcx_data    <= (others => '-');
      mcx_we      <= (others => '0');
    elsif rising_edge(ocp_clk) then
      mcx_state   <= n_mcx_state;
      mcx_cmd_wr  <= n_mcx_cmd_wr;
      mcx_cmd_rd  <= n_mcx_cmd_rd;
      mcx_blen    <= n_mcx_blen;
      mcx_tag     <= n_mcx_tag;
      mcx_blast   <= n_mcx_blast;
      mcx_mst_wr  <= n_mcx_mst_wr;
      mcx_addr    <= n_mcx_addr;
      mcx_data    <= n_mcx_data;
      mcx_we      <= n_mcx_we;
    end if;
  end process;


  gen_n_mcx_state : process(
    mcx_state,
    mcx_cmd_wr,
    mcx_cmd_rd,
    mcx_blen,
    mcx_tag,
    mcx_blast,
    mcx_addr,
    mcx_data,
    mcx_we,
    slv_cmd_valid,
    slv_cmd,
    slv_cmd_q_wr,
    slv_cmd_q_rd,
    slv_wdata_valid,
    slv_wdata,
    mst_cmd_busy,
    rbt_not_full)
  begin
    case mcx_state is
      when mcx_s_idle =>
        -- First data burst in first command, or return to mcx_s_idle
        if (slv_cmd_valid = '1')and(slv_cmd_q_wr = '1')and(slv_wdata_valid = '1')and(mst_cmd_busy = '0') then
          -- Write cycle with 1 data word available
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '1';
          n_mcx_blen   <= slv_cmd.BurstLength;
          n_mcx_tag    <= slv_cmd.Tag;
          n_mcx_cmd_wr <= '1';
          n_mcx_cmd_rd <= '0';
          n_mcx_mst_wr <= '1';
          n_mcx_addr   <= slv_cmd.Addr;  -- Byte addressing
          n_mcx_data   <= slv_wdata.Data;
          n_mcx_we     <= slv_wdata.DataByteEn;

          if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            n_mcx_blast <= '1';
          else
            n_mcx_blast <= '0';
          end if;

        elsif (slv_cmd_valid = '1')and(slv_cmd_q_rd = '1')and(rbt_not_full = '1')and(mst_cmd_busy = '0') then
          -- Read cycle
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '1';
          n_mcx_blen   <= slv_cmd.BurstLength;
          n_mcx_tag    <= slv_cmd.Tag;
          n_mcx_cmd_wr <= '0';
          n_mcx_cmd_rd <= '1';
          n_mcx_mst_wr <= '1';
          n_mcx_addr   <= slv_cmd.Addr;  -- Byte addressing
          n_mcx_data   <= (others => '-');
          n_mcx_we     <= (others => '0');

          if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
            n_mcx_blast <= '1';
          else
            n_mcx_blast <= '0';
          end if;
        else
          -- Return to mcx_s_idle state
          n_mcx_state  <= mcx_s_idle;
          n_mcx_bstart <= '0';
          n_mcx_blen   <= (others => '-');
          n_mcx_tag    <= (others => '-');
          n_mcx_cmd_wr <= '0';
          n_mcx_cmd_rd <= '0';
          n_mcx_mst_wr <= '0';
          n_mcx_addr   <= (others => '-');
          n_mcx_data   <= (others => '-');
          n_mcx_we     <= (others => '0');
          n_mcx_blast  <= '-';
        end if;

      -- Process burst of active command in command FIFO
      when mcx_s_burst =>
        -- First data burst in next command, or next/last data burst in current command
        if (mcx_blast = '1') then
          -- First data burst in next command, or return to mcx_s_idle
          if (slv_cmd_valid = '1')and(slv_cmd_q_wr = '1')and(slv_wdata_valid = '1')and(mst_cmd_busy = '0') then
            -- Write cycle with data word available
            n_mcx_state  <= mcx_s_burst;
            n_mcx_bstart <= '1';
            n_mcx_blen   <= slv_cmd.BurstLength;
            n_mcx_tag    <= slv_cmd.Tag;
            n_mcx_cmd_wr <= '1';
            n_mcx_cmd_rd <= '0';
            n_mcx_mst_wr <= '1';
            n_mcx_addr   <= slv_cmd.Addr;  -- Byte addressing
            n_mcx_data   <= slv_wdata.Data;
            n_mcx_we     <= slv_wdata.DataByteEn;

            if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;
          elsif (slv_cmd_valid = '1')and(slv_cmd_q_rd = '1')and(rbt_not_full = '1')and(mst_cmd_busy = '0') then
            -- Read cycle
            n_mcx_state  <= mcx_s_burst;
            n_mcx_bstart <= '1';
            n_mcx_blen   <= slv_cmd.BurstLength;
            n_mcx_tag    <= slv_cmd.Tag;
            n_mcx_cmd_wr <= '0';
            n_mcx_cmd_rd <= '1';
            n_mcx_mst_wr <= '1';
            n_mcx_addr   <= slv_cmd.Addr;  -- Byte addressing
            n_mcx_data   <= (others => '-');
            n_mcx_we     <= (others => '0');

            if (slv_cmd.BurstLength = std_logic_vector(to_unsigned(1,ADB3_OCP_BURST_WIDTH))) then
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;
          elsif (mst_cmd_busy = '0') then
            -- Return to mcx_s_idle state when conversion is finished
            n_mcx_state  <= mcx_s_idle;
            n_mcx_bstart <= '0';
            n_mcx_blen   <= (others => '-');
            n_mcx_tag    <= (others => '-');
            n_mcx_cmd_wr <= '0';
            n_mcx_cmd_rd <= '0';
            n_mcx_mst_wr <= '0';
            n_mcx_addr   <= (others => '-');
            n_mcx_data   <= (others => '-');
            n_mcx_we     <= (others => '0');
            n_mcx_blast  <= '-';
          else
            -- Wait until conditions are satisfied
            n_mcx_state  <= mcx_s_burst;
            n_mcx_bstart <= '0';
            n_mcx_cmd_wr <= mcx_cmd_wr;
            n_mcx_cmd_rd <= mcx_cmd_rd;
            n_mcx_mst_wr <= '0';
            n_mcx_tag    <= mcx_tag;
            n_mcx_blen   <= mcx_blen;
            n_mcx_addr   <= mcx_addr;
            n_mcx_data   <= mcx_data;
            n_mcx_we     <= mcx_we;
            n_mcx_blast  <= mcx_blast;
          end if;
        else
          n_mcx_state  <= mcx_s_burst;
          n_mcx_bstart <= '0';
          n_mcx_tag    <= mcx_tag;
          n_mcx_cmd_wr <= mcx_cmd_wr;
          n_mcx_cmd_rd <= mcx_cmd_rd;

          -- Next/last data burst in current command, or wait until conditions are satisfied
          if (mcx_cmd_wr = '1')and(slv_wdata_valid = '1')and(mst_cmd_busy = '0') then
            -- Write cycle with data word available
            n_mcx_blen   <= mcx_blen-1;
            n_mcx_addr   <= mcx_addr + 16;  -- byte addressing
            n_mcx_data   <= slv_wdata.Data;
            n_mcx_we     <= slv_wdata.DataByteEn;
            n_mcx_mst_wr <= '1';

            if (mcx_blen = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then -- Last: BL = 2
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;
          elsif (mcx_cmd_rd = '1')and(rbt_not_full = '1')and(mst_cmd_busy = '0') then
            -- Read cycle
            n_mcx_blen   <= mcx_blen-1;
            n_mcx_addr   <= mcx_addr + 16;  -- byte addressing
            n_mcx_data   <= (others => '-');
            n_mcx_we     <= (others => '0');
            n_mcx_mst_wr <= '1';

            if (mcx_blen = std_logic_vector(to_unsigned(2,ADB3_OCP_BURST_WIDTH))) then -- Last: BL = 2
              n_mcx_blast <= '1';
            else
              n_mcx_blast <= '0';
            end if;
          else
            -- Wait until conditions are satisfied
            n_mcx_blen   <= mcx_blen;
            n_mcx_addr   <= mcx_addr;
            n_mcx_data   <= mcx_data;
            n_mcx_we     <= mcx_we;
            n_mcx_mst_wr <= '0';
            n_mcx_blast  <= mcx_blast;
          end if;
        end if;
    end case;
  end process;


  ---------------------------------------------------------------------------
  -- Read resp data path
  ---------------------------------------------------------------------------

  conv_width_resp_p : process(ocp_rst,ocp_clk)
  begin
    if (ocp_rst = '1') then
      slv_resp_sr <= (others => '0');
      slv_q <= (others => '-');
      slv_resp_val <= '0';
    elsif rising_edge(ocp_clk) then
      if (resp_valid_del = '1')and(slv_resp_busy = '0') then -- Convert 1st response
        if (BUS_CMDS = 1) then
          slv_resp_sr <= (others => '0'); -- Full width, always zero
          slv_resp_val <= '1';
        else
          slv_resp_sr <= std_logic_vector(to_unsigned(1,16));
          slv_resp_val <= '0';
        end if;
        slv_q(ADB3_OCP_DATA_WIDTH-1 downto ADB3_OCP_DATA_WIDTH-data_width) <= q;
      elsif (resp_valid_del = '1')and(slv_resp_busy = '1') then -- Convert remaining responses
        slv_resp_sr <= slv_resp_sr(14 downto 0) & not(slv_resp_conv);
        slv_q <= q & slv_q(ADB3_OCP_DATA_WIDTH-1 downto data_width);
        slv_resp_val <= slv_resp_conv;
      else
        slv_resp_sr <= (others => '0');
        slv_q <= (others => '-');
        slv_resp_val <= '0';
      end if;
    end if;
  end process;

  slv_resp_index <= 0 when (BUS_CMDS = 1) else BUS_CMDS-2;
  slv_resp_conv  <= slv_resp_sr(slv_resp_index);
  slv_resp_busy  <= slv_resp_sr(0); -- Full width, always zero


  -- OCP read resp FIFO
  -- Must accept read resp data if available
  -- Advance read pointer if:
  --   read resp FIFO is not empty
  --   slave s2m OCP is ready to receive resp data
  slv_rdata_wr  <= slv_resp_val;
  slv_rdata_d   <= slv_q;
  slv_rdata_adv <= slv_rdata_valid and ocp_m2s.RespAccept;

  slv_rdata_fifo_i : adb3_ocp_bus_if_fifo
  generic map(
    ORDER     => SLV_RDATA_DEPTH,
    WIDTH     => SLV_RDATA_WIDTH)
  port map(
    rst      => ocp_rst,
    clk      => ocp_clk,
    wr       => slv_rdata_wr,
    d        => slv_rdata_d,
    adv      => slv_rdata_adv,
    full     => open,
    not_full => slv_rdata_not_full,
    valid    => slv_rdata_valid,
    q        => slv_rdata_q);

  ocp_s2m.Resp <= OCP_RESP_VALID when (slv_rdata_valid = '1') else OCP_RESP_NONE;
  ocp_s2m.Data <= slv_rdata_q;
  ocp_s2m.Tag  <= rbt_q_tag;

  -- Error checking
  slv_rdata_err_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (slv_rdata_wr = '1')and(slv_rdata_not_full = '0') then
        fifo_err(13) <= '1';
        assert false
          report "(adb3_ocp_simple_bus_if_nb): Attempt to write to slv_rdata_fifo when nearly full"
          severity error;
      end if;
    end if;
  end process;


  -- Debug signals
  debug_info <= fifo_err when DEBUG_ON else (others => '0');
end;