--==============================================================================
--
--     File Name: adb3_ocp_split_nb.vhd
--       Version: 0.3
--          Date: 25 Jul 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 OCP address space de-mux (non-blocking)
--
--   Description: Component giving single OCP slave port control over
--                multiple OCP master ports
--                De-muxes OCP address space using address range table
--                generic input
--
--   Limitations:
--
--         Notes: Non-blocking component
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 02 Dec 2010: Initial design
--                0.2, TW,  09 Feb 2011:
--                     - Rewritten to improve timing and eliminate dead cycles
--                       between bursts. Allows multiple commands to be posted.
--                0.3, KDC, 25 Jul 2011:
--                     - Changed type of addr_range_table from addr_range_table_t
--                       to adb3_ocp_addr_range_table_t
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

--
-- A synchronous FIFO used to buffer commands, command attributes, data, byte
-- enables etc. in 'adb3_ocp_split_nb'.
--
-- 'ORDER' is log2 of the FIFO depth, e.g. ORDER = 4 => depth is 16 locations.
--
-- 'WIDTH' is the width of FIFO locations, in bits.
--
-- 'EMPTY_VAL' is the value driven on 'q(EMPTY_LO:EMPTY_HI)' when 'valid' is
-- not asserted; map both 'EMPTY_LO' and 'EMPTY_HI' to 0 if this feature is
-- not required.
--
entity adb3_ocp_split_nb_fifo is
generic(
  ORDER       : in    natural range 2 to 6;
  WIDTH       : in    natural;
  EMPTY_VAL   : in    std_logic := '0';
  EMPTY_LO    : in    natural := 0;
  EMPTY_HI    : in    natural := 0);
port(
  rst         : in    std_logic;                             -- Asynchronous reset
  clk         : in    std_logic;                             -- Clock
  wr          : in    std_logic;                             -- '1' => write value on 'd' into FIFO
  d           : in    std_logic_vector(WIDTH - 1 downto 0);  -- Item to write into FIFO, qualified by 'wr'
  adv         : in    std_logic;                             -- '1' => present next item on 'q' to next item
  full        : out   std_logic;                             -- '1' => full
  not_full    : out   std_logic;                             -- '1' => not full
  valid       : out   std_logic;                             -- '1' => not empty
  q           : out   std_logic_vector(WIDTH - 1 downto 0)); -- Current item, qualified by 'valid'
end;

architecture syn of adb3_ocp_split_nb_fifo is

  constant DEPTH : natural := 2**ORDER;
  constant MAX_LEVEL : natural := DEPTH - 1;

  subtype ram_data_t is std_logic_vector(WIDTH - 1 downto 0);
  type ram_array_t is array(0 to DEPTH - 1) of ram_data_t;

  signal wptr, rptr : std_logic_vector(ORDER - 1 downto 0);
  signal rn : std_logic_vector(ORDER - 1 downto 0);
  signal not_empty : std_logic;
  signal i_full, i_not_full : std_logic;

  signal qvalid : std_logic;
  signal qce : std_logic;
  signal rd : std_logic;

  signal ram : ram_array_t;
  attribute RAM_STYLE : string;
  attribute RAM_STYLE of ram : signal is "distributed";
  signal sel_ram : ram_data_t;

begin

  valid <= qvalid;
  full <= i_full;
  not_full <= i_not_full;

  write_logic : process(rst, clk)
    variable tmp : std_logic_vector(ORDER - 1 downto 0);
  begin
    if rst = '1' then
      wptr <= (others => '0');
      i_not_full <= '1';
      i_full <= '0';
    elsif rising_edge(clk) then
      if wr = '1' then
        wptr <= wptr + 1;
      end if;

      if wr = '1' then
        assert CONV_INTEGER(wptr - rptr) < MAX_LEVEL severity failure;
        tmp := wptr - rptr + 1;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      else
        tmp := wptr - rptr;
        i_full <= tmp(ORDER - 1);
        i_not_full <= NOT tmp(ORDER - 1);
      end if;
    end if;

    if rising_edge(clk) then
      if wr = '1' then
        ram(conv_integer(wptr)) <= d;
      end if;
    end if;
  end process;

  not_empty <= OR_reduce(rn);
  qce <= adv OR NOT qvalid;
  rd <= (adv OR NOT qvalid) AND not_empty;

  read_logic : process(rst, clk)
  begin
    if rst = '1' then
      rptr <= (others => '0');
    elsif rising_edge(clk) then
      if rd = '1' then
        rptr <= rptr + 1;
      end if;
    end if;

    if rising_edge(clk) then
      if rd = '1' then
        assert CONV_INTEGER(wptr - rptr) > 0 severity failure;
        rn <= wptr - rptr - 1;
      else
        rn <= wptr - rptr;
      end if;
    end if;
  end process;

  output_valid : process(rst, clk)
  begin
    if rst = '1' then
      qvalid <= '0';
    elsif rising_edge(clk) then
      if qce = '1' then
        qvalid <= not_empty;
      end if;
    end if;
  end process;

  sel_ram <= ram(conv_integer(rptr));

  output_data : for i in 0 to WIDTH - 1 generate
    with_reset : if EMPTY_LO <= i AND i < EMPTY_HI generate
      U0 : process(rst, clk)
      begin
        if rst = '1' then
          q(i) <= EMPTY_VAL;
        elsif rising_edge(clk) then
          if qce = '1' then
            if not_empty = '1' then
              q(i) <= sel_ram(i);
            else
              q(i) <= EMPTY_VAL;
            end if;
          end if;
        end if;
      end process;
    end generate;

    without_reset : if NOT(EMPTY_LO <= i AND i < EMPTY_HI) generate
      U0 : process(clk)
      begin
        if rising_edge(clk) then
          if qce = '1' then
            q(i) <= sel_ram(i);
          end if;
        end if;
      end process;
    end generate;
  end generate;

end;

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;

entity adb3_ocp_split_nb is
generic (
  addr_range_table : adb3_ocp_addr_range_table_t;
  error_data       : adb3_ocp_data_s := X"DEADC0DEDEADC0DEDEADC0DEDEADC0DE");
port(
  ocp_rst     : in    std_logic;
  ocp_clk     : in    std_logic;
  -- OCP primary port
  slave_m2s   : in    adb3_ocp_m2sT;
  slave_s2m   : out   adb3_ocp_s2mT;
  -- OCP secondary ports
  masters_m2s : out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1);
  masters_s2m : in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1);
  -- Debug
  debug_info  : out   std_logic_vector(31 downto 0));
end;

architecture syn of adb3_ocp_split_nb is

  -- Zero extend 'v' to 'n' bits
  function extend(
    constant v : std_logic_vector;
    constant n : natural)
  return std_logic_vector is
    variable tmp1 : std_logic_vector(v'length - 1 downto 0);
    variable tmp2 : std_logic_vector(n - 1 downto 0);
  begin
    tmp1 := v;
    if n <= tmp1'length then
      return tmp1(n - 1 downto 0);
    else
      tmp2(v'length - 1 downto 0) := v;
      tmp2(n - 1 downto v'length) := (others => '0');
      return tmp2;
    end if;
  end;

  function conv_string_hexdigit(
    constant val : in std_logic_vector)
  return character is
    variable tmp : std_logic_vector(3 downto 0);
  begin
    assert val'length = 4
      report "conv_string_hex_digit: std_logic_vector passed has invalid length (must be 4)"
      severity failure;

    tmp := val;
    case tmp is
      when "0000" =>
        return '0';
      when "0001" =>
        return '1';
      when "0010" =>
        return '2';
      when "0011" =>
        return '3';
      when "0100" =>
        return '4';
      when "0101" =>
        return '5';
      when "0110" =>
        return '6';
      when "0111" =>
        return '7';
      when "1000" =>
        return '8';
      when "1001" =>
        return '9';
      when "1010" =>
        return 'A';
      when "1011" =>
        return 'B';
      when "1100" =>
        return 'C';
      when "1101" =>
        return 'D';
      when "1110" =>
        return 'E';
      when "1111" =>
        return 'F';
      when others =>
        return 'X';
    end case;
  end;

  function conv_string_hex(
    constant val : in std_logic_vector)
  return string is
    variable u, l : integer;
    variable digit : std_logic_vector(3 downto 0);
    variable s, t : string(1 to 64);
    variable i : natural;
    variable tmp : std_logic_vector(val'length - 1 downto 0);
  begin
    tmp := val;
    i := 0;
    l := tmp'low;
    u := tmp'high + 1;

    while l < u loop
      if (u - l) < 4 then
        digit := extend(tmp(u - 1 downto l), 4);
      else
        digit := tmp(l + 3 downto l);
      end if;
      i := i + 1;
      s(i) := conv_string_hexdigit(digit);
      l := l + 4;
    end loop;

    assert i /= 0 report "conv_string_hex: a bad thing happened with i" severity failure;

    for j in 1 to i loop
      t(j) := s(i - j + 1);
    end loop;

    return t(1 to i);
  end;

  component adb3_ocp_split_nb_fifo
  generic(
    ORDER       : in    natural range 1 to 6;
    WIDTH       : in    natural;
    EMPTY_VAL   : in    std_logic := '0';
    EMPTY_LO    : in    natural := 0;
    EMPTY_HI    : in    natural := 0);
  port(
    rst         : in    std_logic;
    clk         : in    std_logic;
    wr          : in    std_logic;
    d           : in    std_logic_vector(WIDTH - 1 downto 0);
    adv         : in    std_logic;
    full        : out   std_logic;
    not_full    : out   std_logic;
    valid       : out   std_logic;
    q           : out   std_logic_vector(WIDTH - 1 downto 0));
  end component;

  constant NUM_SEC : natural := addr_range_table'length;

  -- One-hot vector representing destination secondary port for incoming command, with the top
  -- bit representing the notional 'address error port'
  subtype port_1h_t is std_logic_vector(NUM_SEC downto 0);
  signal sec_decode_1h : port_1h_t;
  signal sec_decode_error : std_logic := '0';

  -- Slave command FIFO signals
  constant SLV_CMD_ATTR_WIDTH : natural := ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;
  constant SLV_CMD_WIDTH : natural := ADB3_OCP_CMD_WIDTH + SLV_CMD_ATTR_WIDTH;
  signal slv_cmd_wr : std_logic;
  signal slv_cmd_adv : std_logic;
  signal slv_cmd_not_full : std_logic;
  signal slv_cmd_valid : std_logic;
  signal slv_cmd_d, slv_cmd_q : std_logic_vector(SLV_CMD_WIDTH - 1 downto 0);
    type slv_cmd_attr_t is record
    Cmd : ocp_CmdT;
    Addr : adb3_ocp_addr_s;
    BurstLength : adb3_ocp_burst_s;
    Tag : adb3_ocp_tag_s;
  end record;
  signal slv_cmd : slv_cmd_attr_t;

  -- Master command FIFOs
  constant MST_CMD_ATTR_WIDTH : natural := ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH;
  constant MST_CMD_WIDTH : natural := ADB3_OCP_CMD_WIDTH + MST_CMD_ATTR_WIDTH;
  type mst_cmd_d_array_t is array(natural range <>) of std_logic_vector(MST_CMD_WIDTH - 1 downto 0);
  signal mst_cmd_d, mst_cmd_q : mst_cmd_d_array_t(NUM_SEC - 1 downto 0);
  signal mst_cmd_wr : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_cmd_adv : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_cmd_not_full : std_logic_vector(NUM_SEC downto 0); -- one extra bit for address error handling
  signal mst_cmd_valid : std_logic_vector(NUM_SEC - 1 downto 0);

  -- Slave write burst tracking FIFO
  constant WBT_WIDTH : natural := ADB3_OCP_BURST_WIDTH + NUM_SEC + 1;
  type wbt_attr_t is record
    blen : adb3_ocp_burst_s;
    port_1h : port_1h_t;
  end record;
  signal wbt_d, wbt_q : std_logic_vector(WBT_WIDTH - 1 downto 0);
  signal wbt_wr : std_logic;
  signal wbt_adv : std_logic;
  signal wbt_not_full : std_logic;
  signal wbt_valid : std_logic;
  signal wbt_q_rec : wbt_attr_t;

  -- Slave read burst tracking FIFO
  constant RBT_WIDTH : natural := ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH + NUM_SEC + 1;
  type rbt_attr_t is record
    blen : adb3_ocp_burst_s;
    tag : adb3_ocp_tag_s;
    port_1h : port_1h_t;
  end record;
  signal rbt_d, rbt_q : std_logic_vector(RBT_WIDTH - 1 downto 0);
  signal rbt_wr : std_logic;
  signal rbt_adv : std_logic;
  signal rbt_not_full : std_logic;
  signal rbt_valid : std_logic;
  signal rbt_q_rec : rbt_attr_t;

  -- Write data transfer control
  type wdx_state_t is (wdx_s_idle, wdx_s_burst);
  signal wdx_state, n_wdx_state : wdx_state_t;
  signal wdx_blen, n_wdx_blen : adb3_ocp_burst_s;
  signal wdx_blast, n_wdx_blast : std_logic;
  signal wdx_port_1h, n_wdx_port_1h : port_1h_t;
  signal n_wdx_bstart : std_logic;

  -- Read data transfer control
  type rdx_state_t is (rdx_s_idle, rdx_s_burst);
  signal rdx_state, n_rdx_state : rdx_state_t;
  signal rdx_blen, n_rdx_blen : adb3_ocp_burst_s;
  signal rdx_blast, n_rdx_blast : std_logic;
  signal rdx_tag, n_rdx_tag : adb3_ocp_tag_s;
  signal rdx_port_1h, n_rdx_port_1h : port_1h_t;
  signal n_rdx_bstart : std_logic;

  -- Slave write data FIFO
  constant SLV_WDATA_WIDTH : natural := ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;
  signal slv_wdata_d, slv_wdata_q : std_logic_vector(SLV_WDATA_WIDTH - 1 downto 0);
  signal slv_wdata_wr : std_logic;
  signal slv_wdata_adv : std_logic;
  signal slv_wdata_not_full : std_logic;
  signal slv_wdata_valid : std_logic;

  -- Master write data FIFOs
  constant MST_WDATA_WIDTH : natural := ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH;
  type mst_wdata_d_array_t is array(natural range <>) of std_logic_vector(MST_WDATA_WIDTH - 1 downto 0);
  signal mst_wdata_d, mst_wdata_q : mst_wdata_d_array_t(0 to NUM_SEC - 1);
  signal mst_wdata_wr : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_wdata_adv : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_wdata_not_full : std_logic_vector(NUM_SEC downto 0); -- one extra bit for address error handling
  signal mst_wdata_valid : std_logic_vector(NUM_SEC - 1 downto 0);

  -- Slave read data FIFO
  constant SLV_RDATA_WIDTH : natural := ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH;
  signal slv_rdata_d, slv_rdata_q : std_logic_vector(SLV_RDATA_WIDTH - 1 downto 0);
  signal slv_rdata_wr : std_logic;
  signal slv_rdata_adv : std_logic;
  signal slv_rdata_not_full : std_logic;
  signal slv_rdata_valid : std_logic;

  -- Master read data FIFOs
  constant MST_RDATA_WIDTH : natural := ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH;
  type mst_rdata_d_array_t is array(natural range <>) of std_logic_vector(MST_RDATA_WIDTH - 1 downto 0);
  signal mst_rdata_d : mst_rdata_d_array_t(0 to NUM_SEC - 1);
  signal mst_rdata_q : mst_rdata_d_array_t(0 to NUM_SEC); -- one extra element for address error handling
  signal mst_rdata_q_sel : std_logic_vector(MST_RDATA_WIDTH - 1 downto 0);
  signal mst_rdata_wr : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_rdata_adv : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_rdata_not_full : std_logic_vector(NUM_SEC - 1 downto 0);
  signal mst_rdata_valid : std_logic_vector(NUM_SEC downto 0); -- one extra bit for address error handling

begin

  debug_info <= (others => '0');

  ---------------------------------------------------------------------------
  -- Command path
  ---------------------------------------------------------------------------

  slv_cmd_wr <= '1' when slv_cmd_not_full = '1' AND slave_m2s.Cmd /= OCP_CMD_IDLE else '0';
  slv_cmd_d <= slave_m2s.Cmd & slave_m2s.Tag & slave_m2s.BurstLength & slave_m2s.Addr;
  slv_cmd_adv <= slv_cmd_valid AND OR_reduce(sec_decode_1h AND mst_cmd_not_full) AND rbt_not_full AND wbt_not_full;

  -- Slave command FIFO
  slv_cmd_fifo : adb3_ocp_split_nb_fifo
  generic map(
    ORDER => 3,
    WIDTH => SLV_CMD_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => slv_cmd_wr,
    d => slv_cmd_d,
    adv => slv_cmd_adv,
    full => open,
    not_full => slv_cmd_not_full,
    valid => slv_cmd_valid,
    q => slv_cmd_q);

  slave_s2m.CmdAccept <= slv_cmd_not_full;

  slv_cmd <= (Cmd => slv_cmd_q(SLV_CMD_WIDTH - 1 downto SLV_CMD_ATTR_WIDTH),
              Tag => slv_cmd_q(SLV_CMD_ATTR_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH),
              BurstLength => slv_cmd_q(ADB3_OCP_ADDR_WIDTH + ADB3_OCP_BURST_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH),
              Addr => slv_cmd_q(ADB3_OCP_ADDR_WIDTH - 1 downto 0));

  -- Decode address to determine destination secondary port (as a one-hot vector)
  decode_addr : process(slv_cmd, sec_decode_error)
  begin
    for i in 0 to NUM_SEC - 1 loop
      -- Asserted one-hot depending on OCP address of incoming command
      sec_decode_1h(i) <= AND_reduce((slv_cmd.Addr AND NOT addr_range_table(i).addr_mask) xnor
                                     (addr_range_table(i).addr_base AND NOT addr_range_table(i).addr_mask));
    end loop;
    sec_decode_1h(NUM_SEC) <= sec_decode_error;
  end process;

  -- Catch addresses that don't hit any of the address ranges in 'addr_range_table'
  check_addr : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      sec_decode_error <= '0';
    elsif rising_edge(ocp_clk) then
      if slv_cmd_valid = '1' AND OR_reduce(sec_decode_1h) = '0' AND rbt_not_full = '1' AND wbt_not_full = '1' then
        assert false report "adb3_ocp_split_nb: slave port address 0x" & conv_string_hex(slv_cmd.Addr) & " does not match any address range" severity warning;
        sec_decode_error <= '1';
      else
        sec_decode_error <= '0';
      end if;
    end if;
  end process;

  -- Master command FIFOs
  gen_mst_cmd_fifos : for i in 0 to NUM_SEC - 1 generate
    mst_cmd_d(i) <= slv_cmd.Cmd & slv_cmd.Tag & slv_cmd.BurstLength & slv_cmd.Addr;
    mst_cmd_wr(i) <= slv_cmd_valid AND sec_decode_1h(i) AND mst_cmd_not_full(i) AND rbt_not_full AND wbt_not_full;
    mst_cmd_adv(i) <= mst_cmd_valid(i) AND masters_s2m(i).CmdAccept;

    U0 : adb3_ocp_split_nb_fifo
    generic map(
      ORDER => 3,
      WIDTH => MST_CMD_WIDTH,
      EMPTY_VAL => '0',
      EMPTY_LO => MST_CMD_ATTR_WIDTH,
      EMPTY_HI => MST_CMD_WIDTH)
    port map(
      rst => ocp_rst,
      clk => ocp_clk,
      wr => mst_cmd_wr(i),
      d => mst_cmd_d(i),
      adv => mst_cmd_adv(i),
      full => open,
      not_full => mst_cmd_not_full(i),
      valid => mst_cmd_valid(i),
      q => mst_cmd_q(i));

    masters_m2s(i).Addr <= mst_cmd_q(i)(ADB3_OCP_ADDR_WIDTH - 1 downto 0);
    masters_m2s(i).BurstLength <= mst_cmd_q(i)(ADB3_OCP_BURST_WIDTH + ADB3_OCP_ADDR_WIDTH - 1 downto ADB3_OCP_ADDR_WIDTH);
    masters_m2s(i).Tag <= mst_cmd_q(i)(ADB3_OCP_TAG_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_ADDR_WIDTH - 1 downto ADB3_OCP_BURST_WIDTH + ADB3_OCP_ADDR_WIDTH);
    masters_m2s(i).Cmd <= mst_cmd_q(i)(ADB3_OCP_CMD_WIDTH + ADB3_OCP_TAG_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_ADDR_WIDTH - 1 downto ADB3_OCP_TAG_WIDTH + ADB3_OCP_BURST_WIDTH + ADB3_OCP_ADDR_WIDTH);
  end generate;

  -- The notional 'error command FIFO' is never full
  mst_cmd_not_full(NUM_SEC) <= '1';

  -- Generate write burst tracking FIFO
  wbt_d <= sec_decode_1h & slv_cmd.BurstLength;
  wbt_wr <= '1' when slv_cmd_adv = '1' AND (slv_cmd.Cmd = OCP_CMD_WRITE OR slv_cmd.Cmd = OCP_CMD_WRITENONPOST) else '0';
  wbt_adv <= n_wdx_bstart;

  wbt_fifo : adb3_ocp_split_nb_fifo
  generic map(
    ORDER => 3,
    WIDTH => WBT_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => wbt_wr,
    d => wbt_d,
    adv => wbt_adv,
    full => open,
    not_full => wbt_not_full,
    valid => wbt_valid,
    q => wbt_q);

  wbt_q_rec <= (port_1h => wbt_q(NUM_SEC + ADB3_OCP_BURST_WIDTH downto ADB3_OCP_BURST_WIDTH),
                blen => wbt_q(ADB3_OCP_BURST_WIDTH - 1 downto 0));

  rbt_d <= sec_decode_1h & slv_cmd.Tag & slv_cmd.BurstLength;
  rbt_wr <= '1' when slv_cmd_adv = '1' AND slv_cmd.Cmd = OCP_CMD_READ else '0';
  rbt_adv <= n_rdx_bstart;

  -- Generate read burst tracking FIFO
  rbt_fifo : adb3_ocp_split_nb_fifo
  generic map(
    ORDER => 3,
    WIDTH => RBT_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => rbt_wr,
    d => rbt_d,
    adv => rbt_adv,
    full => open,
    not_full => rbt_not_full,
    valid => rbt_valid,
    q => rbt_q);

  rbt_q_rec <= (port_1h => rbt_q(NUM_SEC + ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH downto ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH),
                tag => rbt_q(ADB3_OCP_BURST_WIDTH + ADB3_OCP_TAG_WIDTH - 1 downto ADB3_OCP_BURST_WIDTH),
                blen => rbt_q(ADB3_OCP_BURST_WIDTH - 1 downto 0));

  ---------------------------------------------------------------------------
  -- Write data path
  ---------------------------------------------------------------------------

  gen_wdx_state : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      wdx_state <= wdx_s_idle;
      wdx_blen <= (others => '-');
      wdx_blast <= '-';
      wdx_port_1h <= (others => '0');
    elsif rising_edge(ocp_clk) then
      wdx_state <= n_wdx_state;
      wdx_blen <= n_wdx_blen;
      wdx_blast <= n_wdx_blast;
      wdx_port_1h <= n_wdx_port_1h;
    end if;
  end process;

  gen_n_wdx_state : process(
    wdx_state,
    wdx_blen,
    wdx_blast,
    wdx_port_1h,
    slv_wdata_adv,
    wbt_valid,
    wbt_q_rec)
  begin
    case wdx_state is
      when wdx_s_idle =>
        if wbt_valid = '1' then
          n_wdx_state <= wdx_s_burst;
          n_wdx_blen <= wbt_q_rec.blen;
          n_wdx_blast <= NOT OR_reduce(wbt_q_rec.blen(wbt_q_rec.blen'high downto 1));
          n_wdx_port_1h <= wbt_q_rec.port_1h;
          n_wdx_bstart <= '1';
        else
          n_wdx_state <= wdx_s_idle;
          n_wdx_blen <= (others => '-');
          n_wdx_blast <= '-';
          n_wdx_port_1h <= (others => '0');
          n_wdx_bstart <= '0';
        end if;

      when wdx_s_burst =>
        if slv_wdata_adv = '1' then
          if wdx_blast = '1' then
            if wbt_valid = '1' then
              n_wdx_state <= wdx_s_burst;
              n_wdx_blen <= wbt_q_rec.blen;
              n_wdx_blast <= NOT OR_reduce(wbt_q_rec.blen(wbt_q_rec.blen'high downto 1));
              n_wdx_port_1h <= wbt_q_rec.port_1h;
              n_wdx_bstart <= '1';
            else
              n_wdx_state <= wdx_s_idle;
              n_wdx_blen <= (others => '-');
              n_wdx_blast <= '-';
              n_wdx_port_1h <= (others => '0');
              n_wdx_bstart <= '0';
            end if;
          else
            n_wdx_state <= wdx_s_burst;
            n_wdx_blen <= wdx_blen - 1;
            n_wdx_blast <= NOT OR_reduce(wdx_blen(wdx_blen'high downto 2) & wdx_blen(0));
            n_wdx_port_1h <= wdx_port_1h;
            n_wdx_bstart <= '0';
          end if;
        else
          n_wdx_state <= wdx_s_burst;
          n_wdx_blen <= wdx_blen;
          n_wdx_blast <= wdx_blast;
          n_wdx_port_1h <= wdx_port_1h;
          n_wdx_bstart <= '0';
        end if;
    end case;
  end process;

  slv_wdata_wr <= slv_wdata_not_full AND slave_m2s.DataValid;
  slv_wdata_d <= slave_m2s.DataByteEn & slave_m2s.Data;
  slv_wdata_adv <= slv_wdata_valid AND OR_reduce(wdx_port_1h AND mst_wdata_not_full);

  -- Slave write data FIFO
  slv_wdata_fifo : adb3_ocp_split_nb_fifo
  generic map(
    ORDER => 4,
    WIDTH => SLV_WDATA_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => slv_wdata_wr,
    d => slv_wdata_d,
    adv => slv_wdata_adv,
    full => open,
    not_full => slv_wdata_not_full,
    valid => slv_wdata_valid,
    q => slv_wdata_q);

  slave_s2m.DataAccept <= slv_wdata_not_full;

  -- Master write data FIFOs
  gen_mst_wdata_fifos : for i in 0 to NUM_SEC - 1 generate
    mst_wdata_d(i) <= slv_wdata_q;
    mst_wdata_wr(i) <= slv_wdata_valid AND wdx_port_1h(i) AND mst_wdata_not_full(i);
    mst_wdata_adv(i) <= mst_wdata_valid(i) AND masters_s2m(i).DataAccept;

    U0 : adb3_ocp_split_nb_fifo
    generic map(
      ORDER => 4,
      WIDTH => MST_WDATA_WIDTH)
    port map(
      rst => ocp_rst,
      clk => ocp_clk,
      wr => mst_wdata_wr(i),
      d => mst_wdata_d(i),
      adv => mst_wdata_adv(i),
      full => open,
      not_full => mst_wdata_not_full(i),
      valid => mst_wdata_valid(i),
      q => mst_wdata_q(i));

      masters_m2s(i).Data <= mst_wdata_q(i)(ADB3_OCP_DATA_WIDTH - 1 downto 0);
      masters_m2s(i).DataByteEn <= mst_wdata_q(i)(ADB3_OCP_BE_WIDTH + ADB3_OCP_DATA_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH);
      masters_m2s(i).DataValid <= mst_wdata_valid(i);
  end generate;

  -- The notional 'error port' always accepts data
  mst_wdata_not_full(NUM_SEC) <= '1';

  ---------------------------------------------------------------------------
  -- Read data path
  ---------------------------------------------------------------------------

  gen_rdx_state : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      rdx_state <= rdx_s_idle;
      rdx_blen <= (others => '-');
      rdx_blast <= '-';
      rdx_tag <= (others => '-');
      rdx_port_1h <= (others => '0');
    elsif rising_edge(ocp_clk) then
      rdx_state <= n_rdx_state;
      rdx_blen <= n_rdx_blen;
      rdx_blast <= n_rdx_blast;
      rdx_tag <= n_rdx_tag;
      rdx_port_1h <= n_rdx_port_1h;
    end if;
  end process;

  gen_n_rdx_state : process(
    rdx_state,
    rdx_blen,
    rdx_blast,
    rdx_tag,
    rdx_port_1h,
    slv_rdata_wr,
    rbt_valid,
    rbt_q_rec)
  begin
    case rdx_state is
      when rdx_s_idle =>
        if rbt_valid = '1' then
          n_rdx_state <= rdx_s_burst;
          n_rdx_blen <= rbt_q_rec.blen;
          n_rdx_blast <= NOT OR_reduce(rbt_q_rec.blen(rbt_q_rec.blen'high downto 1));
          n_rdx_tag <= rbt_q_rec.tag;
          n_rdx_port_1h <= rbt_q_rec.port_1h;
          n_rdx_bstart <= '1';
        else
          n_rdx_state <= rdx_s_idle;
          n_rdx_blen <= (others => '-');
          n_rdx_blast <= '-';
          n_rdx_tag <= (others => '-');
          n_rdx_port_1h <= (others => '0');
          n_rdx_bstart <= '0';
        end if;

      when rdx_s_burst =>
        if slv_rdata_wr = '1' then
          if rdx_blast = '1' then
            if rbt_valid = '1' then
              n_rdx_state <= rdx_s_burst;
              n_rdx_blen <= rbt_q_rec.blen;
              n_rdx_blast <= NOT OR_reduce(rbt_q_rec.blen(rbt_q_rec.blen'high downto 1));
              n_rdx_tag <= rbt_q_rec.tag;
              n_rdx_port_1h <= rbt_q_rec.port_1h;
              n_rdx_bstart <= '1';
            else
              n_rdx_state <= rdx_s_idle;
              n_rdx_blen <= (others => '-');
              n_rdx_blast <= '-';
              n_rdx_tag <= (others => '-');
              n_rdx_port_1h <= (others => '0');
              n_rdx_bstart <= '0';
            end if;
          else
            n_rdx_state <= rdx_s_burst;
            n_rdx_blen <= rdx_blen - 1;
            n_rdx_blast <= NOT OR_reduce(rdx_blen(rdx_blen'high downto 2) & rdx_blen(0));
            n_rdx_tag <= rdx_tag;
            n_rdx_port_1h <= rdx_port_1h;
            n_rdx_bstart <= '0';
          end if;
        else
          n_rdx_state <= rdx_s_burst;
          n_rdx_blen <= rdx_blen;
          n_rdx_blast <= rdx_blast;
          n_rdx_tag <= rdx_tag;
          n_rdx_port_1h <= rdx_port_1h;
          n_rdx_bstart <= '0';
        end if;
    end case;
  end process;

  -- Master read data FIFOs
  gen_mst_rdata_fifos : for i in 0 to NUM_SEC - 1 generate
    mst_rdata_wr(i) <= '1' when mst_rdata_not_full(i) = '1' AND masters_s2m(i).Resp = OCP_RESP_VALID else '0';
    mst_rdata_d(i) <= masters_s2m(i).Tag & masters_s2m(i).Data;
    mst_rdata_adv(i) <= rdx_port_1h(i) AND mst_rdata_valid(i) AND slv_rdata_not_full;

    mst_rdata_fifo : adb3_ocp_split_nb_fifo
    generic map(
      ORDER => 4,
      WIDTH => MST_RDATA_WIDTH)
    port map(
      rst => ocp_rst,
      clk => ocp_clk,
      wr => mst_rdata_wr(i),
      d => mst_rdata_d(i),
      adv => mst_rdata_adv(i),
      full => open,
      not_full => mst_rdata_not_full(i),
      valid => mst_rdata_valid(i),
      q => mst_rdata_q(i));

    masters_m2s(i).RespAccept <= mst_rdata_not_full(i);
  end generate;

  mst_rdata_valid(NUM_SEC) <= '1';
  mst_rdata_q(NUM_SEC) <= rdx_tag & error_data;

  gen_mst_rdata_q_sel : process(mst_rdata_q, rdx_port_1h)
    variable tmp : std_logic_vector(MST_RDATA_WIDTH - 1 downto 0);
  begin
    tmp := (others => '0');
    for i in 0 to NUM_SEC loop
      if rdx_port_1h(i) = '1' then
        tmp := tmp OR mst_rdata_q(i);
      end if;
    end loop;
    mst_rdata_q_sel <= tmp;
  end process;

  -- Verify that the tag returned by a master port for a read is the same as
  -- the tag from the original command received on the slave port.
  check_tags : process(ocp_clk)
    variable exp_tag : adb3_ocp_tag_s;
  begin
    if rising_edge(ocp_clk) then
      if slv_rdata_wr = '1' then
        exp_tag := slv_rdata_d(ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH);
        assert exp_tag = rdx_tag
          report "adb3_ocp_split_nb: tag mismatch entering data into slave read data FIFO, expected=0x" & conv_string_hex(exp_tag) & " actual=0x" & conv_string_hex(rdx_tag)
          severity warning;
      end if;
    end if;
  end process;

  -- Slave read data FIFO
  slv_rdata_wr <= slv_rdata_not_full AND OR_reduce(rdx_port_1h AND mst_rdata_valid);
  slv_rdata_d <= mst_rdata_q_sel;
  slv_rdata_adv <= slv_rdata_valid AND slave_m2s.RespAccept;

  U0 : adb3_ocp_split_nb_fifo
  generic map(
    ORDER => 4,
    WIDTH => SLV_RDATA_WIDTH)
  port map(
    rst => ocp_rst,
    clk => ocp_clk,
    wr => slv_rdata_wr,
    d => slv_rdata_d,
    adv => slv_rdata_adv,
    full => open,
    not_full => slv_rdata_not_full,
    valid => slv_rdata_valid,
    q => slv_rdata_q);

  slave_s2m.Data <= slv_rdata_q(ADB3_OCP_DATA_WIDTH - 1 downto 0);
  slave_s2m.Tag <= slv_rdata_q(ADB3_OCP_TAG_WIDTH + ADB3_OCP_DATA_WIDTH - 1 downto ADB3_OCP_DATA_WIDTH);
  slave_s2m.Resp <= "0" & slv_rdata_valid;

end;
