--==============================================================================
--
--     File Name: adb3_ocp_tb_pkg.vhd
--       Version: 0.5
--          Date: 10 Oct 2011
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package adb3_ocp_tb_pkg
--
--   Description: Package of components, procedures and functions for applying
--                testbench stimulus to a target FPGA using ADB3 OCP.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 19 Oct 2009: Initial design
--                0.2, KDC, 05 May 2011:
--                     - Changed procedure adb3_target_sim_burst_write to
--                       allow OCP data to be accepted before OCP command
--                       has been accepted
--                0.3, KDC, 17 May 2011:
--                     - Changed address to be of type adb3_ocp_addr_s
--                     - Changed adb3_target_sim_read32, adb3_target_sim_write32
--                       procedures to correctly align data to 16-byte address
--                       boundary
--                     - Changed adb3_target_sim_read, adb3_target_sim_write
--                       procedures to expect 16-byte aligned address and data
--                0.4, KDC, 29 Aug 2011:
--                     - Removed all component definitions from adb3_target_tb_pkg
--                     - Renamed adb3_target_* procedures adb3_ocp_*
--                     - Renamed adb3_wait_cycles adb3_ocp_sim_wait_cycles
--                       and renamed adb3_ocp_tb_pkg
--                0.5, KDC, 10 Oct 2011:
--                     - Removed adb3_ocp_sim_burst_read procedure
--                     - Replaced adb3_ocp_sim_burst_read with adb3_ocp_sim_read
--                       in adb3_ocp_sim_read_reg32
--                     - Replaced adb3_ocp_sim_burst_write with adb3_ocp_sim_write
--                       in adb3_ocp_sim_write_reg32
--                     - Replaced adb3_ocp_sim_burst_read with adb3_ocp_sim_burst_read_cmd
--                       and adb3_ocp_sim_burst_read_resp in adb3_ocp_sim_read
--
--==============================================================================

--
-- Exported data types:
--
--   byte_t
--   byte_enable_t
--   byte_vector_t
--
-- Exported routines (parenthesis indicated number of overloads):
--
--   conv_byte_vector
--   conv_byte_enable
--   conv_vector
--   conv_string_hex
--   conv_string (2)
--
--   adb3_ocp_sim_read_reg32
--   adb3_ocp_sim_read
--   adb3_ocp_sim_read_cmd
--   adb3_ocp_sim_read_resp
--
--   adb3_ocp_sim_write_reg32
--   adb3_ocp_sim_write
--
--   adb3_ocp_sim_wait_cycles
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;

package adb3_ocp_tb_pkg is

  --------------------------
  -- Testbench data types --
  --------------------------

  -- A byte datatype
  subtype byte_t is std_logic_vector(7 downto 0);

  -- Array of std_logic, intended to hold byte enables in ACTIVE HIGH representation
  type byte_enable_t is array(natural range <>) of std_logic;

  -- Array of bytes, intended to hold data to be written to the
  -- local bus or read from the local bus.
  type byte_vector_t is array(natural range <>) of byte_t;


  ---------------------------
  -- Function declarations --
  ---------------------------

  -- Convert between std_logic_vector, byte_vector and integer datatypes.
  function conv_byte_vector(
    constant x : in std_logic_vector)
  return byte_vector_t;

  function conv_byte_enable(
    constant x : in std_logic_vector)
  return byte_enable_t;

  function conv_vector(
    constant x : in byte_vector_t)
  return std_logic_vector;

  function conv_string_hex(
    constant x : in byte_vector_t)
  return string;

  function conv_string_hex(
    constant x : in std_logic_vector)
  return string;

  function conv_string(
    constant x : in byte_enable_t)
  return string;

  function conv_string(
    constant x : in std_logic_vector)
  return string;


  -- ------------------------------------------------------------
  -- Public routines for performing ADB3 transfers
  -- ------------------------------------------------------------

  -- Perform an ADB3 OCP interface direct slave read.
  --
  --     address     Byte address (4-byte/16-byte aligned).
  --
  --     tag         Identifier for transfer
  --
  --     data        Data to be read. data'length indicates
  --                 desired number of bytes to read.
  --
  --     nxfered     Returns the actual number of bytes transferred.

  procedure adb3_ocp_sim_read_reg32(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  -- Blocking read procedure
  -- Waits until all data has been read before completing
  procedure adb3_ocp_sim_read(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  -- Non-Blocking read command procedure
  -- Produces read commands only
  procedure adb3_ocp_sim_read_cmd(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  -- Non-Blocking read response procedure
  -- waits for read response only
  procedure adb3_ocp_sim_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);


  -- Perform an ADB3 OCP interface direct slave write.
  --
  --     address     Byte address (4-byte/16-byte aligned).
  --
  --     tag         Identifier for transfer.
  --
  --     be          Byte enables. be'length must be the same as
  --                 data'length. Byte enables used are limited
  --                 to data bus width (16) as they are ignored
  --                 thereafter.
  --
  --     data        Data to be written. data'length indicates
  --                 desired number of bytes to write.
  --
  --     nxfered     Returns the actual number of bytes transferred.

  procedure adb3_ocp_sim_write_reg32(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);

  procedure adb3_ocp_sim_write(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT);


  -- Delay for a specified number of clock cycles
  --
  --     cycles      Number of clock cycles of delay.
  --     clk         Clock signal to use for delay

  procedure adb3_ocp_sim_wait_cycles(
           cycles : in    natural;
    signal clk    : in    std_logic);

end;


package body adb3_ocp_tb_pkg is

  -- ---------------------------------------------------------------
  -- Private
  -- ---------------------------------------------------------------

  function conv_string_hexdigit(
    constant val : in std_logic_vector)
  return character is
    variable tmp : std_logic_vector(3 downto 0);
  begin
    assert not(val'length /= 4)
      report "(conv_string_hex_digit): input has invalid length " & integer'image(val'length) & " /= 4"
      severity failure;

    tmp := val;
    case tmp is
      when "0000" => return '0';
      when "0001" => return '1';
      when "0010" => return '2';
      when "0011" => return '3';
      when "0100" => return '4';
      when "0101" => return '5';
      when "0110" => return '6';
      when "0111" => return '7';
      when "1000" => return '8';
      when "1001" => return '9';
      when "1010" => return 'A';
      when "1011" => return 'B';
      when "1100" => return 'C';
      when "1101" => return 'D';
      when "1110" => return 'E';
      when "1111" => return 'F';
      when others => return 'X';
    end case;
  end;


  function conv_std_logic_char(
    constant val: in std_logic)
  return character is
  begin
    case val is
      when 'U' => return 'U';
      when 'X' => return 'X';
      when '0' => return '0';
      when '1' => return '1';
      when 'Z' => return 'Z';
      when 'W' => return 'W';
      when 'L' => return 'L';
      when 'H' => return 'H';
      when '-' => return '-';
    end case;
  end;


  -- ---------------------------------------------------------------
  -- Private ADB3 routines
  -- ---------------------------------------------------------------

  procedure adb3_ocp_sim_burst_read_cmd(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := ADB3_OCP_BE_WIDTH;
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable data_offset  : natural;
    variable burst_len    : natural range 0 to BST_MAX;

    type burst_state_type is (ocp_cmd,ocp_acc,ocp_finished);
    variable state, n_state : burst_state_type;

    variable n_cmd         : ocp_CmdT;
    variable n_addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    variable n_burstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    variable n_datavalid   : std_logic;
    variable n_tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);

  begin
    assert not(address(3 downto 0) /= "0000")
      report "(adb3_ocp_sim_burst_read_cmd): Address is not 16-byte aligned during OCP read"
      severity error;

    state        := ocp_cmd;
    data_offset  := data'low;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;

    while (state /= ocp_finished) loop
      wait until rising_edge(ocp_clk);
      n_datavalid := '0';

      case state is
        when ocp_cmd =>
          n_cmd         := OCP_CMD_READ;
          n_addr        := std_logic_vector(RESIZE(unsigned(address),ADB3_OCP_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,ADB3_OCP_BURST_WIDTH));
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,ADB3_OCP_TAG_WIDTH));
          n_state       := ocp_acc;

        when ocp_acc =>
          if (adb3_ocp_s2m.CmdAccept = '1') then
            n_cmd := OCP_CMD_IDLE;
            n_state := ocp_finished;
          else
            n_state := ocp_acc;
          end if;

        when ocp_finished =>
          n_state := ocp_finished;
      end case;

      state := n_state;

      adb3_ocp_m2s.Cmd         <= n_cmd;
      adb3_ocp_m2s.Addr        <= n_addr;
      adb3_ocp_m2s.BurstLength <= n_burstlength;
      adb3_ocp_m2s.DataValid   <= n_datavalid;
      adb3_ocp_m2s.Tag         <= n_tag;
      adb3_ocp_m2s.DataByteEn  <= (others => '0');
      wait until falling_edge(ocp_clk);
    end loop;

    nxfered := data'length;
  end;


  procedure adb3_ocp_sim_burst_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := ADB3_OCP_BE_WIDTH;
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable data_offset  : natural;
    variable remaining_d  : natural;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m_d          : natural range 0 to OCP_WIDTH;

    variable n_respaccept  : std_logic;

  begin
    data_offset  := data'low;
    remaining_d  := data'length;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;
    adb3_ocp_m2s.RespAccept <= '1';

    while (remaining_d /= 0) loop
      wait until rising_edge(ocp_clk);

      if (adb3_ocp_s2m.Resp = OCP_RESP_VALID) then
        m_d := OCP_WIDTH;
        if (m_d > remaining_d) then
          m_d := remaining_d;
        end if;

        for i in 0 to (OCP_WIDTH - 1) loop
          if (m_d > 0) then
            data(data_offset) := adb3_ocp_s2m.Data(8 * (i + 1) - 1 downto 8 * i);
            data_offset := data_offset + 1;
            remaining_d := remaining_d - 1;
            m_d := m_d - 1;
          end if;
        end loop;
        tag := TO_INTEGER(UNSIGNED(adb3_ocp_s2m.Tag));
      end if;

      wait until falling_edge(ocp_clk);
    end loop;

    adb3_ocp_m2s.RespAccept  <= '0';
    nxfered := data'length - remaining_d;
  end;


  procedure adb3_ocp_sim_burst_write(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    constant OCP_WIDTH    : natural := ADB3_OCP_BE_WIDTH;
    constant BST_MAX      : natural := 2**ADB3_OCP_BURST_WIDTH - 1;

    variable i            : natural;
    variable beat_one     : boolean;
    variable data_offset  : natural;
    variable be_offset    : natural;
    variable remaining    : natural;
    variable next_be      : boolean;
    variable burst_len    : natural range 0 to BST_MAX;
    variable m            : natural range 0 to OCP_WIDTH;

    type burst_state_type is (ocp_cmd,ocp_data,ocp_finished);
    variable state, n_state : burst_state_type;

    variable n_cmd         : ocp_CmdT;
    variable n_addr        : std_logic_vector(ADB3_OCP_ADDR_WIDTH-1 downto 0);
    variable n_burstlength : std_logic_vector(ADB3_OCP_BURST_WIDTH-1 downto 0);
    variable n_datavalid   : std_logic;
    variable n_tag         : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);
    variable n_respaccept  : std_logic;
    variable n_data        : std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0);
    variable n_databyteen  : std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0);

  begin
    assert not(be'length /= data'length)
      report "(adb3_ocp_sim_burst_write): byte enable length (" & integer'image(be'length) &
             ") does not match data length (" & integer'image(data'length) & ")"
      severity failure;

    assert not(address(3 downto 0) /= "0000")
      report "(adb3_ocp_sim_burst_write): Address is not 16-byte aligned during OCP write"
      severity error;

    -- If burst transfer is larger than OCP bus width, then byte enables are ignored for all beats
    -- after the initial one. Check to verify we are not ignoring active enables.
    for i in be'low to be'high loop
      assert not((i >= OCP_WIDTH)and(be(i)='0'))
        report "(adb3_ocp_sim_burst_write): active byte enable (be(" & integer'image(i) &
               ") = " & std_logic'image(be(i)) &
               ") will be ignored during writes with bytes > " & integer'image(OCP_WIDTH)
        severity warning;
    end loop;

    state       := ocp_cmd;
    beat_one    := true;
    data_offset := data'low;
    be_offset   := be'low;
    remaining   := data'length;

    -- remaining =  1..16 => burst_len = 1
    -- remaining = 17..32 => burst_len = 2 etc.
    burst_len := (data'length + OCP_WIDTH - 1) / OCP_WIDTH;

    while (state /= ocp_finished) loop
      wait until rising_edge(ocp_clk);
      n_respaccept := '0';
      next_be      := false;

      case state is
        when ocp_cmd =>
          n_cmd         := OCP_CMD_WRITE;
          n_addr        := std_logic_vector(RESIZE(unsigned(address),ADB3_OCP_ADDR_WIDTH));
          n_burstlength := std_logic_vector(TO_UNSIGNED(burst_len,ADB3_OCP_BURST_WIDTH));
          n_datavalid   := '1';
          n_tag         := std_logic_vector(TO_UNSIGNED(tag,ADB3_OCP_TAG_WIDTH));
          next_be       := true;
          n_state       := ocp_data;

        when ocp_data =>
          n_state := ocp_data;
          if (adb3_ocp_s2m.CmdAccept = '1') then
            n_cmd := OCP_CMD_IDLE;
          end if;
          if (adb3_ocp_s2m.DataAccept = '1') then
            if (remaining /= 0) then
              next_be := true;
            else
              n_datavalid := '0';
              n_state := ocp_finished;
            end if;
          end if;

        when ocp_finished =>    -- State is never executed - used in while test only
          null;
      end case;

      state := n_state;

      if next_be then
        m := OCP_WIDTH;
        if (m > remaining) then
          m := remaining;
        end if;

        -- Bytes > m are set to zero
        for i in 0 to (OCP_WIDTH - 1) loop
          if (m = 0) then
            if (beat_one) then
              n_databyteen(i) := '0';  -- Byte enable used if first beat
            else
              n_databyteen(i) := '1';  -- Byte enables active on subsequent beats
            end if;
            n_data(8 * (i + 1) - 1 downto 8 * i) := (others => '0');
          else
            if (beat_one) then
              n_databyteen(i) := be(be_offset);  -- Byte enable used if first beat
            else
              n_databyteen(i) := '1';            -- Byte enables active on subsequent beats
            end if;
            be_offset := be_offset + 1;
            n_data(8 * (i + 1) - 1 downto 8 * i) := data(data_offset);
            data_offset := data_offset + 1;
            remaining := remaining - 1;
            m := m - 1;
          end if;
        end loop;

        beat_one := false;
      end if;

      adb3_ocp_m2s.Cmd         <= n_cmd;
      adb3_ocp_m2s.Addr        <= n_addr;
      adb3_ocp_m2s.BurstLength <= n_burstlength;
      adb3_ocp_m2s.DataValid   <= n_datavalid;
      adb3_ocp_m2s.Tag         <= n_tag;
      adb3_ocp_m2s.RespAccept  <= n_respaccept;
      adb3_ocp_m2s.Data        <= n_data;
      adb3_ocp_m2s.DataByteEn  <= n_databyteen;
    end loop;

    nxfered := data'length - remaining;
  end;


  -- ---------------------------------------------------------------
  -- Public
  -- ---------------------------------------------------------------

  function conv_byte_vector(
    constant x : in std_logic_vector)
  return byte_vector_t is
    variable n : natural := (x'length + 7) / 8;         -- Size of x in bytes (min size of 1)
    variable m : std_logic_vector(8 * n - 1 downto 0);  -- New vector of size n bytes
    variable t : byte_vector_t(0 to n - 1);             -- New byte_vector of size n bytes
  begin
    m := std_logic_vector(RESIZE(unsigned(x), 8 * n));

    for i in 0 to n - 1 loop
      t(i) := m(8 * (i + 1) - 1 downto 8 * i);
    end loop;

    return t;
  end;

  function conv_byte_enable(
    constant x : in std_logic_vector)
  return byte_enable_t is
    variable n : natural := x'length;               -- Size of x in bits (min size of 1)
    variable m : std_logic_vector(n - 1 downto 0);  -- New vector of size n bits
    variable t : byte_enable_t(0 to n - 1);         -- New byte_enable of size n bits
  begin
    m := x;

    for i in 0 to n - 1 loop
      t(i) := m(i);
    end loop;

    return t;
  end;

  function conv_vector(
    constant x : in byte_vector_t)
  return std_logic_vector is
    variable v          : std_logic_vector((8*x'length)-1 downto 0);
    variable next_byte  : byte_t;
    variable j          : natural;
  begin
    for i in x'range loop
      next_byte := x(i);
      j := i - x'low;
      v(j*8+7 downto j*8) := next_byte;
    end loop;

    return v;
  end;

  function conv_string_hex(
    constant x : in byte_vector_t)
  return string is
    variable j         : natural := 2 * x'length;
    variable s         : string(1 to j);
    variable next_byte : byte_t;
  begin
    for i in x'range loop
      next_byte := x(i);
      s(j)      := conv_string_hexdigit(next_byte(3 downto 0));
      s(j-1)    := conv_string_hexdigit(next_byte(7 downto 4));
      j         := j - 2;
    end loop;

    return s;
  end;

  function conv_string_hex(
    constant x : in std_logic_vector)
  return string is
    variable i : natural := x'length mod 4;
    variable j : natural := x'length + (4-i) mod 4;
    variable s : string(1 to j/4);
    variable y : std_logic_vector(j-1 downto 0) := (others => '0');
  begin
    y(x'length - 1 downto 0) := x;

    for k in (j/4)-1 downto 0 loop
      s(j/4 - k) := conv_string_hexdigit(y((k*4)+3 downto k*4));
    end loop;

    return s;
  end;

  function conv_string(
    constant x : in byte_enable_t)
  return string is
    variable j        : natural;
    variable s        : string(1 to x'length);
    variable next_bit : std_logic;
  begin
    j := x'length;

    for i in x'range loop
      next_bit := x(i);
      s(j) := conv_std_logic_char(next_bit);
      j    := j - 1;
    end loop;

    return s;
  end;


  -- Calculate the string width required to represent
  -- a decimal number in decimal format
  function conv_dec_width(
    decimal : in natural)
  return natural is
  begin
    assert not(decimal < 0)
      report "(conv_dec_width): Cannot use 32-bit natural"
      severity failure;

    case decimal is
      when 0 to 9 => return 1;
      when 10 to 99 => return 2;
      when 100 to 999 => return 3;
      when 1000 to 9999 => return 4;
      when 10000 to 99999 => return 5;
      when 100000 to 999999 => return 6;
      when 1000000 to 9999999 => return 7;
      when 10000000 to 99999999 => return 8;
      when 100000000 to 999999999 => return 9;
      when 1000000000 to integer'high => return 10;
    end case;
  end function;


  function conv_string(
    constant x : in std_logic_vector)
  return string is
    variable j        : natural := TO_INTEGER(unsigned(x));
    variable s        : string(1 to conv_dec_width(j));
  begin
    s := integer'image(j);

    return s;
  end;


  -- ---------------------------------------------------------------
  -- Public ADB3 routines
  -- ---------------------------------------------------------------

  -- Data:    4-byte read (blocking)
  -- Address: 4-byte aligned
  -- Blocking read procedure waits until all data has been read before completing
  procedure adb3_ocp_sim_read_reg32(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    variable offset : natural range 0 to 3 := TO_INTEGER(UNSIGNED(address(3 downto 2)));
    variable addr_a : adb3_ocp_addr_s;
    variable data_a : byte_vector_t(0 to 15);
    variable xfer_a : natural;
  begin
    assert not(address(1 downto 0) /= "00")
      report "(adb3_ocp_sim_read_reg32): Address is not 4-byte aligned during 32-bit register read"
      severity error;

    assert not(data'length /= 4)
      report "(adb3_ocp_sim_read_reg32): Data is not 4-bytes"
      severity error;

    -- Convert 4-byte aligned address to 16-byte aligned address
    addr_a := address(ADB3_OCP_ADDR_WIDTH-1 downto 4) & "0000";

    adb3_ocp_sim_read(addr_a, tag, data_a, xfer_a, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);

    -- Convert 16-byte aligned data to 4-byte data
    for i in 0 to 3 loop
      data(i) := data_a(4*offset+i);
    end loop;

    if (xfer_a = 16) then
      nxfered := 4;
    else
      nxfered := 0;
      assert false
        report "(adb3_ocp_sim_read_reg32): Bytes transferred = " & integer'image(xfer_a) &
               " (/=16) during 4-byte register read."
        severity error;
    end if;
  end;

  -- Data:    4-byte write
  -- Address: 4-byte aligned
  procedure adb3_ocp_sim_write_reg32(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    variable offset : natural range 0 to 3 := TO_INTEGER(UNSIGNED(address(3 downto 2)));
    variable addr_a : adb3_ocp_addr_s;
    variable be_a   : byte_enable_t(0 to 15) := (others => '0');
    variable data_a : byte_vector_t(0 to 15) := (others => (others => '0'));
    variable xfer_a : natural;
  begin
    assert not(address(1 downto 0) /= "00")
      report "(adb3_ocp_sim_write_reg32): Address is not 4-byte aligned during 32-bit register write"
      severity error;

    assert not(be'length /= 4)
      report "(adb3_ocp_sim_write_reg32): BE is not 4-bits"
      severity error;

    assert not(data'length /= 4)
      report "(adb3_ocp_sim_write_reg32): Data is not 4-bytes"
      severity error;

    -- Convert 4-byte aligned address to 16-byte aligned address
    addr_a := address(ADB3_OCP_ADDR_WIDTH-1 downto 4) & "0000";

    -- Convert 4-bit be to 16-bit aligned be
    -- Convert 4-byte data to 16-byte aligned data
    for i in 0 to 3 loop
      be_a(4*offset+i)   := be(i);
      data_a(4*offset+i) := data(i);
    end loop;

    adb3_ocp_sim_write(addr_a, tag, be_a, data_a, xfer_a, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);

    if (xfer_a = 16) then
      nxfered := 4;
    else
      nxfered := 0;
      assert false
        report "(adb3_ocp_sim_write_reg32): Bytes transferred = " & integer'image(xfer_a) &
               " (/=16) during 4-byte register write."
        severity error;
    end if;
  end;

  -- Data:    n-byte read (blocking)
  -- Address: 16-byte aligned
  -- Blocking read procedure waits until all data has been read before completing
  procedure adb3_ocp_sim_read(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
    variable tag_out : natural;
  begin
    adb3_ocp_sim_burst_read_cmd(address, tag, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
    adb3_ocp_sim_burst_read_resp(data, tag_out, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Address: 16-byte aligned
  -- Non-Blocking read command procedure. Produces read commands only
  procedure adb3_ocp_sim_read_cmd(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           data         : out   byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_ocp_sim_burst_read_cmd(address, tag, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Data:    n-byte read (non-blocking)
  -- Non-Blocking read response procedure. Waits for read response only
  procedure adb3_ocp_sim_read_resp(
           data         : out   byte_vector_t;
           tag          : out   natural;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_ocp_sim_burst_read_resp(data, tag, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;

  -- Data:    n-byte write
  -- Address: 16-byte aligned
  procedure adb3_ocp_sim_write(
           address      : in    adb3_ocp_addr_s;
           tag          : in    natural;
           be           : in    byte_enable_t;
           data         : in    byte_vector_t;
           nxfered      : out   natural;
    signal ocp_clk      : in    std_logic;
    signal adb3_ocp_m2s : out   adb3_ocp_m2sT;
    signal adb3_ocp_s2m : in    adb3_ocp_s2mT)
  is
  begin
    adb3_ocp_sim_burst_write(address, tag, be, data, nxfered, ocp_clk, adb3_ocp_m2s, adb3_ocp_s2m);
  end;


  procedure adb3_ocp_sim_wait_cycles(
           cycles : in    natural;
    signal clk    : in    std_logic)
  is
  begin
    for i in 0 to (cycles-1) loop
      wait until rising_edge(clk);
    end loop;
  end;

end;
