--==============================================================================
--
--     File Name: adb3_ocp_transaction_probe_sim.vhd
--       Version: 0.2
--          Date: 05 May 2011
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 OCP transaction test probe
--
--   Description: Monitors ADB3 OCP transactions and reports on:
--                  Cumulative number of read commands (read_cmd_cnt)
--                  Cumulative number of completed read commands (read_dat_cnt)
--                  Read command active (read_active)
--                  Latency (cycles) between read command and first returned data
--                  Duration (cycles) of read command data burst
--                  Read data with incorrect tag for current read command (read_tag_err_cnt,read_tag_err)
--                  Read data for read command which has completed (read_cmp_err_cnt,read_cmp_err)
--
--                  OCP_RESP_FAILED detection (resp_failed)
--                  OCP_RESP_ERROR detection (resp_error)
--
--                  Cumulative number of write commands (write_cmd_cnt)
--                  Cumulative number of completed write commands (write_dat_cnt)
--                  Write command active (write_active)
--                  Duration (cycles) of write command data burst
--                  Write data for write command which has completed (write_cmp_err_cnt,write_cmp_err)
--                  Write data with invalid DataByteEn value (write_be_err_cnt,write_be_err)
--
--                  Invalid command detection (inv_cmd)
--                  Invalid address detection (inv_addr)
--                  Invalid burst length detection (inv_bl)
--                  Average data throughput in MB/s
--                  Average data efficiency in % (data beat cycles/total cycles)
--
--   Limitations:
--
--         Notes: Component of adb3_probe_pkg Library
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 25 Jun 09: Initial design
--                0.2, KDC, 05 May 2011:
--                       -- Changed "Write Data accepted without open
--                          OCP_CMD_WRITE command" error to warning
--                       -- Added illegal OCP command check
--
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library std;
use std.textio.all;

library work;
use work.adb3_ocp.all;
use work.adb3_probe_pkg.all;


entity adb3_ocp_transaction_probe is
generic(
  enable_logging   : boolean := true;                   -- Enable use of log file for info/warnings/errors
  sel_int_log_file : boolean := true;                   -- Select between internal name and external name for log file
  int_log_filename : string  := "ocp_probe";            -- Internal filename for log file if selected and enabled
  addr_align_bits  : natural := ADB3_OCP_BE_WIDTH_LOG2; -- Set number of unused address LSBs for checking
  addr_width_max   : natural := ADB3_OCP_ADDR_WIDTH;    -- Set maximum address width for checking
  data_burst_max   : natural := MAX_NUM_OF_BURSTS;      -- Set maximum burst length for checking
  enable_tag_check : boolean := true);                  -- Enable checking of OCP_CMD_READ tag with read data tag
port(
  ocp_clk : in  std_logic;
  ocp_m2s : in  adb3_ocp_m2sT;
  ocp_s2m : in  adb3_ocp_s2mT;
  status  : out adb3_ocp_probe_status_r);
end adb3_ocp_transaction_probe;


architecture sim of adb3_ocp_transaction_probe is

  constant INT_FILE         : string  := int_log_filename & ".txt";
  constant SEL_EXT_LOG_FILE : boolean := not(sel_int_log_file);
  constant BL_ZERO          : natural := 0;
  constant ARRAY_MAX        : natural := 127;  -- Size of array for read commands in flight (arbitrary)

  subtype burst_s is natural range 0 to (MAX_NUM_OF_BURSTS-1);

  type command_r is record
    tag          : natural;
    cycle        : natural;
    burst_length : burst_s;
  end record;

  constant COMMAND_0 : command_r :=
    (tag           => 0,
     cycle         => 0,
     burst_length  => 0);

  type data_r is record
    burst_count : burst_s;
  end record;

  constant DATA_0 : data_r :=
    (burst_count  => 0);

  type ocp_cmd_t is (CMD_IDLE,
                     CMD_WRITE,
                     CMD_READ,
                     CMD_WRITENONPOST,
                     CMD_INVALID);
  signal ocp_cmd : ocp_cmd_t;

  type ocp_rsp_t is (RESP_NONE,
                     RESP_VALID,
                     RESP_FAILED,
                     RESP_ERROR,
                     RESP_INVALID);
  signal ocp_rsp : ocp_rsp_t;

  type adb3_ocp_probe_active_r is record
    read_active   : std_logic;  -- Read command issued but not yet complete
    write_active  : std_logic;  -- Write command issued but not yet complete
  end record;

  signal ocp_active : adb3_ocp_probe_active_r;

  type cmd_array_t is array (natural range 0 to ARRAY_MAX) of command_r;
  type dat_array_t is array (natural range 0 to ARRAY_MAX) of data_r;

  shared variable read_cmd_array : cmd_array_t := (others => COMMAND_0);
  shared variable read_dat_array : dat_array_t := (others => DATA_0);

  shared variable act_write_bytes : natural;
  shared variable max_write_bytes : natural;
  shared variable max_read_bytes  : natural;

  shared variable write_cmd_cnt,read_cmd_cnt : natural range 0 to ARRAY_MAX := 0;
  shared variable write_dat_cnt,read_dat_cnt : natural range 0 to ARRAY_MAX := 0;

  signal clock_cycle : natural  := 0;

  signal read_tag_err_cnt  : natural := 0;
  signal read_cmp_err_cnt  : natural := 0;
  signal write_cmp_err_cnt : natural := 0;
  signal write_be_err_cnt  : natural := 0;
  signal resp_failed_cnt   : natural := 0;
  signal resp_error_cnt    : natural := 0;
  signal inv_cmd_cnt       : natural := 0;
  signal inv_addr_cnt      : natural := 0;
  signal inv_bl_cnt        : natural := 0;

  signal data_beat_cnt      : natural := 0;
  signal data_mb_per_second : natural := 0;
  signal data_efficiency    : natural := 0;

  file file_out_id : text;


begin

  -- Use externally opened log file with id as defined in adb3_probe_pkg or
  -- Open probe log file using open_file procedure defined in adb3_probe_pkg
  --
  open_file_p : process
  begin
    if (enable_logging and sel_int_log_file) then
      open_file(INT_FILE,WRITE_MODE,file_out_id);
    end if;
    wait;
  end process;


  clock_cycle_p : process(ocp_clk)
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      clock_cycle <= clock_cycle + 1;
    end if;
  end process;


  -- Read command/data reporting
  -- Detect occurrences of OCP_CMD_READ and store tag, burst length, and cycle values
  -- Detect occurrences of OCP_RESP_VALID and increment burst counter for current OCP_CMD_READ
  -- Check if OCP_RESP_VALID found with tag that does not belong to current OCP_CMD_READ (read_tag_err_cnt, read_tag_err)
  -- Check if OCP_RESP_VALID found with tag belonging to completed OCP_CMD_READ (read_cmp_err_cnt, read_cmp_err)
  -- Count total number of OCP_CMD_READs and total number completed
  -- Determine the latency in cycles between OCP_CMD_READ and first OCP_RESP_VALID
  -- Determine length in cycles between first OCP_RESP_VALID and last OCP_RESP_VALID
  --
  -- ocp_CmdT : OCP_CMD_IDLE, OCP_CMD_WRITE, OCP_CMD_READ, (OCP_CMD_WRITENONPOST)
  -- ocp_RespT: OCP_RESP_NONE, OCP_RESP_VALID, OCP_RESP_FAILED, OCP_RESP_ERROR
  --
  cmd_read_report_p : process(ocp_clk)
    variable print_line      : line;
    variable read_cmd_cycle  : natural;
    variable first_dat_cycle : natural;
    variable read_cmd_tag    : natural;
    variable read_cmd_bl     : natural;
    variable read_dat_tag    : natural;
    variable read_latency    : natural;
    variable read_transfer   : natural;
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if (ocp_m2s.Cmd = OCP_CMD_READ) and (ocp_s2m.CmdAccept = '1') then
        read_cmd_tag  := TO_INTEGER(UNSIGNED(ocp_m2s.Tag));
        read_cmd_bl   := TO_INTEGER(UNSIGNED(ocp_m2s.BurstLength));
        write(print_line,string'("OCP_CMD_READ Command (Tag: ")); write(print_line,read_cmd_tag);
        write(print_line,string'(", Burst: ")); write(print_line,read_cmd_bl);
        write(print_line,string'(", Addr: ")); hex_write(print_line,ocp_m2s.Addr,right,0);
        write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
        if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
        read_cmd_array(read_cmd_cnt).tag          := read_cmd_tag;
        read_cmd_array(read_cmd_cnt).cycle        := clock_cycle;
        read_cmd_array(read_cmd_cnt).burst_length := read_cmd_bl;
        if (read_cmd_cnt = ARRAY_MAX) then
          read_cmd_cnt := 0;
        else
          read_cmd_cnt := read_cmd_cnt + 1;
        end if;
        read_cmd_array(read_cmd_cnt) := COMMAND_0;
        max_read_bytes := 0;
      end if;

      if (ocp_m2s.RespAccept ='1') and (ocp_s2m.Resp = OCP_RESP_VALID) then
        read_cmd_tag   := read_cmd_array(read_dat_cnt).tag;
        read_dat_tag   := TO_INTEGER(UNSIGNED(ocp_s2m.Tag));
        read_cmd_cycle := read_cmd_array(read_dat_cnt).cycle;
        read_cmd_bl    := read_cmd_array(read_dat_cnt).burst_length;
        max_read_bytes := max_read_bytes + BYTES_PER_BEAT;
        if (read_dat_tag /= read_cmd_tag)and(enable_tag_check) then
          deallocate(print_line);
          write(print_line,string'("  ERROR: Read Data (Tag: ")); write(print_line,read_dat_tag);
          write(print_line,string'(") mismatch with OCP_CMD_READ command (Tag: ")); write(print_line,read_cmd_tag);
          write(print_line,string'(")"));
          assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
          assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity error;
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          read_tag_err_cnt <= read_tag_err_cnt + 1;
        else
          if (read_dat_array(read_dat_cnt).burst_count = 0) then
            read_latency := clock_cycle - read_cmd_cycle;
            deallocate(print_line);
            write(print_line,string'("  First Read Data (Tag: ")); write(print_line,read_dat_tag);
            write(print_line,string'(", Burst: ")); write(print_line,read_cmd_bl);
            write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
            write(print_line,string'(" (Latency: ")); write(print_line,read_latency);
            write(print_line,string'(" cycles)"));
            if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
            first_dat_cycle := clock_cycle;
          end if;
          -- Increment read burst counter for this tag
          read_dat_array(read_dat_cnt).burst_count := read_dat_array(read_dat_cnt).burst_count + 1;
          -- If read burst counter = size of burst in read command then last beat of read
          if (read_dat_array(read_dat_cnt).burst_count = read_cmd_bl) then
            read_transfer := clock_cycle - first_dat_cycle + 1;
            deallocate(print_line);
            write(print_line,string'("  Last  Read Data (Tag: ")); write(print_line,read_dat_tag);
            write(print_line,string'(", Burst: ")); write(print_line,read_cmd_bl);
            write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
            write(print_line,string'(" (Xfer: ")); write(print_line,read_transfer);
            write(print_line,string'(" cycles"));
            write(print_line,string'(", Bytes: ")); write(print_line,max_read_bytes);
            write(print_line,string'(")"));
            if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
            -- Finished with this tag so increment pointer
            if (read_dat_cnt = ARRAY_MAX) then
              read_dat_cnt := 0;
            else
              read_dat_cnt := read_dat_cnt + 1;
            end if;
            read_dat_array(read_dat_cnt) := DATA_0;  -- Initialise burst count to 0
          end if;
          if (read_dat_array(read_dat_cnt).burst_count > read_cmd_bl) then
            deallocate(print_line);
            write(print_line,string'("  ERROR: Read Data (Tag: ")); write(print_line,read_cmd_tag);
            write(print_line,string'(") found without open OCP_CMD_READ command"));
            write(print_line,string'(" (burst_count: ")); write(print_line,read_dat_array(read_dat_cnt).burst_count);
            write(print_line,string'(" > read_cmd_bl: ")); write(print_line,read_cmd_bl);
            write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
            assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
            assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity error;
            if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
            read_cmp_err_cnt <= read_cmp_err_cnt + 1;
          end if;
        end if;
      end if;

      -- Test to see if number of read commands matches number of completed reads
      if (read_dat_cnt = read_cmd_cnt) then
        ocp_active.read_active <= '0';
      else
        ocp_active.read_active <= '1';
      end if;
      if (read_tag_err_cnt = 0) then
        status.read_tag_err <= '0';
      else
        status.read_tag_err <= '1';
      end if;
      if (read_cmp_err_cnt = 0) then
        status.read_cmp_err <= '0';
      else
        status.read_cmp_err <= '1';
      end if;
    end if;
  end process;


  -- Detect occurrences of OCP_RESP_FAILED and count
  -- Detect occurrences of OCP_RESP_ERROR and count
  --
  resp_error_p : process(ocp_clk)
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if (ocp_m2s.RespAccept = '1') and (ocp_s2m.Resp = OCP_RESP_FAILED) then
        resp_failed_cnt <= resp_failed_cnt + 1;
      end if;

      if (ocp_m2s.RespAccept = '1') and (ocp_s2m.Resp = OCP_RESP_ERROR) then
        resp_error_cnt <= resp_error_cnt + 1;
      end if;

      if (resp_failed_cnt = 0) then
        status.resp_failed <= '0';
      else
        status.resp_failed <= '1';
      end if;

      if (resp_error_cnt = 0) then
        status.resp_error <= '0';
      else
        status.resp_error <= '1';
      end if;
    end if;
  end process;


  -- Write command/data reporting
  -- Detect occurrences of OCP_CMD_WRITE and store tag, burst length, and cycle values
  -- Detect occurrences of OCP_CMD_WRITE data and increment burst counter
  -- Check if write data found found with tag belonging to completed OCP_CMD_WRITE (write_cmp_err_cnt, write_cmp_err)
  -- Count total number of OCP_CMD_WRITEs and total number completed
  -- Determine length in cycles between OCP_CMD_WRITEs and last OCP_CMD_WRITE data
  -- Check DataByteEn for invalid values (write_be_err_cnt, write_be_err)
  --
  -- ocp_CmdT : OCP_CMD_IDLE, OCP_CMD_WRITE, OCP_CMD_READ, (OCP_CMD_WRITENONPOST)
  -- ocp_RespT: OCP_RESP_NONE, OCP_RESP_VALID, OCP_RESP_FAILED, OCP_RESP_ERROR
  --
  cmd_write_report_p : process(ocp_clk)
    variable print_line      : line;
    variable write_cmd_cycle : natural;
    variable write_cmd_tag   : natural;
    variable write_cmd_bl    : natural;
    variable write_dat_bc    : natural;
    variable write_bytes_en  : natural;
    variable write_transfer  : natural;
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if (ocp_m2s.Cmd = OCP_CMD_WRITE) and (ocp_s2m.CmdAccept = '1') then
        write_cmd_tag   := TO_INTEGER(UNSIGNED(ocp_m2s.Tag));
        write_cmd_bl    := TO_INTEGER(UNSIGNED(ocp_m2s.BurstLength));
        write_cmd_cycle := clock_cycle;
        deallocate(print_line);
        write(print_line,string'("OCP_CMD_WRITE Command (Tag: ")); write(print_line,write_cmd_tag);
        write(print_line,string'(", Burst: ")); write(print_line,write_cmd_bl);
        write(print_line,string'(", Addr: ")); hex_write(print_line,ocp_m2s.Addr,right,0);
        write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
        if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
        if (write_cmd_cnt = ARRAY_MAX) then
          write_cmd_cnt := 0;
        else
          write_cmd_cnt := write_cmd_cnt + 1;
        end if;
        write_dat_bc := 0;
        act_write_bytes := 0; max_write_bytes := 0;
      end if;

      if (ocp_m2s.DataValid ='1') and (ocp_s2m.DataAccept = '1') then
        write_bytes_en := enables_to_bytes(ocp_m2s.DataByteEn);
        act_write_bytes := act_write_bytes + write_bytes_en;
        max_write_bytes := max_write_bytes + BYTES_PER_BEAT;
        write_dat_bc := write_dat_bc + 1;
        -- If write burst counter = size of burst in write command then last beat of write
        if (write_dat_bc = write_cmd_bl) then
          write_transfer := clock_cycle - write_cmd_cycle;
          deallocate(print_line);
          write(print_line,string'("  Last Write Data (Tag: ")); write(print_line,write_cmd_tag);
          write(print_line,string'(", Burst: ")); write(print_line,write_cmd_bl);
          write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
          write(print_line,string'(" (Xfer: ")); write(print_line,write_transfer);
          write(print_line,string'(" cycles"));
          write(print_line,string'(", Bytes: ")); write(print_line,act_write_bytes);
          write(print_line,string'("/")); write(print_line,max_write_bytes);
          write(print_line,string'(")"));
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          if (write_dat_cnt = ARRAY_MAX) then
            write_dat_cnt := 0;
          else
            write_dat_cnt := write_dat_cnt + 1;
          end if;
        end if;
        if (write_dat_bc > write_cmd_bl) then
          deallocate(print_line);
          write(print_line,string'("  WARNING: Write Data (Tag: ")); write(print_line,write_cmd_tag);
          write(print_line,string'(", Burst: ")); write(print_line,write_cmd_bl);
          write(print_line,string'(") accepted without open OCP_CMD_WRITE command"));
          write(print_line,string'(" on cycle: ")); write(print_line,clock_cycle);
          assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
          assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity warning;
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          write_cmp_err_cnt <= write_cmp_err_cnt + 1;
        end if;
        -- Only test for zero byte enables on first write of command
        if (write_bytes_en = 0)and(max_write_bytes = BYTES_PER_BEAT) then
          deallocate(print_line);
          write(print_line,string'("  WARNING: Write Data (Tag: ")); write(print_line,write_cmd_tag);
          write(print_line,string'(", Burst: ")); write(print_line,write_cmd_bl);
          write(print_line,string'(") accepted with ")); write(print_line,write_bytes_en);
          write(print_line,string'(" Bytes enabled"));
          write(print_line,string'(" on cycle: ")); write(print_line,clock_cycle);
          assert not(SEL_EXT_LOG_FILE) report print_line.all severity warning;
          assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity warning;
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          write_be_err_cnt <= write_be_err_cnt + 1;
        end if;
      end if;

      -- Test to see if number of write commands matches number of completed writes
      if (write_dat_cnt = write_cmd_cnt) then
        ocp_active.write_active <= '0';
      else
        ocp_active.write_active <= '1';
      end if;
      if (write_cmp_err_cnt = 0) then
        status.write_cmp_err <= '0';
      else
        status.write_cmp_err <= '1';
      end if;
      if (write_be_err_cnt = 0) then
        status.write_be_err <= '0';
      else
        status.write_be_err <= '1';
      end if;
    end if;
  end process;


  -- Write non posted command/data reporting
  --
  cmd_writeonpost_report_p : process(ocp_clk)
    variable print_line      : line;
    variable write_cmd_tag   : natural;
    variable write_cmd_bl    : natural;
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if ocp_m2s.Cmd = OCP_CMD_WRITENONPOST and ocp_s2m.CmdAccept = '1' then
        write_cmd_tag  := TO_INTEGER(UNSIGNED(ocp_m2s.Tag));
        write_cmd_bl   := TO_INTEGER(UNSIGNED(ocp_m2s.BurstLength));
        deallocate(print_line);
        write(print_line,string'("OCP_CMD_WRITENONPOST Command (Tag: ")); write(print_line,write_cmd_tag);
        write(print_line,string'(", Burst: ")); write(print_line,write_cmd_bl);
        write(print_line,string'(", Addr: ")); hex_write(print_line,ocp_m2s.Addr,right,0);
        write(print_line,string'(") on cycle: ")); write(print_line,clock_cycle);
        assert false report print_line.all severity warning;
        if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
      end if;
    end if;
  end process;


  -- Check all command values for validity
  -- Valid commands: OCP_CMD_IDLE, OCP_CMD_WRITE, OCP_CMD_WRITENONPOST, OCP_CMD_READ
  --
  cmd_check_p : process(ocp_clk)
    variable print_line : line;
    variable command    : natural;
  begin
    if ocp_clk'event and ocp_clk = '1' then
      if (ocp_m2s.Cmd = OCP_CMD_ILLEGAL1) or (ocp_m2s.Cmd = OCP_CMD_ILLEGAL2) or
         (ocp_m2s.Cmd = OCP_CMD_ILLEGAL3) or (ocp_m2s.Cmd = OCP_CMD_ILLEGAL4) then
        command := TO_INTEGER(UNSIGNED(ocp_m2s.Cmd));
        deallocate(print_line);
        write(print_line,string'("  ERROR: Invalid Command (CMD: ")); write(print_line,command);
        write(print_line,string'(") asserted on cycle: ")); write(print_line,clock_cycle);
        assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
        assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity error;
        if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
        inv_cmd_cnt <= inv_cmd_cnt + 1;
      end if;

      if (inv_cmd_cnt = 0) then
        status.inv_cmd <= '0';
      else
        status.inv_cmd <= '1';
      end if;
    end if;
  end process;


  -- Check all command address values for validity
  -- Check LSBs defined by generic addr_align_bits are zero
  -- Check address width <= addr_width_max
  --
  cmd_check_addr_p : process(ocp_clk)
    variable print_line : line;
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if ((ocp_m2s.Cmd = OCP_CMD_WRITE) or (ocp_m2s.Cmd = OCP_CMD_WRITENONPOST) or (ocp_m2s.Cmd = OCP_CMD_READ)) and
          (ocp_s2m.CmdAccept = '1') then
        if (addr_align_bits > 0)and(or_reduce(ocp_m2s.Addr(addr_align_bits-1 downto 0)) /= '0') then
          deallocate(print_line);
          write(print_line,string'("  ERROR: Address: ")); hex_write(print_line,ocp_m2s.Addr,right,0);
          write(print_line,string'(" not ")); write(print_line,2**addr_align_bits);
          write(print_line,string'(" byte aligned on cycle: ")); write(print_line,clock_cycle);
          assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
          assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity error;
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          inv_addr_cnt <= inv_addr_cnt + 1;
        end if;
        if (or_reduce(ocp_m2s.Addr(ADB3_OCP_ADDR_WIDTH-1 downto addr_width_max)) /= '0') then
          deallocate(print_line);
          write(print_line,string'("  ERROR: Address: ")); hex_write(print_line,ocp_m2s.Addr,right,0);
          write(print_line,string'(" exceeds width addr_width_max = ")); write(print_line,addr_width_max);
          write(print_line,string'(" on cycle: ")); write(print_line,clock_cycle);
          assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
          assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity error;
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          inv_addr_cnt <= inv_addr_cnt + 1;
        end if;
      end if;

      if (inv_addr_cnt = 0) then
        status.inv_addr <= '0';
      else
        status.inv_addr <= '1';
      end if;
    end if;
  end process;


  -- Check all burst length values for validity
  -- Check BL_ZERO < bl <= data_burst_max
  --
  bl_check_p : process(ocp_clk)
    variable print_line : line;
    variable cmd_tag    : natural;
    variable cmd_bl     : natural;
  begin
    if ocp_clk'event and ocp_clk = '1' then
      if ((ocp_m2s.Cmd = OCP_CMD_WRITE) or (ocp_m2s.Cmd = OCP_CMD_WRITENONPOST) or (ocp_m2s.Cmd = OCP_CMD_READ)) and
          (ocp_s2m.CmdAccept = '1') then
        cmd_tag  := TO_INTEGER(UNSIGNED(ocp_m2s.Tag));
        cmd_bl   := TO_INTEGER(UNSIGNED(ocp_m2s.BurstLength));
        if ((cmd_bl = BL_ZERO)or(cmd_bl > data_burst_max)) then
          deallocate(print_line);
          write(print_line,string'("  ERROR: Command:"));
          if (ocp_m2s.Cmd = "001") then write(print_line,string'(" OCP_CMD_WRITE")); end if;
          if (ocp_m2s.Cmd = "010") then write(print_line,string'(" OCP_CMD_READ")); end if;
          if (ocp_m2s.Cmd = "101") then write(print_line,string'(" OCP_CMD_WRITENONPOST")); end if;
          write(print_line,string'(" (Tag: ")); write(print_line,cmd_tag);
          write(print_line,string'(", Addr: ")); hex_write(print_line,ocp_m2s.Addr,right,0);
          write(print_line,string'(") with invalid BurstLength: ")); write(print_line,cmd_bl);
          write(print_line,string'(" on cycle: ")); write(print_line,clock_cycle);
          assert not(SEL_EXT_LOG_FILE) report print_line.all severity error;
          assert not(sel_int_log_file) report print_line.all & " (File: " & INT_FILE & ")" severity error;
          if (enable_logging) then writeline_int_ext(file_out_id,print_line,sel_int_log_file); end if;
          inv_bl_cnt <= inv_bl_cnt + 1;
        end if;
      end if;

      if (inv_bl_cnt = 0) then
        status.inv_bl <= '0';
      else
        status.inv_bl <= '1';
      end if;
    end if;
  end process;


  -- Count number of cycles of read/write data
  -- Calculate data transfer rate in MB/s
  --
  utilisation_p : process(ocp_clk)
  begin
    if ocp_clk'event and ocp_clk = '1' then
      if (ocp_m2s.RespAccept ='1') and (ocp_s2m.Resp = OCP_RESP_VALID) then
        data_beat_cnt <= data_beat_cnt + 1;
      end if;

      if (ocp_m2s.DataValid ='1') and (ocp_s2m.DataAccept = '1') then
        data_beat_cnt <= data_beat_cnt + 1;
      end if;

      -- Convert beats to MBytes: data_beat_cnt / 65536
      -- Convert cycles to seconds: clock_cycle / 250,000,000
      -- MBytes/second = (data_beat_cnt/clock_cycle)*(250,000,000/65536))
      if (clock_cycle > 0) then
        data_mb_per_second <= (data_beat_cnt*3815)/clock_cycle;
        data_efficiency    <= (data_beat_cnt*100)/clock_cycle;
      end if;
    end if;
  end process;


  -- Convert OCP command to enumerated type
  decode_cmd_p : process(ocp_clk)
    variable vec_cmd : ocp_CmdT;
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if (ocp_s2m.CmdAccept = '1') then
        vec_cmd := ocp_m2s.Cmd;

        case vec_cmd is
          when OCP_CMD_IDLE         => ocp_cmd <= CMD_IDLE;
          when OCP_CMD_WRITE        => ocp_cmd <= CMD_WRITE;
          when OCP_CMD_READ         => ocp_cmd <= CMD_READ;
          when OCP_CMD_WRITENONPOST => ocp_cmd <= CMD_WRITENONPOST;
          when others               => ocp_cmd <= CMD_INVALID;
        end case;
      end if;
    end if;
  end process;

  -- Convert OCP response to enumerated type
  decode_rsp_p : process(ocp_clk)
    variable vec_rsp : ocp_RespT;
  begin
    if (ocp_clk'event and ocp_clk = '1') then
      if (ocp_m2s.RespAccept ='1') then
        vec_rsp := ocp_s2m.Resp;

        case vec_rsp is
          when OCP_RESP_NONE   => ocp_rsp <= RESP_NONE;
          when OCP_RESP_VALID  => ocp_rsp <= RESP_VALID;
          when OCP_RESP_FAILED => ocp_rsp <= RESP_FAILED;
          when OCP_RESP_ERROR  => ocp_rsp <= RESP_ERROR;
          when others          => ocp_rsp <= RESP_INVALID;
        end case;
      end if;
    end if;
  end process;

end;