--==============================================================================
--
--     File Name: adb3_probe_pkg.vhd
--       Version: 0.2
--          Date: 25 Jun 2009
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: adb3 test probe package
--
--   Description: adb3 test probe package
--
--   Limitations:
--
--         Notes: Component of adb3_test Library
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 25 Jun 2009: Initial design
--                0.2, KDC, 05 May 2011: Added illegal OCP command constants
--
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library std;
use std.textio.all;

library work;
use work.adb3_ocp.all;


package adb3_probe_pkg is

  constant MAX_NUM_OF_TAGS    : natural := 2**ADB3_OCP_TAG_WIDTH;
  constant MAX_NUM_OF_BURSTS  : natural := 2**ADB3_OCP_BURST_WIDTH;
  constant BYTES_PER_BEAT     : natural := ADB3_OCP_DATA_WIDTH/8;

  constant PROBE_K28_0 : std_logic_vector(7 downto 0) := "00011100"; -- x"1C"
  constant PROBE_K28_1 : std_logic_vector(7 downto 0) := "00111100"; -- x"3C"
  constant PROBE_K28_2 : std_logic_vector(7 downto 0) := "01011100"; -- x"5C"
  constant PROBE_K28_3 : std_logic_vector(7 downto 0) := "01111100"; -- x"7C"
  constant PROBE_K28_4 : std_logic_vector(7 downto 0) := "10011100"; -- x"9C"
  constant PROBE_K28_5 : std_logic_vector(7 downto 0) := "10111100"; -- x"BC"
  constant PROBE_K28_6 : std_logic_vector(7 downto 0) := "11011100"; -- x"DC"
  constant PROBE_K28_7 : std_logic_vector(7 downto 0) := "11111100"; -- x"FC"
  constant PROBE_K23_7 : std_logic_vector(7 downto 0) := "11110111"; -- x"F7"
  constant PROBE_K27_7 : std_logic_vector(7 downto 0) := "11111011"; -- x"FB"
  constant PROBE_K29_7 : std_logic_vector(7 downto 0) := "11111101"; -- x"FD"
  constant PROBE_K30_7 : std_logic_vector(7 downto 0) := "11111110"; -- x"FE"

  constant OCP_CMD_ILLEGAL1 : ocp_CmdT := "011";
  constant OCP_CMD_ILLEGAL2 : ocp_CmdT := "100";
  constant OCP_CMD_ILLEGAL3 : ocp_CmdT := "110";
  constant OCP_CMD_ILLEGAL4 : ocp_CmdT := "111";



  -- Status records
  type adb3_ocp_probe_status_r is record
    resp_failed   : std_logic;  -- OCP_RESP_FAILED detected (sticky)
    resp_error    : std_logic;  -- OCP_RESP_ERROR detected (sticky)
    inv_cmd       : std_logic;  -- invalid command detected (sticky)
    inv_addr      : std_logic;  -- invalid address detected (sticky)
    inv_bl        : std_logic;  -- invalid burst length detected (sticky)
    read_tag_err  : std_logic;  -- Read command/data tag mismatch error (sticky)
    read_cmp_err  : std_logic;  -- Read data for completed read command error (sticky)
    write_cmp_err : std_logic;  -- Write data for completed write command error (sticky)
    write_be_err  : std_logic;  -- Write data byte enable error (sticky)
  end record;

  -- Define ststus pass
  constant ADB3_OCP_PROBE_STATUS_OK : adb3_ocp_probe_status_r :=
    (resp_failed   => '0',
     resp_error    => '0',
     inv_cmd       => '0',
     inv_addr      => '0',
     inv_bl        => '0',
     read_tag_err  => '0',
     read_cmp_err  => '0',
     write_cmp_err => '0',
     write_be_err  => '0');

--  type adb3_dma_probe_status_r is record
--    desc_err      : std_logic;  -- PCIE/MPTL descriptor read error (sticky)
--    ctrl_err      : std_logic;  -- Control setup error (sticky)
--    write_err     : std_logic;  -- PCIe/MPTL data write error (sticky)
--    fifo_err      : std_logic;  -- fifo fail error (sticky)
--    read_err      : std_logic;  -- PCIe/MPTL data read error (sticky)
--    dma_complete  : boolean;    -- DMA FSMs are idle with read and write complete
--  end record;

  -- File id that must be used to write to combined log file
  file ext_file_out_id : text;


  -- Function definitions
  function vectorise(
    s : in std_logic)
  return std_logic_vector;

  function enables_to_bytes(
    byte_en : in std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0))
  return natural;

  function conv_nibble_hex(
    nibble : in std_logic_vector(3 downto 0))
  return character;


  -- Procedure definitions
  procedure open_file(
    file_name    : in string;
    file_mode    : in file_open_kind;
    file file_id : text);

  procedure close_file(
    file file_id : text);

  procedure writeline_int_ext(
    file int_file_out_id : text;
    file_out_line        : inout line;
    use_int_log_file     : in boolean);

  procedure hex_write(
    hex_line  : inout line;
    vector    : in    std_logic_vector;
    justified : in    side;
    field     : in    natural);


  -- Component definitions
  component adb3_ocp_transaction_probe
  generic(
    enable_logging   : boolean := true;
    sel_int_log_file : boolean := true;
    int_log_filename : string  := "ocp_probe";
    addr_align_bits  : natural := ADB3_OCP_BE_WIDTH_LOG2;
    addr_width_max   : natural := ADB3_OCP_ADDR_WIDTH;
    data_burst_max   : natural := MAX_NUM_OF_BURSTS;
    enable_tag_check : boolean := true);
  port(
    ocp_clk : in    std_logic;
    ocp_m2s : in    adb3_ocp_m2sT;
    ocp_s2m : in    adb3_ocp_s2mT;
    status  : out   adb3_ocp_probe_status_r);
  end component;

  component adb3_mptl64_packet_probe
  generic(
    enable_logging   : boolean := true;
    sel_int_log_file : boolean := true;
    int_log_filename : string  := "mptl_probe";
    number_of_ports  : natural := 4;
    monitor_channel  : natural := 4);
  port(
    -- MPTL Interface
    clk               : in  std_logic;
    mptl64_data       : in  std_logic_vector(63 downto 0);
    mptl64_valid      : in  std_logic);
  end component;

  component adb3_mptl128_packet_probe
  generic(
    enable_logging   : boolean := true;
    sel_int_log_file : boolean := true;
    int_log_filename : string  := "mptl_probe";
    number_of_ports  : natural := 4;
    monitor_channel  : natural := 4);
  port(
    -- MPTL Interface
    clk               : in  std_logic;
    mptl128_data      : in  std_logic_vector(127 downto 0);
    mptl128_valid     : in  std_logic);
  end component;

--  component adb3_dma_test_probe
--  generic(
--    log_file_name : string := "dma_probe");
--  port(
--    -- OCP Interface
--    ocp_clk      : in  std_logic;
--    irq          : in  std_logic;
--    pcie_ocp_m2s : in  adb3_ocp_m2sT;
--    pcie_ocp_s2m : in  adb3_ocp_s2mT;
--    mptl_ocp_m2s : in  adb3_ocp_m2sT;
--    mptl_ocp_s2m : in  adb3_ocp_s2mT;
--    smap_ocp_m2s : in  adb3_ocp_m2sT;
--    smap_ocp_s2m : in  adb3_ocp_s2mT;
--    dma_control  : in  adb3_dma_control_register_type;
--    dma_status   : in  adb3_dma_status_register_type;
--    status       : out adb3_dma_probe_status_r);
--  end component;

end package;


package body adb3_probe_pkg is

  -- Function to convert std_logic signal to 1-bit std_logic_vector
  -- Use in type conversion functions as std_logic cannot be used
  function vectorise(
    s: std_logic)
  return std_logic_vector is
    variable v : std_logic_vector(0 downto 0);
  begin
    v(0) := s;
    return v;
  end;


  -- Function to return the enabled number of bytes in byte enable vector
  function enables_to_bytes(
    byte_en: std_logic_vector(ADB3_OCP_DATA_WIDTH/8-1 downto 0))
  return natural is
    variable bytes : natural;
  begin
    bytes := 0;

    for i in 0 to (ADB3_OCP_DATA_WIDTH/8-1)
      loop
        if (byte_en(i) = '1') then
          bytes := bytes + 1;
        end if;
      end loop;

    return bytes;
  end;


  -- Calculate the string equivalent of
  -- a std_logic_vector nibble
  function conv_nibble_hex(
    nibble : in std_logic_vector(3 downto 0))
  return character is
    variable found_w : boolean := false;
    variable found_z : boolean := false;
    variable found_h : boolean := false;
    variable found_l : boolean := false;
    variable found_x : boolean := false;
    variable found_u : boolean := false;
    variable found_d : boolean := false;
  begin
    -- Detect presence of non '0','1' values
    -- In order of importance
    for i in 0 to 3 loop
      case nibble(i) is
        when 'U' => found_u := true;
        when 'X' => found_x := true;
        when '0' => null;
        when '1' => null;
        when 'Z' => found_z := true;
        when 'W' => found_w := true;
        when 'L' => found_l := true;
        when 'H' => found_h := true;
        when '-' => found_d := true;
      end case;
    end loop;

    -- Assign character for all nibble values
    -- Non-hex character results are prioritised U,X,Z,W,L,H,-
    case nibble is
      when "0000" => return '0';
      when "0001" => return '1';
      when "0010" => return '2';
      when "0011" => return '3';
      when "0100" => return '4';
      when "0101" => return '5';
      when "0110" => return '6';
      when "0111" => return '7';
      when "1000" => return '8';
      when "1001" => return '9';
      when "1010" => return 'A';
      when "1011" => return 'B';
      when "1100" => return 'C';
      when "1101" => return 'D';
      when "1110" => return 'E';
      when "1111" => return 'F';
      when others => if    (found_u) then return 'U';
                     elsif (found_x) then return 'X';
                     elsif (found_z) then return 'Z';
                     elsif (found_w) then return 'W';
                     elsif (found_l) then return 'L';
                     elsif (found_h) then return 'H';
                     elsif (found_d) then return '-';
                     else return '?';
                     end if;
    end case;
  end function;


  ------------------------------------------------------------------------------
  -- Procedure Name: open_file
  -- Dependencies: package std.textio
  -- Procedure to open file file_name in mode file_mode
  -- Tests file_mode for validity and asserts failure if invalid
  -- Decodes file_status and asserts failure in all cases except:
  --   OPEN_OK      : file was opened successfully for first time
  --   STATUS_ERROR : file was previously opened (warning)
  -- Requires file definition in calling process or architecture:
  --   file <file_id_name> : text;
  -- Procedure call:
  --   open_file(<file_name>,<file_mode>,<file_id_name>);
  ------------------------------------------------------------------------------

  procedure open_file(
    file_name : in string;
    file_mode : in file_open_kind;
    file file_id : text)
  is
    variable file_status : file_open_status;
    variable mode_line   : line;
    variable status_line : line;
    variable report_line : line;
  begin
    case file_mode is
      when READ_MODE   => write(mode_line,string'("'Read Mode'"));
      when WRITE_MODE  => write(mode_line,string'("'Write Mode'"));
      when APPEND_MODE => write(mode_line,string'("'Append Mode'"));
      when others      => assert false report "Illegal file mode" severity failure;
    end case;

    file_open(file_status,file_id,file_name,file_mode);

    case file_status is
      when OPEN_OK      => write(status_line,string'("'File opened successfully'"));
      when STATUS_ERROR => write(status_line,string'("'File previously opened'"));
      when NAME_ERROR   => if (file_mode = WRITE_MODE) then
                             write(status_line,string'("'Cannot create file'"));
                           else
                             write(status_line,string'("'File does not exist'"));
                           end if;
      when MODE_ERROR   => write(status_line,string'("'File exists but cannot be opened in specific mode'"));
      when others       => write(status_line,string'("'Illegal file status'"));
    end case;

    write(report_line,string'("Opened file '" & file_name & "' in " & mode_line.all & ", Status: " & status_line.all));

    case file_status is
      when OPEN_OK      => assert false report report_line.all severity note;
      when STATUS_ERROR => assert false report report_line.all severity warning;
      when NAME_ERROR   => assert false report report_line.all severity failure;
      when MODE_ERROR   => assert false report report_line.all severity failure;
      when others       => assert false report report_line.all severity failure;
    end case;
  end;

  procedure close_file(
    file file_id : text)
  is
  begin
    file_close(file_id);
  end;

  procedure writeline_int_ext(
    file int_file_out_id : TEXT;
    file_out_line    : inout LINE;
    use_int_log_file : in boolean)
  is
  begin
    if (use_int_log_file) then
      writeline(int_file_out_id,file_out_line);
    else
      writeline(ext_file_out_id,file_out_line);
    end if;
  end;


  -- Convert the std_logic_vector input 'vector' to hex character string
  -- and write to line 'hex_line'
  -- Format <length>x<nibble>..<nibble>, eg 11x3FF
  -- Use justified input (left,right) to position output string in field width <field>
  -- if <field> is greater than width of output string
  procedure hex_write(
    hex_line  : inout line;
    vector    : in    std_logic_vector;
    justified : in    side;
    field     : in    natural)
  is
    variable full_nibbles  : natural;
    variable vector_space  : integer;
    variable nibble_top    : natural;
    variable vector_top    : natural;
    variable vector_bot    : natural;
    variable vector_low    : natural;
    variable vector_len    : natural;
    variable width_line    : line;
    variable width_len     : natural;
    variable nibble        : std_logic_vector(3 downto 0);
    variable nibble_hex    : character;
  begin
    -- Get lower bound and length of vector input
    vector_low := vector'low;
    vector_len := vector'length;

    -- Round up vector length to next nearest nibble
    full_nibbles := (vector_len + 4 - 1) / 4;

    -- Get length of string containing decimal vector length
    deallocate(width_line);
    write(width_line,integer'image(vector_len));
    width_len := width_line'length;

    -- Calculate difference between field size and length of hex string
    -- String length = Width chars + "x" + full nibbles
    vector_space := field-(width_len+1+full_nibbles);

    -- Insert space before hex string if appropriate
    if (vector_space > 0)and(justified=right) then
      for i in 1 to vector_space loop
        write(hex_line,string'(" "));
      end loop;
    end if;

    write(hex_line,integer'image(vector_len));
    write(hex_line,string'("x"));

    -- Process bits one at a time starting at msb
    for i in (full_nibbles-1) downto 0 loop
      vector_bot := i*4 + vector_low;
      if (i = (full_nibbles-1)) then                      -- Write top nibble bits (may be < 4 bits)
        nibble_top := (vector'length-1);
        vector_top := nibble_top + vector_low;
        nibble     := "0000";
        nibble((nibble_top rem 4) downto 0) := vector(vector_top downto vector_bot);
      else                                                -- Write remaining nibble bits (4 bits)
        vector_top := ((i+1)*4)-1 + vector_low;
        nibble     := vector(vector_top downto vector_bot);
      end if;
      nibble_hex := conv_nibble_hex(nibble);
      write(hex_line,nibble_hex);
    end loop;

    -- Insert space after hex string if appropriate
    if (vector_space > 0)and(justified=left) then
      for i in 1 to vector_space loop
        write(hex_line,string'(" "));
      end loop;
    end if;
  end procedure;

end package body;

