--==============================================================================
--
--     File Name: adb3_target_inc_pkg_sim_mptl_6t1.vhd
--       Version: 0.4
--          Date: 25 Aug 2011
--         Model: Simulation only
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package adb3_target_inc_pkg
--
--   Description: Package used in target FPGAs during Full MPTL simulation
--
--   Limitations: ADM-XRC-6T1 Full-MPTL simulation only
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 01 Jun 2010: Initial design
--                0.2, KDC, 26 May 2011:
--                  - Changed DDR3_BANK_ROW_WIDTH to 14 to support 1Gib and
--                    2Gib devices
--                  - Added LCLK_DOUBLE constant to support double ended LCLK
--                0.3, KDC, 25 Jul 2011:
--                  - Removed BOARD_TYPE constant
--                  - Added clks_in_t, clks_mgt_in_t, clks_out_t
--                  - Added mptl_b2t_t, mptl_t2b_t
--                0.4, KDC, 25 Aug 2011:
--                  - Added xrm_gpio_t, pn4_gpio_t, pn6_gpio_t, gpio_inout_t
--                  - Added ddr3_addr_out_t, ddr3_ctrl_out_t, ddr3_data_inout_t,
--                    ddr3_clk_out_t, ddr3_addr_out_array_t, ddr3_ctrl_out_array_t,
--                    ddr3_data_inout_array_t, ddr3_clk_out_array_t, DDR3_BANK0,
--                    mem_addr_out_t, mem_ctrl_out_t, mem_data_inout_t, mem_clk_out_t
--                  - Renamed Pn4 and Pn6 max range constants to SE and DE.
--                  - Removed XRM_GPIO_VALID, PN4_GPIO_VALID, PN6_GPIO_VALID,
--                    DDR3_16_BYTE_ADDR_WIDTH
--                  - Added XRM_GPIO_WIDTH, mem_byte_addr_width_array_t and
--                    MEM_BYTE_ADDR_WIDTH_ARRAY
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;


package adb3_target_inc_pkg is

  -------------------------
  -- General definitions --
  -------------------------

  -- Select target use
  -- SIM_OCP:  Generate simulation code. Direct OCP to OCP connections between bridge (TB) and target (ie bypass MPTL)
  -- SIM_MPTL: Generate simulation code. Full OCP to MPTL to OCP connections between bridge (TB) and target
  -- SYN_NGC:  Generate synthesis code.  Instantiate .ngc core for OCP-MPTL interface to bridge
  constant TARGET_USE : target_use_t := SIM_MPTL;

  -- Record type for double ended signals
  type std_logic_dbl_t is record
    p : std_logic;
    n : std_logic;
  end record;


  -----------------------
  -- Clock definitions --
  -----------------------

  -- Target FPGA clock inputs
  constant REF_CLK_FREQ_HZ : natural := 200000000;

  type clks_in_t is record
    ref_clk      : std_logic_dbl_t;
    lclk         : std_logic_dbl_t;
    xrm_gclk_m2c : std_logic_dbl_t;
  end record;

  -- Target FPGA MGT clock inputs
  -- MGT113 CLK0 (buffered) = 250MHz oscillator reference clock
  -- MGT113 CLK1 (buffered) = 250MHz PCIe reference clock (derived from 100MHz input on P5)
  -- MGT114 CLK0 (buffered) = 250MHz PCIe reference clock (derived from 100MHz input on P5)
  -- MGT115 CLK0            = 250MHz oscillator reference clock (reserved for MPTL interface use)
  -- MGT116 CLK0 (buffered) = 250MHz oscillator reference clock
  -- MGT117 CLK0 (buffered) = XRM reference clock

  -- Select MGT clock inputs to buffer                        18   17   16   15   14   13   12   11   10
  constant MGT_CLKS_VALID : std_logic_vector(17 downto 0) := "00"&"01"&"01"&"00"&"01"&"11"&"00"&"00"&"00"; -- "clk1 clk0"

  type clks_mgt_in_t is record
    mgt113_clk0 : std_logic_dbl_t;
    mgt113_clk1 : std_logic_dbl_t;
    mgt114_clk0 : std_logic_dbl_t;
    mgt115_clk0 : std_logic_dbl_t;
    mgt116_clk0 : std_logic_dbl_t;
    mgt117_clk0 : std_logic_dbl_t;
  end record;

  -- Target FPGA clock outputs
  type clks_out_t is record
    xrm_mgtclk_c2m : std_logic_dbl_t;
  end record;


  ----------------------
  -- GPIO definitions --
  ----------------------

  -- Target GPIO interfaces
  constant XRM_GPIO_WIDTH : natural range 0 to MAX_GPIO_FR_WIDTH := 146;
  constant PN4_GPIO_WIDTH : natural range 0 to MAX_GPIO_DE_WIDTH := 32; -- 0/32 Build option (p)
  constant PN6_GPIO_WIDTH : natural range 0 to MAX_GPIO_SE_WIDTH := 38;

  type xrm_gpio_t is record
    xrm_da_p    : std_logic_vector(15 downto 0);
    xrm_da_n    : std_logic_vector(15 downto 0);
    xrm_db_p    : std_logic_vector(15 downto 0);
    xrm_db_n    : std_logic_vector(15 downto 0);
    xrm_dc_p    : std_logic_vector(15 downto 0);
    xrm_dc_n    : std_logic_vector(15 downto 0);
    xrm_dd_p    : std_logic_vector(15 downto 0);
    xrm_dd_n    : std_logic_vector(15 downto 0);
    xrm_da_cc_p : std_logic;
    xrm_da_cc_n : std_logic;
    xrm_db_cc_p : std_logic;
    xrm_db_cc_n : std_logic;
    xrm_dc_cc_p : std_logic;
    xrm_dc_cc_n : std_logic;
    xrm_dd_cc_p : std_logic;
    xrm_dd_cc_n : std_logic;
    xrm_sa      : std_logic_vector(1 downto 0);
    xrm_sb      : std_logic_vector(1 downto 0);
    xrm_sc      : std_logic_vector(1 downto 0);
    xrm_sd      : std_logic_vector(3 downto 0);
  end record;

  type pn4_gpio_t is record
    gpio_p : std_logic_vector(PN4_GPIO_WIDTH downto 1);
    gpio_n : std_logic_vector(PN4_GPIO_WIDTH downto 1);
  end record;

  type pn6_gpio_t is record
    gpio : std_logic_vector(PN6_GPIO_WIDTH downto 1);
  end record;

  -- GPIO interface
  type gpio_inout_t is record
    xrm_gpio : xrm_gpio_t;
    pn4_gpio : pn4_gpio_t;
    pn6_gpio : pn6_gpio_t;
  end record;


  ------------------------
  -- Memory definitions --
  ------------------------

  -- Target FPGA DDR3 SDRAM interface
  constant DDR3_BANKS : natural range 0 to MAX_MEM_BANKS := 4;

  -- Interface row address width
--  constant DDR3_BANK_ROW_WIDTH : natural := 13; -- Up to 1 Gib parts =  8K row x 1K col x 8 banks x 16 bits
  constant DDR3_BANK_ROW_WIDTH : natural := 14; -- Up to 2 Gib parts = 16K row x 1K col x 8 banks x 16 bits
--  constant DDR3_BANK_ROW_WIDTH : natural := 15; -- Up to 4 Gib parts = 32K row x 1K col x 8 banks x 16 bits

  -- Interface data width
  constant DDR3_BANK_DATA_WIDTH : natural := 32; -- 2 off 16-bit parts

  -- Interface address width = Bank(3)+Row(13/14/15)+Col(10) and convert to byte addressing (+2)
  constant DDR3_BYTE_ADDR_WIDTH : natural := 3 + DDR3_BANK_ROW_WIDTH + 10 + 2;

  -- DDR3 memory interface types
  type ddr3_addr_out_t is record
    a : std_logic_vector(DDR3_BANK_ROW_WIDTH-1 downto 0);
  end record;

  type ddr3_ctrl_out_t is record
    ras_l   : std_logic;
    cas_l   : std_logic;
    we_l    : std_logic;
    cs_l    : std_logic_vector(0 downto 0);
    dm      : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1  downto 0);
    ba      : std_logic_vector(2 downto 0);
    reset_l : std_logic;
    odt     : std_logic_vector(0 downto 0);
    cke     : std_logic_vector(0 downto 0);
  end record;

  type ddr3_data_inout_t is record
    dq    : std_logic_vector(DDR3_BANK_DATA_WIDTH-1 downto 0);
    dqs_p : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1 downto 0);
    dqs_n : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1 downto 0);
  end record;

  type ddr3_clk_out_t is record
    clk_p : std_logic_vector(0 downto 0);
    clk_n : std_logic_vector(0 downto 0);
  end record;

  type ddr3_addr_out_array_t   is array (0 to DDR3_BANKS-1) of ddr3_addr_out_t;
  type ddr3_ctrl_out_array_t   is array (0 to DDR3_BANKS-1) of ddr3_ctrl_out_t;
  type ddr3_data_inout_array_t is array (0 to DDR3_BANKS-1) of ddr3_data_inout_t;
  type ddr3_clk_out_array_t    is array (0 to DDR3_BANKS-1) of ddr3_clk_out_t;


  -- Target FPGA memory interface
  constant MEM_BANKS  : natural range 0 to MAX_MEM_BANKS := DDR3_BANKS;
  constant DDR3_BANK0 : natural range 0 to (MEM_BANKS - DDR3_BANKS) := 0;

  -- Target FPGA memory interface bank address width array
  type mem_byte_addr_width_array_t is array (0 to MEM_BANKS-1) of natural;
  constant MEM_BYTE_ADDR_WIDTH_ARRAY : mem_byte_addr_width_array_t := ((DDR3_BANKS+DDR3_BANK0-1) downto DDR3_BANK0 => DDR3_BYTE_ADDR_WIDTH);

  type mem_addr_out_t is record
    ddr3_addr_out : ddr3_addr_out_array_t;
  end record;

  type mem_ctrl_out_t is record
    ddr3_ctrl_out : ddr3_ctrl_out_array_t;
  end record;

  type mem_data_inout_t is record
    ddr3_data_inout : ddr3_data_inout_array_t;
  end record;

  type mem_clk_out_t is record
    ddr3_clk_out : ddr3_clk_out_array_t;
  end record;


  --------------------------------
  -- MPTL interface definitions --
  --------------------------------

  -- Target FPGA MPTL interface OCP channels
  constant DS_CHANNELS  : natural range 1 to MAX_DS_CHANNELS  := 1;
  constant DMA_CHANNELS : natural range 1 to MAX_DMA_CHANNELS := 4;
  constant DM_CHANNELS  : natural range 1 to MAX_DM_CHANNELS  := 1;

  -- Target FPGA MPTL interface OCP channel address widths
  -- Direct Slave:      22 bits =   4 MiB (expandable using bridge paging)
  -- DMA address space: 39 bits = 512 GiB
  -- Direct Master:     32 bits =   4 GiB
  constant DS_ADDR_WIDTH  : natural range 1 to ADB3_OCP_ADDR_WIDTH := 22;
  constant DMA_ADDR_WIDTH : natural range 1 to ADB3_OCP_ADDR_WIDTH := 39;
  constant DM_ADDR_WIDTH  : natural range 1 to ADB3_OCP_ADDR_WIDTH := 32;

  -- Target FPGA MPTL interface serial width
  constant MPTL_SER_WIDTH : natural range 1 to MAX_MPTL_SER_WIDTH := 4;

  -- MPTL data interface
  type mptl_pins_t is record
    p : std_logic_vector(MPTL_SER_WIDTH-1 downto 0);
    n : std_logic_vector(MPTL_SER_WIDTH-1 downto 0);
  end record;

  subtype mptl_b2t_t is mptl_pins_t;
  subtype mptl_t2b_t is mptl_pins_t;

  -- MPTL sideband interface (b2t)
  type mptl_sb_b2t_t is record
    mptl_bridge_gtp_online_l : std_logic;
  end record;

  -- MPTL sideband interface (t2b)
  type mptl_sb_t2b_t is record
    mptl_target_configured_l : std_logic;
    mptl_target_gtp_online_l : std_logic;
  end record;

end;


package body adb3_target_inc_pkg is

end;