--==============================================================================
--
--     File Name: adb3_target_inc_pkg_syn_ngc_6tadv8_pcie.vhd
--       Version: 0.1
--          Date: 19 Aug 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package adb3_target_inc_pkg
--
--   Description: Package used in target FPGAs during synthesis
--
--   Limitations: ADM-XRC-6TADV8 synthesis only
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 19 Aug 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;


package adb3_target_inc_pkg is

  -------------------------
  -- General definitions --
  -------------------------

  -- Select target use
  -- SIM_OCP:  Generate simulation code. Direct OCP to OCP connections between bridge (TB) and target (ie bypass MPTL)
  -- SIM_MPTL: Generate simulation code. Full OCP to MPTL to OCP connections between bridge (TB) and target
  -- SYN_NGC:  Generate synthesis code.  Instantiate .ngc core for OCP-MPTL interface to bridge
  constant TARGET_USE : target_use_t := SYN_NGC;

  -- Record type for double ended signals
  type std_logic_dbl_t is record
    p : std_logic;
    n : std_logic;
  end record;


  -----------------------
  -- Clock definitions --
  -----------------------

  -- Target FPGA clock inputs
  constant REF_CLK_FREQ_HZ : natural := 200000000;

  type clks_in_t is record
    ref_clk : std_logic_dbl_t;
  end record;

  -- Target FPGA MGT clock inputs
  -- MGT110 CLK0 (buffered) = Reference clock 156.25MHz
  -- MGT112 CLK0            = PCIe Reference clock 100 MHz (reserved for PCIe interface use)
  -- MGT113 CLK0 (buffered) = Reference clock 156.25MHz
  -- MGT115 CLK0 (buffered) = Reference clock 156.25MHz
  -- MGT117 CLK0 (buffered) = Reference clock 156.25MHz

  -- Select MGT clock inputs to buffer                        18   17   16   15   14   13   12   11   10
  constant MGT_CLKS_VALID : std_logic_vector(17 downto 0) := "00"&"01"&"00"&"01"&"00"&"01"&"00"&"00"&"01"; -- "clk1 clk0"

  type clks_mgt_in_t is record
    mgt110_clk0 : std_logic_dbl_t;
    mgt112_clk0 : std_logic_dbl_t;
    mgt113_clk0 : std_logic_dbl_t;
    mgt115_clk0 : std_logic_dbl_t;
    mgt117_clk0 : std_logic_dbl_t;
  end record;

  -- Target FPGA clock outputs
  type clks_out_t is record
    unused : std_logic_dbl_t;
  end record;


  ----------------------
  -- GPIO definitions --
  ----------------------

  -- Target GPIO interfaces
  constant XRM_GPIO_WIDTH : natural range 0 to MAX_GPIO_FR_WIDTH := 146;
  constant PN4_GPIO_WIDTH : natural range 0 to MAX_GPIO_DE_WIDTH := 0;
  constant PN6_GPIO_WIDTH : natural range 0 to MAX_GPIO_SE_WIDTH := 8;

  type pn6_gpio_t is record
    gpio : std_logic_vector(PN6_GPIO_WIDTH downto 1);
  end record;

  -- GPIO interface
  type gpio_inout_t is record
    pn6_gpio : pn6_gpio_t;
  end record;


  ------------------------
  -- Memory definitions --
  ------------------------

  -- Target FPGA DDR3 SDRAM interface
  constant DDR3_BANKS : natural range 0 to MAX_MEM_BANKS := 4;

  -- Interface row address width
--  constant DDR3_BANK_ROW_WIDTH : natural := 13; -- Up to 1 Gib parts =  8K row x 1K col x 8 banks x 16 bits
  constant DDR3_BANK_ROW_WIDTH : natural := 14; -- Up to 2 Gib parts = 16K row x 1K col x 8 banks x 16 bits
--  constant DDR3_BANK_ROW_WIDTH : natural := 15; -- Up to 4 Gib parts = 32K row x 1K col x 8 banks x 16 bits

  -- Interface data width
  constant DDR3_BANK_DATA_WIDTH : natural := 32; -- 2 off 16-bit parts

  -- Interface address width = Bank(3)+Row(13/14/15)+Col(10) and convert to byte addressing (+2)
  constant DDR3_BYTE_ADDR_WIDTH : natural := 3 + DDR3_BANK_ROW_WIDTH + 10 + 2;

  -- DDR3 memory interface types
  type ddr3_addr_out_t is record
    a : std_logic_vector(DDR3_BANK_ROW_WIDTH-1 downto 0);
  end record;

  type ddr3_ctrl_out_t is record
    ras_l   : std_logic;
    cas_l   : std_logic;
    we_l    : std_logic;
    cs_l    : std_logic_vector(0 downto 0);
    dm      : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1  downto 0);
    ba      : std_logic_vector(2 downto 0);
    reset_l : std_logic;
    odt     : std_logic_vector(0 downto 0);
    cke     : std_logic_vector(0 downto 0);
  end record;

  type ddr3_data_inout_t is record
    dq    : std_logic_vector(DDR3_BANK_DATA_WIDTH-1 downto 0);
    dqs_p : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1 downto 0);
    dqs_n : std_logic_vector((DDR3_BANK_DATA_WIDTH/8)-1 downto 0);
  end record;

  type ddr3_clk_out_t is record
    clk_p : std_logic_vector(0 downto 0);
    clk_n : std_logic_vector(0 downto 0);
  end record;

  type ddr3_addr_out_array_t   is array (0 to DDR3_BANKS-1) of ddr3_addr_out_t;
  type ddr3_ctrl_out_array_t   is array (0 to DDR3_BANKS-1) of ddr3_ctrl_out_t;
  type ddr3_data_inout_array_t is array (0 to DDR3_BANKS-1) of ddr3_data_inout_t;
  type ddr3_clk_out_array_t    is array (0 to DDR3_BANKS-1) of ddr3_clk_out_t;

  -- Target FPGA memory interface
  constant MEM_BANKS  : natural range 0 to MAX_MEM_BANKS := DDR3_BANKS;
  constant DDR3_BANK0 : natural range 0 to (MEM_BANKS - DDR3_BANKS) := 0;

  -- Target FPGA memory interface bank address width array
  type mem_byte_addr_width_array_t is array (0 to MEM_BANKS-1) of natural;
  constant MEM_BYTE_ADDR_WIDTH_ARRAY : mem_byte_addr_width_array_t := ((DDR3_BANKS+DDR3_BANK0-1) downto DDR3_BANK0 => DDR3_BYTE_ADDR_WIDTH);

  type mem_addr_out_t is record
    ddr3_addr_out : ddr3_addr_out_array_t;
  end record;

  type mem_ctrl_out_t is record
    ddr3_ctrl_out : ddr3_ctrl_out_array_t;
  end record;

  type mem_data_inout_t is record
    ddr3_data_inout : ddr3_data_inout_array_t;
  end record;

  type mem_clk_out_t is record
    ddr3_clk_out : ddr3_clk_out_array_t;
  end record;


  --------------------------------
  -- PCIe interface definitions --
  --------------------------------

  -- Target FPGA PCIe interface OCP channels
  constant DS_CHANNELS  : natural range 1 to MAX_DS_CHANNELS  := 1;
  constant DMA_CHANNELS : natural range 1 to MAX_DMA_CHANNELS := 2;
  constant DM_CHANNELS  : natural range 1 to MAX_DM_CHANNELS  := 1;

  -- Target FPGA PCIe interface OCP channel address widths
  -- Direct Slave:      22 bits =   4 MiB (expandable using bridge paging)
  -- DMA address space: 39 bits = 512 GiB
  -- Direct Master:     32 bits =   4 GiB
  constant DS_ADDR_WIDTH  : natural range 1 to ADB3_OCP_ADDR_WIDTH := 22;
  constant DMA_ADDR_WIDTH : natural range 1 to ADB3_OCP_ADDR_WIDTH := 39;
  constant DM_ADDR_WIDTH  : natural range 1 to ADB3_OCP_ADDR_WIDTH := 32;

  -- Target FPGA PCIe interface serial width
  constant PCIE_SER_WIDTH : natural range 1 to MAX_PCIE_SER_WIDTH := 4; -- x4 interface
--  constant PCIE_SER_WIDTH : natural range 1 to MAX_PCIE_SER_WIDTH := 8; -- x8 interface

  -- PCIe data interface
  type pcie_pins_t is record
    p : std_logic_vector(PCIE_SER_WIDTH-1 downto 0);
    n : std_logic_vector(PCIE_SER_WIDTH-1 downto 0);
  end record;

  subtype pcie_h2t_t is pcie_pins_t;
  subtype pcie_t2h_t is pcie_pins_t;


  ---------------------------
  -- Custom IO definitions --
  ---------------------------

  -- PCIe core custom interfaces
  constant adb3_flash_data_width    : natural := 16;
  constant adb3_flash_address_width : natural := 25;

  type adb3_model_in_t is record
    -- Flash Interface
    mvmro_mp      : std_logic;
    -- AVR interface
    avr_tx        : std_logic;
    avr_fpga_sgnl : std_logic;
    avr_alarm_l   : std_logic;
  end record;

  type adb3_model_out_t is record
    -- AVR interface
    avr_rx      : std_logic;
    avr_mp_sgnl : std_logic;
  end record;

  type adb3_model_inout_t is record
    -- Flash Interface
    fl_a      : std_logic_vector(adb3_flash_address_width downto 1);
    fl_sel_n  : std_logic;
    fl_oe_n   : std_logic;
    fl_we_n   : std_logic;
    fl_rst_n  : std_logic;
    fl_d      : std_logic_vector(adb3_flash_data_width-1 downto 0);
    fl_adv_n  : std_logic;
    fl_clk    : std_logic;
    fl_wait_n : std_logic;
    -- Power
    -- IIC0 / System Monitor
    iic0_scl  : std_logic;
    iic0_sda  : std_logic;
    -- IIC1
    iic1_scl  : std_logic;
    iic1_sda  : std_logic;
  end record;

  type custom_in_t is record
    adb3_model : adb3_model_in_t;
  end record;

  type custom_out_t is record
    adb3_model : adb3_model_out_t;
  end record;

  type custom_inout_t is record
    adb3_model : adb3_model_inout_t;
  end record;

end;


package body adb3_target_inc_pkg is

end;