--==============================================================================
--
--     File Name: adb3_target_tb_inc_pkg_6tge.vhd
--       Version: 0.1
--          Date: 18 Jul 2011
--         Model: Simulation only
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Package adb3_target_inc_pkg
--
--   Description: Package used in target testbench
--
--   Limitations: ADM-XRC-6TGE testbench
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 18 Jul 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.ddr3_sdram_pkg.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;


package adb3_target_tb_inc_pkg is

  constant BOARD_TYPE : string(1 to 12) := "ADM-XRC-6TGE";

  -- Target FPGA clock outputs expected frequency
  type clks_out_exp_t is record
    xrm_mgtclk_c2m : natural;
  end record;

  -- DDR3 component models
  constant DDR3_1G_PART : part_t := MT41J64M16_187E;
  constant DDR3_2G_PART : part_t := MT41J128M16_187E;
  constant DDR3_4G_PART : part_t := MT41J256M16_187E;

  -------------------
  -- Build options --
  -------------------

  -- Option z: Target device
  -- Option y: Target device speed

  -- Option m: DDR3 Memory size
  -- 1/2/4 Gib DDR3 SDRAM parts fitted
  constant OPTION_M : part_t := DDR3_1G_PART;

  -- Option c: Cooling

  -- Option p: Pn4 Connector
  -- True:     Pn4 connector fitted
  -- False: No Pn4 connector fitted
  constant OPTION_P : boolean := false; -- unused

  -- Option e: Ethernet link select
  -- True:                 No Ethernet Link - x8 MGT link on Pn6
  -- False: 10/100/1000Base-T Ethernet link - x4 MGT on Pn6
  constant OPTION_E : boolean := false; -- unused

  -- Option g: Ref Clk 1 select
  -- True:  no MGT clock input on Pn6 - 46 GPIO on Pn6
  -- False:    MGT clock input on Pn6 - 44 GPIO on Pn6
  constant OPTION_G : boolean := false;


  ------------------------------
  -- Clock period definitions --
  ------------------------------

  constant CLK_666_PER : time :=  1.50 ns;
  constant CLK_250_PER : time :=  4.00 ns;
  constant CLK_200_PER : time :=  5.00 ns;
  constant CLK_125_PER : time :=  8.00 ns;
  constant CLK_100_PER : time := 10.00 ns;


  -- Board clocks
  -- Pn6 reference
  constant PN6_REFCLK : time := CLK_100_PER; -- (Build option g false)

  -- XRM MGT reference
  constant XRM_MGTCLK_M2C : time := CLK_125_PER;

  -- U11 input P5 PCIe reference = 100MHz
  -- Output is controlled by switch SW1H
  constant PCIEREFCLK_250_0 : time := CLK_250_PER; -- Bridge clock
  constant PCIEREFCLK_250_1 : time := CLK_250_PER; -- Target clock
  constant PCIEREFCLK_VAR_0 : time := CLK_125_PER; -- Target clock (SW1H = closed)
--  constant PCIEREFCLK_VAR_0 : time := CLK_250_PER; -- Target clock (SW1H = open)

  -- U227 input 25MHz 25ppm oscillator
  constant PROG_CLK_0A : time := CLK_250_PER; -- Bridge MPTL clock
  constant PROG_CLK_0B : time := CLK_250_PER; -- Target MPTL clock
  constant PROG_CLK_1  : time := CLK_200_PER; -- Target clock (variable)
  constant PROG_CLK_2  : time := CLK_200_PER; -- Target clock (variable)
  constant PROG_CLK_3A : time := CLK_125_PER; -- Bridge clock (variable) (single ended)
  constant PROG_CLK_3B : time := CLK_125_PER; -- Ethernet Phy (variable) (single ended)

  -- U38 input 200MHz 50ppm oscillator
  constant REFCLK200M_0 : time := CLK_200_PER; -- Bridge reference clock
  constant REFCLK200M_1 : time := CLK_200_PER; -- Target reference clock


  -- Bridge clocks
  constant BRIDGE_REF_CLK_PER : time := CLK_250_PER; -- Use 250MHz clock for faster simulation
--  constant BRIDGE_REF_CLK_PER : time := REFCLK200M_0;

  -- Bridge MGT clocks
  constant BRIDGE_MPTL_CLK_PER : time := PROG_CLK_0A; -- Osc. programmable clock 0A (MPTL clock)


  -- Target clocks
  constant TARGET_REF_CLK_PER   : time := REFCLK200M_1;
  constant TARGET_XRM_CLKIN_PER : time := CLK_125_PER;

  -- Target MGT clocks
  constant MGT112_CLK1_PER : time := PCIEREFCLK_VAR_0; -- PCIe Reference clock 0 125/250 MHz (sw1h)
  constant MGT113_CLK0_PER : time := PROG_CLK_2;       -- Osc. programmable clock 2
  constant MGT113_CLK1_PER : time := PN6_REFCLK;       -- P6   Reference clock (Build option g)
  constant MGT114_CLK0_PER : time := PCIEREFCLK_250_1; -- PCIe Reference clock 1 250 MHz
  constant MGT115_CLK0_PER : time := PROG_CLK_0B;      -- Osc. programmable clock 0B (MPTL clock)
  constant MGT116_CLK0_PER : time := PROG_CLK_1;       -- Osc. programmable clock 1
  constant MGT117_CLK0_PER : time := XRM_MGTCLK_M2C;   -- XRM  Reference clock (variable)


  ---------------------
  -- On-board memory --
  ---------------------

  -- Select DDR3 part for board simulation
  constant BOARD_DDR3_PART : part_t := OPTION_M;

  -- DDR3 SDRAM address widths
  constant BOARD_DDR3_BANK_ROW_WIDTH : natural := BOARD_DDR3_PART.part_size.arow_width;
  constant BOARD_DDR3_BANK_COL_WIDTH : natural := BOARD_DDR3_PART.part_size.acol_width;
  constant BOARD_DDR3_BANK_BNK_WIDTH : natural := BOARD_DDR3_PART.part_size.bank_width;

  -- DDR3 SDRAM address width = Bank+Row+Col and convert to byte addressing (+2)
  constant BOARD_DDR3_BYTE_ADDR_WIDTH : natural := BOARD_DDR3_BANK_BNK_WIDTH + BOARD_DDR3_BANK_ROW_WIDTH + BOARD_DDR3_BANK_COL_WIDTH + 2;

  -- Address width for each on-board memory bank
  constant BOARD_MEM_BYTE_ADDR_WIDTH_ARRAY : mem_byte_addr_width_array_t := ((DDR3_BANKS+DDR3_BANK0-1) downto DDR3_BANK0 => BOARD_DDR3_BYTE_ADDR_WIDTH);

end;


package body adb3_target_tb_inc_pkg is

end;