--==============================================================================
--
--     File Name: mptl_if_target_wrap_6tl.vhd
--       Version: 0.1
--          Date: 18 Jul 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Target MPTL interface wrapper
--
--   Description: Instantiate full-MPTL simulation netlist or synthesis
--                core for target MPTL interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 18 Jul 2011:
--                     - ADM-XRC-6TL specific version of mptl_if_target_wrap.
--                       moved from hdl/vhdl/common/mptl
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;

entity mptl_if_target_wrap is
port(
  -- OCP Interface
  ocp_clk                  : in    std_logic;
  direct_slave_m2s         : out   adb3_ocp_m2sT;
  direct_slave_s2m         : in    adb3_ocp_s2mT;
  dma_channels_m2s         : out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  dma_channels_s2m         : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  dma_abort                : in    std_logic_vector(DMA_CHANNELS-1 downto 0);
  -- MPTL Interface
  mptl_b2t                 : in    mptl_b2t_t;
  mptl_t2b                 : out   mptl_t2b_t;
  mptl_clk                 : in    std_logic_dbl_t;
  -- MPTL Initialisation
  ocp_ready                : in    std_logic;
  mptl_sb_b2t              : in    mptl_sb_b2t_t;
  mptl_sb_t2b              : out   mptl_sb_t2b_t;
  -- MPTL Sideband
  gpio_t2b                 : in    std_logic_vector(15 downto 0);
  gpio_b2t                 : out   std_logic_vector(15 downto 0));
end;


architecture syn_6tl of mptl_if_target_wrap is

  -- Unused IO
  signal mptl_clk_out : std_logic;

  -- Internal OCP
  signal ds_int_m2s  : adb3_ocp_m2sT;
  signal ds_int_s2m  : adb3_ocp_s2mT;
  signal dma_int_m2s : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  signal dma_int_s2m : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);

  -- MPTL sideband
  signal mptl_target_config_n     : std_logic;
  signal mptl_target_gtp_online_n : std_logic;
  signal mptl_bridge_gtp_online_n : std_logic;

  component mptl64par_interface_target_6tl
  port(
    -- OCP Interface
    ocp_clk                  : in    std_logic;
    direct_slave_m2s         : out   adb3_ocp_m2sT;
    direct_slave_s2m         : in    adb3_ocp_s2mT;
    dma_channels_m2s         : out   adb3_ocp_m2sT_array(0 to 1);
    dma_channels_s2m         : in    adb3_ocp_s2mT_array(0 to 1);
    -- MPTL Interface
    mptl_tx                  : out   mptl_pins_t;
    mptl_rx                  : in    mptl_pins_t;
    mptl_clk                 : in    std_logic_dbl_t;
    mptl_clk_out             : out   std_logic;
    -- Bridge to Target Sideband Signals
    mptl_target_configured_n : out   std_logic;
    mptl_target_gtp_online_n : out   std_logic;
    mptl_bridge_gtp_online_n : in    std_logic;
    -- MPTL64 Sideband Interface
    gpio_t2b                 : in    std_logic_vector(15 downto 0);
    gpio_b2t                 : out   std_logic_vector(15 downto 0));
  end component;

begin

  syn_ngc_g : if (TARGET_USE = SYN_NGC) generate
    -- Direct connecton during synthesis
    direct_slave_m2s <= ds_int_m2s;
    ds_int_s2m       <= direct_slave_s2m;

    dma_channels_m2s <= dma_int_m2s;
    dma_int_s2m      <= dma_channels_s2m;
  end generate;

  sim_mptl_g : if (TARGET_USE = SIM_MPTL) generate
    -- Define data during MPTL simulation
    direct_slave_m2s <= ds_int_m2s;
    ds_int_s2m       <= make_defined_s2m(direct_slave_s2m);

    dma_g : for n in 0 to DMA_CHANNELS-1 generate
      dma_channels_m2s(n) <= dma_int_m2s(n);
      dma_int_s2m(n)      <= make_defined_s2m(dma_channels_s2m(n));
    end generate;
  end generate;


  -- Hold off MPTL target configured until all OCP clocks are stable (resets inactive)
  mptl_sb_t2b.mptl_target_configured_l <= mptl_target_config_n or not(ocp_ready);

  -- Connect MPTL initialisation signals
  mptl_sb_t2b.mptl_target_gtp_online_l <= mptl_target_gtp_online_n;
  mptl_bridge_gtp_online_n <= mptl_sb_b2t.mptl_bridge_gtp_online_l;

  -- dma_abort on pins
  mptl_sb_t2b.mptl_dma_abort_l <= not(dma_abort);

  -- MPTL simulation netlist or synthesis .NGC core
  mptl64par_interface_target_6tl_i : mptl64par_interface_target_6tl
  port map(
    -- OCP Interface
    ocp_clk                  => ocp_clk,                  -- in    std_logic
    direct_slave_m2s         => ds_int_m2s,               -- out   adb3_ocp_m2sT
    direct_slave_s2m         => ds_int_s2m,               -- in    adb3_ocp_s2mT
    dma_channels_m2s         => dma_int_m2s,              -- out   adb3_ocp_m2sT_array(0 to 1)
    dma_channels_s2m         => dma_int_s2m,              -- in    adb3_ocp_s2mT_array(0 to 1)
    -- MPTL Interface
    mptl_tx                  => mptl_t2b,                 -- out   mptl_pins_t
    mptl_rx                  => mptl_b2t,                 -- in    mptl_pins_t
    mptl_clk                 => mptl_clk,                 -- in    mptl_clk_type
    mptl_clk_out             => mptl_clk_out,             -- out   std_logic
    -- MPTL Sideband Signals
    mptl_target_configured_n => mptl_target_config_n,     -- out   std_logic_dbl_t
    mptl_target_gtp_online_n => mptl_target_gtp_online_n, -- out   std_logic
    mptl_bridge_gtp_online_n => mptl_bridge_gtp_online_n, -- in    std_logic
    -- MPTL Sideband Interface
    gpio_t2b                 => gpio_t2b,                 -- in    std_logic_vector(15 downto 0)
    gpio_b2t                 => gpio_b2t);                -- out   std_logic_vector(15 downto 0)

end;