--==============================================================================
--
--     File Name: ddr3_if_bank_6tge.vhd
--       Version: 0.2
--          Date: 18 Aug 2011
--         Model: Synthesisable
--
--  Dependencies: Xilinx Coregen MIG DDR3 SDRAM interface
--
--       Company: Alpha Data
--        Design: On-board DDR3 SDRAM interface block using Xiling MIG cores
--
--   Description: OCP to MIG to DDR3 SDRAM interface for ADM-XRC-6TGE
--
--   Limitations: Requires ISE12.3 or later
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 15 Feb 2011: Initial design
--                0.2, KDC, 18 Aug 2011:
--                  - Design based on ddr3_if_bank_v3_6
--                  - Changed entity name to ddr3_if_bank
--                  - Re-organised and re-named MIG constants in mem_if_pkg
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.mem_if_pkg.all;

entity ddr3_if_bank is
generic(
  bank : in    integer range 0 to DDR3_BANKS-1 := 0);
port(
  -- OCP interface
  ocp_rst               : in    std_logic;
  ocp_clk               : in    std_logic;
  ocp_m2s               : in    adb3_ocp_m2sT;
  ocp_s2m               : out   adb3_ocp_s2mT;
  -- Memory interface
  ddr3_rst              : in    std_logic;
  ddr3_ref_clk          : in    std_logic;
  ddr3_clk              : in    std_logic;
  ddr3_iodelay_ctrl_rdy : in    std_logic;
  -- Memory status
  ddr3_if_rdy           : out   std_logic;
  ddr3_if_stat          : out   std_logic_vector(3 downto 0);
  ddr3_if_err           : out   std_logic_vector(3 downto 0);
  -- Physical memory interface
  ddr3_addr_out         : out   ddr3_addr_out_t;
  ddr3_ctrl_out         : out   ddr3_ctrl_out_t;
  ddr3_data_inout       : inout ddr3_data_inout_t;
  ddr3_clk_out          : out   ddr3_clk_out_t;
  -- Debug info
  ddr3_if_debug         : out   std_logic_vector(31 downto 0));
end;


architecture syn_6tge of ddr3_if_bank is

  component infrastructure
  generic(
    TCQ             : integer;
    CLK_PERIOD      : integer;
    nCK_PER_CLK     : integer;
    CLKFBOUT_MULT_F : integer;
    DIVCLK_DIVIDE   : integer;
    CLKOUT_DIVIDE   : integer;
    RST_ACT_LOW     : integer);
  port(
    clk_mem          : out std_logic;
    clk              : out std_logic;
    clk_rd_base      : out std_logic;
    rstdiv0          : out std_logic;
    mmcm_clk         : in  std_logic;
    sys_rst          : in  std_logic;
    iodelay_ctrl_rdy : in  std_logic;
    PSDONE           : out std_logic;
    PSEN             : in  std_logic;
    PSINCDEC         : in  std_logic);
  end component;

  -- Select speed bin constant
  constant MIG_CLOCKS : mig_clocks_t := MIG_CLOCKS_800;

  -- Select DQS column constants for each bank
  constant MIG_BANK0_DQS : mig_dqs_col0_4_t  := MIG_DQS_COL0_4;
  constant MIG_BANK1_DQS : mig_dqs_col0_4_t  := MIG_DQS_COL0_4;
  constant MIG_BANK2_DQS : mig_dqs_col2_4_t  := MIG_DQS_COL2_4;
  constant MIG_BANK3_DQS : mig_dqs_col01_2_t := MIG_DQS_COL01_2;

  -- Constants for MIG simulation options
  constant MIG_SIM_BYPASS_INIT_CAL : string  := conv_sim_bypass_init_cal(TARGET_USE);
  constant MIG_SIM_INIT_OPTION     : string  := conv_sim_init_option(TARGET_USE);
  constant MIG_SIM_CAL_OPTION      : string  := conv_sim_cal_option(TARGET_USE);
  constant MIG_T_RFC_OPTION        : integer := conv_t_rfc_option(DDR3_BANK_ROW_WIDTH);

  -- Constants for MIG generic defaults
  constant PAYLOAD_WIDTH_D : natural := MIG_COMMON.PAYLOAD_WIDTH;
  constant ROW_WIDTH_D     : natural := MIG_COMMON.ROW_WIDTH;
  constant COL_WIDTH_D     : natural := MIG_COMMON.COL_WIDTH;
  constant BANK_WIDTH_D    : natural := MIG_COMMON.BANK_WIDTH;
  constant ADDR_WIDTH_D    : natural := MIG_COMMON.ADDR_WIDTH;
  constant DQS_WIDTH_D     : natural := MIG_COMMON.DQS_WIDTH;
  constant DQS_CNT_WIDTH_D : natural := MIG_COMMON.DQS_CNT_WIDTH;
  constant DQ_WIDTH_D      : natural := MIG_COMMON.DQ_WIDTH;

  constant SYSCLK_PERIOD : integer := MIG_CLOCKS.tCK * MIG_CLOCKS.nCK_PER_CLK;

  -- memc_ui_top types
  subtype app_addr_t     is std_logic_vector(ADDR_WIDTH_D-1 downto 0);
  subtype app_wdf_data_t is std_logic_vector((4*PAYLOAD_WIDTH_D)-1 downto 0);
  subtype app_wdf_mask_t is std_logic_vector((4*PAYLOAD_WIDTH_D)/8-1 downto 0);
  subtype app_cmd_t      is std_logic_vector(2 downto 0);
  subtype app_rd_data_t  is std_logic_vector((4*PAYLOAD_WIDTH_D)-1 downto 0);

  signal mig_rst           : std_logic;
  signal mig_ddr3_if_clk0  : std_logic;
  signal mig_ddr3_if_clk1  : std_logic;
  signal mig_ddr3_if_clk2  : std_logic;
  signal mig_ddr3_if_clk3  : std_logic;
  signal mig_clk_mem       : std_logic;
  signal mig_clk_rd_base   : std_logic;
  signal pd_psdone         : std_logic;
  signal pd_psen           : std_logic;
  signal pd_psincdec       : std_logic;
  signal dfi_init_complete : std_logic;

  -- memc_ui_top signals
  signal app_sz            : std_logic;
  signal app_wdf_wren      : std_logic;
  signal app_wdf_data      : app_wdf_data_t;
  signal app_wdf_mask      : app_wdf_mask_t;
  signal app_wdf_end       : std_logic;
  signal app_addr          : app_addr_t;
  signal app_cmd           : app_cmd_t;
  signal app_en            : std_logic;
  signal app_rdy           : std_logic;
  signal app_wdf_rdy       : std_logic;
  signal app_rd_data       : app_rd_data_t;
  signal app_rd_data_valid : std_logic;

  -- Debug types
  subtype dqs_width_1_t   is std_logic_vector(1*DQS_WIDTH_D-1 downto 0);
  subtype dqs_width_2_t   is std_logic_vector(2*DQS_WIDTH_D-1 downto 0);
  subtype dqs_width_3_t   is std_logic_vector(3*DQS_WIDTH_D-1 downto 0);
  subtype dqs_width_5_t   is std_logic_vector(5*DQS_WIDTH_D-1 downto 0);
  subtype dqs_cnt_width_t is std_logic_vector(DQS_CNT_WIDTH_D-1 downto 0);
  subtype dq_width_4_t    is std_logic_vector(4*DQ_WIDTH_D-1 downto 0);

  -- Debug input signals
  signal dbg_dec_cpt               : std_logic       := '0';
  signal dbg_dec_rd_dqs            : std_logic       := '0';
  signal dbg_dec_rd_fps            : std_logic       := '0';
  signal dbg_inc_cpt               : std_logic       := '0';
  signal dbg_inc_dec_sel           : dqs_cnt_width_t := (others => '0');
  signal dbg_inc_rd_dqs            : std_logic       := '0';
  signal dbg_inc_rd_fps            : std_logic       := '0';
  signal dbg_pd_off                : std_logic       := '0';
  signal dbg_pd_maintain_off       : std_logic       := '0';
  signal dbg_pd_maintain_0_only    : std_logic       := '0';
  signal dbg_wr_dq_tap_set         : dqs_width_5_t   := (others => '0');
  signal dbg_wr_dqs_tap_set        : dqs_width_5_t   := (others => '0');
  signal dbg_wr_tap_set_en         : std_logic       := '0';

  -- Debug output signals
  signal dbg_cpt_first_edge_cnt    : dqs_width_5_t;
  signal dbg_cpt_second_edge_cnt   : dqs_width_5_t;
  signal dbg_cpt_tap_cnt           : dqs_width_5_t;
  signal dbg_dq_tap_cnt            : dqs_width_5_t;
  signal dbg_dqs_tap_cnt           : dqs_width_5_t;
  signal dbg_rd_active_dly         : std_logic_vector(4 downto 0);
  signal dbg_rd_bitslip_cnt        : dqs_width_3_t;
  signal dbg_rd_clkdly_cnt         : dqs_width_2_t;
  signal dbg_rddata                : dq_width_4_t;
  signal dbg_rdlvl_done            : std_logic_vector(1 downto 0);
  signal dbg_rdlvl_err             : std_logic_vector(1 downto 0);
  signal dbg_rdlvl_start           : std_logic_vector(1 downto 0);
  signal dbg_wl_dqs_inverted       : dqs_width_1_t;
  signal dbg_wl_odelay_dq_tap_cnt  : dqs_width_5_t;
  signal dbg_wl_odelay_dqs_tap_cnt : dqs_width_5_t;
  signal dbg_wr_calib_clk_delay    : dqs_width_2_t;
  signal dbg_wrlvl_done            : std_logic;
  signal dbg_wrlvl_err             : std_logic;
  signal dbg_wrlvl_start           : std_logic;

begin

  -- Check MIG interface address widths
  assert not(DDR3_BANK_ROW_WIDTH /= MIG_COMMON.ROW_WIDTH)
    report "(ddr3_if_bank): MIG ROW_WIDTH /= target_inc_pkg DDR3_BANK_ROW_WIDTH"
    severity failure;

  -- Should be Rank(1)+Bank(3)+Row(13/14)+Col(10) and remove Rank (-1) and convert to byte addressing (+2)
  assert not(DDR3_BYTE_ADDR_WIDTH /= (MIG_COMMON.ADDR_WIDTH - 1 + 2))
    report "(ddr3_if_bank): MIG ADDR_WIDTH /= target_inc_pkg DDR3_BYTE_ADDR_WIDTH"
    severity failure;

  assert not(DDR3_BANK_DATA_WIDTH /= MIG_COMMON.DQ_WIDTH)
    report "(ddr3_if_bank): MIG DQ_WIDTH /= target_inc_pkg DDR3_BANK_DATA_WIDTH"
    severity failure;


  -- Instantiate DDR3 MIG bank 0 core
  bank0_g : if (bank = 0) generate
    -- Instantiate OCP to MIG block
    adb3_ocp_ocp2ddr3_nb_i : adb3_ocp_ocp2ddr3_nb
    generic map(
      app_row_width  => ROW_WIDTH_D,   -- natural
      app_col_width  => COL_WIDTH_D,   -- natural
      app_bank_width => BANK_WIDTH_D,  -- natural
      app_addr_width => ADDR_WIDTH_D)  -- natural)
    port map(
      -- OCP Interface
      ocp_rst           => ocp_rst,            -- in    std_logic
      ocp_clk           => ocp_clk,            -- in    std_logic
      ocp_m2s           => ocp_m2s,            -- in    adb3_ocp_m2sT
      ocp_s2m           => ocp_s2m,            -- out   adb3_ocp_s2mT
      -- MIG user interface
      mig_rst           => mig_rst,            -- in    std_logic
      mig_clk           => mig_ddr3_if_clk0,   -- in    std_logic
      phy_init_done     => dfi_init_complete,  -- in    std_logic
      app_rdy           => app_rdy,            -- in    std_logic
      app_wdf_rdy       => app_wdf_rdy,        -- in    std_logic
      app_rd_data       => app_rd_data,        -- in    std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_rd_data_valid => app_rd_data_valid,  -- in    std_logic
      app_en            => app_en,             -- out   std_logic
      app_cmd           => app_cmd,            -- out   std_logic_vector(2 downto 0)
      app_addr          => app_addr,           -- out   std_logic_vector(app_addr_width-1 downto 0);
      app_sz            => app_sz,             -- out   std_logic
      app_wdf_wren      => app_wdf_wren,       -- out   std_logic
      app_wdf_data      => app_wdf_data,       -- out   std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_wdf_mask      => app_wdf_mask,       -- out   std_logic_vector((ADB3_OCP_DATA_WIDTH/8)-1 downto 0)
      app_wdf_end       => app_wdf_end,        -- out   std_logic
      -- Debug info
      debug_info        => ddr3_if_debug);     -- out   std_logic_vector(31 downto 0))

    memc_ui_top_i : entity work.c0_memc_ui_top(arch_c0_memc_ui_top)
    generic map(
      ADDR_CMD_MODE       => MIG_COMMON.ADDR_CMD_MODE,
      BANK_WIDTH          => MIG_COMMON.BANK_WIDTH,
      CK_WIDTH            => MIG_COMMON.CK_WIDTH,
      CKE_WIDTH           => MIG_COMMON.CKE_WIDTH,
      nCK_PER_CLK         => MIG_CLOCKS.nCK_PER_CLK,
      COL_WIDTH           => MIG_COMMON.COL_WIDTH,
      CS_WIDTH            => MIG_COMMON.CS_WIDTH,
      DM_WIDTH            => MIG_COMMON.DM_WIDTH,
      nCS_PER_RANK        => MIG_COMMON.nCS_PER_RANK,
      DEBUG_PORT          => MIG_COMMON.DEBUG_PORT,
      IODELAY_GRP         => MIG_CLOCKS_COMMON.IODELAY_GRP,
      DQ_WIDTH            => MIG_COMMON.DQ_WIDTH,
      DQS_WIDTH           => MIG_COMMON.DQS_WIDTH,
      DQS_CNT_WIDTH       => MIG_COMMON.DQS_CNT_WIDTH,
      ORDERING            => MIG_COMMON.ORDERING,
      OUTPUT_DRV          => MIG_COMMON.OUTPUT_DRV,
      PHASE_DETECT        => MIG_COMMON.PHASE_DETECT,
      RANK_WIDTH          => MIG_COMMON.RANK_WIDTH,
      REFCLK_FREQ         => MIG_CLOCKS.REFCLK_FREQ,
      REG_CTRL            => MIG_COMMON.REG_CTRL,
      ROW_WIDTH           => MIG_COMMON.ROW_WIDTH,
      RTT_NOM             => MIG_COMMON.RTT_NOM,
      RTT_WR              => MIG_COMMON.RTT_WR,
      SIM_BYPASS_INIT_CAL => MIG_SIM_BYPASS_INIT_CAL,
      SIM_CAL_OPTION      => MIG_SIM_CAL_OPTION,
      SIM_INIT_OPTION     => MIG_SIM_INIT_OPTION,
      WRLVL               => MIG_COMMON.WRLVL,
      nDQS_COL0           => MIG_BANK0_DQS.nDQS_COL0,
      nDQS_COL1           => MIG_BANK0_DQS.nDQS_COL1,
      nDQS_COL2           => MIG_BANK0_DQS.nDQS_COL2,
      nDQS_COL3           => MIG_BANK0_DQS.nDQS_COL3,
      DQS_LOC_COL0        => MIG_BANK0_DQS.DQS_LOC_COL0,
      DQS_LOC_COL1        => MIG_BANK0_DQS.DQS_LOC_COL1,
      DQS_LOC_COL2        => MIG_BANK0_DQS.DQS_LOC_COL2,
      DQS_LOC_COL3        => MIG_BANK0_DQS.DQS_LOC_COL3,
      BURST_MODE          => MIG_COMMON.BURST_MODE,
      BM_CNT_WIDTH        => MIG_COMMON.BM_CNT_WIDTH,
      tCK                 => MIG_CLOCKS.tCK,
      tPRDI               => MIG_COMMON.tPRDI,
      tREFI               => MIG_COMMON.tREFI,
      tZQI                => MIG_COMMON.tZQI,
      tRFC                => MIG_T_RFC_OPTION,
      ADDR_WIDTH          => MIG_COMMON.ADDR_WIDTH,
      TCQ                 => MIG_COMMON.TCQ,
      ECC_TEST            => MIG_COMMON.ECC_TEST,
      PAYLOAD_WIDTH       => MIG_COMMON.PAYLOAD_WIDTH)
    port map(
      clk                       => mig_ddr3_if_clk0,           -- in    std_logic
      clk_mem                   => mig_clk_mem,                -- in    std_logic
      clk_rd_base               => mig_clk_rd_base,            -- in    std_logic
      rst                       => mig_rst,                    -- in    std_logic
      ddr_addr                  => ddr3_addr_out.a,            -- out   std_logic_vector(ROW_WIDTH-1 downto 0)
      ddr_ba                    => ddr3_ctrl_out.ba,           -- out   std_logic_vector(BANK_WIDTH-1 downto 0)
      ddr_cas_n                 => ddr3_ctrl_out.cas_l,        -- out   std_logic
      ddr_ck_n                  => ddr3_clk_out.clk_n,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_ck                    => ddr3_clk_out.clk_p,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_cke                   => ddr3_ctrl_out.cke,          -- out   std_logic_vector(CKE_WIDTH-1 downto 0)
      ddr_cs_n                  => ddr3_ctrl_out.cs_l,         -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_dm                    => ddr3_ctrl_out.dm,           -- out   std_logic_vector(DM_WIDTH-1 downto 0)
      ddr_odt                   => ddr3_ctrl_out.odt,          -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_ras_n                 => ddr3_ctrl_out.ras_l,        -- out   std_logic
      ddr_reset_n               => ddr3_ctrl_out.reset_l,      -- out   std_logic
      ddr_parity                => open,                       -- out   std_logic
      ddr_we_n                  => ddr3_ctrl_out.we_l,         -- out   std_logic
      ddr_dq                    => ddr3_data_inout.dq,         -- inout std_logic_vector(DQ_WIDTH-1 downto 0)
      ddr_dqs_n                 => ddr3_data_inout.dqs_n,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      ddr_dqs                   => ddr3_data_inout.dqs_p,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      pd_PSEN                   => pd_psen,                    -- out   std_logic
      pd_PSINCDEC               => pd_psincdec,                -- out   std_logic
      pd_PSDONE                 => pd_psdone,                  -- in    std_logic
      bank_mach_next            => open,                       -- out   std_logic_vector(BM_CNT_WIDTH-1 downto 0)
      dfi_init_complete         => dfi_init_complete,          -- out   std_logic
      app_ecc_multiple_err      => open,                       -- out   std_logic_vector(3 downto 0)
      app_rd_data               => app_rd_data,                -- out   std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_rd_data_end           => open,                       -- out   std_logic
      app_rd_data_valid         => app_rd_data_valid,          -- out   std_logic
      app_rdy                   => app_rdy,                    -- out   std_logic
      app_wdf_rdy               => app_wdf_rdy,                -- out   std_logic
      app_addr                  => app_addr,                   -- in    std_logic_vector(ADDR_WIDTH-1 downto 0)
      app_cmd                   => app_cmd,                    -- in    std_logic_vector(2 downto 0)
      app_en                    => app_en,                     -- in    std_logic
      app_hi_pri                => '0',                        -- in    std_logic
      app_sz                    => app_sz,                     -- in    std_logic
      app_wdf_data              => app_wdf_data,               -- in    std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_wdf_end               => app_wdf_end,                -- in    std_logic
      app_wdf_mask              => app_wdf_mask,               -- in    std_logic_vector(APP_MASK_WIDTH-1 downto 0)
      app_wdf_wren              => app_wdf_wren,               -- in    std_logic
      dbg_wr_dqs_tap_set        => dbg_wr_dqs_tap_set,         -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_dq_tap_set         => dbg_wr_dq_tap_set,          -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_tap_set_en         => dbg_wr_tap_set_en,          -- in    std_logic
      dbg_wrlvl_start           => dbg_wrlvl_start,            -- out   std_logic
      dbg_wrlvl_done            => dbg_wrlvl_done,             -- out   std_logic
      dbg_wrlvl_err             => dbg_wrlvl_err,              -- out   std_logic
      dbg_wl_dqs_inverted       => dbg_wl_dqs_inverted,        -- out   std_logic_vector(DQS_WIDTH-1 downto 0)
      dbg_wr_calib_clk_delay    => dbg_wr_calib_clk_delay,     -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dqs_tap_cnt => dbg_wl_odelay_dqs_tap_cnt,  -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dq_tap_cnt  => dbg_wl_odelay_dq_tap_cnt,   -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rdlvl_start           => dbg_rdlvl_start,            -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_done            => dbg_rdlvl_done,             -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_err             => dbg_rdlvl_err,              -- out   std_logic_vector(1 downto 0)
      dbg_cpt_tap_cnt           => dbg_cpt_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_first_edge_cnt    => dbg_cpt_first_edge_cnt,     -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_second_edge_cnt   => dbg_cpt_second_edge_cnt,    -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rd_bitslip_cnt        => dbg_rd_bitslip_cnt,         -- out   std_logic_vector(3*DQS_WIDTH-1 downto 0)
      dbg_rd_clkdly_cnt         => dbg_rd_clkdly_cnt,          -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_rd_active_dly         => dbg_rd_active_dly,          -- out   std_logic_vector(4 downto 0)
      dbg_pd_off                => dbg_pd_off,                 -- in    std_logic
      dbg_pd_maintain_off       => dbg_pd_maintain_off,        -- in    std_logic
      dbg_pd_maintain_0_only    => dbg_pd_maintain_0_only,     -- in    std_logic
      dbg_inc_cpt               => dbg_inc_cpt,                -- in    std_logic
      dbg_dec_cpt               => dbg_dec_cpt,                -- in    std_logic
      dbg_inc_rd_dqs            => dbg_inc_rd_dqs,             -- in    std_logic
      dbg_dec_rd_dqs            => dbg_dec_rd_dqs,             -- in    std_logic
      dbg_inc_dec_sel           => dbg_inc_dec_sel,            -- in    std_logic_vector(DQS_CNT_WIDTH-1 downto 0)
      dbg_inc_rd_fps            => dbg_inc_rd_fps,             -- in    std_logic
      dbg_dec_rd_fps            => dbg_dec_rd_fps,             -- in    std_logic
      dbg_dqs_tap_cnt           => dbg_dqs_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_dq_tap_cnt            => dbg_dq_tap_cnt,             -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rddata                => dbg_rddata);                -- out   std_logic_vector(4*DQ_WIDTH-1 downto 0)

    -- DDR3 MIG clock generation
    infrastructure_i : infrastructure
      generic map(
      TCQ             => MIG_COMMON.TCQ,                -- integer := 100
      CLK_PERIOD      => SYSCLK_PERIOD,                 -- integer := 3000
      nCK_PER_CLK     => MIG_CLOCKS.nCK_PER_CLK,        -- integer := 2
      CLKFBOUT_MULT_F => MIG_CLOCKS.CLKFBOUT_MULT_F,    -- integer := 2
      DIVCLK_DIVIDE   => MIG_CLOCKS.DIVCLK_DIVIDE,      -- integer := 1
      CLKOUT_DIVIDE   => MIG_CLOCKS.CLKOUT_DIVIDE,      -- integer := 2
      RST_ACT_LOW     => MIG_CLOCKS_COMMON.RST_ACT_LOW) -- integer := 1
    port map(
      -- Clock inputs
      mmcm_clk         => ddr3_clk,              -- in    std_logic
      -- System reset input
      sys_rst          => ddr3_rst,              -- in    std_logic
      -- MMCM/IDELAYCTRL Lock status
      iodelay_ctrl_rdy => ddr3_iodelay_ctrl_rdy, -- in    std_logic
      -- Clock outputs
      clk_mem          => mig_clk_mem,           -- out   std_logic
      clk              => mig_ddr3_if_clk0,      -- out   std_logic
      clk_rd_base      => mig_clk_rd_base,       -- out   std_logic
      -- Reset outputs
      rstdiv0          => mig_rst,               -- out   std_logic
      -- Phase Shift Interface
      PSDONE           => pd_psdone,             -- out   std_logic
      PSEN             => pd_psen,               -- in    std_logic
      PSINCDEC         => pd_psincdec);          -- in    std_logic

    process(mig_ddr3_if_clk0)
    begin
      if rising_edge(mig_ddr3_if_clk0) then
        -- Connect ready
        ddr3_if_rdy <= dfi_init_complete;
      end if;
    end process;
  end generate;

  -- Instantiate DDR3 MIG bank 1 core
  bank1_g : if (bank = 1) generate
    -- Instantiate OCP to MIG block
    adb3_ocp_ocp2ddr3_nb_i : adb3_ocp_ocp2ddr3_nb
    generic map(
      app_row_width  => ROW_WIDTH_D,   -- natural
      app_col_width  => COL_WIDTH_D,   -- natural
      app_bank_width => BANK_WIDTH_D,  -- natural
      app_addr_width => ADDR_WIDTH_D)  -- natural)
    port map(
      -- OCP Interface
      ocp_rst           => ocp_rst,            -- in    std_logic
      ocp_clk           => ocp_clk,            -- in    std_logic
      ocp_m2s           => ocp_m2s,            -- in    adb3_ocp_m2sT
      ocp_s2m           => ocp_s2m,            -- out   adb3_ocp_s2mT
      -- MIG user interface
      mig_rst           => mig_rst,            -- in    std_logic
      mig_clk           => mig_ddr3_if_clk1,   -- in    std_logic
      phy_init_done     => dfi_init_complete,  -- in    std_logic
      app_rdy           => app_rdy,            -- in    std_logic
      app_wdf_rdy       => app_wdf_rdy,        -- in    std_logic
      app_rd_data       => app_rd_data,        -- in    std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_rd_data_valid => app_rd_data_valid,  -- in    std_logic
      app_en            => app_en,             -- out   std_logic
      app_cmd           => app_cmd,            -- out   std_logic_vector(2 downto 0)
      app_addr          => app_addr,           -- out   std_logic_vector(app_addr_width-1 downto 0);
      app_sz            => app_sz,             -- out   std_logic
      app_wdf_wren      => app_wdf_wren,       -- out   std_logic
      app_wdf_data      => app_wdf_data,       -- out   std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_wdf_mask      => app_wdf_mask,       -- out   std_logic_vector((ADB3_OCP_DATA_WIDTH/8)-1 downto 0)
      app_wdf_end       => app_wdf_end,        -- out   std_logic
      -- Debug info
      debug_info        => ddr3_if_debug);     -- out   std_logic_vector(31 downto 0))

    memc_ui_top_i : entity work.c1_memc_ui_top(arch_c1_memc_ui_top)
    generic map(
      ADDR_CMD_MODE       => MIG_COMMON.ADDR_CMD_MODE,
      BANK_WIDTH          => MIG_COMMON.BANK_WIDTH,
      CK_WIDTH            => MIG_COMMON.CK_WIDTH,
      CKE_WIDTH           => MIG_COMMON.CKE_WIDTH,
      nCK_PER_CLK         => MIG_CLOCKS.nCK_PER_CLK,
      COL_WIDTH           => MIG_COMMON.COL_WIDTH,
      CS_WIDTH            => MIG_COMMON.CS_WIDTH,
      DM_WIDTH            => MIG_COMMON.DM_WIDTH,
      nCS_PER_RANK        => MIG_COMMON.nCS_PER_RANK,
      DEBUG_PORT          => MIG_COMMON.DEBUG_PORT,
      IODELAY_GRP         => MIG_CLOCKS_COMMON.IODELAY_GRP,
      DQ_WIDTH            => MIG_COMMON.DQ_WIDTH,
      DQS_WIDTH           => MIG_COMMON.DQS_WIDTH,
      DQS_CNT_WIDTH       => MIG_COMMON.DQS_CNT_WIDTH,
      ORDERING            => MIG_COMMON.ORDERING,
      OUTPUT_DRV          => MIG_COMMON.OUTPUT_DRV,
      PHASE_DETECT        => MIG_COMMON.PHASE_DETECT,
      RANK_WIDTH          => MIG_COMMON.RANK_WIDTH,
      REFCLK_FREQ         => MIG_CLOCKS.REFCLK_FREQ,
      REG_CTRL            => MIG_COMMON.REG_CTRL,
      ROW_WIDTH           => MIG_COMMON.ROW_WIDTH,
      RTT_NOM             => MIG_COMMON.RTT_NOM,
      RTT_WR              => MIG_COMMON.RTT_WR,
      SIM_BYPASS_INIT_CAL => MIG_SIM_BYPASS_INIT_CAL,
      SIM_CAL_OPTION      => MIG_SIM_CAL_OPTION,
      SIM_INIT_OPTION     => MIG_SIM_INIT_OPTION,
      WRLVL               => MIG_COMMON.WRLVL,
      nDQS_COL0           => MIG_BANK1_DQS.nDQS_COL0,
      nDQS_COL1           => MIG_BANK1_DQS.nDQS_COL1,
      nDQS_COL2           => MIG_BANK1_DQS.nDQS_COL2,
      nDQS_COL3           => MIG_BANK1_DQS.nDQS_COL3,
      DQS_LOC_COL0        => MIG_BANK1_DQS.DQS_LOC_COL0,
      DQS_LOC_COL1        => MIG_BANK1_DQS.DQS_LOC_COL1,
      DQS_LOC_COL2        => MIG_BANK1_DQS.DQS_LOC_COL2,
      DQS_LOC_COL3        => MIG_BANK1_DQS.DQS_LOC_COL3,
      BURST_MODE          => MIG_COMMON.BURST_MODE,
      BM_CNT_WIDTH        => MIG_COMMON.BM_CNT_WIDTH,
      tCK                 => MIG_CLOCKS.tCK,
      tPRDI               => MIG_COMMON.tPRDI,
      tREFI               => MIG_COMMON.tREFI,
      tZQI                => MIG_COMMON.tZQI,
      tRFC                => MIG_T_RFC_OPTION,
      ADDR_WIDTH          => MIG_COMMON.ADDR_WIDTH,
      TCQ                 => MIG_COMMON.TCQ,
      ECC_TEST            => MIG_COMMON.ECC_TEST,
      PAYLOAD_WIDTH       => MIG_COMMON.PAYLOAD_WIDTH)
    port map(
      clk                       => mig_ddr3_if_clk1,           -- in    std_logic
      clk_mem                   => mig_clk_mem,                -- in    std_logic
      clk_rd_base               => mig_clk_rd_base,            -- in    std_logic
      rst                       => mig_rst,                    -- in    std_logic
      ddr_addr                  => ddr3_addr_out.a,            -- out   std_logic_vector(ROW_WIDTH-1 downto 0)
      ddr_ba                    => ddr3_ctrl_out.ba,           -- out   std_logic_vector(BANK_WIDTH-1 downto 0)
      ddr_cas_n                 => ddr3_ctrl_out.cas_l,        -- out   std_logic
      ddr_ck_n                  => ddr3_clk_out.clk_n,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_ck                    => ddr3_clk_out.clk_p,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_cke                   => ddr3_ctrl_out.cke,          -- out   std_logic_vector(CKE_WIDTH-1 downto 0)
      ddr_cs_n                  => ddr3_ctrl_out.cs_l,         -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_dm                    => ddr3_ctrl_out.dm,           -- out   std_logic_vector(DM_WIDTH-1 downto 0)
      ddr_odt                   => ddr3_ctrl_out.odt,          -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_ras_n                 => ddr3_ctrl_out.ras_l,        -- out   std_logic
      ddr_reset_n               => ddr3_ctrl_out.reset_l,      -- out   std_logic
      ddr_parity                => open,                       -- out   std_logic
      ddr_we_n                  => ddr3_ctrl_out.we_l,         -- out   std_logic
      ddr_dq                    => ddr3_data_inout.dq,         -- inout std_logic_vector(DQ_WIDTH-1 downto 0)
      ddr_dqs_n                 => ddr3_data_inout.dqs_n,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      ddr_dqs                   => ddr3_data_inout.dqs_p,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      pd_PSEN                   => pd_psen,                    -- out   std_logic
      pd_PSINCDEC               => pd_psincdec,                -- out   std_logic
      pd_PSDONE                 => pd_psdone,                  -- in    std_logic
      bank_mach_next            => open,                       -- out   std_logic_vector(BM_CNT_WIDTH-1 downto 0)
      dfi_init_complete         => dfi_init_complete,          -- out   std_logic
      app_ecc_multiple_err      => open,                       -- out   std_logic_vector(3 downto 0)
      app_rd_data               => app_rd_data,                -- out   std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_rd_data_end           => open,                       -- out   std_logic
      app_rd_data_valid         => app_rd_data_valid,          -- out   std_logic
      app_rdy                   => app_rdy,                    -- out   std_logic
      app_wdf_rdy               => app_wdf_rdy,                -- out   std_logic
      app_addr                  => app_addr,                   -- in    std_logic_vector(ADDR_WIDTH-1 downto 0)
      app_cmd                   => app_cmd,                    -- in    std_logic_vector(2 downto 0)
      app_en                    => app_en,                     -- in    std_logic
      app_hi_pri                => '0',                        -- in    std_logic
      app_sz                    => app_sz,                     -- in    std_logic
      app_wdf_data              => app_wdf_data,               -- in    std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_wdf_end               => app_wdf_end,                -- in    std_logic
      app_wdf_mask              => app_wdf_mask,               -- in    std_logic_vector(APP_MASK_WIDTH-1 downto 0)
      app_wdf_wren              => app_wdf_wren,               -- in    std_logic
      dbg_wr_dqs_tap_set        => dbg_wr_dqs_tap_set,         -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_dq_tap_set         => dbg_wr_dq_tap_set,          -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_tap_set_en         => dbg_wr_tap_set_en,          -- in    std_logic
      dbg_wrlvl_start           => dbg_wrlvl_start,            -- out   std_logic
      dbg_wrlvl_done            => dbg_wrlvl_done,             -- out   std_logic
      dbg_wrlvl_err             => dbg_wrlvl_err,              -- out   std_logic
      dbg_wl_dqs_inverted       => dbg_wl_dqs_inverted,        -- out   std_logic_vector(DQS_WIDTH-1 downto 0)
      dbg_wr_calib_clk_delay    => dbg_wr_calib_clk_delay,     -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dqs_tap_cnt => dbg_wl_odelay_dqs_tap_cnt,  -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dq_tap_cnt  => dbg_wl_odelay_dq_tap_cnt,   -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rdlvl_start           => dbg_rdlvl_start,            -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_done            => dbg_rdlvl_done,             -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_err             => dbg_rdlvl_err,              -- out   std_logic_vector(1 downto 0)
      dbg_cpt_tap_cnt           => dbg_cpt_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_first_edge_cnt    => dbg_cpt_first_edge_cnt,     -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_second_edge_cnt   => dbg_cpt_second_edge_cnt,    -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rd_bitslip_cnt        => dbg_rd_bitslip_cnt,         -- out   std_logic_vector(3*DQS_WIDTH-1 downto 0)
      dbg_rd_clkdly_cnt         => dbg_rd_clkdly_cnt,          -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_rd_active_dly         => dbg_rd_active_dly,          -- out   std_logic_vector(4 downto 0)
      dbg_pd_off                => dbg_pd_off,                 -- in    std_logic
      dbg_pd_maintain_off       => dbg_pd_maintain_off,        -- in    std_logic
      dbg_pd_maintain_0_only    => dbg_pd_maintain_0_only,     -- in    std_logic
      dbg_inc_cpt               => dbg_inc_cpt,                -- in    std_logic
      dbg_dec_cpt               => dbg_dec_cpt,                -- in    std_logic
      dbg_inc_rd_dqs            => dbg_inc_rd_dqs,             -- in    std_logic
      dbg_dec_rd_dqs            => dbg_dec_rd_dqs,             -- in    std_logic
      dbg_inc_dec_sel           => dbg_inc_dec_sel,            -- in    std_logic_vector(DQS_CNT_WIDTH-1 downto 0)
      dbg_inc_rd_fps            => dbg_inc_rd_fps,             -- in    std_logic
      dbg_dec_rd_fps            => dbg_dec_rd_fps,             -- in    std_logic
      dbg_dqs_tap_cnt           => dbg_dqs_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_dq_tap_cnt            => dbg_dq_tap_cnt,             -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rddata                => dbg_rddata);                -- out   std_logic_vector(4*DQ_WIDTH-1 downto 0)

    -- DDR3 MIG clock generation
    infrastructure_i : infrastructure
      generic map(
      TCQ             => MIG_COMMON.TCQ,                -- integer := 100
      CLK_PERIOD      => SYSCLK_PERIOD,                 -- integer := 3000
      nCK_PER_CLK     => MIG_CLOCKS.nCK_PER_CLK,        -- integer := 2
      CLKFBOUT_MULT_F => MIG_CLOCKS.CLKFBOUT_MULT_F,    -- integer := 2
      DIVCLK_DIVIDE   => MIG_CLOCKS.DIVCLK_DIVIDE,      -- integer := 1
      CLKOUT_DIVIDE   => MIG_CLOCKS.CLKOUT_DIVIDE,      -- integer := 2
      RST_ACT_LOW     => MIG_CLOCKS_COMMON.RST_ACT_LOW) -- integer := 1
    port map(
      -- Clock inputs
      mmcm_clk         => ddr3_clk,              -- in    std_logic
      -- System reset input
      sys_rst          => ddr3_rst,              -- in    std_logic
      -- MMCM/IDELAYCTRL Lock status
      iodelay_ctrl_rdy => ddr3_iodelay_ctrl_rdy, -- in    std_logic
      -- Clock outputs
      clk_mem          => mig_clk_mem,           -- out   std_logic
      clk              => mig_ddr3_if_clk1,      -- out   std_logic
      clk_rd_base      => mig_clk_rd_base,       -- out   std_logic
      -- Reset outputs
      rstdiv0          => mig_rst,               -- out   std_logic
      -- Phase Shift Interface
      PSDONE           => pd_psdone,             -- out   std_logic
      PSEN             => pd_psen,               -- in    std_logic
      PSINCDEC         => pd_psincdec);          -- in    std_logic

    process(mig_ddr3_if_clk1)
    begin
      if rising_edge(mig_ddr3_if_clk1) then
        -- Connect ready
        ddr3_if_rdy <= dfi_init_complete;
      end if;
    end process;
  end generate;

  -- Instantiate DDR3 MIG bank 2 core
  bank2_g : if (bank = 2) generate
    -- Instantiate OCP to MIG block
    adb3_ocp_ocp2ddr3_nb_i : adb3_ocp_ocp2ddr3_nb
    generic map(
      app_row_width  => ROW_WIDTH_D,   -- natural
      app_col_width  => COL_WIDTH_D,   -- natural
      app_bank_width => BANK_WIDTH_D,  -- natural
      app_addr_width => ADDR_WIDTH_D)  -- natural)
    port map(
      -- OCP Interface
      ocp_rst           => ocp_rst,            -- in    std_logic
      ocp_clk           => ocp_clk,            -- in    std_logic
      ocp_m2s           => ocp_m2s,            -- in    adb3_ocp_m2sT
      ocp_s2m           => ocp_s2m,            -- out   adb3_ocp_s2mT
      -- MIG user interface
      mig_rst           => mig_rst,            -- in    std_logic
      mig_clk           => mig_ddr3_if_clk2,   -- in    std_logic
      phy_init_done     => dfi_init_complete,  -- in    std_logic
      app_rdy           => app_rdy,            -- in    std_logic
      app_wdf_rdy       => app_wdf_rdy,        -- in    std_logic
      app_rd_data       => app_rd_data,        -- in    std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_rd_data_valid => app_rd_data_valid,  -- in    std_logic
      app_en            => app_en,             -- out   std_logic
      app_cmd           => app_cmd,            -- out   std_logic_vector(2 downto 0)
      app_addr          => app_addr,           -- out   std_logic_vector(app_addr_width-1 downto 0);
      app_sz            => app_sz,             -- out   std_logic
      app_wdf_wren      => app_wdf_wren,       -- out   std_logic
      app_wdf_data      => app_wdf_data,       -- out   std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_wdf_mask      => app_wdf_mask,       -- out   std_logic_vector((ADB3_OCP_DATA_WIDTH/8)-1 downto 0)
      app_wdf_end       => app_wdf_end,        -- out   std_logic
      -- Debug info
      debug_info        => ddr3_if_debug);     -- out   std_logic_vector(31 downto 0))

    memc_ui_top_i : entity work.c2_memc_ui_top(arch_c2_memc_ui_top)
    generic map(
      ADDR_CMD_MODE       => MIG_COMMON.ADDR_CMD_MODE,
      BANK_WIDTH          => MIG_COMMON.BANK_WIDTH,
      CK_WIDTH            => MIG_COMMON.CK_WIDTH,
      CKE_WIDTH           => MIG_COMMON.CKE_WIDTH,
      nCK_PER_CLK         => MIG_CLOCKS.nCK_PER_CLK,
      COL_WIDTH           => MIG_COMMON.COL_WIDTH,
      CS_WIDTH            => MIG_COMMON.CS_WIDTH,
      DM_WIDTH            => MIG_COMMON.DM_WIDTH,
      nCS_PER_RANK        => MIG_COMMON.nCS_PER_RANK,
      DEBUG_PORT          => MIG_COMMON.DEBUG_PORT,
      IODELAY_GRP         => MIG_CLOCKS_COMMON.IODELAY_GRP,
      DQ_WIDTH            => MIG_COMMON.DQ_WIDTH,
      DQS_WIDTH           => MIG_COMMON.DQS_WIDTH,
      DQS_CNT_WIDTH       => MIG_COMMON.DQS_CNT_WIDTH,
      ORDERING            => MIG_COMMON.ORDERING,
      OUTPUT_DRV          => MIG_COMMON.OUTPUT_DRV,
      PHASE_DETECT        => MIG_COMMON.PHASE_DETECT,
      RANK_WIDTH          => MIG_COMMON.RANK_WIDTH,
      REFCLK_FREQ         => MIG_CLOCKS.REFCLK_FREQ,
      REG_CTRL            => MIG_COMMON.REG_CTRL,
      ROW_WIDTH           => MIG_COMMON.ROW_WIDTH,
      RTT_NOM             => MIG_COMMON.RTT_NOM,
      RTT_WR              => MIG_COMMON.RTT_WR,
      SIM_BYPASS_INIT_CAL => MIG_SIM_BYPASS_INIT_CAL,
      SIM_CAL_OPTION      => MIG_SIM_CAL_OPTION,
      SIM_INIT_OPTION     => MIG_SIM_INIT_OPTION,
      WRLVL               => MIG_COMMON.WRLVL,
      nDQS_COL0           => MIG_BANK2_DQS.nDQS_COL0,
      nDQS_COL1           => MIG_BANK2_DQS.nDQS_COL1,
      nDQS_COL2           => MIG_BANK2_DQS.nDQS_COL2,
      nDQS_COL3           => MIG_BANK2_DQS.nDQS_COL3,
      DQS_LOC_COL0        => MIG_BANK2_DQS.DQS_LOC_COL0,
      DQS_LOC_COL1        => MIG_BANK2_DQS.DQS_LOC_COL1,
      DQS_LOC_COL2        => MIG_BANK2_DQS.DQS_LOC_COL2,
      DQS_LOC_COL3        => MIG_BANK2_DQS.DQS_LOC_COL3,
      BURST_MODE          => MIG_COMMON.BURST_MODE,
      BM_CNT_WIDTH        => MIG_COMMON.BM_CNT_WIDTH,
      tCK                 => MIG_CLOCKS.tCK,
      tPRDI               => MIG_COMMON.tPRDI,
      tREFI               => MIG_COMMON.tREFI,
      tZQI                => MIG_COMMON.tZQI,
      tRFC                => MIG_T_RFC_OPTION,
      ADDR_WIDTH          => MIG_COMMON.ADDR_WIDTH,
      TCQ                 => MIG_COMMON.TCQ,
      ECC_TEST            => MIG_COMMON.ECC_TEST,
      PAYLOAD_WIDTH       => MIG_COMMON.PAYLOAD_WIDTH)
    port map(
      clk                       => mig_ddr3_if_clk2,           -- in    std_logic
      clk_mem                   => mig_clk_mem,                -- in    std_logic
      clk_rd_base               => mig_clk_rd_base,            -- in    std_logic
      rst                       => mig_rst,                    -- in    std_logic
      ddr_addr                  => ddr3_addr_out.a,            -- out   std_logic_vector(ROW_WIDTH-1 downto 0)
      ddr_ba                    => ddr3_ctrl_out.ba,           -- out   std_logic_vector(BANK_WIDTH-1 downto 0)
      ddr_cas_n                 => ddr3_ctrl_out.cas_l,        -- out   std_logic
      ddr_ck_n                  => ddr3_clk_out.clk_n,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_ck                    => ddr3_clk_out.clk_p,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_cke                   => ddr3_ctrl_out.cke,          -- out   std_logic_vector(CKE_WIDTH-1 downto 0)
      ddr_cs_n                  => ddr3_ctrl_out.cs_l,         -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_dm                    => ddr3_ctrl_out.dm,           -- out   std_logic_vector(DM_WIDTH-1 downto 0)
      ddr_odt                   => ddr3_ctrl_out.odt,          -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_ras_n                 => ddr3_ctrl_out.ras_l,        -- out   std_logic
      ddr_reset_n               => ddr3_ctrl_out.reset_l,      -- out   std_logic
      ddr_parity                => open,                       -- out   std_logic
      ddr_we_n                  => ddr3_ctrl_out.we_l,         -- out   std_logic
      ddr_dq                    => ddr3_data_inout.dq,         -- inout std_logic_vector(DQ_WIDTH-1 downto 0)
      ddr_dqs_n                 => ddr3_data_inout.dqs_n,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      ddr_dqs                   => ddr3_data_inout.dqs_p,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      pd_PSEN                   => pd_psen,                    -- out   std_logic
      pd_PSINCDEC               => pd_psincdec,                -- out   std_logic
      pd_PSDONE                 => pd_psdone,                  -- in    std_logic
      bank_mach_next            => open,                       -- out   std_logic_vector(BM_CNT_WIDTH-1 downto 0)
      dfi_init_complete         => dfi_init_complete,          -- out   std_logic
      app_ecc_multiple_err      => open,                       -- out   std_logic_vector(3 downto 0)
      app_rd_data               => app_rd_data,                -- out   std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_rd_data_end           => open,                       -- out   std_logic
      app_rd_data_valid         => app_rd_data_valid,          -- out   std_logic
      app_rdy                   => app_rdy,                    -- out   std_logic
      app_wdf_rdy               => app_wdf_rdy,                -- out   std_logic
      app_addr                  => app_addr,                   -- in    std_logic_vector(ADDR_WIDTH-1 downto 0)
      app_cmd                   => app_cmd,                    -- in    std_logic_vector(2 downto 0)
      app_en                    => app_en,                     -- in    std_logic
      app_hi_pri                => '0',                        -- in    std_logic
      app_sz                    => app_sz,                     -- in    std_logic
      app_wdf_data              => app_wdf_data,               -- in    std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_wdf_end               => app_wdf_end,                -- in    std_logic
      app_wdf_mask              => app_wdf_mask,               -- in    std_logic_vector(APP_MASK_WIDTH-1 downto 0)
      app_wdf_wren              => app_wdf_wren,               -- in    std_logic
      dbg_wr_dqs_tap_set        => dbg_wr_dqs_tap_set,         -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_dq_tap_set         => dbg_wr_dq_tap_set,          -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_tap_set_en         => dbg_wr_tap_set_en,          -- in    std_logic
      dbg_wrlvl_start           => dbg_wrlvl_start,            -- out   std_logic
      dbg_wrlvl_done            => dbg_wrlvl_done,             -- out   std_logic
      dbg_wrlvl_err             => dbg_wrlvl_err,              -- out   std_logic
      dbg_wl_dqs_inverted       => dbg_wl_dqs_inverted,        -- out   std_logic_vector(DQS_WIDTH-1 downto 0)
      dbg_wr_calib_clk_delay    => dbg_wr_calib_clk_delay,     -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dqs_tap_cnt => dbg_wl_odelay_dqs_tap_cnt,  -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dq_tap_cnt  => dbg_wl_odelay_dq_tap_cnt,   -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rdlvl_start           => dbg_rdlvl_start,            -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_done            => dbg_rdlvl_done,             -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_err             => dbg_rdlvl_err,              -- out   std_logic_vector(1 downto 0)
      dbg_cpt_tap_cnt           => dbg_cpt_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_first_edge_cnt    => dbg_cpt_first_edge_cnt,     -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_second_edge_cnt   => dbg_cpt_second_edge_cnt,    -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rd_bitslip_cnt        => dbg_rd_bitslip_cnt,         -- out   std_logic_vector(3*DQS_WIDTH-1 downto 0)
      dbg_rd_clkdly_cnt         => dbg_rd_clkdly_cnt,          -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_rd_active_dly         => dbg_rd_active_dly,          -- out   std_logic_vector(4 downto 0)
      dbg_pd_off                => dbg_pd_off,                 -- in    std_logic
      dbg_pd_maintain_off       => dbg_pd_maintain_off,        -- in    std_logic
      dbg_pd_maintain_0_only    => dbg_pd_maintain_0_only,     -- in    std_logic
      dbg_inc_cpt               => dbg_inc_cpt,                -- in    std_logic
      dbg_dec_cpt               => dbg_dec_cpt,                -- in    std_logic
      dbg_inc_rd_dqs            => dbg_inc_rd_dqs,             -- in    std_logic
      dbg_dec_rd_dqs            => dbg_dec_rd_dqs,             -- in    std_logic
      dbg_inc_dec_sel           => dbg_inc_dec_sel,            -- in    std_logic_vector(DQS_CNT_WIDTH-1 downto 0)
      dbg_inc_rd_fps            => dbg_inc_rd_fps,             -- in    std_logic
      dbg_dec_rd_fps            => dbg_dec_rd_fps,             -- in    std_logic
      dbg_dqs_tap_cnt           => dbg_dqs_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_dq_tap_cnt            => dbg_dq_tap_cnt,             -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rddata                => dbg_rddata);                -- out   std_logic_vector(4*DQ_WIDTH-1 downto 0)

    -- DDR3 MIG clock generation
    infrastructure_i : infrastructure
      generic map(
      TCQ             => MIG_COMMON.TCQ,                -- integer := 100
      CLK_PERIOD      => SYSCLK_PERIOD,                 -- integer := 3000
      nCK_PER_CLK     => MIG_CLOCKS.nCK_PER_CLK,        -- integer := 2
      CLKFBOUT_MULT_F => MIG_CLOCKS.CLKFBOUT_MULT_F,    -- integer := 2
      DIVCLK_DIVIDE   => MIG_CLOCKS.DIVCLK_DIVIDE,      -- integer := 1
      CLKOUT_DIVIDE   => MIG_CLOCKS.CLKOUT_DIVIDE,      -- integer := 2
      RST_ACT_LOW     => MIG_CLOCKS_COMMON.RST_ACT_LOW) -- integer := 1
    port map(
      -- Clock inputs
      mmcm_clk         => ddr3_clk,              -- in    std_logic
      -- System reset input
      sys_rst          => ddr3_rst,              -- in    std_logic
      -- MMCM/IDELAYCTRL Lock status
      iodelay_ctrl_rdy => ddr3_iodelay_ctrl_rdy, -- in    std_logic
      -- Clock outputs
      clk_mem          => mig_clk_mem,           -- out   std_logic
      clk              => mig_ddr3_if_clk2,      -- out   std_logic
      clk_rd_base      => mig_clk_rd_base,       -- out   std_logic
      -- Reset outputs
      rstdiv0          => mig_rst,               -- out   std_logic
      -- Phase Shift Interface
      PSDONE           => pd_psdone,             -- out   std_logic
      PSEN             => pd_psen,               -- in    std_logic
      PSINCDEC         => pd_psincdec);          -- in    std_logic

    process(mig_ddr3_if_clk2)
    begin
      if rising_edge(mig_ddr3_if_clk2) then
        -- Connect ready
        ddr3_if_rdy <= dfi_init_complete;
      end if;
    end process;
  end generate;

  -- Instantiate DDR3 MIG bank 3 core
  bank3_g : if (bank = 3) generate
    -- Instantiate OCP to MIG block
    adb3_ocp_ocp2ddr3_nb_i : adb3_ocp_ocp2ddr3_nb
    generic map(
      app_row_width  => ROW_WIDTH_D,   -- natural
      app_col_width  => COL_WIDTH_D,   -- natural
      app_bank_width => BANK_WIDTH_D,  -- natural
      app_addr_width => ADDR_WIDTH_D)  -- natural)
    port map(
      -- OCP Interface
      ocp_rst           => ocp_rst,            -- in    std_logic
      ocp_clk           => ocp_clk,            -- in    std_logic
      ocp_m2s           => ocp_m2s,            -- in    adb3_ocp_m2sT
      ocp_s2m           => ocp_s2m,            -- out   adb3_ocp_s2mT
      -- MIG user interface
      mig_rst           => mig_rst,            -- in    std_logic
      mig_clk           => mig_ddr3_if_clk3,   -- in    std_logic
      phy_init_done     => dfi_init_complete,  -- in    std_logic
      app_rdy           => app_rdy,            -- in    std_logic
      app_wdf_rdy       => app_wdf_rdy,        -- in    std_logic
      app_rd_data       => app_rd_data,        -- in    std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_rd_data_valid => app_rd_data_valid,  -- in    std_logic
      app_en            => app_en,             -- out   std_logic
      app_cmd           => app_cmd,            -- out   std_logic_vector(2 downto 0)
      app_addr          => app_addr,           -- out   std_logic_vector(app_addr_width-1 downto 0);
      app_sz            => app_sz,             -- out   std_logic
      app_wdf_wren      => app_wdf_wren,       -- out   std_logic
      app_wdf_data      => app_wdf_data,       -- out   std_logic_vector(ADB3_OCP_DATA_WIDTH-1 downto 0)
      app_wdf_mask      => app_wdf_mask,       -- out   std_logic_vector((ADB3_OCP_DATA_WIDTH/8)-1 downto 0)
      app_wdf_end       => app_wdf_end,        -- out   std_logic
      -- Debug info
      debug_info        => ddr3_if_debug);     -- out   std_logic_vector(31 downto 0))

    memc_ui_top_i : entity work.c3_memc_ui_top(arch_c3_memc_ui_top)
    generic map(
      ADDR_CMD_MODE       => MIG_COMMON.ADDR_CMD_MODE,
      BANK_WIDTH          => MIG_COMMON.BANK_WIDTH,
      CK_WIDTH            => MIG_COMMON.CK_WIDTH,
      CKE_WIDTH           => MIG_COMMON.CKE_WIDTH,
      nCK_PER_CLK         => MIG_CLOCKS.nCK_PER_CLK,
      COL_WIDTH           => MIG_COMMON.COL_WIDTH,
      CS_WIDTH            => MIG_COMMON.CS_WIDTH,
      DM_WIDTH            => MIG_COMMON.DM_WIDTH,
      nCS_PER_RANK        => MIG_COMMON.nCS_PER_RANK,
      DEBUG_PORT          => MIG_COMMON.DEBUG_PORT,
      IODELAY_GRP         => MIG_CLOCKS_COMMON.IODELAY_GRP,
      DQ_WIDTH            => MIG_COMMON.DQ_WIDTH,
      DQS_WIDTH           => MIG_COMMON.DQS_WIDTH,
      DQS_CNT_WIDTH       => MIG_COMMON.DQS_CNT_WIDTH,
      ORDERING            => MIG_COMMON.ORDERING,
      OUTPUT_DRV          => MIG_COMMON.OUTPUT_DRV,
      PHASE_DETECT        => MIG_COMMON.PHASE_DETECT,
      RANK_WIDTH          => MIG_COMMON.RANK_WIDTH,
      REFCLK_FREQ         => MIG_CLOCKS.REFCLK_FREQ,
      REG_CTRL            => MIG_COMMON.REG_CTRL,
      ROW_WIDTH           => MIG_COMMON.ROW_WIDTH,
      RTT_NOM             => MIG_COMMON.RTT_NOM,
      RTT_WR              => MIG_COMMON.RTT_WR,
      SIM_BYPASS_INIT_CAL => MIG_SIM_BYPASS_INIT_CAL,
      SIM_CAL_OPTION      => MIG_SIM_CAL_OPTION,
      SIM_INIT_OPTION     => MIG_SIM_INIT_OPTION,
      WRLVL               => MIG_COMMON.WRLVL,
      nDQS_COL0           => MIG_BANK3_DQS.nDQS_COL0,
      nDQS_COL1           => MIG_BANK3_DQS.nDQS_COL1,
      nDQS_COL2           => MIG_BANK3_DQS.nDQS_COL2,
      nDQS_COL3           => MIG_BANK3_DQS.nDQS_COL3,
      DQS_LOC_COL0        => MIG_BANK3_DQS.DQS_LOC_COL0,
      DQS_LOC_COL1        => MIG_BANK3_DQS.DQS_LOC_COL1,
      DQS_LOC_COL2        => MIG_BANK3_DQS.DQS_LOC_COL2,
      DQS_LOC_COL3        => MIG_BANK3_DQS.DQS_LOC_COL3,
      BURST_MODE          => MIG_COMMON.BURST_MODE,
      BM_CNT_WIDTH        => MIG_COMMON.BM_CNT_WIDTH,
      tCK                 => MIG_CLOCKS.tCK,
      tPRDI               => MIG_COMMON.tPRDI,
      tREFI               => MIG_COMMON.tREFI,
      tZQI                => MIG_COMMON.tZQI,
      tRFC                => MIG_T_RFC_OPTION,
      ADDR_WIDTH          => MIG_COMMON.ADDR_WIDTH,
      TCQ                 => MIG_COMMON.TCQ,
      ECC_TEST            => MIG_COMMON.ECC_TEST,
      PAYLOAD_WIDTH       => MIG_COMMON.PAYLOAD_WIDTH)
    port map(
      clk                       => mig_ddr3_if_clk3,           -- in    std_logic
      clk_mem                   => mig_clk_mem,                -- in    std_logic
      clk_rd_base               => mig_clk_rd_base,            -- in    std_logic
      rst                       => mig_rst,                    -- in    std_logic
      ddr_addr                  => ddr3_addr_out.a,            -- out   std_logic_vector(ROW_WIDTH-1 downto 0)
      ddr_ba                    => ddr3_ctrl_out.ba,           -- out   std_logic_vector(BANK_WIDTH-1 downto 0)
      ddr_cas_n                 => ddr3_ctrl_out.cas_l,        -- out   std_logic
      ddr_ck_n                  => ddr3_clk_out.clk_n,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_ck                    => ddr3_clk_out.clk_p,         -- out   std_logic_vector(CK_WIDTH-1 downto 0)
      ddr_cke                   => ddr3_ctrl_out.cke,          -- out   std_logic_vector(CKE_WIDTH-1 downto 0)
      ddr_cs_n                  => ddr3_ctrl_out.cs_l,         -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_dm                    => ddr3_ctrl_out.dm,           -- out   std_logic_vector(DM_WIDTH-1 downto 0)
      ddr_odt                   => ddr3_ctrl_out.odt,          -- out   std_logic_vector(CS_WIDTH*nCS_PER_RANK-1 downto 0)
      ddr_ras_n                 => ddr3_ctrl_out.ras_l,        -- out   std_logic
      ddr_reset_n               => ddr3_ctrl_out.reset_l,      -- out   std_logic
      ddr_parity                => open,                       -- out   std_logic
      ddr_we_n                  => ddr3_ctrl_out.we_l,         -- out   std_logic
      ddr_dq                    => ddr3_data_inout.dq,         -- inout std_logic_vector(DQ_WIDTH-1 downto 0)
      ddr_dqs_n                 => ddr3_data_inout.dqs_n,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      ddr_dqs                   => ddr3_data_inout.dqs_p,      -- inout std_logic_vector(DQS_WIDTH-1 downto 0)
      pd_PSEN                   => pd_psen,                    -- out   std_logic
      pd_PSINCDEC               => pd_psincdec,                -- out   std_logic
      pd_PSDONE                 => pd_psdone,                  -- in    std_logic
      bank_mach_next            => open,                       -- out   std_logic_vector(BM_CNT_WIDTH-1 downto 0)
      dfi_init_complete         => dfi_init_complete,          -- out   std_logic
      app_ecc_multiple_err      => open,                       -- out   std_logic_vector(3 downto 0)
      app_rd_data               => app_rd_data,                -- out   std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_rd_data_end           => open,                       -- out   std_logic
      app_rd_data_valid         => app_rd_data_valid,          -- out   std_logic
      app_rdy                   => app_rdy,                    -- out   std_logic
      app_wdf_rdy               => app_wdf_rdy,                -- out   std_logic
      app_addr                  => app_addr,                   -- in    std_logic_vector(ADDR_WIDTH-1 downto 0)
      app_cmd                   => app_cmd,                    -- in    std_logic_vector(2 downto 0)
      app_en                    => app_en,                     -- in    std_logic
      app_hi_pri                => '0',                        -- in    std_logic
      app_sz                    => app_sz,                     -- in    std_logic
      app_wdf_data              => app_wdf_data,               -- in    std_logic_vector(APP_DATA_WIDTH-1 downto 0)
      app_wdf_end               => app_wdf_end,                -- in    std_logic
      app_wdf_mask              => app_wdf_mask,               -- in    std_logic_vector(APP_MASK_WIDTH-1 downto 0)
      app_wdf_wren              => app_wdf_wren,               -- in    std_logic
      dbg_wr_dqs_tap_set        => dbg_wr_dqs_tap_set,         -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_dq_tap_set         => dbg_wr_dq_tap_set,          -- in    std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wr_tap_set_en         => dbg_wr_tap_set_en,          -- in    std_logic
      dbg_wrlvl_start           => dbg_wrlvl_start,            -- out   std_logic
      dbg_wrlvl_done            => dbg_wrlvl_done,             -- out   std_logic
      dbg_wrlvl_err             => dbg_wrlvl_err,              -- out   std_logic
      dbg_wl_dqs_inverted       => dbg_wl_dqs_inverted,        -- out   std_logic_vector(DQS_WIDTH-1 downto 0)
      dbg_wr_calib_clk_delay    => dbg_wr_calib_clk_delay,     -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dqs_tap_cnt => dbg_wl_odelay_dqs_tap_cnt,  -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_wl_odelay_dq_tap_cnt  => dbg_wl_odelay_dq_tap_cnt,   -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rdlvl_start           => dbg_rdlvl_start,            -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_done            => dbg_rdlvl_done,             -- out   std_logic_vector(1 downto 0)
      dbg_rdlvl_err             => dbg_rdlvl_err,              -- out   std_logic_vector(1 downto 0)
      dbg_cpt_tap_cnt           => dbg_cpt_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_first_edge_cnt    => dbg_cpt_first_edge_cnt,     -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_cpt_second_edge_cnt   => dbg_cpt_second_edge_cnt,    -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rd_bitslip_cnt        => dbg_rd_bitslip_cnt,         -- out   std_logic_vector(3*DQS_WIDTH-1 downto 0)
      dbg_rd_clkdly_cnt         => dbg_rd_clkdly_cnt,          -- out   std_logic_vector(2*DQS_WIDTH-1 downto 0)
      dbg_rd_active_dly         => dbg_rd_active_dly,          -- out   std_logic_vector(4 downto 0)
      dbg_pd_off                => dbg_pd_off,                 -- in    std_logic
      dbg_pd_maintain_off       => dbg_pd_maintain_off,        -- in    std_logic
      dbg_pd_maintain_0_only    => dbg_pd_maintain_0_only,     -- in    std_logic
      dbg_inc_cpt               => dbg_inc_cpt,                -- in    std_logic
      dbg_dec_cpt               => dbg_dec_cpt,                -- in    std_logic
      dbg_inc_rd_dqs            => dbg_inc_rd_dqs,             -- in    std_logic
      dbg_dec_rd_dqs            => dbg_dec_rd_dqs,             -- in    std_logic
      dbg_inc_dec_sel           => dbg_inc_dec_sel,            -- in    std_logic_vector(DQS_CNT_WIDTH-1 downto 0)
      dbg_inc_rd_fps            => dbg_inc_rd_fps,             -- in    std_logic
      dbg_dec_rd_fps            => dbg_dec_rd_fps,             -- in    std_logic
      dbg_dqs_tap_cnt           => dbg_dqs_tap_cnt,            -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_dq_tap_cnt            => dbg_dq_tap_cnt,             -- out   std_logic_vector(5*DQS_WIDTH-1 downto 0)
      dbg_rddata                => dbg_rddata);                -- out   std_logic_vector(4*DQ_WIDTH-1 downto 0)

    -- DDR3 MIG clock generation
    infrastructure_i : infrastructure
      generic map(
      TCQ             => MIG_COMMON.TCQ,                -- integer := 100
      CLK_PERIOD      => SYSCLK_PERIOD,                 -- integer := 3000
      nCK_PER_CLK     => MIG_CLOCKS.nCK_PER_CLK,        -- integer := 2
      CLKFBOUT_MULT_F => MIG_CLOCKS.CLKFBOUT_MULT_F,    -- integer := 2
      DIVCLK_DIVIDE   => MIG_CLOCKS.DIVCLK_DIVIDE,      -- integer := 1
      CLKOUT_DIVIDE   => MIG_CLOCKS.CLKOUT_DIVIDE,      -- integer := 2
      RST_ACT_LOW     => MIG_CLOCKS_COMMON.RST_ACT_LOW) -- integer := 1
    port map(
      -- Clock inputs
      mmcm_clk         => ddr3_clk,              -- in    std_logic
      -- System reset input
      sys_rst          => ddr3_rst,              -- in    std_logic
      -- MMCM/IDELAYCTRL Lock status
      iodelay_ctrl_rdy => ddr3_iodelay_ctrl_rdy, -- in    std_logic
      -- Clock outputs
      clk_mem          => mig_clk_mem,           -- out   std_logic
      clk              => mig_ddr3_if_clk3,      -- out   std_logic
      clk_rd_base      => mig_clk_rd_base,       -- out   std_logic
      -- Reset outputs
      rstdiv0          => mig_rst,               -- out   std_logic
      -- Phase Shift Interface
      PSDONE           => pd_psdone,             -- out   std_logic
      PSEN             => pd_psen,               -- in    std_logic
      PSINCDEC         => pd_psincdec);          -- in    std_logic

    process(mig_ddr3_if_clk3)
    begin
      if rising_edge(mig_ddr3_if_clk3) then
        -- Connect ready
        ddr3_if_rdy <= dfi_init_complete;
      end if;
    end process;
  end generate;

  -- Connect status
  ddr3_if_stat(0) <= dbg_wrlvl_done;
  ddr3_if_stat(1) <= dbg_rdlvl_done(0);
  ddr3_if_stat(2) <= dbg_rdlvl_done(1);
  ddr3_if_stat(3) <= dfi_init_complete;
  -- Connect error
  ddr3_if_err(0)  <= dbg_wrlvl_err;
  ddr3_if_err(1)  <= dbg_rdlvl_err(0);
  ddr3_if_err(2)  <= dbg_rdlvl_err(1);
  ddr3_if_err(3)  <= mig_rst;

end;
