--==============================================================================
--
--     File Name: ddr3_sdram_timing_chk.vhd
--       Version: 0.1
--          Date: 28 Dec 2009
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Generic DDR3 SDRAM model
--
--   Description: Model run-time parameter checking
--                Checking of model constants defined in ddr3_sdram_pkg1
--                Checking of model input timing parameters defined in ddr3_sdram_pkg1
--
--   Limitations: Precharge Power Down and Active Power Down not supported
--                OCD Calibration not supported
--                Reads interrupted by reads not supported
--                Writes interrupted by writes not supported
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 28 Dec 09: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;
use ieee.std_logic_arith.all;
use ieee.numeric_std.all;
use ieee.math_real.all;       -- Used for CEIL and LOG2 functions only

library work;
use work.ddr3_sdram_pkg.all;
use work.ddr3_sdram_int_pkg.all;

entity ddr3_sdram_timing_chk is
  generic(
    part : part_t);
  port(
    timing_enabled : in    boolean;
    bank_read_any  : in    boolean;  -- TRUE => current command is a READ/RDAP
    cke_setup      : in    boolean;  -- TRUE => CKE should be checked for setup/hold/validity
    cs_l_setup     : in    boolean;  -- TRUE => CS# should be checked for setup/hold/validity
    ras_l_setup    : in    boolean;  -- TRUE => RAS# should be checked for setup/hold/validity
    cas_l_setup    : in    boolean;  -- TRUE => CAS# should be checked for setup/hold/validity
    we_l_setup     : in    boolean;  -- TRUE => WE# should be checked for setup/hold/validity
    odt_setup      : in    boolean;  -- TRUE => ODT should be checked for setup/hold/validity
    ba_setup       : in    boolean;  -- TRUE => BA should be checked for setup/hold/validity
    a_setup        : in    boolean;  -- TRUE => A except A[10] should be checked for setup/hold/validity
    a10_setup      : in    boolean;  -- TRUE => A[10] should be checked for setup/hold/validity
    dm_setup_r     : in    boolean;  -- TRUE => DM should be checked for setup/hold/validity at CK rising
    dm_setup_f     : in    boolean;  -- TRUE => DM should be checked for setup/hold/validity at CK falling
    t_ck_min       : in    time;
    t_ck_max       : in    time;
    ck_period_sig  : in    time;
    ck             : in    std_logic;
    ck_l           : in    std_logic;
    cke            : in    std_logic;
    cs_l           : in    std_logic;
    ras_l          : in    std_logic;
    cas_l          : in    std_logic;
    we_l           : in    std_logic;
    odt            : in    std_logic;
    dm             : in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
    ba             : in    std_logic_vector(part.part_size.bank_width-1 downto 0);
    a              : in    std_logic_vector(part.part_size.arow_width-1 downto 0);
    dq             : in    std_logic_vector(part.part_size.data_width-1 downto 0);
    dqs            : in    std_logic_vector(part.part_size.data_bytes-1 downto 0);
    dqs_l          : in    std_logic_vector(part.part_size.data_bytes-1 downto 0));
  end ddr3_sdram_timing_chk;


architecture behavioural of ddr3_sdram_timing_chk is

  constant DATA_WIDTH : natural := part.part_size.data_width;
  constant BYTE_WIDTH : natural := part.part_size.byte_width;
  constant AROW_WIDTH : natural := part.part_size.arow_width;
  constant ACOL_WIDTH : natural := part.part_size.acol_width;
  constant DATA_BYTES : natural := part.part_size.data_bytes;
  constant BANK_WIDTH : natural := part.part_size.bank_width;
  constant SDRAM_BITS : natural := part.part_size.sdram_bits;

  constant DATA_BITS  : natural := NATURAL(CEIL(LOG2(REAL(DATA_WIDTH-1))));

  signal t_cl_min : time;
  signal t_cl_max : time;
  signal t_ch_min : time;
  signal t_ch_max : time;

  signal t_dqsl_min : time;
  signal t_dqsh_min : time;
  signal t_dqss     : time := 0 ps;
  signal t_dss_min  : time;
  signal t_dsh_min  : time;

  shared variable ck_rising      : time := 0 ns;
  shared variable ck_falling     : time := 0 ns;
  shared variable ck_high        : time := 0 ns;
  shared variable ck_low         : time := 0 ns;
  shared variable ck_period      : time := 0 ns;

  shared variable ck_l_rising    : time := 0 ns;
  shared variable ck_l_falling   : time := 0 ns;
  shared variable ck_l_high      : time := 0 ns;
  shared variable ck_l_low       : time := 0 ns;
  shared variable ck_l_period    : time := 0 ns;

  signal cke_hold   : boolean := false;
  signal cs_l_hold  : boolean := false;
  signal ras_l_hold : boolean := false;
  signal cas_l_hold : boolean := false;
  signal we_l_hold  : boolean := false;
  signal ba_hold    : boolean := false;
  signal a_hold     : boolean := false;
  signal a10_hold   : boolean := false;
  signal odt_hold   : boolean := false;
  signal dm_hold_r  : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dm_hold_f  : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dq_setup_r : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dq_setup_f : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dq_hold_r  : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dq_hold_f  : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dqs_hold_r : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);
  signal dqs_l_hold_r : boolean_array_t(0 to DATA_BYTES - 1) := (others => false);

  type byte_array_time_t is array (natural range 0 to (DATA_BYTES - 1)) of time;

  shared variable dqs_rising    : byte_array_time_t := (others => 0 ns);
  shared variable dqs_falling   : byte_array_time_t := (others => 0 ns);
  shared variable dqs_high      : byte_array_time_t := (others => 0 ns);
  shared variable dqs_low       : byte_array_time_t := (others => 0 ns);

  shared variable dqs_l_rising  : byte_array_time_t := (others => 0 ns);
  shared variable dqs_l_falling : byte_array_time_t := (others => 0 ns);
  shared variable dqs_l_high    : byte_array_time_t := (others => 0 ns);
  shared variable dqs_l_low     : byte_array_time_t := (others => 0 ns);

  signal ck_delayed   : std_logic;
  signal ck_l_delayed : std_logic;

  alias write_rising  : boolean is dm_setup_r;
  alias write_falling : boolean is dm_setup_f;

  signal write_rising_d1 : boolean := false;

  signal a_except_a10 : std_logic_vector(part.part_size.arow_width-1 downto 0);
  signal a10 : std_logic;

begin

  ----------------------------------
  -- Part size parameter checking --
  ----------------------------------

  -- Check SDRAM size and width parameters defined in package
  assert (DATA_WIDTH = DATA_BYTES * BYTE_WIDTH)
    report "DDR3 SDRAM parameter check (01): Data size definition conflict: " &
    integer'image(DATA_WIDTH) & " /= " &
    integer'image(DATA_BYTES) & " * " &
    integer'image(BYTE_WIDTH)
    severity failure;

  assert (SDRAM_BITS = AROW_WIDTH + ACOL_WIDTH + BANK_WIDTH + DATA_BITS)
    report "DDR3 SDRAM parameter check (02): Address size definition conflict:" &
    integer'image(SDRAM_BITS) & " /= " &
    integer'image(AROW_WIDTH) & " + " &
    integer'image(ACOL_WIDTH) & " + " &
    integer'image(BANK_WIDTH) & " + " &
    integer'image(DATA_BITS)
    severity failure;


  -------------------------------------
  -- Clock periods and skew checking --
  -------------------------------------

  -- Generate t_cl and t_ch values from parameters
  -- time <= time * real
  t_cl_min <= ck_period_sig * part.speed_grade.t_cl_min;
  t_cl_max <= ck_period_sig * part.speed_grade.t_cl_max;
  t_ch_min <= ck_period_sig * part.speed_grade.t_ch_min;
  t_ch_max <= ck_period_sig * part.speed_grade.t_ch_max;

  -- Check input clock ck
  -- Check cycle time
  -- Check high pulse width
  -- Check low pulse width
  check_ck : process(ck)
  begin
    if (ck'event and ck = '1') then
      ck_rising := now;
      ck_low    := ck_rising - ck_falling;
      ck_period := ck_high + ck_low;

      if (timing_enabled) then
        assert not(ck_low < t_cl_min)
          report "DDR3 SDRAM timing check (03): ck clock low pulse too small"
          severity error;
        assert not(ck_low > t_cl_max)
          report "DDR3 SDRAM timing check (04): ck clock low pulse too large"
          severity error;

        if bank_read_any then
          assert not(ck_period < t_ck_min)
            report "DDR3 SDRAM timing check (05): ck clock period too small for selected CL"
            severity error;
          assert not(ck_period > t_ck_max)
            report "DDR3 SDRAM timing check (06): ck clock period too large for selected CL"
            severity error;
        end if;
      end if;
    end if;

    if (ck'event and ck = '0') then
      ck_falling := now;
      ck_high    := ck_falling - ck_rising;

      if (timing_enabled) then
        assert not(ck_high < t_ch_min)
          report "DDR3 SDRAM timing check (07): ck clock high pulse too small"
          severity error;

        assert not(ck_high > t_ch_max)
          report "DDR3 SDRAM timing check (08): ck clock high pulse too large"
          severity error;
      end if;
    end if;
  end process;

  -- Check input clock ck not stuck
  check_ck_stuck : process(timing_enabled)
  begin
    if (timing_enabled) then
      assert not(ck_period = 0 ns)
        report "DDR3 SDRAM timing check (09): ck clock stuck at " & std_logic'image(ck)
        severity error;
    end if;
  end process;

  -- Check input clock ck_l
  -- Check cycle time
  -- Check high pulse width
  -- Check low pulse width
  check_ck_l : process(ck_l)
  begin
    if (ck_l'event and ck_l = '1') then
      ck_l_rising := now;
      ck_l_low    := ck_l_rising - ck_l_falling;
      ck_l_period := ck_l_high + ck_l_low;

      if (timing_enabled) then
        assert not(ck_l_low < t_cl_min)
          report "DDR3 SDRAM timing check (10): ck_l clock low pulse too small"
          severity error;

        assert not(ck_l_low > t_cl_max)
          report "DDR3 SDRAM timing check (11): ck_l clock low pulse too large"
          severity error;

        if bank_read_any then
          assert not(ck_l_period < t_ck_min)
            report "DDR3 SDRAM timing check (12): ck_l clock period too small for selected CL"
            severity error;

          assert not(ck_l_period > t_ck_max)
            report "DDR3 SDRAM timing check (13): ck_l clock period too large for selected CL"
            severity error;
        end if;
      end if;
    end if;

    if (ck_l'event and ck_l = '0') then
      ck_l_falling := now;
      ck_l_high    := ck_l_falling - ck_l_rising;

      if (timing_enabled) then
        assert not(ck_l_high < t_ch_min)
          report "DDR3 SDRAM timing check (14): ck_l clock high pulse too small"
          severity error;

        assert not(ck_l_high > t_ch_max)
          report "DDR3 SDRAM timing check (15): ck_l clock high pulse too large"
          severity error;
      end if;
    end if;
  end process;

  -- Check input clock ck_l not stuck
  check_ck_l_stuck : process(timing_enabled)
  begin
    if (timing_enabled) then
      assert not(ck_l_period = 0 ns)
        report "DDR3 SDRAM timing check (16): ck_l clock stuck at " & std_logic'image(ck)
        severity error;
    end if;
  end process;

  ----------------------------------
  -- Setup and hold time checking --
  ----------------------------------

  -- Check 'cke' hold time referenced to 'ck'
  check_cke_hold : process(cke)
  begin
    assert not timing_enabled or not cke_hold
      report "DDR3 SDRAM timing check (17): ck to cke hold time too small"
      severity error;
  end process;

  -- Check 'cke' setup time referenced to 'k'
  check_cke_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and cke_setup then
      cke_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert cke'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (18): cke to ck setup time too small"
        severity error;

      assert not(cke = 'X')
        report "DDR3 SDRAM timing check (19): cke invalid at ck edge"
        severity error;
    end if;
  end process;

  -- Check 'cs_l' hold time referenced to 'ck'
  check_cs_l_hold : process(cs_l)
  begin
    assert not timing_enabled or not cs_l_hold
      report "DDR3 SDRAM timing check (20): ck to cs_l hold time too small"
      severity error;
  end process;

  -- Check 'cs_l' setup time referenced to 'ck'
  check_cs_l_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and cs_l_setup then
      cs_l_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert cs_l'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (21): cs_l to ck setup time too small"
        severity error;

      assert not(cs_l = 'X')
        report "DDR3 SDRAM timing check (22): cs_l invalid at ck edge"
        severity error;
    end if;
  end process;

  -- Check 'ras_l' hold time referenced to 'ck'
  check_ras_l_hold : process(ras_l)
  begin
    assert not timing_enabled or not ras_l_hold
      report "DDR3 SDRAM timing check (23): ck to ras_l hold time too small"
      severity error;
  end process;

  -- Check 'ras_l' setup time referenced to 'ck'
  check_ras_l_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and ras_l_setup then
      ras_l_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert ras_l'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (24): ras_l to ck setup time too small"
        severity error;

      assert not(ras_l = 'X')
        report "DDR3 SDRAM timing check (25): ras_l invalid at ck edge"
        severity error;
    end if;
  end process;

  -- Check 'cas_l' hold time referenced to 'ck'
  check_cas_l_hold : process(cas_l)
  begin
    assert not timing_enabled or not cas_l_hold
      report "DDR3 SDRAM timing check (26): ck to cas_l hold time too small"
      severity error;
  end process;

  -- Check 'cas_l' setup time referenced to 'ck'
  check_cas_l_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and cas_l_setup then
      cas_l_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert cas_l'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (27): cas_l to ck setup time too small"
        severity error;

      assert not(cas_l = 'X')
        report "DDR3 SDRAM timing check (28): cas_l invalid at ck edge"
        severity error;
    end if;
  end process;

  -- Check 'we_l' hold time referenced to 'ck'
  check_we_l_hold : process(we_l)
  begin
    assert not timing_enabled or not we_l_hold
      report "DDR3 SDRAM timing check (29): ck to we_l hold time too small"
      severity error;
  end process;

  -- Check 'we_l' setup time referenced to 'ck'
  check_we_l_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and we_l_setup then
      we_l_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert we_l'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (30): we_l to ck setup time too small"
        severity error;

      assert not(we_l = 'X')
        report "DDR3 SDRAM timing check (31): we_l invalid at ck edge"
        severity error;
    end if;
  end process;

  a_except_a10 <= a(a'high downto 11) & '0' & a(9 downto 0);

  -- Check 'a' (except a[10]) hold time referenced to 'ck'
  check_a_hold : process(a_except_a10)
  begin
    assert not timing_enabled or not a_hold
      report "DDR3 SDRAM timing check (32): ck to a (minus a[10]) hold time too small"
      severity error;
  end process;

  -- Check 'a' setup time referenced to 'ck'
  check_a_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and a_setup then
      a_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert a_except_a10'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (33): a (minus a[10]) to ck setup time too small"
        severity error;

      assert not(IS_X(a_except_a10))
        report "DDR3 SDRAM timing check (34): a (minus a[10]) invalid at ck edge"
        severity error;
    end if;
  end process;

  a10 <= a(10);

  -- Check 'a[10]' hold time referenced to 'ck'
  check_a10_hold : process(a10)
  begin
    assert not timing_enabled or not a10_hold
      report "DDR3 SDRAM timing check (35): ck to a[10] hold time too small"
      severity error;
  end process;

  -- Check 'a[10]' setup time referenced to 'ck'
  check_a10_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and a10_setup then
      a10_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert a10'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (36): a[10] to ck setup time too small"
        severity error;

      assert not(a10 = 'X')
        report "DDR3 SDRAM timing check (37): a[10] invalid at ck edge"
        severity error;
    end if;
  end process;

  -- Check 'ba' hold time referenced to 'ck'
  check_ba_hold : process(ba)
  begin
    assert not timing_enabled or not ba_hold
      report "DDR3 SDRAM timing check (38): ck to ba hold time too small"
      severity error;
  end process;

  -- Check 'ba' setup time referenced to 'ck'
  check_ba_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and ba_setup then
      ba_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert ba'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (39): ba to ck setup time too small"
        severity error;

      assert not(IS_X(ba))
        report "DDR3 SDRAM timing check (40): ba invalid at ck edge"
        severity error;
    end if;
  end process;

  -- Check 'odt' hold time referenced to 'ck'
  check_odt_hold : process(odt)
  begin
    assert not timing_enabled or not odt_hold
      report "DDR3 SDRAM timing check (41): ck to odt hold time too small"
      severity error;
  end process;

  -- Check 'odt' setup time referenced to 'ck'
  check_odt_setup : process(ck)
  begin
    if rising_edge(ck) and timing_enabled and odt_setup then
      odt_hold <= transport true, false after part.speed_grade.t_ih_min;

      assert odt'stable(part.speed_grade.t_is_min)
        report "DDR3 SDRAM timing check (42): odt to ck setup time too small"
        severity error;

      assert not(odt = 'X')
        report "DDR3 SDRAM timing check (43): odt invalid at ck edge"
        severity error;
    end if;
  end process;

  check_dm : for i in 0 to DATA_BYTES-1 generate
    -- Check 'dm' hold time referenced to 'dqs' rising
    check_dm_hold_r : process(dm(i))
    begin
      assert not timing_enabled or not dm_hold_r(i)
        report "DDR3 SDRAM timing check (44): dqs(" & natural'image(i) & ") rising to dm(" & natural'image(i) & ") hold time too small"
        severity error;
    end process;

    -- Check 'dm' setup time referenced to 'dqs' rising
    check_dm_setup_r : process(dqs(i))
    begin
      if rising_edge(dqs(i)) and timing_enabled and dm_setup_r then
        dm_hold_r(i) <= transport true, false after part.speed_grade.t_dh_min;

        assert dm(i)'stable(part.speed_grade.t_ds_min)
          report "DDR3 SDRAM timing check (45): dm(" & natural'image(i) & ") to dqs(" & natural'image(i) & ") rising setup time too small"
          severity error;

        assert not(IS_X(dm))
          report "DDR3 SDRAM timing check (46): dm(" & natural'image(i) & ") invalid at dqs(" & natural'image(i) & ") rising edge"
          severity error;
      end if;
    end process;

    -- Check 'dm' hold time referenced to 'dqs' falling
    check_dm_hold_f : process(dm(i))
    begin
      assert not timing_enabled or not dm_hold_f(i)
        report "DDR3 SDRAM timing check (47): dqs(" & natural'image(i) & ") falling to dm(" & natural'image(i) & ") hold time too small"
        severity error;
    end process;

    -- Check 'dm' setup time referenced to 'dqs' falling
    check_dm_setup_f : process(dqs(i))
    begin
      if falling_edge(dqs(i)) and timing_enabled and dm_setup_f then
        dm_hold_f(i) <= transport true, false after part.speed_grade.t_dh_min;

        assert dm(i)'stable(part.speed_grade.t_ds_min)
          report "DDR3 SDRAM timing check (48): dm(" & natural'image(i) & ") to dqs(" & natural'image(i) & ") falling setup time too small"
          severity error;

        assert not(IS_X(dm))
          report "DDR3 SDRAM timing check (49): dm(" & natural'image(i) & ") invalid at dqs(" & natural'image(i) & ") falling edge"
          severity error;
      end if;
    end process;
  end generate;

  check_dq : for i in 0 to DATA_BYTES-1 generate
    dq_setup_r(i) <= dm_setup_r and dm(i) = '0';
    dq_setup_f(i) <= dm_setup_f and dm(i) = '0';

    -- Check 'dq' hold time referenced to 'dqs' rising
    check_dq_hold_r : process(dq((i + 1) * BYTE_WIDTH - 1 downto i * BYTE_WIDTH))
      variable u, l : natural;
    begin
      l := i * BYTE_WIDTH;
      u := l + BYTE_WIDTH - 1;
      assert not timing_enabled or not dq_hold_r(i)
        report "DDR3 SDRAM timing check (50): dqs(" & natural'image(i) & ") rising to dq(" & natural'image(u) & ":" & natural'image(l) & ") hold time too small"
        severity error;
    end process;

    -- Check 'dq' setup time referenced to 'dqs' rising
    check_dq_setup_r : process(dqs(i))
      variable u, l : natural;
    begin
      if rising_edge(dqs(i)) and timing_enabled and dq_setup_r(i) then
        dq_hold_r(i) <= transport true, false after part.speed_grade.t_dh_min;

        l := i * BYTE_WIDTH;
        u := l + BYTE_WIDTH - 1;

        assert dq((i + 1) * BYTE_WIDTH - 1 downto i * BYTE_WIDTH)'stable(part.speed_grade.t_ds_min)
          report "DDR3 SDRAM timing check (51): dq(" & natural'image(u) & ":" & natural'image(l) & ") to dqs(" & natural'image(i) & ") rising setup time too small"
          severity error;

        assert not(IS_X(dq(u downto l)))
          report "DDR3 SDRAM timing check (52): dq(" & natural'image(u) & ":" & natural'image(l) & ") invalid at dqs(" & natural'image(i) & ") rising edge"
          severity error;
      end if;
    end process;

    -- Check 'dq' hold time referenced to 'dqs' falling
    check_dq_hold_f : process(dq((i + 1) * BYTE_WIDTH - 1 downto i * BYTE_WIDTH))
      variable u, l : natural;
    begin
      l := i * BYTE_WIDTH;
      u := l + BYTE_WIDTH - 1;
      assert not timing_enabled or not dq_hold_f(i)
        report "DDR3 SDRAM timing check (53): dqs(" & natural'image(i) & ") falling to dq(" & natural'image(u) & ":" & natural'image(l) & ") hold time too small"
        severity error;
    end process;

    -- Check 'dq' setup time referenced to 'dqs' falling
    check_dq_setup_f : process(dqs(i))
      variable u, l : natural;
    begin
      if falling_edge(dqs(i)) and timing_enabled and dq_setup_f(i) then
        dq_hold_f(i) <= transport true, false after part.speed_grade.t_dh_min;

        l := i * BYTE_WIDTH;
        u := l + BYTE_WIDTH - 1;

        assert dq((i + 1) * BYTE_WIDTH - 1 downto i * BYTE_WIDTH)'stable(part.speed_grade.t_ds_min)
          report "DDR3 SDRAM timing check (54): dq(" & natural'image(u) & ":" & natural'image(l) & ") to dqs(" & natural'image(i) & ") falling setup time too small"
          severity error;

        assert not(IS_X(dq(u downto l)))
          report "DDR3 SDRAM timing check (55): dq(" & natural'image(u) & ":" & natural'image(l) & ") invalid at dqs(" & natural'image(i) & ") falling edge"
          severity error;
      end if;
    end process;
  end generate;


  --------------------------------------------------------------
  -- dqs pulse width checking                                 --
  --------------------------------------------------------------

  -- Generate t_dqsl_min and t_dqsh_min values from parameters
  -- time <= time * real
  t_dqsl_min <= ck_period_sig * part.speed_grade.t_dqsl_min;
  t_dqsh_min <= ck_period_sig * part.speed_grade.t_dqsh_min;
  t_dqss     <= ck_period_sig * part.speed_grade.t_dqss;
  t_dss_min  <= ck_period_sig * part.speed_grade.t_dss_min;
  t_dsh_min  <= ck_period_sig * part.speed_grade.t_dsh_min;

  -- Check for DQS edges not inside expected window

  ck_delayed <= transport ck after t_dqss;

  check_dqs_missing : process(ck_delayed)
  begin
    if rising_edge(ck_delayed) and write_rising and timing_enabled then
      for i in 0 to (DATA_BYTES - 1) loop
        assert (now - dqs_rising(i)) < (2 * t_dqss)
          report "DDR3 SDRAM timing check (56): no dqs(" & natural'image(i) & ") rising edge in T_DQSS window"
          severity error;
      end loop;
    end if;

    if falling_edge(ck_delayed) and write_falling and timing_enabled then
      for i in 0 to (DATA_BYTES - 1) loop
        assert (now - dqs_falling(i)) < (2 * t_dqss)
          report "DDR3 SDRAM timing check (57): no dqs(" & natural'image(i) & ") falling edge in T_DQSS window"
          severity error;
      end loop;
    end if;
  end process;

  -- Check for DQS# edges not inside expected window

  ck_l_delayed <= transport ck_l after t_dqss;

  check_dqs_l_ck_l_skew : process(ck_l_delayed)
  begin
    if rising_edge(ck_l_delayed) and write_falling and timing_enabled then
      for i in 0 to (DATA_BYTES - 1) loop
        assert (now - dqs_l_falling(i)) < (2 * t_dqss)
          report "DDR3 SDRAM timing check (58): no dqs_l(" & natural'image(i) & ") rising edge in T_DQSS window"
          severity error;
      end loop;
    end if;

    if falling_edge(ck_l_delayed) and write_rising and timing_enabled then
      for i in 0 to (DATA_BYTES - 1) loop
        assert not((now - dqs_l_rising(i)) > (2 * t_dqss))
          report "DDR3 SDRAM timing check (59): no dqs_l(" & natural'image(i) & ") falling edge in T_DQSS window"
          severity error;
      end loop;
    end if;
  end process;

  delay_write_rising : process(ck)
  begin
    if falling_edge(ck) then
      write_rising_d1 <= write_falling;
    end if;
  end process;

  check_dqs_setup_hold : for i in 0 to DATA_BYTES-1 generate
    -- Check 'dqs' low hold time referenced to 'ck' rising
    check_dqs_hold_r : process(dqs(i))
    begin
      assert not(timing_enabled and dqs_hold_r(i) and falling_edge(dqs(i)))
        report "DDR3 SDRAM timing check (60): dqs(" & natural'image(i) & ") falling to ck rising hold time too small"
        severity error;

      assert not(timing_enabled and dqs_l_hold_r(i) and rising_edge(dqs_l(i)))
        report "DDR3 SDRAM timing check (61): dqs_l(" & natural'image(i) & ") rising to ck rising hold time too small"
        severity error;
    end process;

    -- Check 'dqs' falling edge and 'dqs_l' rising edge setup time referenced to 'ck' rising
    check_dqs_setup_r : process(ck)
    begin
      if rising_edge(ck) and timing_enabled then
        if write_rising then
          dqs_hold_r(i)   <= transport true, false after t_dsh_min;
          dqs_l_hold_r(i) <= transport true, false after t_dsh_min;
        end if;

        if write_rising_d1 then
          assert now - dqs_falling(i) >= t_dss_min
            report "DDR3 SDRAM timing check (62): dqs(" & natural'image(i) & ") falling to ck rising setup time too small"
            severity error;

          assert now - dqs_l_rising(i) >= t_dss_min
            report "DDR3 SDRAM timing check (63): dqs_l(" & natural'image(i) & ") rising to ck rising setup time too small"
            severity error;
        end if;

        if write_rising then
          assert not(dqs(i) = 'X')
            report "DDR3 SDRAM timing check (64): dqs(" & natural'image(i) & ") invalid at ck rising edge"
            severity error;

          assert not(dqs_l(i) = 'X')
            report "DDR3 SDRAM timing check (65): dqs_l(" & natural'image(i) & ") invalid at ck rising edge"
            severity error;
        end if;
      end if;

      if falling_edge(ck) and timing_enabled then
        if write_falling then
          assert not(dqs(i) = 'X')
            report "DDR3 SDRAM timing check (66): dqs(" & natural'image(i) & ") invalid at ck falling edge"
            severity error;

          assert not(dqs_l(i) = 'X')
            report "DDR3 SDRAM timing check (67): dqs_l(" & natural'image(i) & ") invalid at ck falling edge"
            severity error;
        end if;
      end if;
    end process;
  end generate;

  -- Check DQS high and low pulse width
  generate_dqs : for i in 0 to (DATA_BYTES - 1) generate
    check_dqs : process(dqs(i))
    begin
      if (dqs(i)'event and dqs(i) = '1' and write_rising) then   -- dqs rising edge
        dqs_rising(i) := now;
        dqs_low(i)    := dqs_rising(i) - dqs_falling(i);

        if (timing_enabled) then
          assert not(dqs_low(i) < t_dqsl_min)
            report "DDR3 SDRAM timing check (68): dqs low pulse T_DQSL too small"
            severity error;
        end if;
      end if;

      if (dqs(i)'event and dqs(i) = '0' and write_falling) then   -- dqs falling edge
        dqs_falling(i) := now;
        dqs_high(i)    := dqs_falling(i) - dqs_rising(i);

        if (timing_enabled) then
          assert not(dqs_high(i) < t_dqsh_min)
            report "DDR3 SDRAM timing check (69): dqs high pulse T_DQSH too small"
            severity error;
        end if;
      end if;
    end process;
  end generate generate_dqs;


  -- Check DQS# high and low pulse width
  generate_dqs_l : for i in 0 to (DATA_BYTES - 1) generate
    check_dqs_l : process(dqs_l(i))
    begin
      if (dqs_l(i)'event and dqs_l(i) = '0' and write_rising) then   -- dqs_l falling edge = dqs rising edge
        dqs_l_rising(i) := now;
        dqs_l_low(i)    := dqs_l_rising(i) - dqs_l_falling(i);

        if (timing_enabled) then
          assert not(dqs_l_low(i) < t_dqsl_min)
            report "DDR3 SDRAM timing check (70): dqs_l low pulse T_DQSL too small"
            severity error;
        end if;
      end if;

      if (dqs_l(i)'event and dqs_l(i) = '1' and write_falling) then   -- dqs_l rising edge = dqs falling edge
        dqs_l_falling(i) := now;
        dqs_l_high(i)    := dqs_l_falling(i) - dqs_l_rising(i);

        if (timing_enabled) then
          assert not(dqs_l_high(i) < t_dqsh_min)
            report "DDR3 SDRAM timing check (71): dqs_l high pulse T_DQSH too small"
            severity error;
        end if;
      end if;
    end process;

  end generate generate_dqs_l;

  -- Check that ck is valid
  check_ck_valid : process(ck)
  begin
    assert not (timing_enabled and ck = 'X')
      report "DDR3 SDRAM timing check (72): ck invalid"
      severity error;
  end process;

  -- Check that ck_l is valid
  check_ck_l_valid : process(ck_l)
  begin
    assert not (timing_enabled and ck_l = 'X')
      report "DDR3 SDRAM timing check (73): ck_l invalid"
      severity error;
  end process;

end behavioural;
