--==============================================================================
--
--     File Name: test_simple_bit.vhd
--       Version: 0.1
--          Date: 23 Aug 2011
--         Model: Behavioral (NOT SYNTHESISABLE)
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: simple_l
--
--   Description: Illustrates with ADB3:
--                  Register access using direct slave interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Aug 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_tb_pkg.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_tb_inc_pkg.all;
use work.adb3_target_tb_pkg.all;
use work.adb3_probe_pkg.all;

entity test_simple is
end test_simple;

architecture sim of test_simple is

  -- Output clock expected freq (MHz) (unused)
  constant CLKS_OUT_FREQ : clks_out_exp_t := (others => 0);

  -- OCP transaction probe
  constant PROBE_LOG          : boolean := false; -- No logging
  constant DS_ADDR_ALIGN_BITS : natural := 4;     -- 16-byte aligned (128-bit interface)

  -- Test results
  signal simple_complete : boolean;
  signal simple_passed   : boolean;

  signal test_complete   : boolean := false;
  signal test_passed     : boolean := true;

  -- Testbench clocks
  signal ocp_clk         : std_logic;

  -- Probe errors
  signal status_probe_ds : adb3_ocp_probe_status_r := ADB3_OCP_PROBE_STATUS_OK;
  signal probe_ds_err : std_logic;

  --------------------
  -- Target signals --
  --------------------

  -- Clocks
  signal clks_in     : clks_in_t;
  signal clks_mgt_in : clks_mgt_in_t;
  signal clks_out    : clks_out_t;
  signal ref_clk     : std_logic_dbl_t;

  -- Configuration/online signals
  signal interrupt           : std_logic;
  signal pcie_online_cnt     : std_logic_vector(7 downto 0) := (others => '1');
  signal pcie_online_long    : std_logic := '0';
  signal pcie_online_long_d1 : std_logic := '0';

  -- PCIe signals
  signal pcie_rst_l : std_logic := '0';
  signal pcie_clk   : std_logic_dbl_t;
  signal pcie_h2t   : pcie_h2t_t;
  signal pcie_t2h   : pcie_t2h_t;

  -- OCP signals
  signal direct_slave_m2s   : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal direct_slave_s2m   : adb3_ocp_s2mT;
  signal dma_channels_m2s   : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_M2S);
  signal dma_channels_s2m   : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  signal unused_s2m         : adb3_ocp_s2mT_array(0 to DM_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_S2M);

  -- ADV8 specific board interface
  signal model_in    : adb3_model_in_t := (others => '0');
  signal model_inout : adb3_model_inout_t;
  signal model_out   : adb3_model_out_t;

  component simple
  port(
    -- Non-MGT Clock Interface
    ref_clk     : in    std_logic_dbl_t;
    -- PCIe Interface
    pcie_rst_l  : in    std_logic;
    pcie_clk    : in    std_logic_dbl_t;
    pcie_h2t    : in    pcie_h2t_t;
    pcie_t2h    : out   pcie_t2h_t;
    -- Board interface
    model_in    : in    adb3_model_in_t;
    model_inout : inout adb3_model_inout_t;
    model_out   : out   adb3_model_out_t);
  end component;

begin

  -- Report design constants
  report_setup_p : process
  begin
    report "Board Type : " & BOARD_TYPE;                      -- adb3_target_tb_inc_pkg
    report "Target Use : " & target_use_t'image(TARGET_USE);  -- adb3_target_inc_pkg
    report "Waiting for PCIe online....";
    wait;
  end process;


  -------------------------------
  -- OCP direct slave stimulus --
  -------------------------------
  direct_slave_p : process
    variable reg_data_wr    : byte_vector_t(0 to 127);
    variable reg_data_rd    : byte_vector_t(0 to 127);
    variable reg_be         : byte_enable_t(0 to 127);
    variable reg_addr       : adb3_ocp_addr_s;
    variable reg_xfer       : natural;

  begin
    wait until (pcie_online_long = '1');  -- OCP commands ignored until PCIe online
    simple_complete <= false;
    simple_passed <= true;

    --------------------------
    -- Simple Register Test --
    --------------------------
    reg_be(0 to 3) := conv_byte_enable("1111");

    -- Write a pattern to register
    reg_addr := (others => '0');
    reg_data_wr(0 to 3) := conv_byte_vector(X"cafeface");

    adb3_ocp_sim_write_reg32(reg_addr,10,reg_be(0 to 3),reg_data_wr(0 to 3),reg_xfer,ocp_clk,direct_slave_m2s,direct_slave_s2m);
    report "Wrote simple DATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(reg_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));


    -- Read back nibble reversed pattern from register
    reg_addr := (others => '0');
    adb3_ocp_sim_read_reg32(reg_addr,11,reg_data_rd(0 to 3),reg_xfer,ocp_clk,direct_slave_m2s,direct_slave_s2m);
    report "Read simple DATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    simple_complete <= true;
    if (reg_data_rd(0 to 3) = conv_byte_vector(X"ecafefac")) then
      assert false report "Test Simple completed: PASSED." severity note;
    else
      assert false report "Test Simple completed: FAILED." severity error;
      simple_passed <= false;
    end if;

    wait;
  end process;


  -------------------------
  -- Output test results --
  -------------------------

  gen_results_p : process(simple_passed,simple_complete)
  begin
    test_passed   <= simple_passed;
    test_complete <= simple_complete;
  end process;

  test_results_p : process(test_complete,test_passed,probe_ds_err)
  begin
    if (test_complete) then
      assert not(probe_ds_err  = '1') report "Probe OCP Direct Slave status: ERROR."  severity error;

      if (test_passed) then
        assert false report "Test of design SIMPLE completed: PASSED." severity failure;
      else
        assert false report "Test of design SIMPLE completed: FAILED." severity failure;
      end if;
    end if;
  end process;


  ----------------------------
  -- Board Clock Generation --
  ----------------------------

  test_board_clks_i : test_board_clks
  generic map(
    clks_out_freq => CLKS_OUT_FREQ) -- clks_out_exp_t
  port map(
    -- Target clock interface
    target_clks_in     => open,      -- out   clks_in_t
    target_clks_mgt_in => open,      -- out   clks_mgt_in_t
    target_clks_out    => clks_out,  -- in    clks_out_t
    target_ref_clk     => ref_clk,   -- out   std_logic_dbl_t
    target_mptl_clk    => open,      -- out   std_logic_dbl_t
    target_pcie_clk    => pcie_clk,  -- out   std_logic_dbl_t
    -- Bridge clock interface
    bridge_lclk_freq   => 0.0,       -- in    real
    bridge_ocp_clk     => open,      -- out   std_logic
    bridge_mptl_clk    => open,      -- out   std_logic_dbl_t
    -- Test interface
    clks_out_pass      => open);     -- out   boolean


  ------------------------------------
  -- Target Component Instantiation --
  ------------------------------------

  pcie_rst_l <= '1';

  uut : simple
  port map(
    -- Non-MGT Clock Interface
    ref_clk     => ref_clk,      -- in    std_logic_dbl_t
    -- MGT Clock Interface
    pcie_clk    => pcie_clk,     -- in    std_logic_dbl_t
    -- PCIe Interface
    pcie_rst_l  => pcie_rst_l,   -- in    std_logic
    pcie_h2t    => pcie_h2t,     -- in    pcie_h2t_t
    pcie_t2h    => pcie_t2h,     -- out   pcie_t2h_t
    -- Board interface
    model_in    => model_in,     -- in    adb3_model_in_t
    model_inout => model_inout,  -- inout adb3_model_inout_t
    model_out   => model_out);   -- out   adb3_model_out_t)


  -----------------------------------
  -- Host Component Instantiations --
  -----------------------------------

  pcie_if_host_wrap_i : pcie_if_host_wrap
  port map(
    -- OCP Interface
    direct_slave_m2s   => direct_slave_m2s,    -- in    adb3_ocp_m2sT
    direct_slave_s2m   => direct_slave_s2m,    -- out   adb3_ocp_s2mT
    dma_channels_m2s   => dma_channels_m2s,    -- in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m   => dma_channels_s2m,    -- out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    direct_masters_m2s => open,                -- out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1)
    direct_masters_s2m => unused_s2m,          -- in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1)
    interrupt          => interrupt,           -- out   std_logic
    ocp_clk_out        => ocp_clk,             -- out   std_logic
    -- PCIe Interface
    pcie_t2h           => pcie_t2h,            -- in    pcie_t2h_t
    pcie_h2t           => pcie_h2t);           -- out   pcie_h2t_t)

  -- Count cycles of ocp_clk to make sure it is stable
  online_count_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (or_reduce(pcie_online_cnt) = '1') then
        pcie_online_cnt <= pcie_online_cnt - 1;
      end if;
      pcie_online_long    <= not(or_reduce(pcie_online_cnt));
      pcie_online_long_d1 <= pcie_online_long;
    end if;
  end process;


  --------------------------
  -- Testbench OCP probes --
  --------------------------

  -- Direct Slave transaction probe
  ds_probe_i : adb3_ocp_transaction_probe
  generic map(
    enable_logging   => PROBE_LOG,
    int_log_filename => "ds_probe",
    addr_align_bits  => DS_ADDR_ALIGN_BITS,
    addr_width_max   => DS_ADDR_WIDTH)
  port map(
    ocp_clk => ocp_clk,
    ocp_m2s => direct_slave_m2s,
    ocp_s2m => direct_slave_s2m,
    status  => status_probe_ds);

  probe_ds_err <= '0' when status_probe_ds = ADB3_OCP_PROBE_STATUS_OK else '1';

end;