--==============================================================================
--
--     File Name: blk_ds_io_test_6t1.vhd
--       Version: 0.4
--          Date: 23 Aug 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Front/Rear IO test
--
--   Description: Front (XRM 146)/Rear (PN4/PN6) GPIO test
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 20 Apr 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg.
--                  - Added IOB attributes to gpio signals.
--                  - Removed some unnecessary resets.
--                0.3, KDC, 16 Feb 2011:
--                  - Changed range of addr_width to match OCP profile
--                0.4, KDC, 23 Aug 2011:
--                  - Made specific to ADM-XRC-6T1
--                  - Made register read a 2-stage process
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

entity blk_ds_io_test is
generic (
  addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
port(
  -- OCP interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- IO interface
  gpio_inout       : inout gpio_inout_t);
end;


architecture syn_6t1 of blk_ds_io_test is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- First stage register read signals
  signal xrm_ds_q : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_ds_q : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn6_ds_q : std_logic_vector(REG_WIDTH-1 downto 0);

  -- XRM GPIO registers
  signal xrm_gpio_da_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) := (others => '1');
  signal xrm_gpio_db_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) := (others => '1');
  signal xrm_gpio_dc_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) := (others => '1');
  signal xrm_gpio_dd_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) := (others => '1');
  signal xrm_gpio_cs_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0) := (others => '1');
  signal xrm_gpio_da_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_db_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_dc_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_dd_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_cs_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_da_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_db_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_dc_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_dd_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal xrm_gpio_cs_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);

  -- PN4 GPIO registers
  signal pn4_gpio_p_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_p_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_p_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_n_tri_reg   : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_n_wdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pn4_gpio_n_rdata_reg : std_logic_vector(REG_WIDTH-1 downto 0);

  -- PN6 GPIO registers
  signal pn6_gpio_tri_reg      : std_logic_vector((2*REG_WIDTH)-1 downto 0);
  signal pn6_gpio_wdata_reg    : std_logic_vector((2*REG_WIDTH)-1 downto 0);
  signal pn6_gpio_rdata_reg    : std_logic_vector((2*REG_WIDTH)-1 downto 0);

  -- IOB tristate flip-flop signals
  signal xrm_gpio_da_tri_p : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_da_tri_n : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_db_tri_p : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_db_tri_n : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_dc_tri_p : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_dc_tri_n : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_dd_tri_p : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_dd_tri_n : std_logic_vector(15 downto 0) := (others => '1');
  signal xrm_gpio_sa_tri : std_logic_vector(1 downto 0) := (others => '1');
  signal xrm_gpio_sb_tri : std_logic_vector(1 downto 0) := (others => '1');
  signal xrm_gpio_sc_tri : std_logic_vector(1 downto 0) := (others => '1');
  signal xrm_gpio_sd_tri : std_logic_vector(3 downto 0) := (others => '1');
  signal xrm_gpio_da_cc_tri_p : std_logic := '1';
  signal xrm_gpio_da_cc_tri_n : std_logic := '1';
  signal xrm_gpio_db_cc_tri_p : std_logic := '1';
  signal xrm_gpio_db_cc_tri_n : std_logic := '1';
  signal xrm_gpio_dc_cc_tri_p : std_logic := '1';
  signal xrm_gpio_dc_cc_tri_n : std_logic := '1';
  signal xrm_gpio_dd_cc_tri_p : std_logic := '1';
  signal xrm_gpio_dd_cc_tri_n : std_logic := '1';
  signal pn4_gpio_tri_n : std_logic_vector(PN4_GPIO_WIDTH downto 1) := (others => '1');
  signal pn4_gpio_tri_p : std_logic_vector(PN4_GPIO_WIDTH downto 1) := (others => '1');
  signal pn6_gpio_tri : std_logic_vector(PN6_GPIO_WIDTH downto 1) := (others => '1');

  -- IOB output flip-flop signals
  signal xrm_gpio_da_out_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_da_out_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_db_out_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_db_out_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_dc_out_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_dc_out_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_dd_out_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_dd_out_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_sa_out : std_logic_vector(1 downto 0);
  signal xrm_gpio_sb_out : std_logic_vector(1 downto 0);
  signal xrm_gpio_sc_out : std_logic_vector(1 downto 0);
  signal xrm_gpio_sd_out : std_logic_vector(3 downto 0);
  signal xrm_gpio_da_cc_out_p : std_logic;
  signal xrm_gpio_da_cc_out_n : std_logic;
  signal xrm_gpio_db_cc_out_p : std_logic;
  signal xrm_gpio_db_cc_out_n : std_logic;
  signal xrm_gpio_dc_cc_out_p : std_logic;
  signal xrm_gpio_dc_cc_out_n : std_logic;
  signal xrm_gpio_dd_cc_out_p : std_logic;
  signal xrm_gpio_dd_cc_out_n : std_logic;
  signal pn4_gpio_out_n : std_logic_vector(PN4_GPIO_WIDTH downto 1);
  signal pn4_gpio_out_p : std_logic_vector(PN4_GPIO_WIDTH downto 1);
  signal pn6_gpio_out : std_logic_vector(PN6_GPIO_WIDTH downto 1);

  -- IOB input flip-flop signals
  signal xrm_gpio_da_in_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_da_in_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_db_in_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_db_in_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_dc_in_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_dc_in_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_dd_in_p : std_logic_vector(15 downto 0);
  signal xrm_gpio_dd_in_n : std_logic_vector(15 downto 0);
  signal xrm_gpio_sa_in : std_logic_vector(1 downto 0);
  signal xrm_gpio_sb_in : std_logic_vector(1 downto 0);
  signal xrm_gpio_sc_in : std_logic_vector(1 downto 0);
  signal xrm_gpio_sd_in : std_logic_vector(3 downto 0);
  signal xrm_gpio_da_cc_in_p : std_logic;
  signal xrm_gpio_da_cc_in_n : std_logic;
  signal xrm_gpio_db_cc_in_p : std_logic;
  signal xrm_gpio_db_cc_in_n : std_logic;
  signal xrm_gpio_dc_cc_in_p : std_logic;
  signal xrm_gpio_dc_cc_in_n : std_logic;
  signal xrm_gpio_dd_cc_in_p : std_logic;
  signal xrm_gpio_dd_cc_in_n : std_logic;
  signal pn4_gpio_in_n : std_logic_vector(PN4_GPIO_WIDTH downto 1);
  signal pn4_gpio_in_p : std_logic_vector(PN4_GPIO_WIDTH downto 1);
  signal pn6_gpio_in : std_logic_vector(PN6_GPIO_WIDTH downto 1);

  -- IOB attributes for tristate FFs
  attribute IOB : string;
  attribute IOB of xrm_gpio_da_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_da_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_db_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_db_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_sa_tri : signal is "TRUE";
  attribute IOB of xrm_gpio_sb_tri : signal is "TRUE";
  attribute IOB of xrm_gpio_sc_tri : signal is "TRUE";
  attribute IOB of xrm_gpio_sd_tri : signal is "TRUE";
  attribute IOB of xrm_gpio_da_cc_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_da_cc_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_db_cc_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_db_cc_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_cc_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_cc_tri_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_cc_tri_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_cc_tri_n : signal is "TRUE";
  attribute IOB of pn4_gpio_tri_n : signal is "TRUE";
  attribute IOB of pn4_gpio_tri_p : signal is "TRUE";
  attribute IOB of pn6_gpio_tri : signal is "TRUE";

  -- IOB attributes for output FFs
  attribute IOB of xrm_gpio_da_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_da_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_db_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_db_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_sa_out : signal is "TRUE";
  attribute IOB of xrm_gpio_sb_out : signal is "TRUE";
  attribute IOB of xrm_gpio_sc_out : signal is "TRUE";
  attribute IOB of xrm_gpio_sd_out : signal is "TRUE";
  attribute IOB of xrm_gpio_da_cc_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_da_cc_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_db_cc_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_db_cc_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_cc_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_cc_out_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_cc_out_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_cc_out_n : signal is "TRUE";
  attribute IOB of pn4_gpio_out_n : signal is "TRUE";
  attribute IOB of pn4_gpio_out_p : signal is "TRUE";
  attribute IOB of pn6_gpio_out : signal is "TRUE";

  -- IOB attributes for input FFs
  attribute IOB of xrm_gpio_da_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_da_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_db_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_db_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_sa_in : signal is "TRUE";
  attribute IOB of xrm_gpio_sb_in : signal is "TRUE";
  attribute IOB of xrm_gpio_sc_in : signal is "TRUE";
  attribute IOB of xrm_gpio_sd_in : signal is "TRUE";
  attribute IOB of xrm_gpio_da_cc_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_da_cc_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_db_cc_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_db_cc_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_cc_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dc_cc_in_n : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_cc_in_p : signal is "TRUE";
  attribute IOB of xrm_gpio_dd_cc_in_n : signal is "TRUE";
  attribute IOB of pn4_gpio_in_n : signal is "TRUE";
  attribute IOB of pn4_gpio_in_p : signal is "TRUE";
  attribute IOB of pn6_gpio_in : signal is "TRUE";

begin

  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => addr_width, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_WIDTH,  -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 2)          -- natural := 1
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  -- Initialise tri_reg so that output enables are off
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      xrm_gpio_da_tri_reg <= (others => '1');
      xrm_gpio_db_tri_reg <= (others => '1');
      xrm_gpio_dc_tri_reg <= (others => '1');
      xrm_gpio_dd_tri_reg <= (others => '1');
      xrm_gpio_cs_tri_reg <= (others => '1');
      pn4_gpio_p_tri_reg  <= (others => '1');
      pn4_gpio_n_tri_reg  <= (others => '1');
      pn6_gpio_tri_reg <= (others => '1');
      xrm_gpio_da_wdata_reg <= (others => '0');
      xrm_gpio_db_wdata_reg <= (others => '0');
      xrm_gpio_dc_wdata_reg <= (others => '0');
      xrm_gpio_dd_wdata_reg <= (others => '0');
      xrm_gpio_cs_wdata_reg <= (others => '0');
      pn4_gpio_p_wdata_reg  <= (others => '0');
      pn4_gpio_n_wdata_reg  <= (others => '0');
      pn6_gpio_wdata_reg <= (others => '0');
    elsif rising_edge(ocp_clk) then
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          -- XRM write registers
          if (ds_a = XRM_GPIO_DA_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_da_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DA_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_da_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DB_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_db_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DB_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_db_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DC_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dc_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DC_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dc_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DD_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dd_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_DD_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_dd_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_CS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_cs_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = XRM_GPIO_CS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            xrm_gpio_cs_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          -- PN4 write registers
          if (ds_a = PN4_GPIO_P_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_p_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN4_GPIO_N_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_n_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN4_GPIO_P_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_p_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN4_GPIO_N_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            pn4_gpio_n_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
          -- PN6 write registers
          if (ds_a = PN6_GPIO_MS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_tri_reg(32+u downto 32+l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN6_GPIO_MS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_wdata_reg(32+u downto 32+l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN6_GPIO_LS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
        end if;
      end loop;
    end if;
  end process;


  -- Perform multiplex of read registers in a 2 stage process
  -- First, multiplex each group of registers, then multiplex the groups.
  -- This requires the adb3_ocp_simple_bus_if generic read_latency => 2

  -- Multiplex XRM read registers
  xrm_rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      xrm_gpio_da_rdata_reg <= xrm_gpio_da_in_p & xrm_gpio_da_in_n;
      xrm_gpio_db_rdata_reg <= xrm_gpio_db_in_p & xrm_gpio_db_in_n;
      xrm_gpio_dc_rdata_reg <= xrm_gpio_dc_in_p & xrm_gpio_dc_in_n;
      xrm_gpio_dd_rdata_reg <= xrm_gpio_dd_in_p & xrm_gpio_dd_in_n;

      xrm_gpio_cs_rdata_reg(17 downto 0) <= xrm_gpio_dd_cc_in_p  & xrm_gpio_dd_cc_in_n &
                                            xrm_gpio_dc_cc_in_p  & xrm_gpio_dc_cc_in_n &
                                            xrm_gpio_db_cc_in_p  & xrm_gpio_db_cc_in_n &
                                            xrm_gpio_da_cc_in_p  & xrm_gpio_da_cc_in_n &
                                            xrm_gpio_sd_in & xrm_gpio_sc_in & xrm_gpio_sb_in & xrm_gpio_sa_in;

      ds_q_var := (others => '-');

      -- XRM DA read registers
      if (ds_a = XRM_GPIO_DA_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_da_tri_reg;
      end if;
      if (ds_a = XRM_GPIO_DA_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_da_wdata_reg;
      end if;
      if (ds_a = XRM_GPIO_DA_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_da_rdata_reg;
      end if;

      -- XRM DB read registers
      if (ds_a = XRM_GPIO_DB_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_db_tri_reg;
      end if;
      if (ds_a = XRM_GPIO_DB_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_db_wdata_reg;
      end if;
      if (ds_a = XRM_GPIO_DB_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_db_rdata_reg;
      end if;

      -- XRM DC read registers
      if (ds_a = XRM_GPIO_DC_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_dc_tri_reg;
      end if;
      if (ds_a = XRM_GPIO_DC_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_dc_wdata_reg;
      end if;
      if (ds_a = XRM_GPIO_DC_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_dc_rdata_reg;
      end if;

      -- XRM DD read registers
      if (ds_a = XRM_GPIO_DD_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_dd_tri_reg;
      end if;
      if (ds_a = XRM_GPIO_DD_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_dd_wdata_reg;
      end if;
      if (ds_a = XRM_GPIO_DD_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_dd_rdata_reg;
      end if;

      -- XRM CS read registers
      if (ds_a = XRM_GPIO_CS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_cs_tri_reg;
      end if;
      if (ds_a = XRM_GPIO_CS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_gpio_cs_wdata_reg;
      end if;
      if (ds_a = XRM_GPIO_CS_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(17 downto 0) := xrm_gpio_cs_rdata_reg(17 downto 0);
      end if;

      xrm_ds_q <= ds_q_var;
    end if;
  end process;


  -- Multiplex PN4 read registers
  pn4_rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      pn4_gpio_n_rdata_reg <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn4_gpio_in_n), REG_WIDTH));
      pn4_gpio_p_rdata_reg <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn4_gpio_in_p), REG_WIDTH));

      ds_q_var := (others => '-');

      -- PN4 P read registers
      if (ds_a = PN4_GPIO_P_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(PN4_GPIO_WIDTH-1 downto 0) := pn4_gpio_p_tri_reg(PN4_GPIO_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN4_GPIO_P_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(PN4_GPIO_WIDTH-1 downto 0) := pn4_gpio_p_wdata_reg(PN4_GPIO_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN4_GPIO_P_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(PN4_GPIO_WIDTH-1 downto 0) := pn4_gpio_p_rdata_reg(PN4_GPIO_WIDTH-1 downto 0);
      end if;

      -- PN4 N read registers
      if (ds_a = PN4_GPIO_N_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(PN4_GPIO_WIDTH-1 downto 0) := pn4_gpio_n_tri_reg(PN4_GPIO_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN4_GPIO_N_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(PN4_GPIO_WIDTH-1 downto 0) := pn4_gpio_n_wdata_reg(PN4_GPIO_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN4_GPIO_N_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var(PN4_GPIO_WIDTH-1 downto 0) := pn4_gpio_n_rdata_reg(PN4_GPIO_WIDTH-1 downto 0);
      end if;

      pn4_ds_q <= ds_q_var;
    end if;
  end process;


  -- Multiplex PN6 read registers
  pn6_rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      pn6_gpio_rdata_reg   <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn6_gpio_in), 2*REG_WIDTH));

      ds_q_var := (others => '-');

      -- PN6 MS read registers
      if (ds_a = PN6_GPIO_MS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_tri_reg(2*REG_WIDTH-1 downto REG_WIDTH);
      end if;
      if (ds_a = PN6_GPIO_MS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_wdata_reg(2*REG_WIDTH-1 downto REG_WIDTH);
      end if;
      if (ds_a = PN6_GPIO_MS_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_rdata_reg(2*REG_WIDTH-1 downto REG_WIDTH);
      end if;

      -- PN6 LS read registers
      if (ds_a = PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_tri_reg(REG_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN6_GPIO_LS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_wdata_reg(REG_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN6_GPIO_LS_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_rdata_reg(REG_WIDTH-1 downto 0);
      end if;

      pn6_ds_q <= ds_q_var;
    end if;
  end process;


  -- Multiplex read register groups
  rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      ds_q_var := (others => '-');

      -- XRM read registers
      if (ds_a >= XRM_GPIO_DA_DATAO_REG_ADDR(addr_width-1 downto 0)) and
         (ds_a <= XRM_GPIO_CS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := xrm_ds_q;
      end if;

      -- PN4 read registers
      if (ds_a >= PN4_GPIO_P_DATAO_REG_ADDR(addr_width-1 downto 0)) and
         (ds_a <= PN4_GPIO_N_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn4_ds_q;
      end if;

      -- PN6 read registers
      if (ds_a >= PN6_GPIO_MS_DATAO_REG_ADDR(addr_width-1 downto 0)) and
         (ds_a <= PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_ds_q;
      end if;

      ds_q <= ds_q_var;
    end if;
  end process;


  -- Implement IOB input flip-flops
  gen_xrm_iob_iff_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      xrm_gpio_da_in_p    <= gpio_inout.xrm_gpio.xrm_da_p;
      xrm_gpio_da_in_n    <= gpio_inout.xrm_gpio.xrm_da_n;
      xrm_gpio_db_in_p    <= gpio_inout.xrm_gpio.xrm_db_p;
      xrm_gpio_db_in_n    <= gpio_inout.xrm_gpio.xrm_db_n;
      xrm_gpio_dc_in_p    <= gpio_inout.xrm_gpio.xrm_dc_p;
      xrm_gpio_dc_in_n    <= gpio_inout.xrm_gpio.xrm_dc_n;
      xrm_gpio_dd_in_p    <= gpio_inout.xrm_gpio.xrm_dd_p;
      xrm_gpio_dd_in_n    <= gpio_inout.xrm_gpio.xrm_dd_n;
      xrm_gpio_sa_in      <= gpio_inout.xrm_gpio.xrm_sa;
      xrm_gpio_sb_in      <= gpio_inout.xrm_gpio.xrm_sb;
      xrm_gpio_sc_in      <= gpio_inout.xrm_gpio.xrm_sc;
      xrm_gpio_sd_in      <= gpio_inout.xrm_gpio.xrm_sd;
      xrm_gpio_da_cc_in_p <= gpio_inout.xrm_gpio.xrm_da_cc_p;
      xrm_gpio_da_cc_in_n <= gpio_inout.xrm_gpio.xrm_da_cc_n;
      xrm_gpio_db_cc_in_p <= gpio_inout.xrm_gpio.xrm_db_cc_p;
      xrm_gpio_db_cc_in_n <= gpio_inout.xrm_gpio.xrm_db_cc_n;
      xrm_gpio_dc_cc_in_p <= gpio_inout.xrm_gpio.xrm_dc_cc_p;
      xrm_gpio_dc_cc_in_n <= gpio_inout.xrm_gpio.xrm_dc_cc_n;
      xrm_gpio_dd_cc_in_p <= gpio_inout.xrm_gpio.xrm_dd_cc_p;
      xrm_gpio_dd_cc_in_n <= gpio_inout.xrm_gpio.xrm_dd_cc_n;
    end if;
  end process;

  -- Implement IOB output flip-flops
  gen_xrm_iob_off_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      xrm_gpio_da_out_p    <= xrm_gpio_da_wdata_reg(31 downto 16);
      xrm_gpio_da_out_n    <= xrm_gpio_da_wdata_reg(15 downto 0);
      xrm_gpio_db_out_p    <= xrm_gpio_db_wdata_reg(31 downto 16);
      xrm_gpio_db_out_n    <= xrm_gpio_db_wdata_reg(15 downto 0);
      xrm_gpio_dc_out_p    <= xrm_gpio_dc_wdata_reg(31 downto 16);
      xrm_gpio_dc_out_n    <= xrm_gpio_dc_wdata_reg(15 downto 0);
      xrm_gpio_dd_out_p    <= xrm_gpio_dd_wdata_reg(31 downto 16);
      xrm_gpio_dd_out_n    <= xrm_gpio_dd_wdata_reg(15 downto 0);
      xrm_gpio_sa_out      <= xrm_gpio_cs_wdata_reg(1 downto 0);
      xrm_gpio_sb_out      <= xrm_gpio_cs_wdata_reg(3 downto 2);
      xrm_gpio_sc_out      <= xrm_gpio_cs_wdata_reg(5 downto 4);
      xrm_gpio_sd_out      <= xrm_gpio_cs_wdata_reg(9 downto 6);
      xrm_gpio_da_cc_out_n <= xrm_gpio_cs_wdata_reg(10);
      xrm_gpio_da_cc_out_p <= xrm_gpio_cs_wdata_reg(11);
      xrm_gpio_db_cc_out_n <= xrm_gpio_cs_wdata_reg(12);
      xrm_gpio_db_cc_out_p <= xrm_gpio_cs_wdata_reg(13);
      xrm_gpio_dc_cc_out_n <= xrm_gpio_cs_wdata_reg(14);
      xrm_gpio_dc_cc_out_p <= xrm_gpio_cs_wdata_reg(15);
      xrm_gpio_dd_cc_out_n <= xrm_gpio_cs_wdata_reg(16);
      xrm_gpio_dd_cc_out_p <= xrm_gpio_cs_wdata_reg(17);
    end if;
  end process;

  -- Implement IOB tristate flip-flops
  gen_xrm_iob_tff_p : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      xrm_gpio_da_tri_p    <= (others => '1');
      xrm_gpio_da_tri_n    <= (others => '1');
      xrm_gpio_db_tri_p    <= (others => '1');
      xrm_gpio_db_tri_n    <= (others => '1');
      xrm_gpio_dc_tri_p    <= (others => '1');
      xrm_gpio_dc_tri_n    <= (others => '1');
      xrm_gpio_dd_tri_p    <= (others => '1');
      xrm_gpio_dd_tri_n    <= (others => '1');
      xrm_gpio_sa_tri      <= (others => '1');
      xrm_gpio_sb_tri      <= (others => '1');
      xrm_gpio_sc_tri      <= (others => '1');
      xrm_gpio_sd_tri      <= (others => '1');
      xrm_gpio_da_cc_tri_n <= '1';
      xrm_gpio_da_cc_tri_p <= '1';
      xrm_gpio_db_cc_tri_n <= '1';
      xrm_gpio_db_cc_tri_p <= '1';
      xrm_gpio_dc_cc_tri_n <= '1';
      xrm_gpio_dc_cc_tri_p <= '1';
      xrm_gpio_dd_cc_tri_n <= '1';
      xrm_gpio_dd_cc_tri_p <= '1';
    elsif rising_edge(ocp_clk) then
      xrm_gpio_da_tri_p    <= xrm_gpio_da_tri_reg(31 downto 16);
      xrm_gpio_da_tri_n    <= xrm_gpio_da_tri_reg(15 downto 0);
      xrm_gpio_db_tri_p    <= xrm_gpio_db_tri_reg(31 downto 16);
      xrm_gpio_db_tri_n    <= xrm_gpio_db_tri_reg(15 downto 0);
      xrm_gpio_dc_tri_p    <= xrm_gpio_dc_tri_reg(31 downto 16);
      xrm_gpio_dc_tri_n    <= xrm_gpio_dc_tri_reg(15 downto 0);
      xrm_gpio_dd_tri_p    <= xrm_gpio_dd_tri_reg(31 downto 16);
      xrm_gpio_dd_tri_n    <= xrm_gpio_dd_tri_reg(15 downto 0);
      xrm_gpio_sa_tri      <= xrm_gpio_cs_tri_reg(1 downto 0);
      xrm_gpio_sb_tri      <= xrm_gpio_cs_tri_reg(3 downto 2);
      xrm_gpio_sc_tri      <= xrm_gpio_cs_tri_reg(5 downto 4);
      xrm_gpio_sd_tri      <= xrm_gpio_cs_tri_reg(9 downto 6);
      xrm_gpio_da_cc_tri_n <= xrm_gpio_cs_tri_reg(10);
      xrm_gpio_da_cc_tri_p <= xrm_gpio_cs_tri_reg(11);
      xrm_gpio_db_cc_tri_n <= xrm_gpio_cs_tri_reg(12);
      xrm_gpio_db_cc_tri_p <= xrm_gpio_cs_tri_reg(13);
      xrm_gpio_dc_cc_tri_n <= xrm_gpio_cs_tri_reg(14);
      xrm_gpio_dc_cc_tri_p <= xrm_gpio_cs_tri_reg(15);
      xrm_gpio_dd_cc_tri_n <= xrm_gpio_cs_tri_reg(16);
      xrm_gpio_dd_cc_tri_p <= xrm_gpio_cs_tri_reg(17);
    end if;
  end process;

  -- XRM-IO146 bi-directional GPIO
  -- Implement tri-states
  gen_xrm_d_g : for n in 0 to 15 generate
    gpio_inout.xrm_gpio.xrm_da_p(n) <= xrm_gpio_da_out_p(n) when (xrm_gpio_da_tri_p(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_da_n(n) <= xrm_gpio_da_out_n(n) when (xrm_gpio_da_tri_n(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_db_p(n) <= xrm_gpio_db_out_p(n) when (xrm_gpio_db_tri_p(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_db_n(n) <= xrm_gpio_db_out_n(n) when (xrm_gpio_db_tri_n(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_dc_p(n) <= xrm_gpio_dc_out_p(n) when (xrm_gpio_dc_tri_p(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_dc_n(n) <= xrm_gpio_dc_out_n(n) when (xrm_gpio_dc_tri_n(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_dd_p(n) <= xrm_gpio_dd_out_p(n) when (xrm_gpio_dd_tri_p(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_dd_n(n) <= xrm_gpio_dd_out_n(n) when (xrm_gpio_dd_tri_n(n) = '0') else 'Z';
  end generate;

  gen_xrm_s_g : for n in 0 to 1 generate
    gpio_inout.xrm_gpio.xrm_sa(n) <= xrm_gpio_sa_out(n) when (xrm_gpio_sa_tri(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_sb(n) <= xrm_gpio_sb_out(n) when (xrm_gpio_sb_tri(n) = '0') else 'Z';
    gpio_inout.xrm_gpio.xrm_sc(n) <= xrm_gpio_sc_out(n) when (xrm_gpio_sc_tri(n) = '0') else 'Z';
  end generate;

  gen_xrm_sd_g : for n in 0 to 3 generate
    gpio_inout.xrm_gpio.xrm_sd(n) <= xrm_gpio_sd_out(n) when (xrm_gpio_sd_tri(n) = '0') else 'Z';
  end generate;

  gpio_inout.xrm_gpio.xrm_da_cc_p <= xrm_gpio_da_cc_out_p when (xrm_gpio_da_cc_tri_p = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_da_cc_n <= xrm_gpio_da_cc_out_n when (xrm_gpio_da_cc_tri_n = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_db_cc_p <= xrm_gpio_db_cc_out_p when (xrm_gpio_db_cc_tri_p = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_db_cc_n <= xrm_gpio_db_cc_out_n when (xrm_gpio_db_cc_tri_n = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_dc_cc_p <= xrm_gpio_dc_cc_out_p when (xrm_gpio_dc_cc_tri_p = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_dc_cc_n <= xrm_gpio_dc_cc_out_n when (xrm_gpio_dc_cc_tri_n = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_dd_cc_p <= xrm_gpio_dd_cc_out_p when (xrm_gpio_dd_cc_tri_p = '0') else 'Z';
  gpio_inout.xrm_gpio.xrm_dd_cc_n <= xrm_gpio_dd_cc_out_n when (xrm_gpio_dd_cc_tri_n = '0') else 'Z';


  -- Implement IOB input flip-flops
  gen_pn4_iob_iff_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      pn4_gpio_in_n <= gpio_inout.pn4_gpio.gpio_n;
      pn4_gpio_in_p <= gpio_inout.pn4_gpio.gpio_p;
    end if;
  end process;

  -- Implement IOB output flip-flops
  gen_pn4_iob_off_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      pn4_gpio_out_n <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn4_gpio_n_wdata_reg), PN4_GPIO_WIDTH));
      pn4_gpio_out_p <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn4_gpio_p_wdata_reg), PN4_GPIO_WIDTH));
    end if;
  end process;

  -- Implement IOB tristate flip-flops
  gen_pn4_iob_tff_p : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      pn4_gpio_tri_n <= (others => '1');
      pn4_gpio_tri_p <= (others => '1');
    elsif rising_edge(ocp_clk) then
      pn4_gpio_tri_n <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn4_gpio_n_tri_reg), PN4_GPIO_WIDTH));
      pn4_gpio_tri_p <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn4_gpio_p_tri_reg), PN4_GPIO_WIDTH));
    end if;
  end process;

  -- Pn4 bi-directional GPIO
  -- Implement tri-states
  gen_pn4_loop_g : for n in 1 to PN4_GPIO_WIDTH generate
    gpio_inout.pn4_gpio.gpio_p(n) <= pn4_gpio_out_p(n) when (pn4_gpio_tri_p(n) = '0') else 'Z';
    gpio_inout.pn4_gpio.gpio_n(n) <= pn4_gpio_out_n(n) when (pn4_gpio_tri_n(n) = '0') else 'Z';
  end generate;


  -- Implement IOB input flip-flops
  gen_pn6_iob_iff_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      pn6_gpio_in <= gpio_inout.pn6_gpio.gpio;
    end if;
  end process;

  -- Implement IOB output flip-flops
  gen_pn6_iob_off_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      pn6_gpio_out <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn6_gpio_wdata_reg), PN6_GPIO_WIDTH));
    end if;
  end process;

  -- Implement IOB tristate flip-flops
  gen_pn6_iob_tff_p : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      pn6_gpio_tri <= (others => '1');
    elsif rising_edge(ocp_clk) then
      pn6_gpio_tri <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn6_gpio_tri_reg), PN6_GPIO_WIDTH));
    end if;
  end process;

  -- Pn6 bi-directional GPIO
  -- Implement tri-states
  gen_pn6_loop_g : for n in 1 to PN6_GPIO_WIDTH generate
    gpio_inout.pn6_gpio.gpio(n) <= pn6_gpio_out(n) when (pn6_gpio_tri(n) = '0')  else 'Z';
  end generate;

end;