--==============================================================================
--
--     File Name: blk_ds_clk_read_6tadv8.vhd
--       Version: 0.1
--          Date: 22 Aug 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Clock frequency measurement
--
--   Description: Clock frequency measurement using clock_speed blocks
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 22 Aug 2011: Initial design
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

entity blk_ds_clk_read is
generic(
  addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
port(
  -- OCP interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- Clock inputs
  pll_pri_clk      : in    std_logic;
  pll_reg_clk      : in    std_logic;
  pll_ref_clk      : in    std_logic;
  pll_mem_clk      : in    std_logic;
  clk_vec          : in    clk_vec_t);
end;


architecture syn_adv8 of blk_ds_clk_read is

  function conv_ref_clk_tcval(
    constant val: in target_use_t)
  return natural is
  begin
    case val is
      when SIM_OCP  => return REF_CLK_FREQ_HZ / 250000; -- => 4uS measurement period
      when SIM_MPTL => return REF_CLK_FREQ_HZ / 250000; -- => 4uS measurement period
      when SYN_NGC  => return REF_CLK_FREQ_HZ;
    end case;
  end;

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  constant REF_CLK_TCVAL : natural := conv_ref_clk_tcval(TARGET_USE);

  signal pll_pri_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_pri_clk_update  : std_logic;
  signal pll_pri_clk_running : std_logic;

  signal pll_reg_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_reg_clk_update  : std_logic;
  signal pll_reg_clk_running : std_logic;

  signal pll_ref_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_ref_clk_update  : std_logic;
  signal pll_ref_clk_running : std_logic;

  signal pll_mem_clk_freq    : std_logic_vector(REG_WIDTH-1 downto 0);
  signal pll_mem_clk_update  : std_logic;
  signal pll_mem_clk_running : std_logic;

  signal clk_freq            : clk_vec_freq_t;
  signal clk_update          : clk_vec_t;
  signal clk_running         : clk_vec_t;

  -- Clock frequency measurement registers
  signal ctrl_reg        : std_logic_vector(REG_WIDTH-1 downto 0);
  signal stat_reg        : std_logic_vector(REG_WIDTH-1 downto 0);
  signal sel_reg         : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ctrl_clr_update : std_logic;
  signal ctrl_reg_wr     : std_logic_vector(WE_WIDTH-1 downto 0);
  signal ctrl_wr         : std_logic;
  signal sel_clk         : std_logic_vector(4 downto 0);
  signal freq_reg        : std_logic_vector(REG_WIDTH-1 downto 0);

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- Clock frequency measurement
  signal pll_pri_clk_update_sr : std_logic;
  signal pll_reg_clk_update_sr : std_logic;
  signal pll_ref_clk_update_sr : std_logic;
  signal pll_mem_clk_update_sr : std_logic;
  signal clk_update_sr         : clk_vec_t;

begin

  ---------------------------------
  -- Clock frequency measurement --
  ---------------------------------

  -- Use pll_ref_clk as reference clock frequency
  -- Use ocp_clk as read clock frequency

  -- Measure pll_pri_clk frequency
  -- Divider = 2: range = 0 - 400 MHz (smp_clk*2 < REF_CLK_FREQ_HZ*4)
  pll_pri_clk_clock_speed_i : blk_clock_freq
  generic map(
    ref_clk_tcval      => REF_CLK_TCVAL,
    smp_clk_div_stages => 2)
  port map(
    rst                => ocp_rst,
    ref_clk            => pll_ref_clk,
    smp_clk            => pll_pri_clk,
    read_clk           => ocp_clk,
    do                 => '1',
    count              => pll_pri_clk_freq,
    running            => pll_pri_clk_running,
    valid              => open,
    done               => pll_pri_clk_update,
    idle               => open);

  -- Measure pll_reg_clk frequency
  -- Divider = 2: range = 0 - 400 MHz (smp_clk*2 < REF_CLK_FREQ_HZ*4)
  pll_reg_clk_clock_speed_i : blk_clock_freq
  generic map(
    ref_clk_tcval      => REF_CLK_TCVAL,
    smp_clk_div_stages => 2)
  port map(
    rst                => ocp_rst,
    ref_clk            => pll_ref_clk,
    smp_clk            => pll_reg_clk,
    read_clk           => ocp_clk,
    do                 => '1',
    count              => pll_reg_clk_freq,
    running            => pll_reg_clk_running,
    valid              => open,
    done               => pll_reg_clk_update,
    idle               => open);

  -- Measure pll_ref_clk frequency
  -- Divider = 2: range = 0 - 400 MHz (smp_clk*2 < REF_CLK_FREQ_HZ*4)
  pll_ref_clk_clock_speed_i : blk_clock_freq
  generic map(
    ref_clk_tcval      => REF_CLK_TCVAL,
    smp_clk_div_stages => 2)
  port map(
    rst                => ocp_rst,
    ref_clk            => pll_ref_clk,
    smp_clk            => pll_ref_clk,
    read_clk           => ocp_clk,
    do                 => '1',
    count              => pll_ref_clk_freq,
    running            => pll_ref_clk_running,
    valid              => open,
    done               => pll_ref_clk_update,
    idle               => open);

  -- Measure pll_mem_clk frequency
  -- Divider = 3: range = 0 - 800 MHz (smp_clk*2 < REF_CLK_FREQ_HZ*8)
  pll_mem_clk_clock_speed_i : blk_clock_freq
  generic map(
    ref_clk_tcval      => REF_CLK_TCVAL,
    smp_clk_div_stages => 3)
  port map(
    rst                => ocp_rst,
    ref_clk            => pll_ref_clk,
    smp_clk            => pll_mem_clk,
    read_clk           => ocp_clk,
    do                 => '1',
    count              => pll_mem_clk_freq,
    running            => pll_mem_clk_running,
    valid              => open,
    done               => pll_mem_clk_update,
    idle               => open);

  mgt_clks_clock_speed_loop_g : for n in MGT110_CLK0_NUM to MGT118_CLK1_NUM generate
    mgt_clks_clock_speed_g : if (MGT_CLKS_VALID(n-MGT110_CLK0_NUM) = '1') generate
      -- Measure MGT clock frequencies
      -- Divider = 2: range = 0 - 400 MHz (smp_clk*2 < REF_CLK_FREQ_HZ*4)
      mgt_clks_clock_speed_i : blk_clock_freq
      generic map(
        ref_clk_tcval      => REF_CLK_TCVAL,
        smp_clk_div_stages => 2)
      port map(
        rst                => ocp_rst,
        ref_clk            => pll_ref_clk,
        smp_clk            => clk_vec(n),
        read_clk           => ocp_clk,
        do                 => '1',
        count              => clk_freq(n),
        running            => clk_running(n),
        valid              => open,
        done               => clk_update(n),
        idle               => open);
    end generate;
  end generate;


  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => addr_width, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_WIDTH,  -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 1)          -- natural := 1);
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      ctrl_reg    <= (others => '0');
      ctrl_reg_wr <= (others => '0');
      sel_reg     <= (others => '0');
    elsif rising_edge(ocp_clk) then
      ctrl_reg_wr <= (others => '0');
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          if (ds_a = CLK_READ_CTRL_REG_ADDR(addr_width-1 downto 0)) then
            ctrl_reg(u downto l) <= ds_d(u downto l);
            ctrl_reg_wr(i) <= '1';
          end if;
          if (ds_a = CLK_READ_SEL_REG_ADDR(addr_width-1 downto 0)) then
            sel_reg(u downto l) <= ds_d(u downto l);
          end if;
        end if;
      end loop;
    end if;
  end process;

  ctrl_clr_update <= ctrl_reg(31);          -- clear updates
  ctrl_wr         <= AND_reduce(ctrl_reg_wr);
  sel_clk         <= sel_reg(4 downto 0);   -- clock select bits


  -- Multiplex read registers
  rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      ds_q_var := (others => '-');

      if (ds_a = CLK_READ_SEL_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := sel_reg;
      end if;
      if (ds_a = CLK_READ_STAT_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := stat_reg;
      end if;
      if (ds_a = CLK_READ_FREQ_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := freq_reg;
      end if;

      ds_q <= ds_q_var;
    end if;
  end process;


  -- Generate rdata_reg by muxing clock freq counts
  clk_freq_p : process(ocp_rst,ocp_clk)
    variable pll_reg_clk_stat : std_logic_vector(2 downto 0);
    variable pll_pri_clk_stat : std_logic_vector(2 downto 0);
    variable pll_ref_clk_stat : std_logic_vector(2 downto 0);
    variable pll_mem_clk_stat : std_logic_vector(2 downto 0);
    variable clk_freq_stat    : clk_vec_stat_t := (others => "000");
  begin
    if (ocp_rst = '1') then
      pll_reg_clk_update_sr <= '0';
      pll_pri_clk_update_sr <= '0';
      pll_ref_clk_update_sr <= '0';
      pll_mem_clk_update_sr <= '0';
      clk_update_sr         <= (others => '0');
    elsif rising_edge(ocp_clk) then
      -- Fix clock valid for PLL clocks to '1'
      pll_reg_clk_stat := pll_reg_clk_update_sr & '1' & pll_reg_clk_running;
      pll_pri_clk_stat := pll_pri_clk_update_sr & '1' & pll_pri_clk_running;
      pll_ref_clk_stat := pll_ref_clk_update_sr & '1' & pll_ref_clk_running;
      pll_mem_clk_stat := pll_mem_clk_update_sr & '1' & pll_mem_clk_running;

      -- Generate clock valid from MGT_CLKS_VALID(n)
      stat_comb_l : for n in MGT110_CLK0_NUM to MGT118_CLK1_NUM loop
        clk_freq_stat(n) := clk_update_sr(n) & MGT_CLKS_VALID(n-MGT110_CLK0_NUM) & clk_running(n);
      end loop;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_pri_clk_update_sr <= '0';
      elsif (pll_pri_clk_update = '1') then
        pll_pri_clk_update_sr <= '1';
      end if;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_reg_clk_update_sr <= '0';
      elsif (pll_reg_clk_update = '1') then
        pll_reg_clk_update_sr <= '1';
      end if;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_ref_clk_update_sr <= '0';
      elsif (pll_ref_clk_update = '1') then
        pll_ref_clk_update_sr <= '1';
      end if;

      if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
        pll_mem_clk_update_sr <= '0';
      elsif (pll_mem_clk_update = '1') then
        pll_mem_clk_update_sr <= '1';
      end if;

      update_sr_l : for n in MGT110_CLK0_NUM to MGT118_CLK1_NUM loop
        if (ctrl_wr = '1')and(ctrl_clr_update = '1') then
          clk_update_sr(n) <= '0';
        elsif (clk_update(n) = '1') then
          clk_update_sr(n) <= '1';
        end if;
      end loop;
    end if;

    if rising_edge(ocp_clk) then
      case sel_clk is
        -- Internal clocks
        when PLL_REG_CLK_SEL => stat_reg(31 downto 29) <= pll_reg_clk_stat;
        when PLL_PRI_CLK_SEL => stat_reg(31 downto 29) <= pll_pri_clk_stat;
        when PLL_REF_CLK_SEL => stat_reg(31 downto 29) <= pll_ref_clk_stat;
        when PLL_MEM_CLK_SEL => stat_reg(31 downto 29) <= pll_mem_clk_stat;

        -- MGT external clocks
        when MGT110_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT110_CLK0_NUM);
        when MGT112_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT112_CLK0_NUM);
        when MGT113_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT113_CLK0_NUM);
        when MGT115_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT115_CLK0_NUM);
        when MGT117_CLK0_SEL => stat_reg(31 downto 29) <= clk_freq_stat(MGT117_CLK0_NUM);
        when others          => stat_reg <= (others => '0');
      end case;

      case sel_clk is
        -- Internal clocks
        when PLL_REG_CLK_SEL => freq_reg <= pll_reg_clk_freq;
        when PLL_PRI_CLK_SEL => freq_reg <= pll_pri_clk_freq;
        when PLL_REF_CLK_SEL => freq_reg <= pll_ref_clk_freq;
        when PLL_MEM_CLK_SEL => freq_reg <= pll_mem_clk_freq;

        -- MGT external clocks
        when MGT110_CLK0_SEL => freq_reg <= clk_freq(MGT110_CLK0_NUM);
        when MGT112_CLK0_SEL => freq_reg <= clk_freq(MGT112_CLK0_NUM);
        when MGT113_CLK0_SEL => freq_reg <= clk_freq(MGT113_CLK0_NUM);
        when MGT115_CLK0_SEL => freq_reg <= clk_freq(MGT115_CLK0_NUM);
        when MGT117_CLK0_SEL => freq_reg <= clk_freq(MGT117_CLK0_NUM);
        when others          => freq_reg <= (others => '0');
      end case;
    end if;
  end process;

end;