--==============================================================================
--
--     File Name: blk_ds_io_test_6tadv8.vhd
--       Version: 0.1
--          Date: 23 Aug 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Front/Rear IO test
--
--   Description: Front/Rear GPIO test
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 23 Aug 2011: Initial design
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

entity blk_ds_io_test is
generic (
  addr_width : natural range 1 to ADB3_OCP_ADDR_WIDTH);
port(
  -- OCP interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- IO interface
  gpio_inout       : inout gpio_inout_t);
end;


architecture syn_adv8 of blk_ds_io_test is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- PN6 GPIO registers
  signal pn6_gpio_tri_reg      : std_logic_vector((2*REG_WIDTH)-1 downto 0);
  signal pn6_gpio_wdata_reg    : std_logic_vector((2*REG_WIDTH)-1 downto 0);
  signal pn6_gpio_rdata_reg    : std_logic_vector((2*REG_WIDTH)-1 downto 0);

  -- IOB tristate flip-flop signals
  signal pn6_gpio_tri : std_logic_vector(PN6_GPIO_WIDTH downto 1) := (others => '1');

  -- IOB output flip-flop signals
  signal pn6_gpio_out : std_logic_vector(PN6_GPIO_WIDTH downto 1);

  -- IOB input flip-flop signals
  signal pn6_gpio_in : std_logic_vector(PN6_GPIO_WIDTH downto 1);

  -- IOB attributes for tristate FFs
  attribute IOB : string;
  attribute IOB of pn6_gpio_tri : signal is "TRUE";

  -- IOB attributes for output FFs
  attribute IOB of pn6_gpio_out : signal is "TRUE";

  -- IOB attributes for input FFs
  attribute IOB of pn6_gpio_in : signal is "TRUE";

begin

  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => addr_width, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_WIDTH,  -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 1)          -- natural := 1
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  -- Initialise tri_reg so that output enables are off
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      pn6_gpio_tri_reg <= (others => '1');
      pn6_gpio_wdata_reg <= (others => '0');
    elsif rising_edge(ocp_clk) then
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          -- PN6 write registers
          if (ds_a = PN6_GPIO_MS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_tri_reg(32+u downto 32+l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_tri_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN6_GPIO_MS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_wdata_reg(32+u downto 32+l) <= ds_d(u downto l);
          end if;
          if (ds_a = PN6_GPIO_LS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
            pn6_gpio_wdata_reg(u downto l) <= ds_d(u downto l);
          end if;
        end if;
      end loop;
    end if;
  end process;

  -- Multiplex read registers
  rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      pn6_gpio_rdata_reg <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn6_gpio_in), 2*REG_WIDTH));

      ds_q_var := (others => '-');

      -- PN6 MS read registers
      if (ds_a = PN6_GPIO_MS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_tri_reg(2*REG_WIDTH-1 downto REG_WIDTH);
      end if;
      if (ds_a = PN6_GPIO_MS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_wdata_reg(2*REG_WIDTH-1 downto REG_WIDTH);
      end if;
      if (ds_a = PN6_GPIO_MS_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_rdata_reg(2*REG_WIDTH-1 downto REG_WIDTH);
      end if;

      -- PN6 LS read registers
      if (ds_a = PN6_GPIO_LS_TRI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_tri_reg(REG_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN6_GPIO_LS_DATAO_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_wdata_reg(REG_WIDTH-1 downto 0);
      end if;
      if (ds_a = PN6_GPIO_LS_DATAI_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := pn6_gpio_rdata_reg(REG_WIDTH-1 downto 0);
      end if;

      ds_q <= ds_q_var;
    end if;
  end process;


  -- Implement IOB input flip-flops
  gen_pn6_iob_iff_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      pn6_gpio_in <= gpio_inout.pn6_gpio.gpio;
    end if;
  end process;

  -- Implement IOB output flip-flops
  gen_pn6_iob_off_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      pn6_gpio_out <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn6_gpio_wdata_reg), PN6_GPIO_WIDTH));
    end if;
  end process;

  -- Implement IOB tristate flip-flops
  gen_pn6_iob_tff_p : process(ocp_rst, ocp_clk)
  begin
    if ocp_rst = '1' then
      pn6_gpio_tri <= (others => '1');
    elsif rising_edge(ocp_clk) then
      pn6_gpio_tri <= STD_LOGIC_VECTOR(RESIZE(UNSIGNED(pn6_gpio_tri_reg), PN6_GPIO_WIDTH));
    end if;
  end process;

  -- Pn6 bi-directional GPIO
  -- Implement tri-states
  gen_pn6_loop_g : for n in 1 to PN6_GPIO_WIDTH generate
    gpio_inout.pn6_gpio.gpio(n) <= pn6_gpio_out(n) when (pn6_gpio_tri(n) = '0')  else 'Z';
  end generate;

end;