--==============================================================================
--
--     File Name: uber_tb_pkg_6tge.vhd
--       Version: 0.5
--          Date: 25 Aug 2011
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: Package uber_tb_pkg
--
--   Description: Testbench package for uber using ADM-XRC-6TGE
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 25 Mar 2010: Initial design
--                0.2, KDC, 08 Jun 2011: Added TB_DDR3_4G_PART constant
--                0.3, KDC, 27 Jun 2011: Added 666MHz clock
--                0.4, KDC, 25 Jul 2011:
--                     - Imported constants from other test files
--                     - Moved clock generation constants to adb3_target_tb_inc_pkg
--                     - Moved On-board RAM Part Selection to test_board_clks.vhd
--                     - Added test_uber component definitions
--                0.5, KDC, 25 Aug 2011:
--                     - Changed shared variable linti_l_edge to int_edge
--                     - Moved uber component definition to test_uber.vhd
--                     - Added constant CLKS_OUT_FREQ
--                     - Made specific to ADM-XRC-6TGE
--                     - Added TEST_MGT_CLK and TEST_CUS_CLK constants
--                     - Added test_uber_mem component
--
--==============================================================================

--
-- Exported types:
--
--   top_comp_t
--   ds_comp_t
--   dma_comp_t
--   top_pass_t
--   ds_pass_t
--   dma_pass_t
--
-- Exported components:
--
--   test_uber_ds
--   test_uber_dma
--   test_uber_dm
--   test_uber_probes
--   test_uber_mem
--

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_tb_inc_pkg.all;
use work.uber_pkg.all;

package uber_tb_pkg is

  --------------------
  -- Top level test --
  --------------------

  -- Constant for MPTL GPIO loopback test
  constant MPTL_GPIO : std_logic_vector(15 downto 0) := X"F1D0";

  -------------------
  -- DS Clock test --
  -------------------
  -- Define +/- acceptable error for clock frequency measurement (MHz)
  constant CLK_TEST_DIFF : real := 2.0;

  -- Define expected frequencies for Uber PLL clocks in MHz
  -- Used to compare with values read from frequecy measurement registers
  constant PLL_PRI_CLK_FREQ : natural := 200;
  constant PLL_REG_CLK_FREQ : natural := 80;
  constant PLL_MEM_CLK_FREQ : natural := 400;
  constant PLL_REF_CLK_FREQ : natural := 200;

  -- Define expected frequencies for clock outputs in MHz
  constant CLKS_OUT_FREQ : clks_out_exp_t := (xrm_mgtclk_c2m => 80);

  -- Frequency of programmable clock generated by bridge FPGA in Hz
  constant BRIDGE_LCLK_FREQ : real range 5.0e6 to 700.0e6 := 666.0e6; -- Programmable (5 MHz to 700 MHz)
  constant TARGET_LCLK_PER  : time := 1 sec / BRIDGE_LCLK_FREQ;

  -- Define clk_vec clocks to test
  constant TEST_MGT_CLK_SEL  : clk_vec_sel_t := MGT113_CLK0_SEL;
  constant TEST_CUS_CLK_SEL  : clk_vec_sel_t := CUSTOM_CLK0_SEL;

  -- Define expected frequencies for Uber test clocks in MHz
  -- Used to compare with values read from frequecy measurement registers
  -- round(A/B) =- trunc((A+B/2)/B)
  constant TEST_MGT_CLK_FREQ : natural := (1 us + (MGT113_CLK0_PER/2)) / MGT113_CLK0_PER;
  constant TEST_CUS_CLK_FREQ : natural := (1 us + (TARGET_LCLK_PER/2)) / TARGET_LCLK_PER;

  ---------------
  -- DS GPIO test --
  ---------------
  constant PN4_DATA : std_logic_vector(31 downto 0) := X"AABBCCDD";
  constant PN6_DATA : std_logic_vector(63 downto 0) := X"AAAABBBBCCCCDDDD";

  --------------------------
  -- DS on-board RAM test --
  --------------------------
  constant DS_RAM_TEST_BANK : natural range 0 to MEM_BANKS-1 := 1;

  -- Calculate the number of on-board RAM window pages available in the test bank of on-board RAM
  -- Use the byte address width of the on-board RAM bank
  -- DDR3: 2 x 1Gib parts/bank = 128 pages
  -- DDR3: 2 x 2Gib parts/bank = 256 pages
  constant DS_WIN_RAM_PAGES : natural := 2**(BOARD_MEM_BYTE_ADDR_WIDTH_ARRAY(DS_RAM_TEST_BANK) - adb3_ocp_mask_width(RAM_WIN_ADDR_RANGE.addr_mask));

  constant DS_WIN_RAM_BANK     : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(DS_RAM_TEST_BANK,32));
  constant DS_WIN_RAM_PAGE_BOT : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(0,32));
  constant DS_WIN_RAM_PAGE_TOP : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(DS_WIN_RAM_PAGES-1,32));

  --------------------------------
  -- Internal on-board RAM test --
  --------------------------------
  constant INT_RAM_TEST_BANK : natural range 0 to MEM_BANKS-1 := 1;

  -- Words-1 (16-byte words) of each phase of the internal RAM test
  constant INT_RAM_TEST_LENGTH : natural := 256 - 1;
  constant INT_RAM_TEST_LEN : std_logic_vector(31 downto 0) := std_logic_vector(to_unsigned(INT_RAM_TEST_LENGTH,32));

  -- Top address (16-byte words) of RAM bank
  constant INT_RAM_TEST_TOP : natural := (2**(BOARD_MEM_BYTE_ADDR_WIDTH_ARRAY(INT_RAM_TEST_BANK) - 4)) - 1;

  -- Start address (16-byte words) of each phase of the internal RAM test
  -- Set to top address - length of test - top word used by DS part of test
  constant INT_RAM_TEST_START : natural := INT_RAM_TEST_TOP - INT_RAM_TEST_LENGTH - 1;

  -----------------------
  -- DS Interrupt Test --
  -----------------------
  constant MASK_EN_ALL : std_logic_vector(31 downto 0) := X"00000000";

  -- Used in test_uber and test_uber_ds to indicate active edge on interrupt
  shared variable int_edge : std_logic := '0';

  ----------------------
  -- DS Test controls --
  ----------------------
  constant DO_RAM_TEST      : boolean := true;             -- Enable OCP direct slave test of on-board RAM
  constant DO_INT_RAM_TEST  : boolean := true;             -- Enable internal on-board RAM test (blk_mem_app)
  constant DO_CLOCK_TEST    : boolean := true;             -- Enable clock frequency measurement test
  constant DO_INTERRUPT_NUM : natural range 1 to 32 := 32; -- Set length of interrupt test


  --------------
  -- DMA Test --
  --------------

  -- Define DMA OCP channel function
  constant DMA_WRITE_CHANNEL  : natural range 0 to DMA_CHANNELS-1 := 0;
  constant DMA_READ_CHANNEL   : natural range 0 to DMA_CHANNELS-1 := 1;
  constant DMA_SINGLE_CHANNEL : natural range 0 to DMA_CHANNELS-1 := 0;

  -- BRAM is the only part of the DMA address space that is connected to multiple DMA channels
  -- Start DMA at BRAM address 0x100 to avoid clash with DS testing
--  constant DMA_ADDR_WR : adb3_ocp_addr_s := DMA_BRAM_ADDR_BASE + adb3_ocp_base(X"0000000200",DMA_ADDR_WIDTH);
--  constant DMA_ADDR_RD : adb3_ocp_addr_s := DMA_BRAM_ADDR_BASE + adb3_ocp_base(X"0000000200",DMA_ADDR_WIDTH);

  -- On-board RAM is connected to DMA channel 0 only
  -- Start DMA at RAM1 address 0x0200 to avoid clash with DS testing
  constant DMA_ADDR_WR : adb3_ocp_addr_s := DMA_RAM1_ADDR_BASE + adb3_ocp_base(X"0000000200",DMA_ADDR_WIDTH);
  constant DMA_ADDR_RD : adb3_ocp_addr_s := DMA_RAM1_ADDR_BASE + adb3_ocp_base(X"0000000200",DMA_ADDR_WIDTH);

  -- Constants for dma channel 0/1 test
  constant DMA_SIZE     : natural := 4032;  -- Must be multiple of 16
  constant DMA_BL_WRITE : natural := 8;
  constant DMA_BL_READ  : natural := 8;


  ------------------------
  -- Direct Master Test --
  ------------------------

  -- Select Direct Master inclusion
  type master_use_t is (MASTER_ON,MASTER_OFF);

  constant MASTER_USE : master_use_t := MASTER_OFF;
  constant DM_CHANNEL : natural range 0 to DM_CHANNELS-1 := 0;
  constant DM_DATA_WIDTH : natural range 8 to ADB3_OCP_DATA_WIDTH:= 32;

  -- Constants for testbench direct master register interface to simulate host memory
  -- Register offsets from base address in bytes, eg 0,4,8,... for 32 bit registers
  constant HOST_MASTER_REG_NUM  : natural := 5;
  constant HOST_MASTER_REG_OFF  : natural := HOST_MASTER_REG_NUM * (DM_DATA_WIDTH/8);
  constant HOST_MASTER_REG_ADDR : std_logic_vector(DM_ADDR_WIDTH-1 downto 0) := std_logic_vector(TO_UNSIGNED((HOST_MASTER_REG_OFF),DM_ADDR_WIDTH));


  ----------------------
  -- Type Definitions --
  ----------------------

  type top_comp_t is record
    mptl_gpio_complete : boolean;
    dma_abort_complete : boolean;
    clk_out_complete   : boolean;
  end record;

  type ds_comp_t is record
    simple_complete    : boolean;
    master_complete    : boolean;
    clock_complete     : boolean;
    frontio_complete   : boolean;
    reario_complete    : boolean;
    int_complete       : boolean;
    info_complete      : boolean;
    bram_complete      : boolean;
    ram_complete       : boolean;
  end record;

  type dma_comp_t is record
    dma_write_cmd_complete : boolean;
    dma_read_cmd_complete  : boolean;
    dma_read_resp_complete : boolean;
  end record;

  type top_pass_t is record
    mptl_gpio_passed   : boolean;
    dma_abort_passed   : boolean;
    clk_out_passed     : boolean;
  end record;

  type ds_pass_t is record
    simple_passed      : boolean;
    master_passed      : boolean;
    clock_passed       : boolean;
    frontio_passed     : boolean;
    reario_passed      : boolean;
    int_passed         : boolean;
    info_passed        : boolean;
    bram_passed        : boolean;
    ram_passed         : boolean;
  end record;

  type dma_pass_t is record
    dma_write_cmd_passed : boolean;
    dma_read_cmd_passed  : boolean;
    dma_read_resp_passed : boolean;
  end record;


  ---------------------------
  -- Component Definitions --
  ---------------------------

  component test_uber_ds
  port(
    -- OCP Interface
    ocp_clk          : in    std_logic;
    direct_slave_m2s : out   adb3_ocp_m2sT;
    direct_slave_s2m : in    adb3_ocp_s2mT;
    -- Test interface
    start_test       : in    std_logic;
    ds_comp          : out   ds_comp_t;
    ds_pass          : out   ds_pass_t);
  end component;

  component test_uber_dma
  port(
    -- OCP Interface
    ocp_clk          : in    std_logic;
    dma_channels_m2s : out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    -- Test interface
    start_test       : in    std_logic;
    dma_comp         : out   dma_comp_t;
    dma_pass         : out   dma_pass_t);
  end component;

  component test_uber_dm
  port(
    -- OCP Interface
    ocp_clk            : in    std_logic;
    direct_masters_m2s : in    adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : out   adb3_ocp_s2mT_array(0 to DM_CHANNELS-1));
  end component;

  component test_uber_probes
  port(
    -- OCP Interface
    ocp_clk            : in    std_logic;
    direct_slave_m2s   : in    adb3_ocp_m2sT;
    direct_slave_s2m   : in    adb3_ocp_s2mT;
    dma_channels_m2s   : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
    dma_channels_s2m   : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
    direct_masters_m2s : in    adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
    direct_masters_s2m : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
    -- Probe Interface
    probe_ds_err       : out   std_logic;
    probe_dma_err      : out   std_logic;
    probe_dm_err       : out   std_logic);
  end component;

  component test_uber_mem
  generic(
    message_level : natural);
  port(
    -- Memory Interface
    mem_addr_out   : in    mem_addr_out_t;
    mem_ctrl_out   : in    mem_ctrl_out_t;
    mem_data_inout : inout mem_data_inout_t;
    mem_clk_out    : in    mem_clk_out_t);
  end component;

end;


package body uber_tb_pkg is

end;