--==============================================================================
--
--     File Name: blk_bram.vhd
--       Version: 0.4
--          Date: 25 Jul 2011
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: BRAM instantiation block
--
--   Description: Implements a 512 kiB block of memory, organised as 2^15
--                words of 16 bytes, with an OCP interface. Bytes are
--                individually writeable.
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 12 Mar 2010: Initial design
--                0.2, TW, 09 Feb 2011:
--                  - Replicated and registered address signals and increased
--                    read latency in adb3_ocp_simple_bus_if instance to 2, in
--                    order to more easily meet timing.
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg.
--                0.3, TW, 14 Feb 2011:
--                  - Now natively OCP, instead of using an instance of
--                    adb3_ocp_simple_bus_if, for higher performance.
--                0.4, KDC, 25 Jul 2011:
--                  - Replaced native OCP interface with new general purpose
--                    component adb3_ocp_bus_if_nb, still with high performance.
--                  - Replaced mask_vec_width with adb3_ocp_mask_width function
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.uber_pkg.all;

entity blk_bram is
port(
  ocp_rst     : in    std_logic;
  ocp_clk     : in    std_logic;
  -- BRAM OCP Interface
  bram_if_m2s : in    adb3_ocp_m2sT;
  bram_if_s2m : out   adb3_ocp_s2mT);
end;

architecture syn of blk_bram is

  component bram_single_wrap
  generic(
    output_reg : integer := 0);
  port(
    -- Port Interface
    rst  : in    std_logic;
    clk  : in    std_logic;
    addr : in    std_logic_vector(14 downto 0);
    we   : in    std_logic;
    di   : in    std_logic;
    do   : out   std_logic);
  end component;

  constant ADDR_WIDTH : natural := adb3_ocp_mask_width(BRAM_ADDR_RANGE.addr_mask);
  constant DATA_WIDTH : natural := ADB3_OCP_DATA_WIDTH;
  constant BE_WIDTH   : natural := DATA_WIDTH/8;

  signal bus_a  : std_logic_vector(ADDR_WIDTH-1 downto 0);
  signal bus_d  : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal bus_we : std_logic_vector(BE_WIDTH-1 downto 0) := (others => '0');
  signal bus_q  : std_logic_vector(DATA_WIDTH-1 downto 0);

  type bram_addr_array_t is array(0 to ADB3_OCP_BE_WIDTH-1) of std_logic_vector(ADDR_WIDTH-5 downto 0);

  signal bram_a : bram_addr_array_t;
  attribute equivalent_register_removal : string;
  attribute equivalent_register_removal of bram_a : signal is "no";

  signal bram_d  : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal bram_we : std_logic_vector(BE_WIDTH-1 downto 0) := (others => '0');
  signal bram_q  : std_logic_vector(DATA_WIDTH-1 downto 0);

begin

  adb3_ocp_simple_bus_if_nb_i : adb3_ocp_simple_bus_if_nb
  generic map(
    addr_width   => ADDR_WIDTH,  -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => DATA_WIDTH,  -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 3)           -- natural := 1)
  port map(
    -- OCP interface
    ocp_rst    => ocp_rst,     -- in    std_logic
    ocp_clk    => ocp_clk,     -- in    std_logic
    ocp_m2s    => bram_if_m2s, -- in    adb3_ocp_m2sT
    ocp_s2m    => bram_if_s2m, -- out   adb3_ocp_s2mT
    -- Bus interface
    d          => bus_d,       -- out   std_logic_vector(data_width-1 downto 0)
    q          => bus_q,       -- in    std_logic_vector(data_width-1 downto 0)
    a          => bus_a,       -- out   std_logic_vector(addr_width-1 downto 0)
    w          => open,        -- out   std_logic
    r          => open,        -- out   std_logic
    we         => bus_we,      -- out   std_logic_vector((data_width/8)-1 downto 0)
    -- Debug
    debug_info => open);       -- out   std_logic_vector(31 downto 0))

  -- Replicate bus_a address bus to improve timing
  -- Convert bus_a byte address to bram_a 16-byte address
  duplicate_addr : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      bram_d  <= bus_d;
      bram_a  <= (others => bus_a(ADDR_WIDTH-1 downto 4));
      bram_we <= bus_we;
    end if;
  end process;

  -- Generate 32k x 128 bits single port BRAM
  bram_bytes_loop_g : for m in 0 to ADB3_OCP_BE_WIDTH-1 generate
    bram_bits_loop_g : for n in 0 to 7 generate
      bram_single_wrap_i : bram_single_wrap
      generic map(
        output_reg => 1) -- integer := 0);
      port map(
        rst  => ocp_rst,        -- in    std_logic
        clk  => ocp_clk,        -- in    std_logic
        addr => bram_a(m),      -- in    std_logic_vector(14 downto 0)
        we   => bram_we(m),     -- in    std_logic
        di   => bram_d(8*m+n),  -- in    std_logic
        do   => bram_q(8*m+n)); -- out   std_logic)
    end generate;
  end generate;

  bus_q <= bram_q;

end;