--==============================================================================
--
--     File Name: blk_direct_slave.vhd
--       Version: 0.6
--          Date: 24 Aug 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: Direct Slave functional block
--                Functions:
--                  Direct Slave clock domain transfer
--                  Direct Slave address space split
--                  simple test block
--                  clock frequency measurement block
--                  io test block
--                  interrupt test block
--                  info register block
--                  BRAM interface clock domain transfer
--
--   Description: Adds OCP Direct Slave interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 15 Mar 2010: Initial design
--                0.2, TW,  09 Feb 2011:
--                      - Now uses adb3_ocp_comp package instead of adb3_target_pkg.
--                      - Got rid of some unnecessary resets.
--                      - pll_ref_clk is now called pll_pri_clk (primary clock).
--                0.3, KDC, 10 Feb 2011:
--                      - Added new pll_ref_clk measurement based on ref_clk input
--                0.4, KDC, 23 Jun 2011:
--                      - Changed DS address split into 2 stage process
--                      - Split into Registers, BRAM, and RAM, and then convert
--                        to reg clock domain for further splitting.
--                0.5, KDC, 25 Jul 2011:
--                      - Replaced mask_vec_width with adb3_ocp_mask_width function
--                0.6, KDC, 24 Aug 2011:
--                      - Changed blk_ds_int_test interrupt_l signal to finti_l
--                      - Added blk_ds_int_test signal interrupt
--                      - Changed PAGE_ADDR_WIDTH_DDR3 to PAGE_ADDR_WIDTH
--                      - Changed ds_page_ddr3 to ds_page
--                      - Updated blk_ds_mem_reg component instantiation
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

entity blk_direct_slave is
port(
  -- Reset input
  rst                 : in    std_logic;
  -- Clock inputs
  pll_pri_clk         : in    std_logic;
  pll_reg_clk         : in    std_logic;
  pll_ref_clk         : in    std_logic;
  pll_mem_clk         : in    std_logic;
  -- Direct slave primary OCP interface
  direct_slave_m2s    : in    adb3_ocp_m2sT;
  direct_slave_s2m    : out   adb3_ocp_s2mT;
  -- Clock frequency measurement register interface
  clk_vec             : in    clk_vec_t;
  -- GPIO signal test register interface
  gpio_inout          : inout gpio_inout_t;
  -- Interrupt test register interface
  finti_l             : out   std_logic;
  interrupt           : out   std_logic;
  -- Memory interface status
  mem_if_stat         : in    mem_if_stat_array_t;
  mem_if_err          : in    mem_if_err_array_t;
  -- Memory application control
  mem_app_go          : out   mem_app_go_array_t;
  mem_app_offset      : out   mem_app_offset_array_t;
  mem_app_length      : out   mem_app_length_array_t;
  -- Memory application status
  mem_app_done        : in    mem_app_done_array_t;
  mem_app_err         : in    mem_app_err_array_t;
  mem_app_err_ph      : in    mem_app_err_ph_array_t;
  mem_app_err_addr    : in    mem_app_err_addr_array_t;
  -- BRAM secondary OCP interface
  ds_bram_m2s         : out   adb3_ocp_m2sT;
  ds_bram_s2m         : in    adb3_ocp_s2mT;
  -- RAM secondary OCP interface
  ds_ram_m2s          : out   adb3_ocp_m2sT;
  ds_ram_s2m          : in    adb3_ocp_s2mT;
  -- Debug
  bram_mux_err_info   : in    std_logic_vector(31 downto 0);
  ram_mux_err_info    : in    mem_if_debug_array_t;
  mem_if_err_info     : in    mem_if_debug_array_t);
end;


architecture syn of blk_direct_slave is

  -- Size of bank register (DS) to select between banks of memory
  constant BANK_ADDR_WIDTH : natural := 3;

  -- Define bank address signal
  signal ds_bank : std_logic_vector(BANK_ADDR_WIDTH-1 downto 0);

  -- Size of DS memory window BYTE address range = 2M bytes (21 bits)
  constant WIND_ADDR_WIDTH : natural := adb3_ocp_mask_width(RAM_WIN_ADDR_RANGE.addr_mask);

  -- Size of DS PAGE register required to augment direct slave address bits [20:0] to DMA_ADDR_WIDTH - BANK_ADDR_WIDTH bits
  constant PAGE_ADDR_WIDTH : natural := DMA_ADDR_WIDTH - BANK_ADDR_WIDTH - WIND_ADDR_WIDTH;

  -- Define page address signal
  signal ds_page : std_logic_vector(PAGE_ADDR_WIDTH-1 downto 0);

  -- Pad unused OCP address above DMA_ADDR_WIDTH with zeros
  constant OCP_ADDR_PAD : std_logic_vector((ADB3_OCP_ADDR_WIDTH-DMA_ADDR_WIDTH)-1 downto 0) := (others => '0');

  -- Obtain number of slaves from DS address range table
  constant DS_SLAVES : natural := DS_ADDR_RANGE_TABLE'length;

  -- Obtain number of register slaves from DS register address range table
  constant DS_REG_SLAVES : natural := DS_REG_RANGE_TABLE'length;

  -- Decode address widths of register blocks
  constant SIMPLE_ADDR_WIDTH   : natural := adb3_ocp_mask_width(SIMPLE_ADDR_RANGE.addr_mask);
  constant CLK_READ_ADDR_WIDTH : natural := adb3_ocp_mask_width(CLK_READ_ADDR_RANGE.addr_mask);
  constant GPIO_ADDR_WIDTH     : natural := adb3_ocp_mask_width(GPIO_ADDR_RANGE.addr_mask);
  constant INT_ADDR_WIDTH      : natural := adb3_ocp_mask_width(INT_ADDR_RANGE.addr_mask);
  constant RAM_REG_ADDR_WIDTH  : natural := adb3_ocp_mask_width(RAM_REG_ADDR_RANGE.addr_mask);
  constant INFO_ADDR_WIDTH     : natural := adb3_ocp_mask_width(INFO_ADDR_RANGE.addr_mask);

  -- OCP split channels
  signal ds_slaves_m2s : adb3_ocp_m2sT_array(0 to DS_SLAVES-1);
  signal ds_slaves_s2m : adb3_ocp_s2mT_array(0 to DS_SLAVES-1);

  -- OCP register channel in pll_reg_clk domain
  signal ds_reg_m2s : adb3_ocp_m2sT;
  signal ds_reg_s2m : adb3_ocp_s2mT;

  -- OCP register split channels
  signal ds_reg_slaves_m2s : adb3_ocp_m2sT_array(0 to DS_REG_SLAVES-1);
  signal ds_reg_slaves_s2m : adb3_ocp_s2mT_array(0 to DS_REG_SLAVES-1);
  signal ds_split_err_info : std_logic_vector(31 downto 0);

begin

  -- Split DS OCP channel according to function
  -- Uses non-blocking split component (high speed)
  ds_split_nb_ds_i : adb3_ocp_split_nb
  generic map(
    addr_range_table => DS_ADDR_RANGE_TABLE)  -- addr_range_table_t
  port map(
    ocp_rst     => rst,                -- in    std_logic
    ocp_clk     => pll_pri_clk,        -- in    std_logic
    -- Master OCP interface
    slave_m2s   => direct_slave_m2s,   -- in    adb3_ocp_m2sT
    slave_s2m   => direct_slave_s2m,   -- out   adb3_ocp_s2mT
    -- Slave OCP interfaces
    masters_m2s => ds_slaves_m2s,      -- out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1)
    masters_s2m => ds_slaves_s2m,      -- in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1)
    -- Debug
    debug_info  => ds_split_err_info); -- out   std_logic_vector(31 downto 0))


  -- Transfer DS register OCP channel to pll_reg_clk domain
  ds_reg_cross_clk_dom_i : adb3_ocp_cross_clk_dom
  port map(
    -- OCP primary port
    slave_rst  => rst,              -- in    std_logic
    slave_clk  => pll_pri_clk,      -- in    std_logic
    slave_m2s  => ds_slaves_m2s(0), -- in    adb3_ocp_m2sT
    slave_s2m  => ds_slaves_s2m(0), -- out   adb3_ocp_s2mT
    -- OCP secondary port
    master_rst => rst,              -- in    std_logic
    master_clk => pll_reg_clk,      -- in    std_logic
    master_m2s => ds_reg_m2s,       -- out   adb3_ocp_m2sT
    master_s2m => ds_reg_s2m);      -- in    adb3_ocp_s2mT


  -- Split DS OCP register channel according to function
  -- Uses blocking split component (low speed)
  reg_split_b_i : adb3_ocp_split_b
  generic map(
    addr_range_table => DS_REG_RANGE_TABLE)  -- addr_range_table_t)
  port map(
    ocp_rst     => rst,               -- in    std_logic
    ocp_clk     => pll_reg_clk,       -- in    std_logic
    -- OCP primary port
    slave_m2s   => ds_reg_m2s,        -- in    adb3_ocp_m2sT
    slave_s2m   => ds_reg_s2m,        -- out   adb3_ocp_s2mT
    -- OCP secondary ports
    masters_m2s => ds_reg_slaves_m2s, -- out   adb3_ocp_m2sT_array(0 to addr_range_table'length-1)
    masters_s2m => ds_reg_slaves_s2m, -- in    adb3_ocp_s2mT_array(0 to addr_range_table'length-1)
    -- Debug
    debug_info  => open);             -- out   std_logic_vector(31 downto 0))


  -----------------------
  -- Simple test block --
  -----------------------
  blk_ds_simple_test_i : blk_ds_simple_test
  generic map(
    addr_width => SIMPLE_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,                   -- in    std_logic
    ocp_clk          => pll_reg_clk,           -- in    std_logic
    direct_slave_m2s => ds_reg_slaves_m2s(0),  -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_reg_slaves_s2m(0)); -- out   adb3_ocp_s2mT)


  ---------------------------
  -- Clock frequency block --
  ---------------------------
  blk_ds_clk_read_i : blk_ds_clk_read
  generic map(
    addr_width => CLK_READ_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,                  -- in    std_logic
    ocp_clk          => pll_reg_clk,          -- in    std_logic
    direct_slave_m2s => ds_reg_slaves_m2s(1), -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_reg_slaves_s2m(1), -- out   adb3_ocp_s2mT
    -- Clock inputs
    pll_pri_clk      => pll_pri_clk,          -- in    std_logic
    pll_reg_clk      => pll_reg_clk,          -- in    std_logic
    pll_ref_clk      => pll_ref_clk,          -- in    std_logic
    pll_mem_clk      => pll_mem_clk,          -- in    std_logic
    clk_vec          => clk_vec);             -- in    clk_vec_t


  ------------------------------
  -- Front/Rear IO test block --
  ------------------------------
  blk_ds_io_test_i : blk_ds_io_test
  generic map(
    addr_width => GPIO_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,                  -- in    std_logic
    ocp_clk          => pll_reg_clk,          -- in    std_logic
    direct_slave_m2s => ds_reg_slaves_m2s(2), -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_reg_slaves_s2m(2), -- out   adb3_ocp_s2mT
    -- IO interface
    gpio_inout       => gpio_inout);          -- inout gpio_inout_t


  --------------------------
  -- Interrupt test block --
  --------------------------
  blk_ds_int_test_i : blk_ds_int_test
  generic map(
    addr_width => INT_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst          => rst,                  -- in    std_logic
    ocp_clk          => pll_reg_clk,          -- in    std_logic
    direct_slave_m2s => ds_reg_slaves_m2s(3), -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_reg_slaves_s2m(3), -- out   adb3_ocp_s2mT
    -- Interrupt interface
    finti_l          => finti_l,              -- out   std_logic
    interrupt        => interrupt);           -- out   std_logic


  ------------------------
  -- RAM register block --
  ------------------------
  blk_ds_ram_reg_i : blk_ds_mem_reg
  generic map(
    addr_width      => RAM_REG_ADDR_WIDTH, -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
    bank_addr_width => BANK_ADDR_WIDTH,    -- natural
    page_addr_width => PAGE_ADDR_WIDTH)    -- natural
  port map(
    -- OCP interface
    ocp_rst          => rst,                  -- in    std_logic
    ocp_clk          => pll_reg_clk,          -- in    std_logic
    direct_slave_m2s => ds_reg_slaves_m2s(4), -- in    adb3_ocp_m2sT
    direct_slave_s2m => ds_reg_slaves_s2m(4), -- out   adb3_ocp_s2mT
    -- DS RAM interface
    ds_bank          => ds_bank,              -- out   std_logic_vector(bank_addr_width-1 downto 0)
    ds_page          => ds_page,              -- out   std_logic_vector(page_addr_width-1 downto 0)
    -- Memory interface status
    mem_if_stat      => mem_if_stat,          -- in    mem_if_stat_array_t
    mem_if_err       => mem_if_err,           -- in    mem_if_err_array_t
    -- Memory application control
    mem_app_go       => mem_app_go,           -- out   mem_app_go_array_t
    mem_app_offset   => mem_app_offset,       -- out   mem_app_offset_array_t
    mem_app_length   => mem_app_length,       -- out   mem_app_length_array_t
    -- Memory application status
    mem_app_done     => mem_app_done,         -- in    mem_app_done_array_t
    mem_app_err      => mem_app_err,          -- in    mem_app_err_array_t
    mem_app_err_ph   => mem_app_err_ph,       -- in    mem_app_err_ph_array_t
    mem_app_err_addr => mem_app_err_addr,     -- in    mem_app_err_addr_array_t
    -- Debug
    ram_mux_err_info => ram_mux_err_info,     -- in    mem_if_debug_array_t
    mem_if_err_info  => mem_if_err_info);     -- in    mem_if_debug_array_t


  ----------------
  -- Info block --
  ----------------
  blk_ds_info_i : blk_ds_info
  generic map(
    addr_width => INFO_ADDR_WIDTH)  -- natural range 1 to ADB3_OCP_ADDR_WIDTH)
  port map(
    -- OCP interface
    ocp_rst           => rst,                  -- in    std_logic
    ocp_clk           => pll_reg_clk,          -- in    std_logic
    direct_slave_m2s  => ds_reg_slaves_m2s(5), -- in    adb3_ocp_m2sT
    direct_slave_s2m  => ds_reg_slaves_s2m(5), -- out   adb3_ocp_s2mT
    -- Debug
    ds_split_err_info => ds_split_err_info);   -- in    std_logic_vector(31 downto 0))


  --------------------------
  -- BRAM block interface --
  --------------------------

  -- Connect DS BRAM split channel to DS BRAM interface
  ds_bram_m2s <= ds_slaves_m2s(1);
  ds_slaves_s2m(1) <= ds_bram_s2m;


  --------------------------
  -- RAM block interface --
  --------------------------

  -- Translate DS address space memory window to DMA address space DS memory channel
  -- Address translation forms DMA address space address by concatenating:
  --   Bank select address (ds_bank)
  --   Window select address (ds_page)
  --   DS memory window address (ds_slaves_m2s(2).Addr)
  -- Uses full ADB3_OCP_ADDR_WIDTH address range and pad unused bits with zeros
  ds_ram_m2s.Cmd         <= ds_slaves_m2s(2).Cmd;
  ds_ram_m2s.Addr        <= OCP_ADDR_PAD & ds_bank & ds_page & ds_slaves_m2s(2).Addr(WIND_ADDR_WIDTH-1 downto 0);
  ds_ram_m2s.BurstLength <= ds_slaves_m2s(2).BurstLength;
  ds_ram_m2s.Data        <= ds_slaves_m2s(2).Data;
  ds_ram_m2s.DataByteEn  <= ds_slaves_m2s(2).DataByteEn;
  ds_ram_m2s.DataValid   <= ds_slaves_m2s(2).DataValid;
  ds_ram_m2s.RespAccept  <= ds_slaves_m2s(2).RespAccept;
  ds_ram_m2s.Tag         <= ds_slaves_m2s(2).Tag;

  ds_slaves_s2m(2) <= ds_ram_s2m;

end;