--==============================================================================
--
--     File Name: blk_dma.vhd
--       Version: 0.1
--          Date: 12 Mar 2010
--         Model: synth
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: DMA instantiation block
--                Options:
--                  32-bit DMA
--                  64-bit DMA
--                  128-bit DMA
--
--   Description: Adds OCP DMA interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 12 Mar 2010: Initial design
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;
use work.uber_pkg.all;

entity blk_dma is
port(
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  -- DMA OCP Interface
  dma_channels_m2s : in    adb3_ocp_m2sT_array(0 to MAX_DMA_CHANNELS-1);
  dma_channels_s2m : out   adb3_ocp_s2mT_array(0 to MAX_DMA_CHANNELS-1);
  -- BRAM Direct Slave OCP Interface
  ds_bram_m2s      : in    adb3_ocp_m2sT;
  ds_bram_s2m      : out   adb3_ocp_s2mT);
end;


architecture synth of blk_dma is

  constant MUX_CHANNELS : natural := DMA_CHANNELS + 1;

  component bram_single_wrap
  generic(
    output_reg : integer := 0);
  port(
    -- Port Interface
    rst  : in    std_logic;
    clk  : in    std_logic;
    addr : in    std_logic_vector(14 downto 0);
    we   : in    std_logic;
    di   : in    std_logic;
    do   : out   std_logic);
  end component;

  -- OCP mux signals
  signal ocp_m2s       : adb3_ocp_m2sT_array(0 to MUX_CHANNELS-1);
  signal ocp_s2m       : adb3_ocp_s2mT_array(0 to MUX_CHANNELS-1);
  signal mux_status_w  : std_logic_vector(MUX_CHANNELS-1 downto 0);
  signal mux_status_r  : std_logic_vector(MUX_CHANNELS-1 downto 0);
  signal ocp_m2s_memif : adb3_ocp_m2sT;
  signal ocp_s2m_memif : adb3_ocp_s2mT;

  -- Instantiate 32K x 128 Block RAM
  constant DATA_WIDTH : natural := 128;
  constant WE_WIDTH   : natural := DATA_WIDTH/8;
  constant ADDR_WIDTH : natural := ds_vec_width(BRAM_ADDR_RANGE.addr_mask);

  -- BRAM interface
  signal bram_a  : std_logic_vector(ADDR_WIDTH-1 downto 0);
  signal bram_d  : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal bram_q  : std_logic_vector(DATA_WIDTH-1 downto 0);
  signal bram_we : std_logic_vector(WE_WIDTH-1 downto 0);

begin

  ------------------------------
  -- Multiplex OCP interfaces --
  ------------------------------

  -- connect DMA channels to mux
  connect_dma_g : for n in 0 to DMA_CHANNELS-1 generate
    ocp_m2s(n)          <= dma_channels_m2s(n);
    dma_channels_s2m(n) <= ocp_s2m(n);
  end generate;

  -- Connect Direct Slave channel to mux
  ocp_m2s(DMA_CHANNELS) <= ds_bram_m2s;
  ds_bram_s2m           <= ocp_s2m(DMA_CHANNELS);


  adb3_ocp_mux_i : adb3_ocp_mux
  generic map(
    mux_inputs => MUX_CHANNELS)   -- natural := 2
  port map(
    ocp_rst       => ocp_rst,        -- in    std_logic
    ocp_clk       => ocp_clk,        -- in    std_logic
    -- Upstream OCP interfaces
    ocp_m2s       => ocp_m2s,        -- in    adb3_ocp_m2sT_array(0 to mux_inputs-1)
    ocp_s2m       => ocp_s2m,        -- out   adb3_ocp_s2mT_array(0 to mux_inputs-1)
    -- Downstream OCP interface
    ocp_m2s_memif => ocp_m2s_memif,  -- out   adb3_ocp_m2sT
    ocp_s2m_memif => ocp_s2m_memif,  -- in    adb3_ocp_s2mT
    -- Status
    status_w      => mux_status_w,   -- out   std_logic_vector(mux_inputs-1 downto 0)
    status_r      => mux_status_r);  -- out   std_logic_vector(mux_inputs-1 downto 0))

  -------------------------
  -- Muxed OCP interface --
  -------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => ADDR_WIDTH, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => DATA_WIDTH, -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 2)          -- natural := 1
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,        -- in    std_logic
    ocp_clk => ocp_clk,        -- in    std_logic
    ocp_m2s => ocp_m2s_memif,  -- in    adb3_ocp_m2sT
    ocp_s2m => ocp_s2m_memif,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => bram_d,      -- out   std_logic_vector(data_width-1 downto 0)
    q       => bram_q,      -- in    std_logic_vector(data_width-1 downto 0)
    a       => bram_a,      -- out   std_logic_vector(addr_width-1 downto 0)
    w       => open,        -- out   std_logic
    r       => open,        -- out   std_logic
    we      => bram_we);    -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Generate 32k x 128 bits single port BRAM
  bram_bytes_loop_g : for m in 0 to 15 generate
    bram_bits_loop_g : for n in 0 to 7 generate
      bram_single_wrap_i : bram_single_wrap
      generic map(
        output_reg => 1) -- integer := 0);
      port map(
        rst  => ocp_rst,                       -- in    std_logic
        clk  => ocp_clk,                       -- in    std_logic
        addr => bram_a(ADDR_WIDTH-1 downto 4), -- in    std_logic_vector(14 downto 0)
        we   => bram_we(m),                    -- in    std_logic
        di   => bram_d(8*m+n),                 -- in    std_logic
        do   => bram_q(8*m+n));                -- out   std_logic)
    end generate;
  end generate;

end;