--==============================================================================
--
--     File Name: blk_ds_mem_reg.vhd
--       Version: 0.4
--          Date: 29 FSepb 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: On-board memory register interface
--
--   Description: On-board memory register interface
--                Connected to Direct Slave OCP interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 21 Apr 2010: Initial design
--                0.2, TW,  09 Feb 2011:
--                  - Now references adb3_ocp_comp package instead of
--                    adb3_target_pkg.
--                  - Removed some unnecessary resets.
--                0.3, KDC, 16 Feb 2011:
--                  - Changed range of addr_width to match OCP profile
--                0.4, KDC, 29 Sep 2011:
--                  - Changed page_addr_width_ddr3 to page_addr_width
--                  - Changed ds_page_ddr3 to ds_page
--                  - Changed DDR3_PAGE_WIDTH_NIBBLE to PAGE_WIDTH_NIBBLE
--                  - Added MEM_APP_ADDR_WIDTH_BYTE and MEM_BYTE_ADDR_WIDTH_BYTE
--                    for each memory bank
--                  - Changed offset_array, length_array, and err_addr_array to
--                    32-bit vectors
--                  - Changed RAM_BANK0_DDR3_ERR_REG_ADDR to RAM_BANK0_IF_ERR_REG_ADDR
--
--==============================================================================


library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_comp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

entity blk_ds_mem_reg is
generic(
  addr_width      : natural range 1 to ADB3_OCP_ADDR_WIDTH;
  bank_addr_width : natural;
  page_addr_width : natural);
port(
  -- OCP interface
  ocp_rst          : in    std_logic;
  ocp_clk          : in    std_logic;
  direct_slave_m2s : in    adb3_ocp_m2sT;
  direct_slave_s2m : out   adb3_ocp_s2mT;
  -- DS RAM interface
  ds_bank          : out   std_logic_vector(bank_addr_width-1 downto 0);
  ds_page          : out   std_logic_vector(page_addr_width-1 downto 0);
  -- Memory interface status
  mem_if_stat      : in    mem_if_stat_array_t;
  mem_if_err       : in    mem_if_err_array_t;
  -- Memory application control
  mem_app_go       : out   mem_app_go_array_t;
  mem_app_offset   : out   mem_app_offset_array_t;
  mem_app_length   : out   mem_app_length_array_t;
  -- Memory application status
  mem_app_done     : in    mem_app_done_array_t;
  mem_app_err      : in    mem_app_err_array_t;
  mem_app_err_ph   : in    mem_app_err_ph_array_t;
  mem_app_err_addr : in    mem_app_err_addr_array_t;
  -- Debug
  ram_mux_err_info : in    mem_if_debug_array_t;
  mem_if_err_info  : in    mem_if_debug_array_t);
end;


architecture syn of blk_ds_mem_reg is

  constant REG_WIDTH : natural := 32;
  constant WE_WIDTH  : natural := REG_WIDTH/8;

  constant BANK_WIDTH_NIBBLE : std_logic_vector(3 downto 0) := std_logic_vector(to_unsigned(bank_addr_width,4));
  constant PAGE_WIDTH_NIBBLE : std_logic_vector(3 downto 0) := std_logic_vector(to_unsigned(page_addr_width,4));

  type reg_array_t is array (natural range <>) of std_logic_vector(REG_WIDTH-1 downto 0);

  -- Register interface signals
  signal ds_d  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_q  : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_a  : std_logic_vector(addr_width-1 downto 0);
  signal ds_w  : std_logic;
  signal ds_r  : std_logic;
  signal ds_we : std_logic_vector(WE_WIDTH-1 downto 0);

  -- RAM control/status
  signal ds_bank_reg : std_logic_vector(REG_WIDTH-1 downto 0);
  signal ds_page_reg : std_logic_vector(REG_WIDTH-1 downto 0);

  signal ctrl_array     : reg_array_t(0 to MEM_BANKS-1);
  signal offset_array   : reg_array_t(0 to MEM_BANKS-1);
  signal length_array   : reg_array_t(0 to MEM_BANKS-1);
  signal info_array     : reg_array_t(0 to MEM_BANKS-1);
  signal stat_array     : reg_array_t(0 to MEM_BANKS-1);
  signal err_addr_array : reg_array_t(0 to MEM_BANKS-1);

begin

  --------------------------------
  -- Direct Slave OCP interface --
  --------------------------------

  adb3_ocp_simple_bus_if_i : adb3_ocp_simple_bus_if
  generic map(
    addr_width   => addr_width, -- natural range 1 to ADB3_OCP_ADDR_WIDTH
    data_width   => REG_WIDTH,  -- natural range 8 to ADB3_OCP_DATA_WIDTH
    read_latency => 1)          -- natural := 1
  port map(
    -- OCP Interface
    ocp_rst => ocp_rst,           -- in    std_logic
    ocp_clk => ocp_clk,           -- in    std_logic
    ocp_m2s => direct_slave_m2s,  -- in    adb3_ocp_m2sT
    ocp_s2m => direct_slave_s2m,  -- out   adb3_ocp_s2mT
    -- Simple bus interface
    d       => ds_d,        -- out   std_logic_vector(data_width-1 downto 0)
    q       => ds_q,        -- in    std_logic_vector(data_width-1 downto 0)
    a       => ds_a,        -- out   std_logic_vector(addr_width-1 downto 0)
    w       => ds_w,        -- out   std_logic
    r       => ds_r,        -- out   std_logic
    we      => ds_we);      -- out   std_logic_vector((data_width/8)-1 downto 0))


  -- Update write registers
  wr_regs_p : process(ocp_rst,ocp_clk)
    variable u,l : natural;
  begin
    if (ocp_rst = '1') then
      ds_bank_reg  <= (others => '0');
      ds_page_reg  <= (others => '0');
      ctrl_array   <= (others => (others => '0'));
      length_array <= (others => (others => '0'));
      offset_array <= (others => (others => '0'));
    elsif rising_edge(ocp_clk) then
      for i in 0 to WE_WIDTH-1 loop
        u := 8*(i+1)-1;
        l := 8*i;
        if (ds_we(i)='1') then
          -- Direct slave window select registers
          if (ds_a = RAM_DS_BANK_REG_ADDR(addr_width-1 downto 0)) then
            ds_bank_reg(u downto l) <= ds_d(u downto l);
          end if;
          if (ds_a = RAM_DS_PAGE_REG_ADDR(addr_width-1 downto 0)) then
            ds_page_reg(u downto l) <= ds_d(u downto l);
          end if;
          -- Registers for each bank of on-board memory
          for i in 0 to MEM_BANKS-1 loop
            if (ds_a = RAM_BANK0_CTRL_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
              ctrl_array(i)(u downto l) <= ds_d(u downto l);
            end if;
            if (ds_a = RAM_BANK0_OFFSET_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
              offset_array(i)(u downto l) <= ds_d(u downto l);
            end if;
            if (ds_a = RAM_BANK0_LENGTH_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
              length_array(i)(u downto l) <= ds_d(u downto l);
            end if;
          end loop;
        else
          for i in 0 to MEM_BANKS-1 loop
            ctrl_array(i)(u downto l) <= (others => '0');
          end loop;
        end if;
      end loop;
    end if;
  end process;

  -- DS address window bank and page select registers
  -- DS Bank address 000      <=> BRAM
  -- DS Bank address 001..100 <=> RAM banks 0..3
  ds_bank <= ds_bank_reg(bank_addr_width-1 downto 0) + 1;
  ds_page <= ds_page_reg(page_addr_width-1 downto 0);


  mem_banks_g : for n in 0 to MEM_BANKS-1 generate
    constant MEM_APP_ADDR_WIDTH_BYTE  : std_logic_vector(7 downto 0) := std_logic_vector(to_unsigned(MEM_BYTE_ADDR_WIDTH_ARRAY(n)-4,8));
    constant MEM_BYTE_ADDR_WIDTH_BYTE : std_logic_vector(7 downto 0) := std_logic_vector(to_unsigned(MEM_BYTE_ADDR_WIDTH_ARRAY(n),8));
  begin
    -- Memory interface/app control
    mem_app_go(n) <= ctrl_array(n)(8);

    -- Memory app offset (common to all banks)
    mem_app_offset(n) <= offset_array(n);

    -- Memory app length (common to all banks)
    mem_app_length(n) <= length_array(n);

    -- Memory bank info registers
    info_array(n)(7 downto 0)   <= MEM_BYTE_ADDR_WIDTH_BYTE;
    info_array(n)(15 downto 8)  <= MEM_APP_ADDR_WIDTH_BYTE;
    info_array(n)(23 downto 16) <= std_logic_vector(to_unsigned(ADB3_OCP_DATA_WIDTH/8,8));
    info_array(n)(27 downto 24) <= PAGE_WIDTH_NIBBLE;
    info_array(n)(31 downto 28) <= BANK_WIDTH_NIBBLE;

    -- Memory bank/app status registers
    stat_array(n)(3 downto 0)   <= mem_if_stat(n);
    stat_array(n)(11 downto 8)  <= mem_if_err(n);
    stat_array(n)(16)           <= mem_app_done(n);
    stat_array(n)(22 downto 20) <= mem_app_err_ph(n);
    stat_array(n)(23)           <= mem_app_err(n);
    stat_array(n)(31 downto 28) <= std_logic_vector(to_unsigned(n,4));

    -- Memory app error address
    err_addr_array(n) <= mem_app_err_addr(n);
  end generate;


  -- Multiplex read registers
  rd_regs_p : process(ocp_clk)
    variable ds_q_var : std_logic_vector(REG_WIDTH-1 downto 0);
  begin
    if rising_edge(ocp_clk) then
      ds_q_var := (others => '-');

      -- Direct slave window select registers
      if (ds_a = RAM_DS_BANK_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := ds_bank_reg;
      end if;
      if (ds_a = RAM_DS_PAGE_REG_ADDR(addr_width-1 downto 0)) then
        ds_q_var := ds_page_reg;
      end if;

      -- Registers for each bank of on-board memory
      for i in 0 to MEM_BANKS-1 loop
        if (ds_a = RAM_BANK0_OFFSET_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := offset_array(i);
        end if;
        if (ds_a = RAM_BANK0_LENGTH_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := length_array(i);
        end if;
        if (ds_a = RAM_BANK0_INFO_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := info_array(i);
        end if;
        if (ds_a = RAM_BANK0_STAT_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := stat_array(i);
        end if;
        if (ds_a = RAM_BANK0_ERR_ADDR_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := err_addr_array(i);
        end if;
        if (ds_a = RAM_BANK0_MUX_ERR_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := ram_mux_err_info(i);
        end if;
        if (ds_a = RAM_BANK0_IF_ERR_REG_ADDR(addr_width-1 downto 0) + (i*32)) then
          ds_q_var := mem_if_err_info(i);
        end if;
      end loop;

      ds_q <= ds_q_var;
    end if;
  end process;



end;