--==============================================================================
--
--     File Name: blk_mem_app.vhd
--       Version: 0.2
--          Date: 29 Sep 2011
--         Model: Synthesisable
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: On-board memory user code
--
--   Description: User code related to on-board memory
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 29 Oct 2010: Initial design
--                0.2, KDC, 29 Sep 2011:
--                       - Added bank constant MEM_16_BYTE_ADDR_WIDTH
--                       - Made app_offset, app_length, and app_err_addr 32-bit
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;

entity blk_mem_app is
port(
  -- OCP interface
  ocp_rst      : in    std_logic;
  ocp_clk      : in    std_logic;
  app_m2s      : out   adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
  app_s2m      : in    adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
  -- Control interface
  mem_if_rdy   : in    mem_if_rdy_array_t;      -- From DDR3 clock domain
  app_go       : in    mem_app_go_array_t;      -- From register clock domain
  app_offset   : in    mem_app_offset_array_t;  -- From register clock domain
  app_length   : in    mem_app_length_array_t;  -- From register clock domain
  -- Status interface
  app_done     : out   mem_app_done_array_t;
  app_err      : out   mem_app_err_array_t;
  app_err_addr : out   mem_app_err_addr_array_t;
  app_err_ph   : out   mem_app_err_ph_array_t);
end entity;

architecture syn of blk_mem_app is

  component blk_mem_test is
  generic(
    a_width       : in    natural;                                          -- Number of (logical) address bits in memory port
    d_width       : in    natural;                                          -- Number of logical bits in a memory port word
    rd_width      : in    natural;                                          -- Number of physical data pins on memory bank
    tag_base      : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);  -- Tag base value
    tag_incr      : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0);  -- Tag value increment
    tag_mask      : in    std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0)); -- Which tag bits to check
  port(
    -- OCP Interface, to/from memory interface
    ocp_rst       : in    std_logic;
    ocp_clk       : in    std_logic;
    ocp_m2s       : out   adb3_ocp_m2sT;
    ocp_s2m       : in    adb3_ocp_s2mT;
    -- Control interface
    go            : in    std_logic;                              -- Pulse high for at least one cycle to initiate test
    offset        : in    std_logic_vector(a_width - 1 downto 0); -- Starting word address of test
    length        : in    std_logic_vector(a_width - 1 downto 0); -- Number of words to test, minus 1
    done          : out   std_logic;                              -- High indicates finished or idle
    error         : out   std_logic;                              -- Qualified by 'done', high indicates at least one error ocurred
    eaddr         : out   std_logic_vector(a_width - 1 downto 0); -- Qualified by 'done and error', gives word address of first error
    ephase        : out   std_logic_vector(2 downto 0));          -- Qualified by 'done and error', gives phase in which first error occurred
  end component;

  -- Bank 0 uses tags  0 - 15
  -- bank 1 uses tags 16 - 31
  -- bank 2 uses tags 32 - 47
  -- etc.
  constant tag_base : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0) := (others => '0');
  constant tag_incr : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0) := (0 => '1', others => '0');
  constant tag_mask : std_logic_vector(ADB3_OCP_TAG_WIDTH-1 downto 0) := (3 downto 0 => '0', others => '1');

  signal go_d1 : mem_app_go_array_t := (others => '0');
  signal go_d2 : mem_app_go_array_t := (others => '0');
  signal go_d3 : mem_app_go_array_t := (others => '0');
  signal go    : mem_app_go_array_t := (others => '0');

begin

  mem_test_g : for n in 0 to MEM_BANKS-1 generate
    constant MEM_16_BYTE_ADDR_WIDTH : natural := MEM_BYTE_ADDR_WIDTH_ARRAY(n)-4;
  begin
    -- Re-time app_go and mem_if_rdy into app clock domain
    -- Generate go on rising edge
    go_retime_p : process(ocp_clk)
    begin
      if rising_edge(ocp_clk) then
        go_d1(n) <= mem_if_rdy(n) and app_go(n);
        go_d2(n) <= go_d1(n);
        go_d3(n) <= go_d2(n);
        go(n)    <= go_d2(n) and not(go_d3(n));
      end if;
    end process;

    blk_mem_test_i : blk_mem_test
    generic map(
      a_width  => MEM_16_BYTE_ADDR_WIDTH,
      d_width  => ADB3_OCP_DATA_WIDTH,
      rd_width => 32,
      tag_base => tag_base + 16 * n,
      tag_incr => tag_incr,
      tag_mask => tag_mask)
    port map(
      -- Clock interface
      ocp_rst => ocp_rst,
      ocp_clk => ocp_clk,
      ocp_m2s => app_m2s(n),
      ocp_s2m => app_s2m(n),
      -- Control interface
      go      => go(n),
      offset  => app_offset(n)(MEM_16_BYTE_ADDR_WIDTH-1 downto 0),
      length  => app_length(n)(MEM_16_BYTE_ADDR_WIDTH-1 downto 0),
      done    => app_done(n),
      error   => app_err(n),
      eaddr   => app_err_addr(n)(MEM_16_BYTE_ADDR_WIDTH-1 downto 0),
      ephase  => app_err_ph(n));

    app_err_addr(n)(31 downto MEM_16_BYTE_ADDR_WIDTH) <= (others => '0');
  end generate;

end;
