--==============================================================================
--
--     File Name: test_uber_dma_1ch_nb.vhd
--       Version: 0.3
--          Date: 29 Aug 2011
--         Model: Simulation only
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: DMA
--
--   Description: Provides bridge DMA interface (single channel)
--
--   Limitations: Single DMA channel access
--
--         Notes: Non-blocking DMA test
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2010-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 03 Dec 2010: Initial design
--                0.2, KDC, 25 Jul 2011:
--                     - Added signal dma_ch_m2s_int to drive unused DMA channels
--                0.3, KDC, 29 Aug 2011:
--                     - Added adb3_ocp_tb_pkg package
--                     - Changed adb3_target_sim_* to adb3_ocp_sim_*
--                     - Changed mptl_online to start_test
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_tb_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.uber_tb_pkg.all;

entity test_uber_dma is
port(
  -- OCP Interface
  ocp_clk          : in    std_logic;
  dma_channels_m2s : out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  dma_channels_s2m : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  -- Test interface
  start_test       : in    std_logic;
  dma_comp         : out   dma_comp_t;
  dma_pass         : out   dma_pass_t);
end test_uber_dma;

architecture sim of test_uber_dma is

  signal dma_ch_m2s_int : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_M2S);

  signal dma_m2s_cmd  : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal dma_s2m_cmd  : adb3_ocp_s2mT;
  signal dma_m2s_resp : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal dma_s2m_resp : adb3_ocp_s2mT;

begin

  -- Connect dma inputs
  dma_s2m_cmd  <= dma_channels_s2m(DMA_SINGLE_CHANNEL);
  dma_s2m_resp <= dma_channels_s2m(DMA_SINGLE_CHANNEL);


  -- Generate DMA commands to target
  dma_channel_cmd_p : process
    variable addr      : adb3_ocp_addr_s;
    variable data      : byte_vector_t(0 to 511);
    variable be        : byte_enable_t(0 to 511);
    variable xfer      : natural;
    variable val32     : std_logic_vector(31 downto 0);
    variable tx_cnt    : natural;
    variable rx_cnt    : natural;
    variable tag       : natural;
    variable remaining : natural;
    variable chunk     : natural;
  begin
    dma_comp.dma_write_cmd_complete <= false;
    dma_pass.dma_write_cmd_passed   <= true;
    dma_comp.dma_read_cmd_complete  <= false;
    dma_pass.dma_read_cmd_passed    <= true;

    wait until (start_test = '1');  -- OCP commands ignored until MPTL/PCIe is online

    report "DMA write process started (Base address = 0x" & conv_string_hex(DMA_ADDR_WR(DMA_ADDR_WIDTH-1 downto 0)) & ")";

    addr      := DMA_ADDR_WR;
    be        := (others => '1');
    tx_cnt    := 0;
    val32     := X"beef0000";
    tag       := 0;
    remaining := DMA_SIZE;

    while (remaining /= 0) loop
      chunk := DMA_BL_WRITE * ADB3_OCP_BE_WIDTH;
      if (chunk > remaining) then
        chunk := remaining;
      end if;

      for i in 0 to ((chunk/4)-1) loop
        data(4*i to 4*(i+1)-1) := conv_byte_vector(val32 + tx_cnt + i);
      end loop;

      adb3_ocp_sim_write(addr,tag,be(0 to chunk-1),data(0 to chunk-1),xfer,ocp_clk,dma_m2s_cmd,dma_s2m_cmd);
--      report "Wrote data chunk (Tag: " & integer'image(tag) & ", Chunk: " & integer'image(chunk) &
--             ") Transferred Bytes: " & integer'image(xfer) & " on DMA channel 0";

      remaining  := remaining - xfer;
      tx_cnt     := tx_cnt + (xfer/4);
      addr       := addr + chunk;
      tag        := tag + 1;
    end loop;

    report "DMA write process completed";
    report integer'image(tx_cnt*4) & " bytes transferred.";
    dma_comp.dma_write_cmd_complete <= true;

    if (tx_cnt*4 /= DMA_SIZE) then
      dma_pass.dma_write_cmd_passed <= false;
    end if;


    report "DMA read command process started (Base address = 0x" & conv_string_hex(DMA_ADDR_RD(DMA_ADDR_WIDTH-1 downto 0)) & ")";

    addr      := DMA_ADDR_RD;
    rx_cnt    := 0;
    tag       := 0;
    remaining := DMA_SIZE;

    while remaining /= 0 loop
      chunk := DMA_BL_READ * ADB3_OCP_BE_WIDTH;
      if (chunk > remaining) then
        chunk := remaining;
      end if;

      adb3_ocp_sim_read_cmd(addr,tag,data(0 to chunk-1),xfer,ocp_clk,dma_m2s_cmd,dma_s2m_cmd);
      remaining := remaining - xfer;

      rx_cnt := rx_cnt + (xfer/4);
      addr   := addr + chunk;
      tag    := tag + 1;
    end loop;

    report "DMA read command process completed";
    dma_comp.dma_read_cmd_complete <= true;

    if (rx_cnt*4 /= DMA_SIZE) then
      dma_pass.dma_read_cmd_passed <= false;
    end if;

    wait;
  end process;


  -- Read DMA response data from target
  dma_channel_resp_p : process
    variable data      : byte_vector_t(0 to 511);
    variable xfer      : natural;
    variable pat32     : std_logic_vector(31 downto 0);
    variable exp32     : std_logic_vector(31 downto 0);
    variable val32     : std_logic_vector(31 downto 0);
    variable rx_cnt    : natural;
    variable tag       : natural;
    variable exp_tag   : natural;
    variable remaining : natural;
    variable chunk     : natural;
    variable num_err   : natural;
  begin
    dma_comp.dma_read_resp_complete <= false;
    dma_pass.dma_read_resp_passed   <= true;

    wait until (start_test = '1');  -- OCP response ignored until MPTL/PCIe is online

    report "DMA read response data process started";

    exp_tag   := 0;
    rx_cnt    := 0;
    pat32     := X"beef0000";
    num_err   := 0;
    remaining := DMA_SIZE;

    while remaining /= 0 loop
      chunk := DMA_BL_READ * ADB3_OCP_BE_WIDTH;
      if (chunk > remaining) then
        chunk := remaining;
      end if;

      adb3_ocp_sim_read_resp(data(0 to chunk-1),tag,xfer,ocp_clk,dma_m2s_resp,dma_s2m_resp);
--      report "Read data chunk (Tag: " & integer'image(tag) & ", Chunk: " & integer'image(chunk) &
--             ") Transferred Bytes: " & integer'image(xfer) & " on DMA channel 1";
      remaining := remaining - xfer;

      for i in 0 to (xfer/4-1) loop
        val32 := data(4*i+3) & data(4*i+2) & data(4*i+1) & data(4*i);
        exp32 := pat32 + rx_cnt + i;
        if (val32 /= exp32) then
          assert false
            report "(test_uber_dma_1ch_nb) Verification error: expected data = " & conv_string(exp32(15 downto 0)) &
                   " actual data = " & conv_string(val32(15 downto 0))
            severity error;
          num_err := num_err + 1;
        end if;
      end loop;

      if (exp_tag /= tag) then
        report "(test_uber_dma_1ch_nb) Verification error: expected tag = " & integer'image(exp_tag) &
               " actual tag = " & integer'image(tag)
        severity error;
        num_err := num_err + 1;
      end if;

      rx_cnt  := rx_cnt + (xfer/4);
      exp_tag := exp_tag + 1;
    end loop;

    report "DMA read response data process completed";
    report integer'image(rx_cnt*4) & " bytes transferred with " & integer'image(num_err) & " error(s)";
    dma_comp.dma_read_resp_complete <= true;

    if (rx_cnt*4 = DMA_SIZE)and(num_err = 0) then
      assert false report "Test DMA completed: PASSED." severity note;
    else
      assert false report "Test DMA completed: FAILED." severity failure;
      dma_pass.dma_read_resp_passed <= false;
    end if;

    wait;
  end process;


  -- Connect DMA command outputs
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).Cmd         <= dma_m2s_cmd.Cmd;
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).Addr        <= dma_m2s_cmd.Addr;
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).BurstLength <= dma_m2s_cmd.BurstLength;
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).Data        <= dma_m2s_cmd.Data;
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).DataByteEn  <= dma_m2s_cmd.DataByteEn;
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).DataValid   <= dma_m2s_cmd.DataValid;
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).Tag         <= dma_m2s_cmd.Tag;

  -- Connect DMA response outputs
  dma_ch_m2s_int(DMA_SINGLE_CHANNEL).RespAccept  <= dma_m2s_resp.RespAccept;

  -- Connect all DMA channels to output
  -- Unused channels have been initialised
  dma_channels_m2s <= dma_ch_m2s_int;

end;