--==============================================================================
--
--     File Name: test_uber_ds.vhd
--       Version: 0.6
--          Date: 25 Aug 2011
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: Direct Slave
--
--   Description: Provides bridge Direct Slave interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 17 Mar 2009: Initial design
--                0.2, KDC, 05 May 2011:
--                     - Added DO_RAM_TEST constant to control execution of
--                       on-board ram tests
--                     - Added setting of test complete signals if sections
--                       of test are not executed
--                     - Renumbered OCP tags
--                0.3, KDC, 17 May 2011:
--                     - Split variables into register and memory groups
--                       Register for 32-bit accesses, memory for variable
--                       size accesses
--                0.4, KDC, 23 Jun 2011:
--                     - Added ADM-XRC-6TGE option to clock test
--                     - Expanded DS test using on-board memories
--                     - Added DS bank/page reg reads to ensure value changed
--                     - Re-numbered tags
--                0.5, KDC, 25 Jul 2011:
--                     - Moved constants to uber_tb_pkg
--                     - Changed reg_addr to use new register address constants
--                     - Use new constant MGT113_CLK0_FREQ for expected clk freq
--                     - Added new constant INT_RAM_TEST_BANK for internal test
--                0.6, KDC, 25 Aug 2011:
--                     - Changed shared variable linti_l_edge to int_edge
--                     - Changed MEM_VALID to MEM_BANKS > 0
--                     - Added adb3_ocp_tb_pkg package
--                     - Changed adb3_target_sim_* to adb3_ocp_sim_*
--                     - Changed adb3_target_wait_cycles to adb3_ocp_sim_wait_cycles
--                     - Replaced MGT113_CLK0 clock test with TEST_MGT_CLK and
--                       TEST_CUS_CLK tests.
--                     - Changed PN4_GPIO_VALID to PN4_GPIO_WIDTH > 0
--                     - Changed PN6_GPIO_VALID to PN6_GPIO_WIDTH > 0
--                     - Changed XRM_GPIO_VALID to XRM_GPIO_WIDTH = 146
--                     - Change mptl_online to start_test
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.numeric_std.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_tb_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.uber_pkg.all;
use work.today_pkg.all;
use work.uber_tb_pkg.all;

entity test_uber_ds is
port(
  -- OCP Interface
  ocp_clk          : in    std_logic;
  direct_slave_m2s : out   adb3_ocp_m2sT;
  direct_slave_s2m : in    adb3_ocp_s2mT;
  -- Test interface
  start_test       : in    std_logic;
  ds_comp          : out   ds_comp_t;
  ds_pass          : out   ds_pass_t);
end test_uber_ds;

architecture sim of test_uber_ds is

  signal ds_m2s : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal ds_s2m : adb3_ocp_s2mT;

begin
  ----------------------------
  -- Direct Slave interface --
  ----------------------------

  -- Connect direct slave input
  ds_s2m <= direct_slave_s2m;

  direct_slave_p : process
    -- 4-byte register interface
    -- reg_addr is byte address and must be 4-byte aligned
    variable reg_data_wr     : byte_vector_t(0 to 3);
    variable reg_data_rd     : byte_vector_t(0 to 3);
    variable reg_be          : byte_enable_t(0 to 3);
    variable reg_addr        : adb3_ocp_addr_s;
    variable reg_xfer        : natural;
    -- n-byte memory interface
    -- mem_addr is byte address and must be 16-byte aligned
    variable mem_data_wr     : byte_vector_t(0 to 127);
    variable mem_data_rd     : byte_vector_t(0 to 127);
    variable mem_be          : byte_enable_t(0 to 127);
    variable mem_addr        : adb3_ocp_addr_s;
    variable mem_xfer        : natural;

    variable master_idle     : std_logic;
    variable master_timeout  : natural := 64;
    variable master_cnt      : natural := 0;
    variable master_passed   : boolean := true;
    variable clock_running   : std_logic;
    variable ck_passed       : boolean := true;
    variable io_data_ls      : std_logic_vector(31 downto 0);
    variable io_data_ms      : std_logic_vector(31 downto 0);
    variable io_data_x2      : std_logic_vector(63 downto 0);
    variable f_io_passed     : boolean := true;
    variable r_io_passed     : boolean := true;
    variable int_set         : std_logic_vector(31 downto 0);
    variable int_stat        : std_logic_vector(31 downto 0);
    variable int_passed      : boolean := true;
    variable info_passed     : boolean := true;
    variable bram_passed     : boolean := true;
    variable ram_passed      : boolean := true;
    variable ram_init        : std_logic;
    variable ram_done        : std_logic;
    
    -- Factor to convert a clock counter value into MHz; because clock
    -- measurement period is 4 microseconds, and frequencies are displayed in MHz
    constant clk_freq_factor : real := 0.25; 
    -- Expected and actual clock frequencies
    variable clk_freq_expect : real;
    variable clk_freq_actual : real;

  begin
    wait until (start_test = '1');  -- OCP commands ignored until MPTL/PCIE online
    ds_comp.simple_complete  <= false;
    ds_comp.master_complete  <= false;
    ds_comp.clock_complete   <= false;
    ds_comp.frontio_complete <= false;
    ds_comp.reario_complete  <= false;
    ds_comp.int_complete     <= false;
    ds_comp.info_complete    <= false;
    ds_comp.bram_complete    <= false;
    ds_comp.ram_complete     <= false;

    ds_pass.simple_passed    <= true;
    ds_pass.master_passed    <= true;
    ds_pass.clock_passed     <= true;
    ds_pass.frontio_passed   <= true;
    ds_pass.reario_passed    <= true;
    ds_pass.int_passed       <= true;
    ds_pass.info_passed      <= true;
    ds_pass.bram_passed      <= true;
    ds_pass.ram_passed       <= true;


    ---------------------------------------------
    -- Simple Test (32-bit register interface) --
    ---------------------------------------------
    reg_be := conv_byte_enable("1111");

    -- Write a pattern to register
    reg_addr := SIMPLE_WDATA_REG_ADDR;
    reg_data_wr := conv_byte_vector(X"cafeface");
    adb3_ocp_sim_write_reg32(reg_addr,10,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote simple WDATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr) &
           " with enable 0b" & conv_string(reg_be) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read back nibble reversed pattern from register
    reg_addr := SIMPLE_RDATA_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,11,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read simple RDATA " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    ds_comp.simple_complete <= true;
    if (reg_data_rd = conv_byte_vector(X"ecafefac")) then
      assert false report "Test Simple completed: PASSED." severity note;
    else
      assert false report "Test Simple completed: FAILED." severity error;
      ds_pass.simple_passed <= false;
    end if;


    if (MASTER_USE = MASTER_ON) then
      ---------------------------------------------
      -- Master Test (32-bit register interface) --
      ---------------------------------------------
      reg_be := conv_byte_enable("1111");

      -- Read OOR
      report "Read OOR Address (will generate warning)";
      reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
      adb3_ocp_sim_read(reg_addr,12,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr 0 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"DEADC0DE")) then
        master_passed := false;
      end if;

      -- Write direct master write data
      reg_addr := MASTER_WDATA_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"76543210");  -- 32-bit data
      adb3_ocp_sim_write_reg32(reg_addr,13,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master WDATA " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master configuration (1 = write, 0 = read)
      reg_addr := MASTER_SEL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"0000000" & "0001");  -- bit 0 = write
      adb3_ocp_sim_write_reg32(reg_addr,14,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master address and trigger transfer
      reg_addr := MASTER_ADDR_REG_ADDR;
      reg_data_wr := conv_byte_vector(HOST_MASTER_REG_ADDR(DM_DATA_WIDTH-1 downto 0));
      adb3_ocp_sim_write_reg32(reg_addr,15,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master ADDR " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master configuration (1 = write, 0 = read)
      reg_addr := MASTER_SEL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"0000000" & "0000");  -- bit 0 = read
      adb3_ocp_sim_write_reg32(reg_addr,16,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write direct master address and trigger transfer
      reg_addr := MASTER_ADDR_REG_ADDR;
      reg_data_wr := conv_byte_vector(HOST_MASTER_REG_ADDR(DM_DATA_WIDTH-1 downto 0));
      adb3_ocp_sim_write_reg32(reg_addr,17,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote master ADDR " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      master_idle := '0';
      while ((master_idle = '0')and(master_cnt < master_timeout)) loop
        -- Read direct master status
        reg_addr := MASTER_STAT_REG_ADDR;
        adb3_ocp_sim_read_reg32(reg_addr,18,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        master_idle         := reg_data_rd(0)(0);
        master_cnt          := master_cnt + 1;
      end loop;

      if (master_cnt < master_timeout) then
        report "Direct Master transfer complete (" & std_logic'image(master_idle) & ")";
      else
        report "Direct Master transfer timeout (" & std_logic'image(master_idle) & ")";
        master_passed := false;
      end if;

      -- Read direct master read data
      reg_addr := MASTER_RDATA_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,19,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"76543210")) then
        master_passed := false;
      end if;

      ds_comp.master_complete <= true;
      if (master_passed) then
        assert false report "Test Master completed: PASSED." severity note;
      else
        assert false report "Test Master completed: FAILED." severity error;
        ds_pass.master_passed <= false;
      end if;
    else
      ds_comp.master_complete <= true;
    end if;


    if DO_CLOCK_TEST then
      ------------------------------------------------------------------
      -- Clock Frequency Measurement Test (32-bit register interface) --
      ------------------------------------------------------------------
      reg_be := conv_byte_enable("1111");

      -- Read OOR
      report "Read OOR Address (will generate warning)";
      reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
      adb3_ocp_sim_read(reg_addr,20,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr 0 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"DEADC0DE")) then
        ck_passed := false;
      end if;

      -- Clear all flags
      reg_addr := CLK_READ_CTRL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"80000000");
      adb3_ocp_sim_write_reg32(reg_addr,21,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote Clear All CTRL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Select pll_reg_clk
      reg_addr := CLK_READ_SEL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"000000" & "000" & PLL_REG_CLK_SEL);
      adb3_ocp_sim_write_reg32(reg_addr,22,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PLL_REG_CLK_SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_STAT_REG_ADDR;
        adb3_ocp_sim_read_reg32(reg_addr,23,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_FREQ_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,24,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PLL_REG_CLK_FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      clk_freq_expect := real(PLL_REG_CLK_FREQ);
      clk_freq_actual := real(to_integer(unsigned(conv_vector(reg_data_rd)))) * clk_freq_factor;
      
      report "Expected freq = " & real'image(clk_freq_expect) & " MHz +/-" & real'image(CLK_TEST_DIFF) & " MHz";
      report "Actual   freq = " & real'image(clk_freq_actual) & " MHz";

      -- Check for truncated  and +1 frequency values
      if (clk_freq_actual < (clk_freq_expect-CLK_TEST_DIFF)) or (clk_freq_actual > (clk_freq_expect+CLK_TEST_DIFF)) then
        ck_passed := false;
      end if;

      -- Select pll_pri_clk
      reg_addr := CLK_READ_SEL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"000000" & "000" & PLL_PRI_CLK_SEL);
      adb3_ocp_sim_write_reg32(reg_addr,25,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PLL_PRI_CLK_SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_STAT_REG_ADDR;
        adb3_ocp_sim_read_reg32(reg_addr,26,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running       := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_FREQ_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,27,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PLL_PRI_CLK_FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      clk_freq_expect := real(PLL_PRI_CLK_FREQ);
      clk_freq_actual := real(to_integer(unsigned(conv_vector(reg_data_rd)))) * clk_freq_factor;

      report "Expected freq = " & real'image(clk_freq_expect) & " MHz +/-" & real'image(CLK_TEST_DIFF) & " MHz";
      report "Actual   freq = " & real'image(clk_freq_actual) & " MHz";

      -- Check for truncated  and +1 frequency values
      if (clk_freq_actual < (clk_freq_expect-CLK_TEST_DIFF)) or (clk_freq_actual > (clk_freq_expect+CLK_TEST_DIFF)) then
        ck_passed := false;
      end if;

      -- Select TEST_MGT_CLK clock
      reg_addr := CLK_READ_SEL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"000000" & "000" & TEST_MGT_CLK_SEL);
      adb3_ocp_sim_write_reg32(reg_addr,28,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote TEST_MGT_CLK_SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_STAT_REG_ADDR;
        adb3_ocp_sim_read_reg32(reg_addr,29,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running       := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_FREQ_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,30,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read TEST_MGT_CLK_FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      clk_freq_expect := real(TEST_MGT_CLK_FREQ);
      clk_freq_actual := real(to_integer(unsigned(conv_vector(reg_data_rd)))) * clk_freq_factor;

      report "Expected freq = " & real'image(clk_freq_expect) & " MHz +/-" & real'image(CLK_TEST_DIFF) & " MHz";
      report "Actual   freq = " & real'image(clk_freq_actual) & " MHz";

      -- Check for truncated  and +1 frequency values
      if (clk_freq_actual < (clk_freq_expect-CLK_TEST_DIFF)) or (clk_freq_actual > (clk_freq_expect+CLK_TEST_DIFF)) then
        ck_passed := false;
      end if;

      -- Select TEST_CUS_CLK clock
      reg_addr := CLK_READ_SEL_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"000000" & "000" & TEST_CUS_CLK_SEL);
      adb3_ocp_sim_write_reg32(reg_addr,28,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote TEST_CUS_CLK_SEL " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait until clock register has been updated (msb = 1)
      -- Then read clock frequency
      clock_running := '0';
      while (clock_running = '0') loop
        reg_addr := CLK_READ_STAT_REG_ADDR;
        adb3_ocp_sim_read_reg32(reg_addr,29,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        clock_running       := reg_data_rd(3)(7);
      end loop;

      -- Read frequency data
      reg_addr := CLK_READ_FREQ_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,30,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read TEST_CUS_CLK_FREQ " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      clk_freq_expect := real(TEST_CUS_CLK_FREQ);
      clk_freq_actual := real(to_integer(unsigned(conv_vector(reg_data_rd)))) * clk_freq_factor;

      report "Expected freq = " & real'image(clk_freq_expect) & " MHz +/-" & real'image(CLK_TEST_DIFF) & " MHz";
      report "Actual   freq = " & real'image(clk_freq_actual) & " MHz";

      -- Check for truncated  and +1 frequency values
      if (clk_freq_actual < (clk_freq_expect-CLK_TEST_DIFF)) or (clk_freq_actual > (clk_freq_expect+CLK_TEST_DIFF)) then
        ck_passed := false;
      end if;

      ds_comp.clock_complete <= true;
      if (ck_passed) then
        assert false report "Test Clock Read completed: PASSED." severity note;
      else
        assert false report "Test Clock Read completed: FAILED." severity error;
        ds_pass.clock_passed <= false;
      end if;
    else
      ds_comp.clock_complete <= true;
    end if;


    if (XRM_GPIO_WIDTH = 146) then
      -----------------------------------------------
      -- XRM GPIO Test (32-bit register interface) --
      -----------------------------------------------
      reg_be := conv_byte_enable("1111");

      -- Read OOR
      report "Read OOR Address (will generate warning)";
      reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
      adb3_ocp_sim_read(reg_addr,31,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr 0 " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"DEADC0DE")) then
        f_io_passed := false;
      end if;

      -- Write io data
      io_data_ls := X"76543210";

      reg_addr := XRM_GPIO_DD_DATAO_REG_ADDR;
      reg_data_wr := conv_byte_vector(io_data_ls);
      adb3_ocp_sim_write_reg32(reg_addr,32,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote XRM_GPIO_DD DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write io tri-state enable
      reg_addr := XRM_GPIO_DD_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"00000000");
      adb3_ocp_sim_write_reg32(reg_addr,33,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote XRM_GPIO_DD TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait for data to get to outputs
      adb3_ocp_sim_wait_cycles(50,ocp_clk);

      -- Read io data
      reg_addr := XRM_GPIO_DD_DATAI_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,34,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read XRM_GPIO_DD DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (conv_vector(reg_data_rd) /= io_data_ls) then
        f_io_passed := false;
      end if;

      -- Write io tri-state disable
      reg_addr := XRM_GPIO_DD_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"FFFFFFFF");
      adb3_ocp_sim_write_reg32(reg_addr,35,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote XRM_GPIO_DD TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      ds_comp.frontio_complete <= true;
      if (f_io_passed) then
        assert false report "Test Front IO completed: PASSED." severity note;
      else
        assert false report "Test Front IO completed: FAILED." severity error;
        ds_pass.frontio_passed <= false;
      end if;
    else
      ds_comp.frontio_complete <= true;
    end if;


    if (PN4_GPIO_WIDTH > 0) then
      -----------------------------------------------
      -- PN4 GPIO Test (64-bit register interface) --
      -----------------------------------------------
      reg_be := conv_byte_enable("1111");

      -- Write io data
      io_data_ms := PN4_DATA;
      io_data_ls := not(PN4_DATA);

      reg_addr := PN4_GPIO_P_DATAO_REG_ADDR;
      reg_data_wr := conv_byte_vector(io_data_ms);
      adb3_ocp_sim_write_reg32(reg_addr,36,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_P DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := PN4_GPIO_N_DATAO_REG_ADDR;
      reg_data_wr := conv_byte_vector(io_data_ls);
      adb3_ocp_sim_write_reg32(reg_addr,37,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_N DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Write io enable
      reg_addr := PN4_GPIO_P_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"00000000");
      adb3_ocp_sim_write_reg32(reg_addr,38,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_P TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := PN4_GPIO_N_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"00000000");
      adb3_ocp_sim_write_reg32(reg_addr,39,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_N TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait for data to get to outputs
      adb3_ocp_sim_wait_cycles(50,ocp_clk);

      -- Read io data
      reg_addr := PN4_GPIO_P_DATAI_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,40,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN4_GPIO_P DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (conv_vector(reg_data_rd)(PN4_GPIO_WIDTH-1 downto 0) /= io_data_ms(PN4_GPIO_WIDTH-1 downto 0)) then
        assert false report "Test PN4 P Rear IO completed: FAILED." severity error;
        r_io_passed := false;
      end if;

      reg_addr := PN4_GPIO_N_DATAI_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,41,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN4_GPIO_N DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (conv_vector(reg_data_rd)(PN4_GPIO_WIDTH-1 downto 0) /= io_data_ls(PN4_GPIO_WIDTH-1 downto 0)) then
        assert false report "Test PN4 N Rear IO completed: FAILED." severity error;
        r_io_passed := false;
      end if;


      -- Write io tri-state disable
      reg_addr := PN4_GPIO_P_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"FFFFFFFF");
      adb3_ocp_sim_write_reg32(reg_addr,42,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_P TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := PN4_GPIO_N_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"FFFFFFFF");
      adb3_ocp_sim_write_reg32(reg_addr,43,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN4_GPIO_N TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
    end if;


    if (PN6_GPIO_WIDTH > 0) then
      -----------------------------------------------
      -- PN6 GPIO Test (32-bit register interface) --
      -----------------------------------------------
      reg_be := conv_byte_enable("1111");

      -- Write io data
      io_data_ms := PN6_DATA(63 downto 32);
      io_data_ls := PN6_DATA(31 downto 0);

      reg_addr := PN6_GPIO_MS_DATAO_REG_ADDR;
      reg_data_wr := conv_byte_vector(io_data_ms);
      adb3_ocp_sim_write_reg32(reg_addr,44,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_MS DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := PN6_GPIO_LS_DATAO_REG_ADDR;
      reg_data_wr := conv_byte_vector(io_data_ls);
      adb3_ocp_sim_write_reg32(reg_addr,45,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_LS DATAO " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write io enable
      reg_addr := PN6_GPIO_MS_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"00000000");
      adb3_ocp_sim_write_reg32(reg_addr,46,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_MS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := PN6_GPIO_LS_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"00000000");
      adb3_ocp_sim_write_reg32(reg_addr,47,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_LS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Wait for data to get to outputs
      adb3_ocp_sim_wait_cycles(50,ocp_clk);

      -- Read io data
      reg_addr := PN6_GPIO_MS_DATAI_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,48,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN6_GPIO_MS DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      io_data_x2(63 downto 32) := conv_vector(reg_data_rd);

      reg_addr := PN6_GPIO_LS_DATAI_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,49,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read PN6_GPIO_LS DATAI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      io_data_x2(31 downto 0) := conv_vector(reg_data_rd);

      if (io_data_x2(PN6_GPIO_WIDTH-1 downto 0) /= PN6_DATA(PN6_GPIO_WIDTH-1 downto 0)) then
        assert false report "Test PN6 Rear IO completed: FAILED." severity error;
        r_io_passed := false;
      end if;

      -- Write io tri-state disable
      reg_addr := PN6_GPIO_MS_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"FFFFFFFF");
      adb3_ocp_sim_write_reg32(reg_addr,50,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_MS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := PN6_GPIO_LS_TRI_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"FFFFFFFF");
      adb3_ocp_sim_write_reg32(reg_addr,51,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote PN6_GPIO_LS TRI " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      ds_comp.reario_complete <= true;
      if (r_io_passed) then
        assert false report "Test Rear IO completed: PASSED." severity note;
      else
        assert false report "Test Rear IO completed: FAILED." severity error;
        ds_pass.reario_passed <= false;
      end if;
    else
      ds_comp.reario_complete <= true;
    end if;


    --------------------
    -- Interrupt Test --
    --------------------
    reg_be := conv_byte_enable("1111");

    -- Read OOR
    report "Read OOR Address (will generate warning)";
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_ocp_sim_read(reg_addr,52,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr 0 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd /= conv_byte_vector(X"DEADC0DE")) then
      int_passed := false;
    end if;

    -- Write interrupt MASK
    reg_addr := INT_WMASK_REG_ADDR;
    reg_data_wr := conv_byte_vector(MASK_EN_ALL);
    adb3_ocp_sim_write_reg32(reg_addr,53,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote Interrupt MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr) &
           " with enable 0b" & conv_string(reg_be) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read interrupt MASK
    reg_addr := INT_RMASK_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,54,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Interrupt MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd /= conv_byte_vector(MASK_EN_ALL)) then
      int_passed := false;
    end if;

    -- Write interrupt COUNT
    reg_addr := INT_WCOUNT_REG_ADDR;
    reg_data_wr := conv_byte_vector(X"FFFFFFFF");
    adb3_ocp_sim_write_reg32(reg_addr,55,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote Interrupt COUNT " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_wr) &
           " with enable 0b" & conv_string(reg_be) &
           " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read interrupt COUNT
    reg_addr := INT_RCOUNT_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,56,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Interrupt COUNT " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd /= conv_byte_vector(X"FFFFFFFF")) then
      int_passed := false;
    end if;

    -- Cycle through set-detect-clear for all 32 interrupts
    for i in 0 to (DO_INTERRUPT_NUM-1) loop
      int_set := (others => '0');
      int_set(i) := '1';

      -- Write interrupt SET
      reg_addr := INT_SET_REG_ADDR;
      reg_data_wr := conv_byte_vector(int_set);
      adb3_ocp_sim_write_reg32(reg_addr,57,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--      report "Wrote Interrupt SET " & integer'image(reg_xfer) &
--             " bytes 0x" & conv_string_hex(reg_data_wr) &
--             " with enable 0b" & conv_string(reg_be) &
--             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      loop
        exit when (int_edge = '1');   -- Wait for next edge on interrupt (Shared variable defined in uber_tb_pkg)
        adb3_ocp_sim_wait_cycles(1,ocp_clk);
      end loop;

      -- Interrupt response time of 15 cycles
      -- Unrealistically short for purposes of simulation
      adb3_ocp_sim_wait_cycles(15,ocp_clk);

      -- Check interrupt status register
      reg_addr := INT_STAT_REG_ADDR;
      adb3_ocp_sim_read_reg32(reg_addr,58,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--      report "Read Interrupt STAT " & integer'image(reg_xfer) &
--             " bytes 0x" & conv_string_hex(reg_data_rd) &
--             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      int_stat := conv_vector(reg_data_rd);

      -- Mask the STAT register value just read with the value of the MASK register
      int_stat := int_stat and not(MASK_EN_ALL);

      -- Clear the unmasked interrupts
      reg_addr := INT_CLEAR_REG_ADDR;
      reg_data_wr := conv_byte_vector(int_stat);
      adb3_ocp_sim_write_reg32(reg_addr,59,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Interrupt Handler: Cleared interrupt(s), masked STAT = 0x" & conv_string_hex(reg_data_wr);

      -- Re-arm the interrupt logic
      int_edge := '0';
      reg_addr := INT_ARM_REG_ADDR;
      reg_data_wr := conv_byte_vector(X"deadbeef"); -- actual value does not matter for ARM register
      adb3_ocp_sim_write_reg32(reg_addr,60,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    end loop;

    -- Read interrupt STAT data
    reg_addr := INT_STAT_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,61,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Interrupt STAT " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd /= conv_byte_vector(X"00000000")) then
      int_passed := false;
    end if;

    ds_comp.int_complete <= true;
    if (int_passed) then
      assert false report "Test Interrupt completed: PASSED." severity note;
    else
      assert false report "Test Interrupt completed: FAILED." severity error;
      ds_pass.int_passed <= false;
    end if;


    -------------------------------------------
    -- Info Test (32-bit register interface) --
    -------------------------------------------
    reg_be := conv_byte_enable("1111");

    -- Read OOR
    report "Read OOR Address (will generate warning)";
    reg_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_ocp_sim_read(reg_addr,62,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr 0 " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (reg_data_rd /= conv_byte_vector(X"DEADC0DE")) then
      info_passed := false;
    end if;

    -- Read DATE from register
    reg_addr := INFO_DATE_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,63,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info DATE " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd) /= TODAYS_DATE) then
      info_passed := false;
    end if;

    -- Read TIME from register
    reg_addr := INFO_TIME_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,64,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info TIME " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd) /= TODAYS_TIME) then
      info_passed := false;
    end if;

    -- Read BRAM addr base from register
    reg_addr := INFO_BRAM_BASE_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,65,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info BRAM BASE " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd)(DS_ADDR_WIDTH-1 downto 0) /= BRAM_ADDR_BASE(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read BRAM addr mask from register
    reg_addr := INFO_BRAM_MASK_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,66,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info BRAM MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd)(DS_ADDR_WIDTH-1 downto 0) /= BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read RAM addr base from register
    reg_addr := INFO_RAM_BASE_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,67,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info RAM BASE " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd)(DS_ADDR_WIDTH-1 downto 0) /= RAM_WIN_ADDR_BASE(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read BRAM addr mask from register
    reg_addr := INFO_RAM_MASK_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,68,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info RAM MASK " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd)(DS_ADDR_WIDTH-1 downto 0) /= RAM_WIN_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0)) then
      info_passed := false;
    end if;

    -- Read RAM info from register
    reg_addr := INFO_RAM_INFO_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,69,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info RAM INFO " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd)(3 downto 0) /= std_logic_vector(to_unsigned(MEM_BANKS,4))) then
      info_passed := false;
    end if;

    -- Read SDK_VERISON from register
    reg_addr := INFO_SDK_VER_REG_ADDR;
    adb3_ocp_sim_read_reg32(reg_addr,70,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read Info SDK VERSION " & integer'image(reg_xfer) &
           " bytes 0x" & conv_string_hex(reg_data_rd) &
           " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

    if (conv_vector(reg_data_rd) /= SDK_VERSION) then
      info_passed := false;
    end if;

    ds_comp.info_complete <= true;
    if (info_passed) then
      assert false report "Test Info completed: PASSED." severity note;
    else
      assert false report "Test Info completed: FAILED." severity error;
      ds_pass.info_passed <= false;
    end if;


    -----------------------------------
    -- BRAM Test (128-bit interface) --
    -----------------------------------
    mem_be( 0 to 15) := conv_byte_enable("11111111" & "11111111");
    mem_be(16 to 31) := conv_byte_enable("11111111" & "11111111");

    -- Test lower address limit
    -- Write BRAM
    mem_addr := BRAM_ADDR_BASE;
    mem_data_wr(0 to 3) := conv_byte_vector(X"2389EF45");
    adb3_ocp_sim_write_reg32(mem_addr,70,mem_be(0 to 3),mem_data_wr(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr base " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(mem_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    mem_addr := BRAM_ADDR_BASE;
    adb3_ocp_sim_read_reg32(mem_addr,71,mem_data_rd(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr base " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    if (mem_data_rd(0 to 3) /= conv_byte_vector(X"2389EF45")) then
      assert false report "(test_uber_ds): Unexpected test results." severity error;
      bram_passed := false;
    end if;

    -- Test burst length = 1, 1 x 16-byte transfer
    -- Write BRAM
    mem_addr := BRAM_ADDR_BASE;
    mem_data_wr(0 to 15) := conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456");
    adb3_ocp_sim_write(mem_addr,72,mem_be(0 to 15),mem_data_wr(0 to 15),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr base " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_wr(0 to 15)) &
           " with enable 0b" & conv_string(mem_be(0 to 15)) &
           " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    mem_addr := BRAM_ADDR_BASE;
    adb3_ocp_sim_read(mem_addr,73,mem_data_rd(0 to 15),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr base " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_rd(0 to 15)) &
           " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    if (mem_data_rd(0 to 15) /= conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456")) then
      assert false report "(test_uber_ds): Unexpected test results." severity error;
      bram_passed := false;
    end if;

    -- Test burst length > 1, 2 x 16-byte transfer
    -- Write BRAM
    mem_addr := BRAM_ADDR_BASE;
    mem_data_wr( 0 to 15) := conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12");
    mem_data_wr(16 to 31) := conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE");
    adb3_ocp_sim_write(mem_addr,74,mem_be(0 to 31),mem_data_wr(0 to 31),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr base " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_wr(0 to 31)) &
           " with enable 0b" & conv_string(mem_be(0 to 31)) &
           " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    mem_addr := BRAM_ADDR_BASE;
    adb3_ocp_sim_read(mem_addr,75,mem_data_rd(0 to 31),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr base " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_rd(0 to 31)) &
           " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12"))or
       (mem_data_rd(16 to 31) /= conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE")) then
      assert false report "(test_uber_ds): Unexpected test results." severity error;
      bram_passed := false;
    end if;

    -- Test beyond lower address limit (if > 0)
    -- Write BRAM
    if not(OR_reduce(BRAM_ADDR_BASE) = '0') then
      report "Write OOR Address (will generate warning)";
      mem_addr := BRAM_ADDR_BASE - 4;
      mem_data_wr(0 to 3) := conv_byte_vector(X"369CF258");
      adb3_ocp_sim_write_reg32(mem_addr,76,mem_be(0 to 3),mem_data_wr(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote OOR Addr base-4 " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_wr(0 to 3)) &
             " with enable 0b" & conv_string(mem_be(0 to 3)) &
             " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read BRAM
      report "Read OOR Address (will generate warning)";
      mem_addr := BRAM_ADDR_BASE - 4;
      adb3_ocp_sim_read_reg32(mem_addr,77,mem_data_rd(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read OOR Addr base-4 " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_rd(0 to 3)) &
             " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      if (mem_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        bram_passed := false;
      end if;
    end if;

    -- Test beyond upper address limit
    -- Write OOR
    report "Write OOR Address (will generate warning)";
    mem_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    mem_data_wr(0 to 3) := conv_byte_vector(X"258BE147");
    adb3_ocp_sim_write_reg32(mem_addr,78,mem_be(0 to 3),mem_data_wr(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote OOR Addr top+1 " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(mem_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read OOR
    report "Read OOR Address (will generate warning)";
    mem_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_ocp_sim_read_reg32(mem_addr,79,mem_data_rd(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr top+1 " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    if (mem_data_rd(0 to 3) /= conv_byte_vector(X"DEADC0DE")) then
      assert false report "(test_uber_ds): Unexpected test results." severity error;
      bram_passed := false;
    end if;

    -- Test burst length > 1, 2 x 16-byte transfer
    -- Write OOR
    report "Write OOR Address (will generate warning)";
    mem_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    mem_data_wr( 0 to 15) := conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12");
    mem_data_wr(16 to 31) := conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE");
    adb3_ocp_sim_write(mem_addr,80,mem_be(0 to 31),mem_data_wr(0 to 31),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote OOR Addr top+1 " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_wr(0 to 31)) &
           " with enable 0b" & conv_string(mem_be(0 to 31)) &
           " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read OOR
    report "Read OOR Address (will generate warning)";
    mem_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) + 1;
    adb3_ocp_sim_read(mem_addr,81,mem_data_rd(0 to 31),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read OOR Addr top+1 " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_rd(0 to 31)) &
           " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE"))or
       (mem_data_rd(16 to 31) /= conv_byte_vector(X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE" & X"DEADC0DE")) then
      assert false report "(test_uber_ds): Unexpected test results." severity error;
      bram_passed := false;
    end if;

    -- Test upper address limit
    -- Write BRAM
    mem_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
    mem_data_wr(0 to 3) := conv_byte_vector(X"147AD036");
    adb3_ocp_sim_write_reg32(mem_addr,82,mem_be(0 to 3),mem_data_wr(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Wrote BRAM Addr top " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_wr(0 to 3)) &
           " with enable 0b" & conv_string(mem_be(0 to 3)) &
           " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    -- Read BRAM
    mem_addr := BRAM_ADDR_BASE + BRAM_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
    adb3_ocp_sim_read_reg32(mem_addr,83,mem_data_rd(0 to 3),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
    report "Read BRAM Addr top " & integer'image(mem_xfer) &
           " bytes 0x" & conv_string_hex(mem_data_rd(0 to 3)) &
           " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

    if (mem_data_rd(0 to 3) /= conv_byte_vector(X"147AD036")) then
      assert false report "(test_uber_ds): Unexpected test results." severity error;
      bram_passed := false;
    end if;


    ds_comp.bram_complete <= true;
    if (bram_passed) then
      assert false report "Test BRAM completed: PASSED." severity note;
    else
      assert false report "Test BRAM completed: FAILED." severity error;
      ds_pass.bram_passed <= false;
    end if;


    if (MEM_BANKS > 0) and DO_RAM_TEST then
      -------------------------------------
      -- RAM Test (32/128-bit interface) --
      -------------------------------------
      -- Start memory test
      reg_be := conv_byte_enable("1111");

      report "Waiting for on-board RAM bank " & integer'image(DS_RAM_TEST_BANK) & " to initialise...";

      -- Wait until DS RAM bank init done
      ram_init := '0';
      while (ram_init = '0') loop
        reg_addr := RAM_BANK0_STAT_REG_ADDR + DS_RAM_TEST_BANK*32;
        adb3_ocp_sim_read_reg32(reg_addr,84,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--        report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
--               " bytes 0x" & conv_string_hex(reg_data_rd) &
--               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
        ram_init := reg_data_rd(0)(3);
      end loop;

      report "On-board RAM bank " & integer'image(DS_RAM_TEST_BANK) & " initialised";

      -- Read RAM Bank Info register
      reg_addr := RAM_BANK0_INFO_REG_ADDR + DS_RAM_TEST_BANK*32;
      adb3_ocp_sim_read_reg32(reg_addr,85,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Bank Info Reg " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));


      if DO_INT_RAM_TEST then
        report "Waiting for on-board RAM bank " & integer'image(INT_RAM_TEST_BANK) & " to initialise...";

        -- Wait until internal RAM bank init done
        ram_init := '0';
        while (ram_init = '0') loop
          reg_addr := RAM_BANK0_STAT_REG_ADDR + INT_RAM_TEST_BANK*32;
          adb3_ocp_sim_read_reg32(reg_addr,84,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--          report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
--                 " bytes 0x" & conv_string_hex(reg_data_rd) &
--                 " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
          ram_init := reg_data_rd(0)(3);
        end loop;

        report "On-board RAM bank " & integer'image(INT_RAM_TEST_BANK) & " initialised";

        -- Read RAM Bank Info register
        reg_addr := RAM_BANK0_INFO_REG_ADDR + INT_RAM_TEST_BANK*32;
        adb3_ocp_sim_read_reg32(reg_addr,85,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Read RAM Bank Info Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Write RAM test offset (16-byte words)
        -- Do not use last 16-byte word as this is will conflict with DS PAGE TOP test below
        reg_addr := RAM_BANK0_OFFSET_REG_ADDR + INT_RAM_TEST_BANK*32;
        reg_data_wr := conv_byte_vector(std_logic_vector(to_unsigned(INT_RAM_TEST_START,32)));
        adb3_ocp_sim_write_reg32(reg_addr,86,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Wrote RAM Bank Offset Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_wr) &
               " with enable 0b" & conv_string(reg_be) &
               " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Read RAM test offset
        adb3_ocp_sim_read_reg32(reg_addr,87,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Read RAM Bank Offset Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Write RAM test length-1 (16-byte words)
        reg_addr := RAM_BANK0_LENGTH_REG_ADDR + INT_RAM_TEST_BANK*32;
        reg_data_wr := conv_byte_vector(INT_RAM_TEST_LEN);
        adb3_ocp_sim_write_reg32(reg_addr,88,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Wrote RAM Bank Length Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_wr) &
               " with enable 0b" & conv_string(reg_be) &
               " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Read RAM test length
        adb3_ocp_sim_read_reg32(reg_addr,89,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Read RAM Bank Length Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        -- Start internal test
        -- Write ram ctrl
        reg_addr := RAM_BANK0_CTRL_REG_ADDR + INT_RAM_TEST_BANK*32;
        reg_data_wr := conv_byte_vector(X"00000100");
        adb3_ocp_sim_write_reg32(reg_addr,90,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
        report "Wrote RAM Bank Ctrl Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_wr) &
               " with enable 0b" & conv_string(reg_be) &
               " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
      end if;

      -- Set up access to RAMs using DS window
      reg_be := conv_byte_enable("1111");

      -- Write DS bank
      reg_addr := RAM_DS_BANK_REG_ADDR;
      reg_data_wr := conv_byte_vector(DS_WIN_RAM_BANK);
      adb3_ocp_sim_write_reg32(reg_addr,91,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Bank Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read DS bank (to ensure register has been written before RAM access)
      adb3_ocp_sim_read_reg32(reg_addr,92,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM DS Bank Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Write DS bottom page
      reg_addr := RAM_DS_PAGE_REG_ADDR;
      reg_data_wr := conv_byte_vector(DS_WIN_RAM_PAGE_BOT);
      adb3_ocp_sim_write_reg32(reg_addr,93,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read DS bottom page (to ensure register has been written before RAM access)
      adb3_ocp_sim_read_reg32(reg_addr,94,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Test lower address limit
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      reg_data_wr := conv_byte_vector(X"349AF056");
      adb3_ocp_sim_write_reg32(reg_addr,95,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      adb3_ocp_sim_read_reg32(reg_addr,96,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"349AF056")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Write DS top page
      reg_addr := RAM_DS_PAGE_REG_ADDR;
      reg_data_wr := conv_byte_vector(DS_WIN_RAM_PAGE_TOP);
      adb3_ocp_sim_write_reg32(reg_addr,97,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read DS top page (to ensure register has been written before RAM access)
      adb3_ocp_sim_read_reg32(reg_addr,98,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Test upper address limit
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE + RAM_WIN_ADDR_MASK(DS_ADDR_WIDTH-1 downto 0) - 3;
      reg_data_wr := conv_byte_vector(X"47AD0369");
      adb3_ocp_sim_write_reg32(reg_addr,99,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr top " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      adb3_ocp_sim_read_reg32(reg_addr,100,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr top " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"47AD0369")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Write DS bottom page
      reg_addr := RAM_DS_PAGE_REG_ADDR;
      reg_data_wr := conv_byte_vector(DS_WIN_RAM_PAGE_BOT);
      adb3_ocp_sim_write_reg32(reg_addr,101,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read DS bottom page (to ensure register has been written before RAM access)
      adb3_ocp_sim_read_reg32(reg_addr,102,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM DS Page Reg Addr " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));


      mem_be( 0 to 15) := conv_byte_enable("11111111" & "11111111");
      mem_be(16 to 31) := conv_byte_enable("11111111" & "11111111");
      mem_be(32 to 47) := conv_byte_enable("11111111" & "11111111");
      mem_be(48 to 63) := conv_byte_enable("11111111" & "11111111");
      mem_be(64 to 79) := conv_byte_enable("11111111" & "11111111");
      mem_be(80 to 95) := conv_byte_enable("11111111" & "11111111");

      -- Test even burst length, 32-byte aligned, BL 6
      -- Write RAM
      mem_addr := RAM_WIN_ADDR_BASE;
      mem_data_wr( 0 to 15) := conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12");
      mem_data_wr(16 to 31) := conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE");
      mem_data_wr(32 to 47) := conv_byte_vector(X"23456789" & X"ABCDEF12" & X"3456789A" & X"BCDEF123");
      mem_data_wr(48 to 63) := conv_byte_vector(X"EDCBA987" & X"654321FE" & X"DCBA9876" & X"54321FED");
      mem_data_wr(64 to 79) := conv_byte_vector(X"3456789A" & X"BCDEF123" & X"456789AB" & X"CDEF1234");
      mem_data_wr(80 to 95) := conv_byte_vector(X"DCBA9876" & X"54321FED" & X"CBA98765" & X"4321FEDC");

      adb3_ocp_sim_write(mem_addr,103,mem_be(0 to 95),mem_data_wr(0 to 95),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_wr(0 to 95)) &
             " with enable 0b" & conv_string(mem_be(0 to 95)) &
             " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      mem_addr := RAM_WIN_ADDR_BASE;
      adb3_ocp_sim_read(mem_addr,104,mem_data_rd(0 to 95),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_rd(0 to 95)) &
             " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"12345678" & X"9ABCDEF1" & X"23456789" & X"ABCDEF12"))or
         (mem_data_rd(16 to 31) /= conv_byte_vector(X"FEDCBA98" & X"7654321F" & X"EDCBA987" & X"654321FE"))or
         (mem_data_rd(32 to 47) /= conv_byte_vector(X"23456789" & X"ABCDEF12" & X"3456789A" & X"BCDEF123"))or
         (mem_data_rd(48 to 63) /= conv_byte_vector(X"EDCBA987" & X"654321FE" & X"DCBA9876" & X"54321FED"))or
         (mem_data_rd(64 to 79) /= conv_byte_vector(X"3456789A" & X"BCDEF123" & X"456789AB" & X"CDEF1234"))or
         (mem_data_rd(80 to 95) /= conv_byte_vector(X"DCBA9876" & X"54321FED" & X"CBA98765" & X"4321FEDC")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Test odd burst length, 32-byte aligned, BL 5
      -- Write RAM
      mem_addr := RAM_WIN_ADDR_BASE;
      mem_data_wr( 0 to 15) := conv_byte_vector(X"456789AB" & X"CDEF1234" & X"56789ABC" & X"DEF12345");
      mem_data_wr(16 to 31) := conv_byte_vector(X"CBA98765" & X"4321FEDC" & X"BA987654" & X"321FEDCB");
      mem_data_wr(32 to 47) := conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456");
      mem_data_wr(48 to 63) := conv_byte_vector(X"BA987654" & X"321FEDCB" & X"A9876543" & X"21FEDCBA");
      mem_data_wr(64 to 79) := conv_byte_vector(X"6789ABCD" & X"EF123456" & X"789ABCDE" & X"F1234567");

      adb3_ocp_sim_write(mem_addr,105,mem_be(0 to 79),mem_data_wr(0 to 79),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_wr(0 to 79)) &
             " with enable 0b" & conv_string(mem_be(0 to 79)) &
             " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      mem_addr := RAM_WIN_ADDR_BASE;
      adb3_ocp_sim_read(mem_addr,106,mem_data_rd(0 to 79),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_rd(0 to 79)) &
             " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"456789AB" & X"CDEF1234" & X"56789ABC" & X"DEF12345"))or
         (mem_data_rd(16 to 31) /= conv_byte_vector(X"CBA98765" & X"4321FEDC" & X"BA987654" & X"321FEDCB"))or
         (mem_data_rd(32 to 47) /= conv_byte_vector(X"56789ABC" & X"DEF12345" & X"6789ABCD" & X"EF123456"))or
         (mem_data_rd(48 to 63) /= conv_byte_vector(X"BA987654" & X"321FEDCB" & X"A9876543" & X"21FEDCBA"))or
         (mem_data_rd(64 to 79) /= conv_byte_vector(X"6789ABCD" & X"EF123456" & X"789ABCDE" & X"F1234567")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Test even burst length, non 32-byte aligned, BL 2 and BL 4
      -- Write RAM
      mem_addr := RAM_WIN_ADDR_BASE+16;
      mem_data_wr( 0 to 15) := conv_byte_vector(X"789ABCDE" & X"F1234567" & X"89ABCDEF" & X"12345678");
      mem_data_wr(16 to 31) := conv_byte_vector(X"A9876543" & X"21FEDCBA" & X"98765432" & X"1FEDCBA9");

      adb3_ocp_sim_write(mem_addr,107,mem_be(0 to 31),mem_data_wr(0 to 31),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_wr(0 to 31)) &
             " with enable 0b" & conv_string(mem_be(0 to 31)) &
             " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      mem_addr := RAM_WIN_ADDR_BASE+16;
      adb3_ocp_sim_read(mem_addr,108,mem_data_rd(0 to 31),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_rd(0 to 31)) &
             " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"789ABCDE" & X"F1234567" & X"89ABCDEF" & X"12345678"))or
         (mem_data_rd(16 to 31) /= conv_byte_vector(X"A9876543" & X"21FEDCBA" & X"98765432" & X"1FEDCBA9"))then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Write RAM
      mem_addr := RAM_WIN_ADDR_BASE+16;
      mem_data_wr( 0 to 15) := conv_byte_vector(X"89ABCDEF" & X"12345678" & X"9ABCDEF1" & X"23456789");
      mem_data_wr(16 to 31) := conv_byte_vector(X"98765432" & X"1FEDCBA9" & X"87654321" & X"FEDCBA98");
      mem_data_wr(32 to 47) := conv_byte_vector(X"9ABCDEF1" & X"23456789" & X"ABCDEF12" & X"3456789A");
      mem_data_wr(48 to 63) := conv_byte_vector(X"87654321" & X"FEDCBA98" & X"7654321F" & X"EDCBA987");
      adb3_ocp_sim_write(mem_addr,109,mem_be(0 to 63),mem_data_wr(0 to 63),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_wr(0 to 63)) &
             " with enable 0b" & conv_string(mem_be(0 to 63)) &
             " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      mem_addr := RAM_WIN_ADDR_BASE+16;
      adb3_ocp_sim_read(mem_addr,110,mem_data_rd(0 to 63),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_rd(0 to 63)) &
             " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"89ABCDEF" & X"12345678" & X"9ABCDEF1" & X"23456789"))or
         (mem_data_rd(16 to 31) /= conv_byte_vector(X"98765432" & X"1FEDCBA9" & X"87654321" & X"FEDCBA98"))or
         (mem_data_rd(32 to 47) /= conv_byte_vector(X"9ABCDEF1" & X"23456789" & X"ABCDEF12" & X"3456789A"))or
         (mem_data_rd(48 to 63) /= conv_byte_vector(X"87654321" & X"FEDCBA98" & X"7654321F" & X"EDCBA987"))then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Test odd burst length, non 32-byte aligned, BL 5
      -- Write RAM
      mem_addr := RAM_WIN_ADDR_BASE+16;
      mem_data_wr( 0 to 15) := conv_byte_vector(X"ABCDEF12" & X"3456789A" & X"BCDEF123" & X"456789AB");
      mem_data_wr(16 to 31) := conv_byte_vector(X"7654321F" & X"EDCBA987" & X"654321FE" & X"DCBA9876");
      mem_data_wr(32 to 47) := conv_byte_vector(X"BCDEF123" & X"456789AB" & X"CDEF1234" & X"56789ABC");
      mem_data_wr(48 to 63) := conv_byte_vector(X"654321FE" & X"DCBA9876" & X"54321FED" & X"CBA98765");
      mem_data_wr(64 to 79) := conv_byte_vector(X"CDEF1234" & X"56789ABC" & X"DEF12345" & X"6789ABCD");

      adb3_ocp_sim_write(mem_addr,111,mem_be(0 to 79),mem_data_wr(0 to 79),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_wr(0 to 79)) &
             " with enable 0b" & conv_string(mem_be(0 to 79)) &
             " to byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      mem_addr := RAM_WIN_ADDR_BASE+16;
      adb3_ocp_sim_read(mem_addr,112,mem_data_rd(0 to 79),mem_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(mem_xfer) &
             " bytes 0x" & conv_string_hex(mem_data_rd(0 to 79)) &
             " from byte address 0x" & conv_string_hex(mem_addr(DS_ADDR_WIDTH-1 downto 0));

      if (mem_data_rd( 0 to 15) /= conv_byte_vector(X"ABCDEF12" & X"3456789A" & X"BCDEF123" & X"456789AB"))or
         (mem_data_rd(16 to 31) /= conv_byte_vector(X"7654321F" & X"EDCBA987" & X"654321FE" & X"DCBA9876"))or
         (mem_data_rd(32 to 47) /= conv_byte_vector(X"BCDEF123" & X"456789AB" & X"CDEF1234" & X"56789ABC"))or
         (mem_data_rd(48 to 63) /= conv_byte_vector(X"654321FE" & X"DCBA9876" & X"54321FED" & X"CBA98765"))or
         (mem_data_rd(64 to 79) /= conv_byte_vector(X"CDEF1234" & X"56789ABC" & X"DEF12345" & X"6789ABCD")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      -- Test byte enables
      -- Write RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be := conv_byte_enable("1000");
      reg_data_wr := conv_byte_vector(X"45000000");
      adb3_ocp_sim_write_reg32(reg_addr,113,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be := conv_byte_enable("0100");
      reg_data_wr := conv_byte_vector(X"00AB0000");
      adb3_ocp_sim_write_reg32(reg_addr,114,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be := conv_byte_enable("0010");
      reg_data_wr := conv_byte_vector(X"00000100");
      adb3_ocp_sim_write_reg32(reg_addr,115,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      reg_addr := RAM_WIN_ADDR_BASE;
      reg_be := conv_byte_enable("0001");
      reg_data_wr := conv_byte_vector(X"00000067");
      adb3_ocp_sim_write_reg32(reg_addr,116,reg_be,reg_data_wr,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Wrote RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_wr) &
             " with enable 0b" & conv_string(reg_be) &
             " to byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      -- Read RAM
      reg_addr := RAM_WIN_ADDR_BASE;
      adb3_ocp_sim_read_reg32(reg_addr,117,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
      report "Read RAM Win Addr base " & integer'image(reg_xfer) &
             " bytes 0x" & conv_string_hex(reg_data_rd) &
             " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

      if (reg_data_rd /= conv_byte_vector(X"45AB0167")) then
        assert false report "(test_uber_ds): Unexpected test results." severity error;
        ram_passed := false;
      end if;

      if DO_INT_RAM_TEST then
        report "Waiting for internal test of on-board RAM bank " & integer'image(INT_RAM_TEST_BANK) & " to complete...";

        -- Wait until internal RAM test done by reading done status
        ram_done := '0';
        while (ram_done = '0') loop
          reg_addr := RAM_BANK0_STAT_REG_ADDR + INT_RAM_TEST_BANK*32;
          adb3_ocp_sim_read_reg32(reg_addr,118,reg_data_rd,reg_xfer,ocp_clk,ds_m2s,ds_s2m);
--          report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
--                 " bytes 0x" & conv_string_hex(reg_data_rd) &
--                 " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));
          ram_done := reg_data_rd(2)(0);
        end loop;

        report "Read RAM Bank Stat Reg " & integer'image(reg_xfer) &
               " bytes 0x" & conv_string_hex(reg_data_rd) &
               " from byte address 0x" & conv_string_hex(reg_addr(DS_ADDR_WIDTH-1 downto 0));

        report "Internal test of on-board RAM bank " & integer'image(INT_RAM_TEST_BANK) & " complete";

        -- Test stat bit 23 = mem_app_err
        if (reg_data_rd(2)(7) = '1') then
          report "Internal test of on-board RAM bank " & integer'image(INT_RAM_TEST_BANK) & " failed";
          ram_passed := false;
        end if;
      end if;

      ds_comp.ram_complete <= true;
      if (ram_passed) then
        assert false report "Test RAM completed: PASSED." severity note;
      else
        assert false report "Test RAM completed: FAILED." severity error;
        ds_pass.ram_passed <= false;
      end if;
    else
      ds_comp.ram_complete <= true;
    end if;

    wait;
  end process;

  -- Connect direct slave output
  direct_slave_m2s <= ds_m2s;

end;