--==============================================================================
--
--     File Name: test_uber_pcie.vhd
--       Version: 0.1
--          Date: 19 Aug 2011
--         Model: Simulation only
--
--  Dependencies:
--
--       Company: Alpha Data
--        Design: ADB3 target testbench design using PCIe
--
--   Description: Illustrates with ADB3:
--                  Register access using direct slave interface
--                  BRAM access using direct slave interface
--                  Multi-channel DMA
--                  Direct master interface
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 19 Aug 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_unsigned.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_ocp_tb_pkg.all;
use work.adb3_target_types_pkg.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_tb_inc_pkg.all;
use work.adb3_target_tb_pkg.all;
use work.uber_tb_pkg.all;

entity test_uber is
end test_uber;

architecture sim of test_uber is

  -- Test results
  signal top_comp : top_comp_t;
  signal top_pass : top_pass_t;
  signal ds_comp  : ds_comp_t;
  signal ds_pass  : ds_pass_t;
  signal dma_comp : dma_comp_t;
  signal dma_pass : dma_pass_t;

  signal test_complete : boolean := false;
  signal test_passed   : boolean := true;

  -- Host clocks
  signal ocp_clk : std_logic;

  -- Host test signals from target
  signal dma_abort : std_logic_vector(DMA_CHANNELS-1 downto 0);
  signal interrupt : std_logic;

  -- Probe errors
  signal probe_ds_err  : std_logic;
  signal probe_dm_err  : std_logic;
  signal probe_dma_err : std_logic;

  -- DDR3 signals
  signal mem_addr_out   : mem_addr_out_t;
  signal mem_ctrl_out   : mem_ctrl_out_t;
  signal mem_data_inout : mem_data_inout_t;
  signal mem_clk_out    : mem_clk_out_t;

  --------------------
  -- Target signals --
  --------------------

  -- Clocks
  signal clks_in     : clks_in_t;
  signal clks_out    : clks_out_t;
  signal clks_mgt_in : clks_mgt_in_t;

  -- Configuration/online signals
  signal pcie_online_cnt     : std_logic_vector(7 downto 0) := (others => '1');
  signal pcie_online_long    : std_logic := '0';
  signal pcie_online_long_d1 : std_logic := '0';

  -- PCIe signals
  signal pcie_rst_l : std_logic := '0';
  signal pcie_h2t   : pcie_h2t_t;
  signal pcie_t2h   : pcie_t2h_t;

  -- OCP signals
  signal direct_slave_m2s   : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal direct_slave_s2m   : adb3_ocp_s2mT;
  signal dma_channels_m2s   : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_M2S);
  signal dma_channels_s2m   : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  signal direct_masters_m2s : adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
  signal direct_masters_s2m : adb3_ocp_s2mT_array(0 to DM_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_S2M);

  -- ADV8 specific board interface
  signal custom_in    : custom_in_t;
  signal custom_out   : custom_out_t;
  signal custom_inout : custom_inout_t;

  -- GPIO Interface
  signal gpio_inout : gpio_inout_t;

  -- clk_out test
  signal clks_out_pass : boolean;

begin

  -- Report design constants
  report_setup_p : process
  begin
    report "Board Type : " & BOARD_TYPE;                      -- adb3_target_tb_inc_pkg
    report "Target Use : " & target_use_t'image(TARGET_USE);  -- adb3_target_inc_pkg
    report "Master Use : " & master_use_t'image(MASTER_USE);  -- uber_tb_pkg
    report "Waiting for PCIe online....";
    wait;
  end process;


  -------------------------------
  -- OCP direct slave stimulus --
  -------------------------------
  test_uber_ds_i : test_uber_ds
  port map(
    -- OCP Interface
    ocp_clk          => ocp_clk,           -- in    std_logic
    direct_slave_m2s => direct_slave_m2s,  -- out   adb3_ocp_m2sT
    direct_slave_s2m => direct_slave_s2m,  -- in    adb3_ocp_s2mT
    -- Test interface
    start_test       => pcie_online_long,  -- in    std_logic
    ds_comp          => ds_comp,           -- out   ds_comp_r
    ds_pass          => ds_pass);          -- out   ds_pass_r)

  ----------------------
  -- OCP DMA stimulus --
  ----------------------
  test_uber_dma_i : test_uber_dma
  port map(
    -- OCP Interface
    ocp_clk          => ocp_clk,           -- in    std_logic
    dma_channels_m2s => dma_channels_m2s,  -- out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m => dma_channels_s2m,  -- in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    -- Test interface
    start_test       => pcie_online_long,  -- in    std_logic
    dma_comp         => dma_comp,          -- out   dma_comp_t
    dma_pass         => dma_pass);         -- out   dma_pass_t)

  --------------------------------
  -- OCP direct master stimulus --
  --------------------------------
  dm_g : if (MASTER_USE = MASTER_ON) generate
    test_uber_dm_i : test_uber_dm
    port map(
      -- OCP Interface
      ocp_clk            => ocp_clk,             -- in    std_logic
      direct_masters_m2s => direct_masters_m2s,  -- in    adb3_ocp_m2sT_array(0 to MASTER_CHANNELS-1)
      direct_masters_s2m => direct_masters_s2m); -- out   adb3_ocp_s2mT_array(0 to MASTER_CHANNELS-1))
  end generate;


  -- Test dma_abort bus continuously
  dma_abort_p : process
  begin
    wait until (pcie_online_long = '1');  -- OCP commands ignored until PCIe online
    adb3_ocp_sim_wait_cycles(50,ocp_clk); -- Wait for data to get to outputs

    top_comp.dma_abort_complete <= true;
    top_pass.dma_abort_passed   <= true;

    loop
      if (OR_reduce(dma_abort(DMA_CHANNELS-1 downto 0)) = '0') then
--        assert false report "Test dma_abort completed: PASSED." severity note;
        null;
      else
        assert false report "Test dma_abort completed: FAILED." severity error;
        top_pass.dma_abort_passed <= false;
        exit;
      end if;

      adb3_ocp_sim_wait_cycles(1,ocp_clk);
    end loop;
  end process;


  -- Test clk_out period
  clk_out_p : process
  begin
    wait until (pcie_online_long = '1');  -- OCP commands ignored until PCIe online

    top_comp.clk_out_complete <= true;
    top_pass.clk_out_passed   <= true;

    loop
      if (clks_out_pass) then
--        assert false report "Test clk_out completed: PASSED." severity note;
        null;
      else
        assert false report "Test clk_out completed: FAILED." severity error;
        top_pass.clk_out_passed <= false;
        exit;
      end if;

      adb3_ocp_sim_wait_cycles(1,ocp_clk);
    end loop;
  end process;


  -- Monitor interrupt input and detect rising edges
  interrupt_monitor_p : process
  begin
    loop
      -- Wait for a positive edge on 'interrupt'
      wait until rising_edge(interrupt);
      report "Interrupt Monitor: Detected rising edge on interrupt";
      -- Shared variable defined in uber_tb_pkg
      int_edge := '1';
    end loop;
  end process;


  -------------------------
  -- Output test results --
  -------------------------

  gen_results_p : process(top_pass,top_comp,ds_pass,ds_comp,dma_pass,dma_comp)
  begin
    test_passed <= top_pass.dma_abort_passed and
                   top_pass.clk_out_passed and
                   dma_pass.dma_write_cmd_passed and dma_pass.dma_read_cmd_passed and dma_pass.dma_read_resp_passed and
                   ds_pass.simple_passed and
                   ds_pass.master_passed and
                   ds_pass.clock_passed and
                   ds_pass.frontio_passed and ds_pass.reario_passed and
                   ds_pass.int_passed and
                   ds_pass.info_passed and
                   ds_pass.bram_passed and
                   ds_pass.ram_passed;

    test_complete <= top_comp.dma_abort_complete and
                     top_comp.clk_out_complete and
                     dma_comp.dma_write_cmd_complete and dma_comp.dma_read_cmd_complete and dma_comp.dma_read_resp_complete and
                     ds_comp.simple_complete and
                     ds_comp.master_complete and
                     ds_comp.clock_complete and
                     ds_comp.frontio_complete and ds_comp.reario_complete and
                     ds_comp.int_complete and
                     ds_comp.info_complete and
                     ds_comp.bram_complete and
                     ds_comp.ram_complete;
  end process;


  test_results_p : process(test_complete,test_passed,probe_ds_err,probe_dm_err,probe_dma_err)
  begin
    if (test_complete) then
      assert not(probe_ds_err  = '1') report "Probe OCP Direct Slave status: ERROR."  severity error;
      assert not(probe_dm_err  = '1') report "Probe OCP Direct Master status: ERROR." severity error;
      assert not(probe_dma_err = '1') report "Probe OCP DMA status: ERROR."           severity error;

      if (test_passed) then
        assert false report "Test of design UBER completed: PASSED." severity failure;
      else
        assert false report "Test of design UBER completed: FAILED." severity failure;
      end if;
    end if;
  end process;


  ---------------------------------
  -- Board Clock Generation/Test --
  ---------------------------------

  test_board_clks_i : test_board_clks
  generic map(
    clks_out_freq => CLKS_OUT_FREQ) -- clks_out_exp_t
  port map(
    -- Target clock interface
    target_clks_in     => clks_in,         -- out   clks_in_t
    target_clks_mgt_in => clks_mgt_in,     -- out   clks_mgt_in_t
    target_clks_out    => clks_out,        -- in    clks_out_t
    target_ref_clk     => open,            -- out   std_logic_dbl_t
    target_mptl_clk    => open,            -- out   std_logic_dbl_t
    target_pcie_clk    => open,            -- out   std_logic_dbl_t
    -- Bridge clock interface
    bridge_lclk_freq   => 0.0,             -- in    real
    bridge_ocp_clk     => open,            -- out   std_logic
    bridge_mptl_clk    => open,            -- out   std_logic_dbl_t
    -- Test interface
    clks_out_pass      => clks_out_pass);  -- out   boolean


  ------------------------------------
  -- Target Component Instantiation --
  ------------------------------------

  pcie_rst_l <= '1';
  custom_in.adb3_model <= (others => '0');

  uut : uber
  port map(
    -- Non-MGT Clock Interface
    clks_in        => clks_in,         -- in    clks_in_t
    clks_out       => clks_out,        -- out   clks_out_t
    -- MGT Clock Interface
    clks_mgt_in    => clks_mgt_in,     -- in    clks_mgt_in_t
    -- PCIe Interface
    pcie_rst_l     => pcie_rst_l,      -- in    std_logic
    pcie_h2t       => pcie_h2t,        -- in    pcie_h2t_t
    pcie_t2h       => pcie_t2h,        -- out   pcie_t2h_t
    -- GPIO Interface
    gpio_inout     => gpio_inout,      -- inout gpio_inout_t
    -- Custom Interface
    custom_in      => custom_in,       -- in    custom_in_t
    custom_out     => custom_out,      -- out   custom_out_t
    custom_inout   => custom_inout,    -- inout custom_inout_t
    -- Memory Interface
    mem_addr_out   => mem_addr_out,    -- out   mem_addr_out_t
    mem_ctrl_out   => mem_ctrl_out,    -- out   mem_ctrl_out_t
    mem_data_inout => mem_data_inout,  -- inout mem_data_inout_t
    mem_clk_out    => mem_clk_out);    -- out   mem_clk_out_t)


  -------------------------------------
  -- Bridge Component Instantiations --
  -------------------------------------

  -- testbench interface
  pcie_if_host_wrap_i : pcie_if_host_wrap
  port map(
    -- OCP Interface
    direct_slave_m2s   => direct_slave_m2s,    -- in    adb3_ocp_m2sT
    direct_slave_s2m   => direct_slave_s2m,    -- out   adb3_ocp_s2mT
    dma_channels_m2s   => dma_channels_m2s,    -- in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m   => dma_channels_s2m,    -- out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    direct_masters_m2s => direct_masters_m2s,  -- out   adb3_ocp_m2sT_array(0 to DM_CHANNELS-1)
    direct_masters_s2m => direct_masters_s2m,  -- in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1)
    dma_abort          => dma_abort,           -- out   std_logic_vector(DMA_CHANNELS-1 downto 0)
    interrupt          => interrupt,           -- out   std_logic
    ocp_clk_out        => ocp_clk,             -- out   std_logic
    -- PCIe Interface
    pcie_t2h           => pcie_t2h,            -- in    pcie_t2b_t
    pcie_h2t           => pcie_h2t);           -- out   pcie_b2t_t)


  -- Count cycles of ocp_clk to make sure it is stable
  online_count_p : process(ocp_clk)
  begin
    if rising_edge(ocp_clk) then
      if (or_reduce(pcie_online_cnt) = '1') then
        pcie_online_cnt <= pcie_online_cnt - 1;
      end if;
      pcie_online_long    <= not(or_reduce(pcie_online_cnt));
      pcie_online_long_d1 <= pcie_online_long;
    end if;
  end process;


  ------------------------
  -- Bridge OCP probes --
  ------------------------

  test_uber_probes_i : test_uber_probes
  port map(
     -- OCP Interface
    ocp_clk            => ocp_clk,             -- in    std_logic
    direct_slave_m2s   => direct_slave_m2s,    -- in    adb3_ocp_m2sT
    direct_slave_s2m   => direct_slave_s2m,    -- in    adb3_ocp_s2mT
    dma_channels_m2s   => dma_channels_m2s,    -- in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m   => dma_channels_s2m,    -- in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    direct_masters_m2s => direct_masters_m2s,  -- in    adb3_ocp_m2sT_array(0 to MASTER_CHANNELS-1)
    direct_masters_s2m => direct_masters_s2m,  -- in    adb3_ocp_s2mT_array(0 to MASTER_CHANNELS-1)
    -- Probe Interface
    probe_ds_err       => probe_ds_err,        -- out   std_logic;
    probe_dma_err      => probe_dma_err,       -- out   std_logic;
    probe_dm_err       => probe_dm_err);       -- out   std_logic)


  ---------------------------------
  -- On-board Memory Bank Models --
  ---------------------------------

  test_uber_mem_i : test_uber_mem
  generic map(
    message_level => 0) -- natural);
  port map(
    -- Memory Interface
    mem_addr_out   => mem_addr_out,    -- in    mem_addr_out_t
    mem_ctrl_out   => mem_ctrl_out,    -- in    mem_ctrl_out_t
    mem_data_inout => mem_data_inout,  -- inout mem_data_inout_t
    mem_clk_out    => mem_clk_out);    -- in    mem_clk_out_t)

end;