--==============================================================================
--
--     File Name: test_uber_probes.vhd
--       Version: 0.2
--          Date: 17 May 2011
--         Model: Simulation only
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 Target Design Test: OCP probes
--
--   Description: Provides bridge OCP probes
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2009-2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 17 Mar 2009: Initial design
--                0.2, KDC, 17 May 2011:
--                     - Corrected values of DS_ADDR_ALIGN_BITS and
--                       DM_ADDR_ALIGN_BITS to 4
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;
use ieee.std_logic_misc.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_inc_pkg.all;
use work.uber_tb_pkg.all;
use work.adb3_probe_pkg.all;

entity test_uber_probes is
port(
  -- OCP Interface
  ocp_clk            : in    std_logic;
  direct_slave_m2s   : in    adb3_ocp_m2sT;
  direct_slave_s2m   : in    adb3_ocp_s2mT;
  dma_channels_m2s   : in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  dma_channels_s2m   : in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1);
  direct_masters_m2s : in    adb3_ocp_m2sT_array(0 to DM_CHANNELS-1);
  direct_masters_s2m : in    adb3_ocp_s2mT_array(0 to DM_CHANNELS-1);
  -- Probe Interface
  probe_ds_err       : out   std_logic;
  probe_dma_err      : out   std_logic;
  probe_dm_err       : out   std_logic);
end test_uber_probes;

architecture sim of test_uber_probes is

  -- Constant to control OCP transaction probe file logging
  constant PROBE_LOG : boolean := false;

  constant DS_ADDR_ALIGN_BITS  : natural := 4; -- 16-byte aligned (128-bit interface)
  constant DM_ADDR_ALIGN_BITS  : natural := 4; -- 16-byte aligned (128-bit interface)
  constant DMA_ADDR_ALIGN_BITS : natural := 4; -- 16-byte aligned (128 bit interface)

  type status_probe_dma_t is array(0 to DMA_CHANNELS-1) of adb3_ocp_probe_status_r;

  signal status_probe_ds  : adb3_ocp_probe_status_r := ADB3_OCP_PROBE_STATUS_OK;
  signal status_probe_dm  : adb3_ocp_probe_status_r := ADB3_OCP_PROBE_STATUS_OK;
  signal status_probe_dma : status_probe_dma_t := (others => ADB3_OCP_PROBE_STATUS_OK);

  signal status_dma_err : std_logic_vector(DMA_CHANNELS-1 downto 0);

begin

  --------------------------------------
  -- Direct Slave OCP channel probes --
  --------------------------------------

  -- Direct Slave transaction probe
  ds_probe_i : adb3_ocp_transaction_probe
  generic map(
    enable_logging   => PROBE_LOG,
    int_log_filename => "ds_probe",
    addr_align_bits  => DS_ADDR_ALIGN_BITS,
    addr_width_max   => DS_ADDR_WIDTH)
  port map(
    ocp_clk => ocp_clk,
    ocp_m2s => direct_slave_m2s,
    ocp_s2m => direct_slave_s2m,
    status  => status_probe_ds);

  probe_ds_err <= '0' when status_probe_ds = ADB3_OCP_PROBE_STATUS_OK else '1';


  --------------------------------------
  -- Direct Master OCP channel probes --
  --------------------------------------

  dm_probe_g : if (MASTER_USE = MASTER_ON) generate
    -- Direct Master transaction probe
    dm_probe_i : adb3_ocp_transaction_probe
    generic map(
      enable_logging   => PROBE_LOG,
      int_log_filename => "dm_probe",
      addr_align_bits  => DM_ADDR_ALIGN_BITS,
      addr_width_max   => DM_ADDR_WIDTH)
     port map(
      ocp_clk => ocp_clk,
      ocp_m2s => direct_masters_m2s(DM_CHANNEL),
      ocp_s2m => direct_masters_s2m(DM_CHANNEL),
      status  => status_probe_dm);
  end generate;

  probe_dm_err <= '0' when status_probe_dm = ADB3_OCP_PROBE_STATUS_OK else '1';


  ----------------------------
  -- DMA OCP channel probes --
  ----------------------------

  dma_probe_g : for n in 0 to DMA_CHANNELS-1 generate
    constant DMA_FILENAME : string  := "dma" & integer'image(n) & "_probe";
  begin
    -- Disable read tag check for DMA as read command tag is not used by read data
    dma_probe_i : adb3_ocp_transaction_probe
    generic map(
      enable_logging   => PROBE_LOG,
      int_log_filename => DMA_FILENAME,
      addr_align_bits  => DMA_ADDR_ALIGN_BITS,
      addr_width_max   => DMA_ADDR_WIDTH,
      enable_tag_check => false)
    port map(
      ocp_clk => ocp_clk,
      ocp_m2s => dma_channels_m2s(n),
      ocp_s2m => dma_channels_s2m(n),
      status  => status_probe_dma(n));

    status_dma_err(n) <= '0' when (status_probe_dma(n) = ADB3_OCP_PROBE_STATUS_OK) else '1';
  end generate;

  probe_dma_err <= OR_reduce(status_dma_err);

end;