--==============================================================================
--
--     File Name: uber_pcie.vhd
--       Version: 0.1
--          Date: 19 Aug 2011
--         Model: Synthesisable
--
--  Dependencies: Set simulator resolution to ps for accurate simulation
--
--       Company: Alpha Data
--        Design: ADB3 target design using PCIe
--
--   Description: Illustrates:
--                  OCP Direct Slave port:
--                    Register access using DS port
--                    BRAM access using DS port
--                    On-board RAM access using DS port
--                  OCP DMA ports:
--                    BRAM access using all DMA ports
--                    On-board RAM access using DMA port 0
--
--   Limitations:
--
--         Notes:
--
--    Disclaimer: THESE DESIGNS ARE PROVIDED "AS IS" WITH NO WARRANTY
--                WHATSOEVER AND ALPHA DATA SPECIFICALLY DISCLAIMS ANY
--                IMPLIED WARRANTIES OF MERCHANTABILITY, FITNESS FOR
--                A PARTICULAR PURPOSE, OR AGAINST INFRINGEMENT.
--
--                Copyright (C) 2011 Alpha Data, All rights reserved
--
--     Revisions: 0.1, KDC, 19 Aug 2011: Initial design
--
--==============================================================================

library ieee;
use ieee.std_logic_1164.all;

library work;
use work.adb3_ocp.all;
use work.adb3_target_inc_pkg.all;
use work.adb3_target_pkg.all;
use work.uber_pkg.all;

entity uber is
port(
  -- Non-MGT Clock Interface
  clks_in        : in    clks_in_t;
  clks_out       : out   clks_out_t;
  -- MGT Clock Interface
  clks_mgt_in    : in    clks_mgt_in_t;
  -- PCIe Interface
  pcie_rst_l     : in    std_logic;
  pcie_h2t       : in    pcie_h2t_t;
  pcie_t2h       : out   pcie_t2h_t;
  -- GPIO Interface
  gpio_inout     : inout gpio_inout_t;
  -- Custom Interface
  custom_in      : in    custom_in_t;
  custom_out     : out   custom_out_t;
  custom_inout   : inout custom_inout_t;
  -- Memory Interface
  mem_addr_out   : out   mem_addr_out_t;
  mem_ctrl_out   : out   mem_ctrl_out_t;
  mem_data_inout : inout mem_data_inout_t;
  mem_clk_out    : out   mem_clk_out_t);
end;


architecture syn of uber is

  -- Chipscope instantiation control
  constant CHIPSCOPE_ON : boolean := false;

  ------------------
  -- Reset signal --
  ------------------
  signal rst : std_logic;

  -------------------
  -- Clock signals --
  -------------------
  signal pll_pri_clk    : std_logic;
  signal pll_reg_clk    : std_logic;
  signal pll_mem_clk    : std_logic;
  signal pll_ref_clk    : std_logic;
  signal clk_vec        : clk_vec_t;
  signal pcie_clk       : std_logic_dbl_t;

  ----------------------------
  -- PCIe interface signals --
  ----------------------------
  signal direct_slave_m2s : adb3_ocp_m2sT;
  signal direct_slave_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;
  signal dma_channels_m2s : adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1);
  signal dma_channels_s2m : adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1) := (others => ADB3_OCP_DEFAULT_S2M);
  signal dma_abort        : std_logic_vector(DMA_CHANNELS-1 downto 0);

  -- Debug info
  signal bram_mux_err_info : std_logic_vector(31 downto 0);
  signal ram_mux_err_info  : mem_if_debug_array_t;

  signal empty_m2s : adb3_ocp_m2sT := ADB3_OCP_DEFAULT_M2S;
  signal empty_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  signal ocp_ch0_trig : std_logic_vector(7 downto 0) := (others => '0');
  signal ocp_ch1_trig : std_logic_vector(7 downto 0) := (others => '0');
  signal ocp_ch2_trig : std_logic_vector(7 downto 0) := (others => '0');

  --------------------------------
  -- Register interface signals --
  --------------------------------

  -- Interrupt interface
  signal interrupt : std_logic;

  -- BRAM direct slave OCP interface
  signal ds_bram_m2s : adb3_ocp_m2sT;
  signal ds_bram_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  -- On-board RAM sirect slave OCP interface
  signal ds_ram_m2s  : adb3_ocp_m2sT;
  signal ds_ram_s2m  : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  -- BRAM OCP interface
  signal bram_if_m2s : adb3_ocp_m2sT;
  signal bram_if_s2m : adb3_ocp_s2mT := ADB3_OCP_DEFAULT_S2M;

  -- On-board memory OCP interface
  signal ram_if_m2s  : adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
  signal ram_if_s2m  : adb3_ocp_s2mT_array(0 to MEM_BANKS-1) := (others => ADB3_OCP_DEFAULT_S2M);

  signal mem_if_stat     : mem_if_stat_array_t;
  signal mem_if_err      : mem_if_err_array_t;
  signal mem_if_rdy      : mem_if_rdy_array_t;
  signal mem_if_err_info : mem_if_debug_array_t;

  -- On-board memory application interface
  signal mem_app_m2s  : adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
  signal mem_app_s2m  : adb3_ocp_s2mT_array(0 to MEM_BANKS-1) := (others => ADB3_OCP_DEFAULT_S2M);

  signal mem_app_go       : mem_app_go_array_t;
  signal mem_app_done     : mem_app_done_array_t;
  signal mem_app_err      : mem_app_err_array_t;
  signal mem_app_err_ph   : mem_app_err_ph_array_t;
  signal mem_app_offset   : mem_app_offset_array_t;
  signal mem_app_length   : mem_app_length_array_t;
  signal mem_app_err_addr : mem_app_err_addr_array_t;

begin

  ---------------------
  -- Clock interface --
  ---------------------
  blk_clocks_i : blk_clocks
  port map(
    -- Clock inputs
    clks_in     => clks_in,      -- in    clks_in_t
    clks_out    => clks_out,     -- out   clks_out_t
    clks_mgt_in => clks_mgt_in,  -- in    clks_mgt_in_t
    -- Reset output
    pll_rst     => rst,          -- out   std_logic
    -- Clock outputs
    pll_pri_clk => pll_pri_clk,  -- out   std_logic
    pll_reg_clk => pll_reg_clk,  -- out   std_logic
    pll_mem_clk => pll_mem_clk,  -- out   std_logic
    pll_ref_clk => pll_ref_clk,  -- out   std_logic
    clk_vec     => clk_vec,      -- out   clk_vec_t
    -- PCIe outputs
    mptl_clk    => open,         -- out   std_logic_dbl_t
    pcie_clk    => pcie_clk);    -- out   std_logic_dbl_t


  --------------------
  -- PCIe interface --
  --------------------

  -- dma_abort tied inactive
  dma_abort(DMA_CHANNELS-1 downto 0) <= (others => '0');

  -- Wrapper for pcie target interfaces
  pcie_if_target_wrap_i : pcie_if_target_wrap
  port map(
    ref_clk          => pll_ref_clk,              -- in    std_logic
    -- OCP interface
    ocp_clk          => pll_pri_clk,              -- in    std_logic
    direct_slave_m2s => direct_slave_m2s,         -- out   adb3_ocp_m2sT
    direct_slave_s2m => direct_slave_s2m,         -- in    adb3_ocp_s2mT
    dma_channels_m2s => dma_channels_m2s,         -- out   adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m => dma_channels_s2m,         -- in    adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    dma_abort        => dma_abort,                -- in    std_logic_vector(DMA_CHANNELS-1 downto 0)
    interrupt        => interrupt,                -- in    std_logic
    -- PCIe Interface
    pcie_rst_l       => pcie_rst_l,               -- in    std_logic
    pcie_clk         => pcie_clk,                 -- in    std_logic_dbl_t
    pcie_h2t         => pcie_h2t,                 -- in    pcie_h2t_t
    pcie_t2h         => pcie_t2h,                 -- out   pcie_t2h_t
    -- Board IO
    model_in         => custom_in.adb3_model,     -- in    adb3_model_in_t
    model_out        => custom_out.adb3_model,    -- out   adb3_model_out_t
    model_inout      => custom_inout.adb3_model); -- inout adb3_model_inout_t


  ------------------------------
  -- OCP Direct Slave Channel --
  ------------------------------

  blk_direct_slave_i : blk_direct_slave
  port map(
    -- Reset input
    rst               => rst,                  -- in    std_logic
    -- Clock inputs
    pll_pri_clk       => pll_pri_clk,          -- in    std_logic
    pll_reg_clk       => pll_reg_clk,          -- in    std_logic
    pll_ref_clk       => pll_ref_clk,          -- in    std_logic
    pll_mem_clk       => pll_mem_clk,          -- in    std_logic
    -- Direct slave primary OCP interface
    direct_slave_m2s  => direct_slave_m2s,     -- in    adb3_ocp_m2sT
    direct_slave_s2m  => direct_slave_s2m,     -- out   adb3_ocp_s2mT
    -- Clock frequency measurement register interface
    clk_vec           => clk_vec,              -- in    clk_vec_t
    -- GPIO signal test register interface
    gpio_inout        => gpio_inout,           -- inout gpio_inout_t
    -- Interrupt test register interface
    finti_l           => open,                 -- out   std_logic
    interrupt         => interrupt,            -- out   std_logic
    -- Memory interface status
    mem_if_stat       => mem_if_stat,          -- in    mem_if_stat_array_t
    mem_if_err        => mem_if_err,           -- in    mem_if_err_array_t
    -- Memory application control
    mem_app_go        => mem_app_go,           -- out   mem_app_go_array_t
    mem_app_offset    => mem_app_offset,       -- out   mem_app_offset_array_t
    mem_app_length    => mem_app_length,       -- out   mem_app_length_array_t
    -- Memory application status
    mem_app_done      => mem_app_done,         -- in    mem_app_done_array_t
    mem_app_err       => mem_app_err,          -- in    mem_app_err_array_t
    mem_app_err_ph    => mem_app_err_ph,       -- in    mem_app_err_ph_array_t
    mem_app_err_addr  => mem_app_err_addr,     -- in    mem_app_err_addr_array_t
    -- BRAM secondary OCP interface
    ds_bram_m2s       => ds_bram_m2s,          -- out   adb3_ocp_m2sT
    ds_bram_s2m       => ds_bram_s2m,          -- in    adb3_ocp_s2mT
    -- RAM secondary OCP interface
    ds_ram_m2s        => ds_ram_m2s,           -- out   adb3_ocp_m2sT
    ds_ram_s2m        => ds_ram_s2m,           -- in    adb3_ocp_s2mT
    -- Debug info
    bram_mux_err_info => bram_mux_err_info,    -- in   std_logic_vector(31 downto 0)
    ram_mux_err_info  => ram_mux_err_info,     -- in   mem_if_debug_array_t
    mem_if_err_info   => mem_if_err_info);     -- in   mem_if_debug_array_t

  ----------------------
  -- OCP DMA Channels --
  ----------------------

  blk_dma_switch_i : blk_dma_switch
  port map(
    ocp_rst           => rst,                  -- in    std_logic
    ocp_clk           => pll_pri_clk,          -- in    std_logic
    -- Source: BRAM direct slave OCP interface
    ds_bram_m2s       => ds_bram_m2s,          -- in    adb3_ocp_m2sT
    ds_bram_s2m       => ds_bram_s2m,          -- out   adb3_ocp_s2mT
    -- Source: On-board RAM DS window OCP interface
    ds_ram_m2s        => ds_ram_m2s,           -- in    adb3_ocp_m2sT
    ds_ram_s2m        => ds_ram_s2m,           -- out   adb3_ocp_s2mT
    -- Source: DMA OCP interface
    dma_channels_m2s  => dma_channels_m2s,     -- in    adb3_ocp_m2sT_array(0 to DMA_CHANNELS-1)
    dma_channels_s2m  => dma_channels_s2m,     -- out   adb3_ocp_s2mT_array(0 to DMA_CHANNELS-1)
    -- Source: On-board RAM app OCP interface
    mem_app_m2s       => mem_app_m2s,          -- in    adb3_ocp_m2sT_array(0 to MEM_BANKS-1)
    mem_app_s2m       => mem_app_s2m,          -- out   adb3_ocp_s2mT_array(0 to MEM_BANKS-1)
    -- Dest: BRAM OCP interface
    bram_if_m2s       => bram_if_m2s,          -- out   adb3_ocp_m2sT
    bram_if_s2m       => bram_if_s2m,          -- in    adb3_ocp_s2mT
    -- Dest: On-board RAM OCP interface
    ram_if_m2s        => ram_if_m2s,           -- out   adb3_ocp_m2sT_array(0 to MEM_BANKS-1)
    ram_if_s2m        => ram_if_s2m,           -- in    adb3_ocp_s2mT_array(0 to MEM_BANKS-1)
      -- Debug
    bram_mux_err_info => bram_mux_err_info,    -- out   std_logic_vector(31 downto 0);
    ram_mux_err_info  => ram_mux_err_info);    -- out   mem_if_info_array_t);


  --------------------
  -- BRAM Interface --
  --------------------

  blk_bram_i : blk_bram
  port map(
    ocp_rst     => rst,          -- in    std_logic
    ocp_clk     => pll_pri_clk,  -- in    std_logic
    -- BRAM OCP interface
    bram_if_m2s => bram_if_m2s,  -- in    adb3_ocp_m2sT
    bram_if_s2m => bram_if_s2m); -- out   adb3_ocp_s2mT)


  -------------------------------
  -- On-board Memory Interface --
  -------------------------------

  blk_mem_if_i : blk_mem_if
  port map(
    -- OCP interface
    ocp_rst         => rst,              -- in    std_logic
    ocp_clk         => pll_pri_clk,      -- in    std_logic
    mem_if_m2s      => ram_if_m2s,       -- in    adb3_ocp_m2sT_array(0 to MEM_BANKS-1)
    mem_if_s2m      => ram_if_s2m,       -- out   adb3_ocp_s2mT_array(0 to MEM_BANKS-1)
    -- Memory clocking
    mem_rst         => rst,              -- in    std_logic
    mem_ref_clk     => pll_ref_clk,      -- in    std_logic
    mem_clk         => pll_mem_clk,      -- in    std_logic
    -- Memory status
    mem_if_rdy      => mem_if_rdy,       -- out   mem_if_rdy_array_t
    mem_if_stat     => mem_if_stat,      -- out   mem_if_stat_array_t
    mem_if_err      => mem_if_err,       -- out   mem_if_err_array_t
    -- Memory physical interface
    mem_addr_out    => mem_addr_out,     -- out   mem_addr_out_t
    mem_ctrl_out    => mem_ctrl_out,     -- out   mem_ctrl_out_t
    mem_data_inout  => mem_data_inout,   -- inout mem_data_inout_t
    mem_clk_out     => mem_clk_out,      -- out   mem_clk_out_t
    -- Debug info
    mem_if_err_info => mem_if_err_info); -- out   mem_if_debug_array_t)


  ---------------------------------
  -- On-Board Memory Application --
  ---------------------------------

  blk_mem_app_i : blk_mem_app
  port map(
    -- OCP interface
    ocp_rst       => rst,              -- in    std_logic;
    ocp_clk       => pll_pri_clk,      -- in    std_logic;
    app_m2s       => mem_app_m2s,      -- out   adb3_ocp_m2sT_array(0 to MEM_BANKS-1);
    app_s2m       => mem_app_s2m,      -- in    adb3_ocp_s2mT_array(0 to MEM_BANKS-1);
    -- Control interface
    mem_if_rdy    => mem_if_rdy,       -- in    mem_if_rdy_array_t
    app_go        => mem_app_go,       -- in    mem_app_go_array_t
    app_offset    => mem_app_offset,   -- in    mem_app_offset_array_t;
    app_length    => mem_app_length,   -- in    mem_app_length_array_t;
    app_done      => mem_app_done,     -- out   mem_app_done_array_t
    app_err       => mem_app_err,      -- out   mem_app_err_array_t
    app_err_addr  => mem_app_err_addr, -- out   mem_app_err_addr_array_t;
    app_err_ph    => mem_app_err_ph);  -- out   app_err_ph_array_t


  -------------------------
  -- Chipscope Interface --
  -------------------------

  blk_chipscope_i : blk_chipscope
  generic map(
    instantiate => CHIPSCOPE_ON)  -- boolean := true
  port map(
    -- OCP channel 0 interface
    ocp_ch0_clk   => pll_pri_clk,             -- in    std_logic
    ocp_ch0_m2s   => dma_channels_m2s(0),     -- in    adb3_ocp_m2sT
    ocp_ch0_s2m   => dma_channels_s2m(0),     -- in    adb3_ocp_s2mT
    ocp_ch0_trig  => ocp_ch0_trig,            -- in    std_logic_vector(7 downto 0)
    -- OCP channel 1 interface
    ocp_ch1_clk   => pll_pri_clk,             -- in    std_logic
    ocp_ch1_m2s   => direct_slave_m2s,        -- in    adb3_ocp_m2sT
    ocp_ch1_s2m   => direct_slave_s2m,        -- in    adb3_ocp_s2mT
    ocp_ch1_trig  => ocp_ch1_trig,            -- in    std_logic_vector(7 downto 0)
    -- OCP channel 2 interface
    ocp_ch2_clk   => pll_pri_clk,             -- in    std_logic
    ocp_ch2_m2s   => empty_m2s,               -- in    adb3_ocp_m2sT
    ocp_ch2_s2m   => empty_s2m,               -- in    adb3_ocp_s2mT
    ocp_ch2_trig  => ocp_ch2_trig);           -- in    std_logic_vector(7 downto 0))

end;